/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/** 
    Generation of the declaration of a signal as a parameter of a read/write function.
    @param InOutSignal (in) a signal
    @param anInput (in) at "true" for a input signal, "false" for an output.
*/
  _VIRTUAL_ void
cGenerator::genInputOutputParameter (ptexpcal InOutSignal, boolean anInput)
{
  TypeType t;
  ptident id;

  if (EM->getNumberOfDimensions (InOutSignal) == 0)
    {
      t = EM->getType (InOutSignal);
      gentype (t);
      id = EM->getName (InOutSignal);
      TextForm->printidentstar (id, anInput && (t != StringType));
    }
  else
    genSignalDeclaration (InOutSignal);
}

/**
  Generation of  header declaration or header definition of a read operation.
  @param bdef (in)  when it is pk_true, it implies the generation of the header definition, 
  the generation of the header declaration otherwise.
  @param sig (in) a signal from which one the following text is generated :
  - "extern logical r_foo_SIGNAME(SIGTYPE *SIGNAME);" when not bdef
  - "extern logical r_foo_SIGNAME(SIGTYPE *SIGNAME)" when bdef

  where SIGNAME denotes the name of SIG, and SIGTYPE the type of SIG
*/
_VIRTUAL_ void
cGenerator::genReadHeaderExterncall (ptexpcal sig, boolean bdef)
{
  functionDeclaration = pk_true;
  TextForm->printDclEnvReadFuncBegin (getIOCompleteIdentifier (sig), progName);
  genInputOutputParameter (sig, pk_true);
  if (bdef)
    TextForm->printRPline ();
  else
    TextForm->printRPSemiColonLine ();
  functionDeclaration = pk_false;
}

/**
  Generation of  header declaration or header definition of a write operation.
  @param bdef (in)   when it is pk_true, it implies the generation of the header definition, 
  the generation of the header declaration otherwise.
  @param sig (in) a signal from which one the following text is generated :

  - "extern void w_foo_SIGNAME(SIGTYPE SIGNAME);" when not bdef
  - "extern void w_foo_SIGNAME(SIGTYPE SIGNAME)" when bdef

  where SIGNAME denotes the name of SIG, and SIGTYPE the type of SIG
*/
_VIRTUAL_ void
cGenerator::genWriteHeaderExterncall (ptexpcal sig, boolean bdef)
{
  ptmemnode mn;

  functionDeclaration = pk_true;
  getindielemdim (&sig, &mn);
  TextForm->printDclEnvWriteFuncBegin (getIOCompleteIdentifier (sig), progName);
  genInputOutputParameter (sig, pk_false);
  if (bdef)
    TextForm->printRPline ();
  else
    TextForm->printRPSemiColonLine ();
  functionDeclaration = pk_false;
}


/**
  It generates the body of the reading function of a signal.
  @param aSig (in) an input signal. 
  \see genReadType
*/
_VIRTUAL_ void
cGenerator::genDefReadingSignal (ptexpcal aSig)
{
  ptident vcdname, ids;
  TypeType t;

  TextForm->blockIndentOpening ();
  ids = EM->getName (aSig);
  t = EM->getType (aSig);
  if (TypeTestArrayClass (t))
    genLocalIndexLstDeclaration (InternalIndexLst, -1);
  genReadWriteType (aSig, NILEXPCAL, t, pk_true);
  if (EM->getAttribute (aSig, vcdmark, (attributeValue *) & vcdname))
    TextForm->printVCDValue (t, ids, vcdname, pk_true);
  TextForm->printreturn (pk_true);
  TextForm->blockIndentClosing ();
}


/**
  Generation of  the body of the writing function of a signal.
  @param aSig (in) an output signal. 
  \see genWriteType
*/
_VIRTUAL_ void
cGenerator::genDefWritingSignal (ptexpcal aSig)
{
  ptident vcdname, ids;
  TypeType t;

  TextForm->blockIndentOpening ();
  ids = EM->getName (aSig);
  t = EM->getType (aSig);
  if (TypeTestArrayClass (t))
    genLocalIndexLstDeclaration (InternalIndexLst, -1);
  genReadWriteType (aSig, NILEXPCAL, t, pk_false);
  if (EM->getAttribute (aSig, vcdmark, (attributeValue *) & vcdname))
    TextForm->printVCDValue (t, ids, vcdname, pk_false);
  TextForm->printbodywritefinal (ids);	// line, flush
  TextForm->blockIndentClosing ();
}

/** Generation of  the "reading function" entry point of a signal. 
   @param aSig (in) a input signal.
*/
_VIRTUAL_ void
cGenerator::genDeclAndDefReadingSignal (ptexpcal aSig)
{
  genReadHeaderExterncall (aSig, pk_true);
  genDefReadingSignal (aSig);
}

/** Generation of  the "writing function" entry point of a signal. 
   @param aSig (in) a output signal.
*/
_VIRTUAL_ void
cGenerator::genDeclAndDefWritingSignal (ptexpcal aSig)
{
  genWriteHeaderExterncall (aSig, pk_true);
  genDefWritingSignal (aSig);
}


/** 
    It generates the definition of a read instruction for an identifier (ids), taking
    into account a set of fields (fields) and an array context (isArray).

    examples:
    - foo
    - &(*foo).f1.f2 for two fields (f1, f2) and isArray=false.
    - &(*foo).f1.f2 for two fields (f1, f2) and isArray=false.
*/
_VIRTUAL_ void
cGenerator::genbodyread (ptident ids, ptexpcal fields, boolean isArray)
{
  if (fields != NILEXPCAL)
    {
      TextForm->printkwref ();
      TextForm->printLP ();
      if (isArray)
	TextForm->printident (ids);
      else
	{
	  TextForm->printLP ();
	  TextForm->printstar ();
	  TextForm->printident (ids);
	  TextForm->printRP ();
	}
      genFieldsList (fields);
      TextForm->printRP ();
    }
  else
    TextForm->printident (ids);
}


/** 
    It generates the definition of a write instruction for an identifier (ids), taking
    into account a set of fields (fields).
*/
_VIRTUAL_ void
cGenerator::genbodywrite (ptident ids, ptexpcal addedExp)
{

  TextForm->printident (ids);
  genFieldsList (addedExp);
  TextForm->printValueSeparator ();
}


/**
   It generates the expansion of the lpart:= rpart in lpart.e1....eN = rpart.f1.....fm  for each field of lpart 
   for a signal of structure type, where
   - e1....eN are given by lfields (optional) 
   - f1.....fm are given by lfieldsdef (optional).
   - t is the the type of lpart.e1....eN.
   It calls genAssignString() for the string type.
*/
_VIRTUAL_ void
cGenerator::genReadWriteType (ptexpcal vsig, ptexpcal addedExp, TypeType t, boolean bread)
{
  ptexpcal addedExp1, e;
  NameList NL;
  TypeList fieldsType;
  TypeType TT;
  TypeSubList NSubLstType;
  boolean isArray, Goon;
  NameSubList NSubLst;
  ptrepstr fieldName;
  int n;
  ptident ids;

  if (TypeTestNamedClass (t))
    t = NamedTypeGetTerm (t);
  if (TypeTestStructureClass (t))
    {
      NL = StructureTypeGetNameLst (t);	// field Name
      NSubLst = NameListCast2SubList (NL);
      fieldsType = StructureTypeGetTypeLst (t);
      NSubLstType = TypeListCast2SubList (fieldsType);
      Goon = pk_true;
      while (Goon)
	{
	  fieldName = NameSubListGetHead (NSubLst);
	  e = EM->mkConstantExp (opnom, fieldName);
	  TT = TypeSubListGetHead (NSubLstType);
	  addedExp1 = EM->mkNaryExp (oplesign, e, addedExp, NILBACKANNOTATION);
	  genReadWriteType (vsig, addedExp1, TT, bread);
	  if (NameSubListTestLast (NSubLst))
	    Goon = pk_false;
	  else
	    {
	      NSubLst = NameSubListGetTail (NSubLst);
	      NSubLstType = TypeSubListGetTail (NSubLstType);
	    }
	}
    }
  else if (TypeTestArrayClass (t))
    {
      n = genOpenLoopForDimLst (ArrayMatrixTypeGetDimLst (t));
      if (getExprLastIndexes (n, &e))
	addedExp = EM->mkNaryExp (oplesign, e, addedExp, NILBACKANNOTATION);
      TextForm->printblockopening ();
      genReadWriteType (vsig, addedExp, ArrayMatrixTypeGetTerm (t), bread);
      TextForm->printBlockClosingBackIndent ();
      genCloseIteration (n);
    }
  else if ((t == ComplexType) || (t == DComplexType))
    {
      if (t == ComplexType)
	TT = RealType;
      else
	TT = DRealType;
      e = EM->mkConstantExp (opnom, string2repstr ("re"));
      addedExp1 = EM->mkNaryExp (oplesign, e, addedExp, NILBACKANNOTATION);
      genReadWriteType (vsig, addedExp1, TT, bread);
      e = EM->mkConstantExp (opnom, string2repstr ("im"));
      addedExp1 = EM->mkNaryExp (oplesign, e, addedExp, NILBACKANNOTATION);
      genReadWriteType (vsig, addedExp1, TT, bread);
    }
  else
    {
      ids = EM->getName (vsig);
      if (bread)
	{
	  TT = EM->getType (vsig);
	  if (TypeTestNamedClass (TT))
	    TT = NamedTypeGetTerm (TT);
	  isArray = TypeTestArrayClass (TT);
	  TextForm->printbodyreadhead (ids, t);
	  genbodyread (ids, addedExp, isArray);
	  TextForm->printdclreadtail (ids);
	}
      else
	{
	  TextForm->printbodywritehead (ids, t);
	  genbodywrite (ids, addedExp);
	  TextForm->printbodywritetail (ids);
	}
    }
}

/** 
   Generation of  reading/writing operations for input/output signals of the program.
   @param ins (in) list of the input nodes.
   @param outs (in) list of the output nodes.
*/
_VIRTUAL_ void
cGenerator::genIOReadingWriting (ptlINode ins, ptlINode outs)
{
  ptINode n;

  while (ptlINode_get_elem (&ins, &n))
    genDeclAndDefReadingSignal (NM->get_io_arg (n));

  while (ptlINode_get_elem (&outs, &n))
    genDeclAndDefWritingSignal (NM->get_io_arg (n));
}

/** 
    Generation of  file descriptor declarations.
    One by input/output node of the program.
*/
_VIRTUAL_ void
cGenerator::genIONodesFileDeclarations (ptlINode ln, boolean InOrOut, boolean * bfirst)
{
  ptINode n;
  ptexpcal exp;

  while (ptlINode_get_elem (&ln, &n))
    {
      exp = NM->get_io_arg (n);
      TextForm->printdclfile (getIOCompleteIdentifier (exp), InOrOut, bfirst);
    }
}

/**
   Generation of  file descriptor declarations.
   One by input/output signals of the program using the following conventions : 
   for an input X (resp. output), the file descriptor is called "frX" (resp. "fwX")
   @param ins (in) list of the input nodes.
   @param outs (in) list of the output nodes.
*/
_VIRTUAL_ void
cGenerator::genIOFileDeclarations (ptlINode ins, ptlINode outs)
{
  boolean bfirst;

  bfirst = pk_true;
  genIONodesFileDeclarations (ins, pk_true, &bfirst);
  genIONodesFileDeclarations (outs, pk_false, &bfirst);
  if (!bfirst)
    printptvlinebi (1);
  TextForm->printVCDDeclarations ();
  TextForm->printLine ();
}


/**
   It generates the closing of the files for input/outputs operations (ln).
   It is a set of reading node, when InOrOut, or writing nodes otherwise.
*/
_VIRTUAL_ void
cGenerator::genIONodesClosing (ptlINode ln, boolean InOrOut)
{
  ptexpcal exp;
  ptINode n;

  while (ptlINode_get_elem (&ln, &n))
    {
      exp = NM->get_io_arg (n);
      TextForm->printfileclose (getIOCompleteIdentifier (exp), InOrOut);
    }
}

/**
  Generation of  the entry point Foo_endio().c of the file foo_io.c, where Foo denotes the name
  of the current graph. The definition of this entry is a set of file closing (genIONodesClosing()).
  @param ins (in) list of the input nodes.
  @param outs (in) list of the output nodes.
*/
_VIRTUAL_ void
cGenerator::genIOClosing (ptlINode ins, ptlINode outs)
{
  TextForm->printdclendiofunchead ();
  genIONodesClosing (ins, pk_true);
  genIONodesClosing (outs, pk_false);
  TextForm->genIOClosingVCDFile ();
  TextForm->printdclfuncend ();
}

/**
   It generates the opening of the files for input/outputs operations (ln).
   It is a set of reading node, when InOrOut, or writing nodes otherwise.
*/
_VIRTUAL_ int
cGenerator::genIONodesOpeningFiles (ptlINode ln, boolean InOrOut, int cpt, ptident idvcd)
{
  ptexpcal exp;
  ptINode n;
  ptident vcdname, id;
  TypeType t;

  while (ptlINode_get_elem (&ln, &n))
    {
      exp = NM->get_io_arg (n);
      id = getIOCompleteIdentifier (exp);
      TextForm->printfileopen (id, InOrOut);
      cpt++;
      t = EM->getType (exp);
      vcdname = nameManager->new_exclusiveIdent (idvcd, cpt);
      EM->addAttribute (exp, vcdmark, (attributeValue) vcdname);
      TextForm->printdclVCDvar (t, id, vcdname);
    }
  return cpt;
}

/** 
   Generation of the entry point foo_begio().c of the file foo_io.c. 
   The definition of this entry is a set of file openings (genIONodesOpeningFiles()).
   @param ins (in) list of the input nodes.
   @param outs (in) list of the output nodes.
*/
_VIRTUAL_ void
cGenerator::genIOOpening (ptlINode ins, ptlINode outs)
{
  ptident idvcd;

  idvcd = string2ident ("<");
  TextForm->printdclbegiofunchead ();
  TextForm->genIOOpeningVCDFile ();
  genIONodesOpeningFiles (outs, pk_false, genIONodesOpeningFiles (ins, pk_true, 0, idvcd), idvcd);
  TextForm->printVCDHeaderEnd ();
  TextForm->printdclfuncend ();
}

/**
   It opens the file (foo_io.c) for the code generation of the io module, where foo is the name of the current graph.
*/
_VIRTUAL_ void
cGenerator::geniomoduleInit ()
{
  TextForm->printdcliofilehead (progName);
  /* Aug 2014 (LBesnard) :The declaration of the indexes have been moved in every read/write operation for code generation with threads.
     genIndexLstDeclaration(InternalIndexLst, -1); */

}

/**
   It closes the file (foo_io.c), previously opened in geniomoduleInit().
*/
_VIRTUAL_ void
cGenerator::geniomoduleFinalize ()
{
  TextForm->closefile ();
}

/**
   It generates the I/O function (where foo is the name of the application):
   - foo_stepIO_begin(); 
   - foo_stepIO_end();
   These functions are called before (resp. after) the foo_iterate().
   It allows to connect the result of a simuilation with a VCD viewer (gtkwave,...)
*/
_VIRTUAL_ void
cGenerator::genIOVCDFunctions ()
{
  // foo_stepIO_begin();
  TextForm->printdclstepiobeginfunchead ();
  TextForm->printstepiobeginbody ();
  TextForm->printdclfuncend ();
  // foo_stepIO_end();
  TextForm->printdclstepioendfunchead ();
  TextForm->printdclfuncend ();
}

/** 
   Entry point of the generating of the i/o module. 
   Code (file declarations, file opening, file closing, i/o operations) is generated  in the file foo_io.c, 
   where foo is the name of the current graph.

   It initializes the io code generation 
   then it genrates 
     - the declaration of the io-files: genIOFileDeclarations()
     - the opening of the io-files for read/write operations : genIOOpening()
     - the closing of the io-files for read/write operations : genIOClosing()
     - the code for each read (input)/write (=output) of the interface of the Signal program: genIOReadingWriting()

   Then it closes the file (foo_io.c).

   \note Conventions: for an input x (resp. output), the attached file is "Rx.dat" (resp. "Wx.dat").

*/
_VIRTUAL_ void
cGenerator::geniomodule ()
{
  ptlINode ins, outs;


  NM->getInputs (&ins);		// inputs
  NM->getOutputs (&outs);	// outputs
  geniomoduleInit ();
  genIOFileDeclarations (ins, outs);
  genIOOpening (ins, outs);
  genIOClosing (ins, outs);
  genIOReadingWriting (ins, outs);
  genIOVCDFunctions ();
  geniomoduleFinalize ();
}
