/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Specific to the management of the nodes of delay tree.
   It generates the initial value for each indexes assigned to the node MN.
   if i>0, the definition of the index assigned with mn is already generated. This value can be different of the initial value index.
*/
  _VIRTUAL_ void
cGenerator::genDelayIndexesInit (ptmemnode mn, int i /* boolean b */ )
{
  ptlmemnode lmn;
  int taille, InitvalueIndex;
  ptmemnode mn1;

  mn1 = mn;
  lmn = ptmemnode_get_equiv (mn);
  while (mn != NILMEMNODE)
    {
      taille = ptmemnode_get_size (mn);
      InitvalueIndex = taille - ptmemnode_get_initial_index (mn) - 2;
      if (ptmemnode_get_delay_exp (mn) != NILEXPCAL)
	InitvalueIndex = InitvalueIndex + 1;
      InitvalueIndex = InitvalueIndex % taille;
      if ((i < 0) || ((i >= 0) && (i != InitvalueIndex)))
	genAssignDelay (mn, InitvalueIndex);
      if (!ptlmemnode_getNaryArgs (&lmn, &mn))
	mn = NILMEMNODE;
    }
}

/** 
    It generates the assignment to the bottom value (pk_false) for each signals of a list (lsig) of signals.
    Some particular cases are filtered ( signal marked by BELONG_MEMORY mark is not initialized,
    When a signal is initialized, it is marked by an internal mark (INIT_DONE). 
*/
_VIRTUAL_ void
cGenerator::genInitBooleanClocks (ptlinstances lsig)
{
  ptexpcal aSig, expd;
  ptINode node;

  while (ptlinstances_get_next (&lsig, &aSig))
    {
      aSig = getExpressionToGenerate (aSig);
      if (EM->IsSignal (aSig))
	{
	  node = EM->getNode (aSig);
	  if (!NM->isMarked (node, BELONG_MEMORY))
	    if (EM->getInitialValue (aSig) == NILEXPCAL)
	      if (!NM->is_delay (node))
		if (!NM->is_cellNode (node))
		  if (!isNodeAtInitialClock (node, &expd))
		    if (!EM->isMarked (aSig, INIT_DONE))
		      {
			EM->addMark (aSig, INIT_DONE);
			genBottomValue (aSig);
		      }
	}
    }
}

/**
   It generates the calling to the initialisation of an external object (externalObj), 
   using a data area (vdata), implemented in a specified language (vl).

   Example: for \em foo  value of externalObj, \em myData  value of vdata, it generates:
   - For C language as implementation language, <em>  foo_initialize (&myData); </em>
   - For C++ language as implementation language, <em>  myData.foo_initialize(); </em>
   
   \note
   - If vl is nilstr, C language is used.
   - Such a code is is produced by Polychrony for separated compiling.
   - currently, only C and C++ are allowed as implementation language.
*/
_VIRTUAL_ void
cGenerator::genInitSignalNodeCall (ptrepstr externalObj, ptrepstr vdata, ptrepstr vl)
{
  ptident id;
  id = repstr2ident (externalObj);
  if ((vl == nilstr) || (vl == string2repstr ("C")))
    {
      TextForm->printInitCallModuleBegin (id);
      TextForm->printkwref ();
      genPrefix ();
      TextForm->printstr (vdata);
    }
  else if (vl == string2repstr ("C++"))
    {
      TextForm->printstr (vdata);
      TextForm->printDot ();
      TextForm->printInitCallModuleBegin (id);
    }
  else
    {
      pK_messenger->errorStrStr (errInterfaceToLanguage, string2repstr ("C"), vl);
      TextForm->printkwnotimpl ();
      TextForm->printInitCallModuleBegin (id);
      TextForm->printkwref ();
      TextForm->printstr (vdata);
    }
  TextForm->printRPSemiColonLine ();
}

/**
   For node (node) referencing a "GreyBox" or "BlackBox" model, it generates (genInitSignalNodeCall()) 
   if it not already done (using lstr list) the calling to the entry point defining of the initialisation of the model.
   It uses the pragma associated with node and the getClassAndObjectExternalCall() for extracting informations of this pragma.

   lstr (out) lstr + the name of the object associated with the node.
*/
_VIRTUAL_ void
cGenerator::genInitNodeCall (ptINode node, ptlrepstr * lstr)
{
  ptrepstr externalObj, vdata, vl;
  ptpragma valPragma;
  ntrees vlang;

  if (isAnnotatedGreyBox (node, &valPragma) || isAnnotatedBlackBoxGraph (node, &valPragma))
    {
      externalObj = getClassAndObjectExternalCall (valPragma, &vdata);
      if (ptlMark_insert_elem (lstr, vdata))
	{

	  if (!NM->getAttribute (node, EXTERN_LANGUAGE_IMPL, (attributeValue *) & vlang))
	    SET_NILTREE (vlang);
	  vl = getStrLanguageExternalDef (vlang);
	  genInitSignalNodeCall (externalObj, vdata, vl);
	}
    }

}

/** 
    It generates the call of the initializing of external referenced graphs.
    The considered nodes are annotated by "GreyBox" or "BlackBox" informations (this code is produced 
    by Polychrony for separated compiling).

    It calls genInitNodeCall() for each node call node and each procedure node call.
*/
_VIRTUAL_ void
cGenerator::geninitNodeCall ()
{
  ptlINode lNodeCall;
  ptINode aNodeCall;
  ptlrepstr lstr;

  /* Initilisation des externes "vus" en tant que boites Noires/Grises (exploitation du pragma Grey_Box "toto") */
  lstr = NILLREPSTR;
  if (NM->getNodeCalls (&lNodeCall))
    while (ptlINode_get_elem (&lNodeCall, &aNodeCall))
      genInitNodeCall (aNodeCall, &lstr);

  if (NM->getProcCalls (&lNodeCall))
    while (ptlINode_get_elem (&lNodeCall, &aNodeCall))
      genInitNodeCall (aNodeCall, &lstr);
}


/**
   This method generates the code of constant signals.
*/
_VIRTUAL_ void
cGenerator::genConstantSignals ()
{
  ptINode anEquation;
  ptexpcal expg, expd;
  ptlINode lis;

  lis = lNodesHinit;
  while (ptlINode_get_elem (&lis, &anEquation))
    if (mustBeGenerated (anEquation))
      {
	NM->get_equation_arg (anEquation, &expg, &expd);
	genAssignSynchronous (expg, expd);
      }
}

/**
   @return "true" if a signal (exp) is a node of a delay tree, "false" otherwise.
   For a node of a delay tree, it generates the initialisation (when it is initiliazed) of the signal (exp).
   
   For that, it calls 
   - genAssignDelay()
   - genAssignSynchronous()

   and always calls  genDelayIndexesInit()

   Moreover, the signal (node of a delay tree) exp is added to the list lexp if this node is a Signal delay node. This list
   will be used to generate (genAssignEquivalence()) the equivalences (see note below).

   
   \note When a signal is a node of a delay tree, it will be replaced for the code generation by an access to 
   a global area (array that allows to represent all the signals of the same tree of delays) using a 
   indice to this array.
*/
_VIRTUAL_ boolean
cGenerator::genInitTreeDelaySignal (ptexpcal exp, ptlinstances * lexp)
{
  ptexpcal expinit;
  ptmemnode mn;
  int taille, i, nbic;
  //  boolean b;

  exp = EM->getCloneClassHead (exp);
  if (!ptexpcal_satisfies_node_tree_delay (exp, &mn))
    return pk_false;
  genParametrizedComment (comm_SignalInit, ident2repstr (EM->getName (exp)));
  expinit = EM->getInitialValue (exp);
  if (expinit != NILEXPCAL)
    {
      taille = ptmemnode_get_size (mn);

      i = (taille - ptmemnode_get_initial_index (mn) - 1) % taille;
      genAssignDelay (mn, i);
      // b = pk_true;
      if (getexpcalnbdim (exp) == getexpcalnbdim (expinit))
	genAssignSynchronous (exp, expinit);
      else
	{
	  nbic = ptmemnode_get_nb_indexes (mn);
	  ptmemnode_set_nb_indexes (mn, nbic + 1);
	  genAssignSynchronous (exp, expinit);
	  ptmemnode_set_nb_indexes (mn, nbic);
	}
    }
  else
    i = -1;			// b = pk_false;
  genDelayIndexesInit (mn, i /* b */ );
  // genAssignEquivalence(exp); must be generated later (order may not be a good one)
  if (i > 0 /*b */ )
    *lexp = ptlinstances_insert (*lexp, exp);
  genComment (comm_Done);
  return pk_true;
}


boolean cGenerator::mustBeInitialized(ptexpcal exp)
{
  ptINode aNode;
  ptexpcal s;

  aNode = EM->getNode (exp);
  // Grrr : this filter is NOT equivalent to mustbeGenerated()
  if (!NM->isMarked (aNode, GencodeMark)) return pk_false;
  if (EM->getAttribute (exp, OPTIMIZED_MEMORY_ATTRIBUTE, (attributeValue *) & s))
    {
      if (!NM->is_delay (aNode)) return pk_false;
    }
  else
    {
      if (NM->isMarked (aNode, substituteGenCodeMark))	return pk_false;
      if (NM->is_equation (aNode))
	if (EM->getAttribute (exp, OPTIMIZED_ATTRIBUTE, (attributeValue *) & s))
	  return pk_false;
      if (EM->isAggregatedSignal (exp)) return pk_false;
    }
  return pk_true;
}


/** 
    It generates the initializing of a signal (exp) filtering optimisation attributes.
    It calls genInitTreeDelaySignal() method when exp is a node of a delay tree.
*/
_VIRTUAL_ void
cGenerator::genSignalInit (ptexpcal exp)
{
  ptexpcal expinit;
  ptINode aNode;
  ptlinstances l, ls, lexp;
  ptexpcal s;

  if (EM->isStateVar (exp)) return; // Init of x is reported to x variable
  if (! mustBeInitialized(exp)) return;

  lexp = NILLINSTANCES;
  if (!genInitTreeDelaySignal (exp, &lexp))
    {
      expinit = EM->getInitialValue (exp);
      if (expinit != NILEXPCAL)
	{
	  exp = getExpressionToGenerate (exp);
	  if (!EM->isMarked (exp, INIT_DONE))
	    {
	      EM->addMark (exp, INIT_DONE);
	      if (EM->getOperator (expinit) == oplelevec)
		genListVectorElem (exp, expinit, getexpcalnbdim (exp) != getexpcalnbdim (expinit));
	      else
		genAssignSynchronous (exp, expinit);
	    }
	}
    }
  // Struct/bundle --> indexes in tree delays
  aNode = EM->getNode (exp);
  if (NM->is_lfields (aNode))
    {
      ls = NM->get_lfields_arg (aNode);
      while (ptlinstances_get_next (&ls, &s))
	genInitTreeDelaySignal (s, &lexp);
      
      if (lexp != NILLINSTANCES)
	{
	  // genComment(comm_DelayTreeEq);
	  l = lexp;
	  while (ptlinstances_get_next (&lexp, &s))
	    genAssignEquivalence (s);
	  ptlinstances_free (l);
	}
    }
}


/** 
    It generates the initializing of the list of signals (lsig).
    It calls genSignalInit() for each of them.
*/
_VIRTUAL_ void
cGenerator::genSignalsInit (ptlinstances lsig)
{
  ptexpcal sig;
  while (ptlinstances_get_next (&lsig, &sig))
    genSignalInit (sig);
}


/** 
    This method generates the initializing of the non-substitued parameters.
*/
_VIRTUAL_ void
cGenerator::genParameterInit (void)
{
  ptlinstances liste;
  ptexpcal exp, expg, expd;
  ptINode n;

  liste = Constants;
  while (ptlinstances_get_next (&liste, &exp))
    {
      n = EM->getNode (exp);
      if (mustBeGenerated (n))
	if (!TypeTestPredef (EM->getUnamedType (exp)))
	  {
	    NM->get_equation_arg (n, &expg, &expd);
	    genAssignSynchronous (expg, expd);
	  }
    }
}

/** 
    It generates initialisations for the state variables. 
    The initial value of a state variable X is reported to 
    the "X" identifier that denotes the previous value of X (available by AttributeNameGencode attribute).
*/
void
cGenerator::genStatesInit ()
{
  ptlinstances lexpr;
  ptexpcal exp, stateVar;
  ptident idPrevious;

  if (!EM->getExprList (opvvarin, &lexpr))
    return;
  // lexpr = Sets of ?x expressions
  while (ptlinstances_get_next (&lexpr, &exp))
    if (EM->getAttribute (exp, AttributeNameGencode, (attributeValue *) & idPrevious))
      {
	stateVar = EM->getUnaryArg (exp);
	genAssignSynchronous (exp, EM->getInitialValue (stateVar));
      }
}

/**
   It generates the initialisation of each affine_sample undersampling. 
   For such an element (B := affine_sample{pahse, d} (C)), 
   a counter has been assigned, and declared, previously in the genAffineSampleCounterDeclaration() method.
   This counter is initialized here to the phase value. 
*/
_VIRTUAL_ void
cGenerator::geninitAffineSample ()
{
  ptINode node;
  ptlINode ln;
  ptexpcal linputs, loutputs, par, boolClock, phase;
  ptident nom, idCounter;

  ln = lAffineSample;
  while (ptlINode_get_elem (&ln, &node))
    {
      NM->get_procCall_arg (node, &nom, &par, &linputs, &loutputs);
      EM->getNaryArgs (&loutputs, &boolClock);
      if (EM->getAttribute (boolClock, affineSampleAttribute, (attributeValue *) & idCounter))
	{
	  EM->getNaryArgs (&par, &phase);
	  printPrefixedIdent (idCounter);
	  TextForm->printassign ();
	  genexpression (phase, opnullary);
	  TextForm->printSemiColonLine ();
	}
    }
}



_PROTECTED_
/**
   It initialises the generation of the foo_initialize entry point.
   It is isolated to allow specialisations.
*/
  _VIRTUAL_ void
cGenerator::genInitialisationsInitialise ()
{
  TextForm->printdclinitializefunchead (SEPARATE);
}


/**
   It finalizes the generation of the foo_initialize entry point.
   It is isolated to allow specialisations.
*/
_VIRTUAL_ void
cGenerator::genInitialisationsFinalyze ()
{
  TextForm->printtruefuncend ();
}



_VIRTUAL_ void
cGenerator::genInitClocks()
 {
  TextForm->printcallentrypointstepinitialize (progName, SEPARATE);
 }


/** 
    This method is the entry point of the generation of the initialisations.
    An entry point Foo_initialize(), where Foo denotes the name of the current graph, is generated which contains these initializings
    (parameters, memories, constant signals) and aslo the calls to the initial entry point of node calls.
    
    It calls:
    - genInitialisationsInitialise()
    - genParameterInit()
    - genSignalsInit()
    - genStatesInit()
    - genConstantSignals()
    - geninitNodeCall()
    - geninitAffineSample()
    - genInitBooleanClocks()
    - printcallentrypointstepinitialize()
    - genInitialisationsFinalyze()
*/
_VIRTUAL_ void
cGenerator::genInitialisations ()
{
  genInitialisationsInitialise ();
  genParameterInit ();
  genSignalsInit (Inputs);
  genSignalsInit (Outputs);
  genSignalsInit (Locals);
  genStatesInit ();
  genConstantSignals ();
  geninitNodeCall ();
  geninitAffineSample ();
  genInitBooleanClocks (lClocksNotToBeInitiliazed);
  genInitClocks();
  genInitialisationsFinalyze ();
}
