/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/


_PROTECTED_
/**
   It returns the name of a task(node).
   It is assumed that this method is called after calling mkTaskName() that fixes the names of the task.
 */
  ptident cGeneratorTasks::getTaskName (ptINode node)
{
  if (TaskNamesTable.find (node) == TaskNamesTable.end ())
    return string2ident ("UNDEFINED_IDENT");	// error
  return TaskNamesTable[node];
}

/**
   it sets a name to a task (n). The unicity of the name is guaranteed.
   Assumed : n is a cluster or an io operation.
*/
void
cGeneratorTasks::mkTaskName (ptINode n)
{
  ptident ori_id, id;
  ptexpcal lsig;
  TaskNamesTableType::iterator it;
  int i;
  boolean bcontinue, found;

  if (NM->is_ioNode (n))
    {
      id = getIOCompleteIdentifier (NM->get_io_arg (n));
      id = makeIdent_ident_ident (progName, id);
    }
  else if (NM->isProcessCall (n))	// _present_XX, pK_receive for Rendez-Vous
    {
      lsig = NM->get_leftOperand (n);
      if (lsig == NILEXPCAL) lsig = NM->get_rightOperand (n);
      id = EM->getName (EM->getNaryArgs (&lsig));
    }
  else if (NM->is_trainCluster (n))
    NM->get_clusterName (n, &id);
  else
    {
      printf ("DEBUG. cGeneratorTasks::mkTaskName: NILIDENT for the name of a task !!!\n");
      id = string2ident ("UNDEFINED_IDENT");
    }
  // unicity.
  bcontinue = pk_true;
  i = 0;
  ori_id = id;
  while (bcontinue)
    {
      it = TaskNamesTable.begin ();
      found = pk_false;
      while (!found && it != TaskNamesTable.end ())
	{
	  found = eq_idents (id, (*it).second);
	  it++;
	}
      if (found)
	{
	  id = makeIdent_ident_int (ori_id, i);
	  i++;
	}
      else
	bcontinue = pk_false;
    }
  TaskNamesTable[n] = id;
}


/**
   It generates the code that signals, to the successors of a task (aTask), the end of its execution.
*/
_VIRTUAL_ void
cGeneratorTasks::genEndTask (ptINode aTask)
{
  int n;
  ptldependence ldep;
  ptdependence aDep;
  ptINode vFrom, vTo;
  ptinsth h;

  int numTask = getTaskNumber(aTask);
  // Signaling to the successors the end of the task.
  if (aTask == delayCluster)
    {
      TextForm->printTaskSignal (SemSynchroEndClusters, numTask);	// signal the end of the step to the "iterate task".
    }
  else
    {
      // Signaling to the successors
      ldep = NM->get_successors (aTask);
      while (ldep->get_next (&aDep, &ldep))
	{
	  h = aDep->getDependence (&vFrom, &vTo);
	  if (vTo != delayCluster)
	    {
	      n = getTaskNumber (vTo);
	      if (n >= 0)
		TextForm->printTaskSignal (n, numTask);
	    }
	}

      // When it exists a delay cluster, the clusters without successors becomes predecessors of the delay cluster.
      if (NM->get_successors (aTask)->getSize () == 0)
	{
	  if (delayCluster != NILINODE)
	    TextForm->printTaskSignal (getTaskNumber (delayCluster), numTask);
	  else
	    TextForm->printTaskSignal (SemSynchroEndClusters, numTask);
	}
    }
}

/**
   It generates the code of a thread. This code has the following form:
   \code
   while(pk_true)
      {
      Wait(); // waiting for the start order.
      specific code of the task;
      Signal(); // signaling to my successors the end of my step.
      }
   \endcode
   \see genEndTask
*/
_VIRTUAL_ void
cGeneratorTasks::genCodeTask (ptINode aTask)
{
  ptident id;
  int i, n, iTask, numTask;

  numTask = getTaskNumber (aTask);
  genParametrizedComment (comm_TaskNumber, integer2repstr (numTask));
  id = getTaskName (aTask);

  // while (pk_true) {
  // Waiting() n times, n = nb predecessors of the task.
  iTask = getTaskNumber (aTask);

  TextForm->printTaskHeaderDef (id, numTask);

  n = semIniTable[iTask];
  if (n == 0) n++;
  for (i = 0; i < n; i++)
    {
      TextForm->printTaskWait (iTask);
    }

  TextForm->printTraceTaskAfterWait (id, numTask);
  // Specific code of the task.
  /* note pour LBesnard:
     read : Il faut dans le module io avoir : Wait(plein(x)); x : =read(x); signal(vide(x)) 
     Ici, on fait appel à la lecture (qui peut etre bloquante)
     write : Il faut dans le module io avoir : Wait(vide(x)); write(x); signal(plein(x))
     Ici, on fait appel à l'écriture (qui peut etre bloquante)
   */
  initContextGenerator (tick);
  genNode (aTask, pk_false, pk_true);
  genBlockOpeningOf (tick);	// Closing all the blocks of the clock hierachy.

  genEndTask (aTask);

  TextForm->printTaskHeaderDefTail (id, numTask);

}

/**
   It generates the code for the tasks of the application.
   \see genCodeTask().
*/
_VIRTUAL_ void
cGeneratorTasks::genCodeTasks ()
{
  ptINode aTask;
  ptlINode lNodes;

  lNodes = lTasks;
  while (ptlINode_get_elem (&lNodes, &aTask))
    genCodeTask (aTask);
}

/**
   It generates the wait of the end of the step for the iterate task.
   If a "delay cluster" exists, the "iterate task" wait for the end of the execution of the
   "delay cluster", otherwise it waits for all the tasks without successors.
*/
_VIRTUAL_ void
cGeneratorTasks::genWaitEndOfstep ()
{
  int i;
  for (i = 1; i <= nbPredIterateTask; i++)
    TextForm->printTaskWait (SemSynchroEndClusters);
}


/**
   Do nothing for the thread generation.
   It is redefined in cGeneratorDist.h class.
*/
_VIRTUAL_ void
cGeneratorTasks::genPrintInfosStep ()
{
  // Don't delete me!
}

/**
   Do nothing for the thread generation.
   It is redefined in cGeneratorDist.h class.
*/
_VIRTUAL_ void
cGeneratorTasks::genDeclStepCounter()
{
  // Don't delete me!
}


/**
   It generates the foo_iterate() entry point. 
   It is a task. The code has the following form:
   \code
   while(pk_true)
      {
      signals to the clusters without predecessor that they can start their execution
      wait the signal of the "delay cluster" (end of an iterate).
      }
   \endcode  It is a task that
*/
_VIRTUAL_ void
cGeneratorTasks::genEntryPointIterate ()
{
  int i;

  genParametrizedComment (comm_TaskNumber, integer2repstr (numberOfTasks));
  TextForm->printTaskHeaderDef (idScheduler, numberOfTasks);

  genWaitEndOfstep ();
  genPrintInfosStep ();
  TextForm->printTraceTaskAfterWait (NILIDENT, numberOfTasks);
  // Authorizing the execution of the clusters without predecessors 
  for (i = 0; i < numberOfTasks; i++)
    {
      if (semIniTable[i] == 0)
	{
	  TextForm->printTaskSignal (i, numberOfTasks);
	}
    }
  TextForm->printTaskHeaderDefTail (NILIDENT, numberOfTasks);
}

_VIRTUAL_ void
cGeneratorTasks::genMainTask ()
{
  TextForm->printTasksMainEntry (SemSynchroMainTask, pk_false);
}


/**
   It generates the code of the suspending of the tasks of the application.
*/
_VIRTUAL_ void
cGeneratorTasks::genEntryPointSuspend ()
{
  ptrepstr id;
  id = kwOsPrimitives[3];
  TextForm->printkwFunctionHeadNoParameters (id, undefinedType, pk_true);
  genCallOsPrimitive (id);
  TextForm->blockclose ();
}


/**
   It generates the code of the resuming of the tasks of the application.
*/
_VIRTUAL_ void
cGeneratorTasks::genEntryPointResume ()
{
  ptrepstr id;
  id = kwOsPrimitives[2];
  TextForm->printkwFunctionHeadNoParameters (id, undefinedType, pk_true);
  genCallOsPrimitive (id);
  TextForm->blockclose ();	// TextForm->printResumeTailDef();
}


/** 
    Generation of the entry point FOO_iterate_Task() dynamic scheduler of the program.
*/
_VIRTUAL_ void
cGeneratorTasks::genScheduler ()
{
  genEntryPointIterate ();
  genCodeTasks ();
  genTerminationTask ();
  genStart ();
  genEntryPointSuspend ();
  genEntryPointResume ();

  genMainTask ();
}


_VIRTUAL_ ptlINode
cGeneratorTasks::getTasksToBeGenerated (ptlINode ln)
{
  ptINode node;
  ptlINode l;

  lTasks = cGenerator::getNodesToBeGenerated (ln);
  // keeping the real tasks.
  l = lTasks;
  while (ptlINode_get_elem (&l, &node))
    {
      if (!NM->is_ioNode (node) && !NM->isProcessCall (node) /* _present_XX */  && !NM->is_trainCluster (node))
	ptlINode_remove_elem (&lTasks, node);
    }
  return lTasks;
}

/**
   Fixing a number to each task, a name to each task.
 */
_VIRTUAL_ void
cGeneratorTasks::fixingTaskAttributes ()
{
  ptINode node;
  ptlINode l;

  l = lTasks;
  bTermTaskGen = pk_false;
  while (ptlINode_get_elem (&l, &node))
    {
      if (existPredefinedPragma (NM->getPragmas (node), PRAGMA_DELAYCLUSTER))
	delayCluster = node;
      if (NM->is_readNode (node))
	bTermTaskGen = pk_true;
      TaskNumTable[node] = numberOfTasks;
      numberOfTasks++;
      mkTaskName (node);
    }
}

/**
   It defines the number of tasks (\ref cGeneratorTasks#numberOfTasks "numberOfTasks") of the application 
   and it identifies the delay cluster (\ref cGeneratorTasks#delayCluster "delayCluster").

   A number, available by the \ref cGeneratorTasks#taskAttribute "taskAttribute" attribute, is assigned to each cluster.
   - the indexes 0 to numberOfTasks-1 are the clusters of the applications
   - the value numberOfTasks will be associated with the scheduller (iterate).
   \see getTaskNumber().
*/
_VIRTUAL_ void
cGeneratorTasks::defineNumberOfTasks ()
{
  lTasks = getTasksToBeGenerated (SubGraphTable[0]->ln);	// SubGraphTable[0]->ldelays is empty, by consruction.
  fixingTaskAttributes ();
}

/**
   It defines the number of semaphores useful to define the synchronisations between the tasks of the application.
   There are one semaphore by task (the foo_iterate is a task).

   \note 
   - this method must be called after the defineNumberOfTasks().
   - Currently, the number of semaphores = number of tasks.
*/
_VIRTUAL_ void
cGeneratorTasks::defineNumberOfSemaphores ()
{
  // One semaphore by task + the following ones.
  SemSynchroEndClusters = numberOfTasks;	// It is also the index of the iterate task 
  SemSynchroMainTask = SemSynchroEndClusters;
  numberSemaphores = SemSynchroMainTask;
  if (bTermTaskGen)
    {
      SemTerminationTask = SemSynchroEndClusters + 1;
      numberSemaphores++;
    }
}

/**
   @returns the number associated with a node (node), the value -1 when it is unknown.
   This number is available after the call to the defineNumberOfTasks() method.
*/
int
cGeneratorTasks::getTaskNumber (ptINode node)
{
  if (TaskNumTable.find (node) == TaskNumTable.end ())
    return -1;			// error
  return TaskNumTable[node];
}


/**
   @return the number of predecessors of a task (node), taking into account the cluster of delays
   (Recall: the dependencies are inverted for the delays).
 */
_VIRTUAL_ int
cGeneratorTasks::getNbPredecessorsCluster (ptINode node)
{
  int nbpred;
  ptldependence ldep;
  ptdependence aDep;
  ptINode vFrom, vTo;
  ptinsth h;

  nbpred = 0;
  ldep = NM->get_predecessors (node);
  while (ldep->get_next (&aDep, &ldep))
    {
      h = aDep->getDependence (&vFrom, &vTo);
      if (!NM->is_inComposedNode (vFrom))	// not internal to a cluster.
	if (vFrom != delayCluster)
	  { 
	    if (NM->is_subgraph (vFrom))
	      nbpred++;
	    else if (NM->isProcessCall (vFrom))	// _present_XX()
	      nbpred++;
	    else
	      // I/O tasks.
	      if (NM->is_writeNode (vFrom))
		nbpred++;		// strange, it may exist a dependence from an output to an another cluster !!!
	      else
		// only the "real" inputs, not them produced by the nodecall nodes.
		if (NM->is_readNode (vFrom) && !NM->is_outputNodeCall (vFrom, &vTo))
		  nbpred++;
	  }
    }
  return nbpred;
}

/**
   It computes the initialisations of the semaphores.
   A table semIniTable[numberOfTasks] is used to keep this value.
   
   if semIniTable[i]=m, it specifies that the i-th task has m predecessors.
   For a task without predecessor, the value is set to 1 : it is the scheduller that signals to 
   such a task it can start.

   \note this method must be called after defineNumberOfSemaphores().
*/
_VIRTUAL_ void
cGeneratorTasks::initsemaphores ()
{
  int i;
  ptINode aTask;
  ptlINode lNodes;

  for (i = 0; i <= numberOfTasks; i++)
    semIniTable.push_back (0);
  if (delayCluster != NILINODE)
    i = getTaskNumber (delayCluster);
  lNodes = lTasks;
  while (ptlINode_get_elem (&lNodes, &aTask))
    if (aTask != delayCluster)
      {
	semIniTable[getTaskNumber (aTask)] = getNbPredecessorsCluster (aTask);
	if (NM->get_successors (aTask)->getSize () == 0)
	  {
	    if (delayCluster != NILINODE)
	      {
		semIniTable[i]++;
		nbPredIterateTask = 1;
	      }
	    else
	      nbPredIterateTask++;
	  }
      }
  semIniTable[numberOfTasks] = 0;	// end of a step, used to synchronize the delayCluster and the iterate.
}

/**
   It prepares the generation of the code of the tasks by
   - defining the number of tasks of the application: defineNumberOfTasks()
   - defining the number of semaphores of the application:defineNumberOfSemaphores()
   - computing the initialisation of each semaphore: initsemaphores()
*/
_VIRTUAL_ void
cGeneratorTasks::initGenCodeTasks ()
{
  kwOsPrimitives.push_back (string2repstr ("create"));
  kwOsPrimitives.push_back (string2repstr ("cancel"));
  kwOsPrimitives.push_back (string2repstr ("resume"));
  kwOsPrimitives.push_back (string2repstr ("suspend"));
  kwOsPrimitives.push_back (string2repstr ("start"));
  idScheduler = TextForm->getIterateName ();
  idTerminate = TextForm->getTerminateName ();
  defineNumberOfTasks ();
  defineNumberOfSemaphores ();
  initsemaphores ();
}

/** 
    It generates in the file FOO_Tasks.h the declarations of
    - the number of tasks
    - the tasks
    - the semaphores
    - the headers of the tasks (genTasksHeader()).
*/
_VIRTUAL_ void
cGeneratorTasks::genTasksDeclarations ()
{
  TextForm->printTaskDeclarations (progName);	/* opening the file FOO_Tasks.h */
  TextForm->printNumberOfTasks (numberSemaphores + 1);	// numberOfTasks is not the real number, it is numberSemaphores.
  TextForm->printDclOfSemaphores (numberSemaphores + 1);
  genTasksHeader ();
  TextForm->closefile ();	/* closing the file. */
}

/**
   It generates the code of the application with threads.
   The communications of the application (for distributed code), are generated by the geniomodule() method.
   
   \see {genExternalDeclarationsFile(), dclTypes(), genTasksDeclarations(), genBody(), geniomodule()}
*/
_VIRTUAL_ void
cGeneratorTasks::gencode ()
{
  genExternalDeclarationsFile ();	/* file foo_externals.h */
  dclTypes ();			/* file foo_types.h  except the externals */
  // genMain();
  genTasksDeclarations ();
  genBody ();
  geniomodule ();
}

/**
   It generates the initialisations of the semaphores.
   All the semaphores are initialized to the value 0.
*/
_VIRTUAL_ void
cGeneratorTasks::genSemaphoresIni ()
{
  int i;

  for (i = 0; i < numberOfTasks; i++)
    TextForm->printSemaphoreIni (i, 0);

  TextForm->printSemaphoreIni (SemSynchroEndClusters, 0);	// Synchro delayCluster-iterate.
  if (bTermTaskGen)
    TextForm->printSemaphoreIni (SemTerminationTask, 0);	// Termination task if it exists.
}


/**
   It generates the headers of the tasks of the application.
   \see printTaskHeaderDcl()
*/
_VIRTUAL_ void
cGeneratorTasks::genTasksHeader ()
{
  ptINode aTask;
  ptlINode lNodes;

  lNodes = lTasks;
  while (ptlINode_get_elem (&lNodes, &aTask))
    TextForm->printTaskHeaderDcl (getTaskName (aTask));
  TextForm->printTaskHeaderDcl (idScheduler);

  if (bTermTaskGen)
    TextForm->printTaskHeaderDcl (idTerminate);
}

/**
   It generates the code of the creating of the tasks of the application.
   \see printTaskCreate()
*/
_VIRTUAL_ void
cGeneratorTasks::genTasksCreate ()
{
  ptrepstr id;
  id = kwOsPrimitives[0];
  genCallOsPrimitive (id);
  if (bTermTaskGen)
    TextForm->printTaskCreate (SemTerminationTask, idTerminate, id);
}

/**
   It finalizes the generation of the foo_initialize entry point.
*/
_VIRTUAL_ void
cGeneratorTasks::genInitialisationsFinalyze ()
{
  genSemaphoresIni ();
  genTasksCreate ();
  cGenerator::genInitialisationsFinalyze ();
}


/**
   @copydoc cGenerator::gendeclarationsBody()
   Moreover, it adds the declarations associated with the tasks.
   \see printIncludeTaskDeclarations().
*/
_VIRTUAL_ void
cGeneratorTasks::gendeclarationsBody ()
{
  cGenerator::gendeclarationsBody ();
  genDeclStepCounter();
  TextForm->printIncludeTaskDeclarations (progName);
}


/**
   It generates the code of a read environment. The thread terminates when the read function returns a "false" value.
   The form of the call is :
   \code
   if (!r_<processName>_<ident1)> (ident2)) pK_Task_exit();
   \endcode
   where ident1(resp. ident2) denotes the identifier of the signal exp (resp. expEff).
   @param exp (in) a signal.
   @param expEff (in) a signal.
   @param processName (in) an identifier.
*/
_VIRTUAL_ void
cGeneratorTasks::genCodeReadEnv (ptexpcal exp, ptident processName, ptexpcal expEff)
{
  TextForm->printkwif ();
  TextForm->printLP ();
  TextForm->printkwnot ();
  TextForm->printReadFunctionLP (getIOCompleteIdentifier (exp), processName);
  genReadEnvSignal (expEff);
  TextForm->printRP ();
  TextForm->printRP ();
  TextForm->printOpenBlockIndent ();
  TextForm->printTaskSignal (SemTerminationTask, -1);	// It frees the termination task.
  // TextForm->printTaskSuicide();
  TextForm->printkwreturnNull ();
  TextForm->printCloseBlockIndent (2);
}


_VIRTUAL_ void
cGeneratorTasks::genCallOsPrimitive (ptrepstr kw)
{
  ptINode aTask;
  ptlINode lNodes;

  lNodes = lTasks;
  while (ptlINode_get_elem (&lNodes, &aTask))
    {
      TextForm->printCallOsPrimitive (getTaskNumber (aTask), getTaskName (aTask), kw);
    }
  TextForm->printCallOsPrimitive (SemSynchroEndClusters, idScheduler, kw);
}

/**
   It generates the code of the end of the application (termination task). 
   The thread is first waiting until a read function returns a "false" value, in this case the read task send a signal() to
   free this termintion task.
   Then this task cancels the other tasks, before sending a signal() to the main task.
*/
_VIRTUAL_ void
cGeneratorTasks::genTerminationTask ()
{
  if (bTermTaskGen)
    {
      TextForm->printTerminationTaskHeaderDef ();
      TextForm->printTaskWait (SemTerminationTask);
      genCallOsPrimitive (kwOsPrimitives[1]);
      TextForm->printCancelTask ();
      //      TextForm->printExit (1);
      TextForm->printCloseBlockIndent (0);
    }
}


/**
   It generates the start entrypoint of the application.
   It assumes the threads/semaphores are created and initialized.
*/
_VIRTUAL_ void
cGeneratorTasks::genStart ()
{
  int i;

  TextForm->printkwFunctionHeadNoParameters (kwOsPrimitives[4], undefinedType, pk_true);
  for (i = 1; i <= nbPredIterateTask; i++)
    TextForm->printTaskSignal (SemSynchroEndClusters, -1);
  TextForm->blockclose ();
}


