/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
#define STRING_ITERATE "step"
#define DATA_AREA_SEPARATE_PREFIX "data->"
/**
   It signals an internal error (operation not yet implemented in the target language,....).
   The tag someProblem is set to "true". It is the returned code of the generator.
*/
void
cGenerator::setSomeProblem()
{
  someProblem = pk_true;
}

/**
   @return the string associated with an external specification (vlang).
   In the current context, it is the implementation language of an external node.
*/
ptrepstr cGenerator::getStrLanguageExternalDef(ntrees vlang)
{
  return ExtractString(STree_getExternalValue(vlang));
}

/**
   For a pragma (valPragma), assumed to be a pragma associated with an external object generated by Polychrony (Black box, Grey box).
   The information is set at the production of the graph. 
   Such a pragma is composed of the name of the external reference(class or structure), and the name of the current reference (vdata). 
   This current reference can be:
   - an object of the class given by the external reference, when the implementation language is a object language, 
   - or a declaration of a structured variable, the type is given by the external reference, when the implemetation language is C for example
   (it allows to have reentrant code).

   @return the name of the external reference (class).
*/
ptrepstr cGenerator::getClassAndObjectExternalCall(ptpragma valPragma, ptrepstr * vdata)
{
  ptrepstr ithCall, externalName;
  ntrees statement;

  statement = getPragmaStatement(valPragma);
  externalName = getPragmaTerm(statement, 1);
  ithCall = getPragmaTerm(statement, 2);
  strconc2str(externalName, ithCall, pk_false, vdata);
  return externalName;
}

/**
   @return the identifier(trees) of the signal operand s
   s can be a reference to a signal or a field of a bundle.
*/
ptident cGenerator::getIOCompleteIdentifier(ptexpcal s)
{
  ptINode n, n1;
  ptident id;

  n = EM->getNode(s);
  id = EM->getName(s);
  if (!NM->is_outputNodeCall(n, &n1))
    return id;
  if (NM->isNodeCall(n1))
    return id;
  return id;
}

/** 
    Generation of  the entry foo_STEP_initialize().
    It initializes some clocks to undefined.
*/
_VIRTUAL_ void cGenerator::genEntryPointStepInitialize()
{
  TextForm->printentrypointstepinitialize(progName, SEPARATE);
  set_booleanClocks_undefined(tick);
  TextForm->printdclfuncend();
}

/** 
    Generating the  the step initialisation call if it is not done. 
*/
_VIRTUAL_ void cGenerator::printCallEntrypointStepInitialize()
{
  if (!bCallEntrypointStepInitialize)
    {
      bCallEntrypointStepInitialize = pk_true;
      TextForm->printcallentrypointstepinitialize(progName, SEPARATE);
    }
}

/**
   It generates the definition of the "micro step" clock if 
   the the graph is an endochronous with an internal clock root.
*/
_VIRTUAL_ void cGenerator::genMicroStepDef()
{
  if (bEndochronousWithInternalClockRoot)
    {
      printPrefixedIdent(getMicroStepName());
      TextForm->printassign();
      TextForm->printkwnot();
      TextForm->printLP();
      genClockExpression(vhComplete, opnullary);
      TextForm->printRP();
      TextForm->printSemiColon();
      TextForm->printLine();
    }
}

_VIRTUAL_ void cGenerator::StepFinalize(boolean modeBlackBox)
{
  TextForm->printStepFinalyze(modeBlackBox);
}

/** 
    Generation of the entry foo_STEP_finalize().
    @param ldelays (in) set of delays of the program;
    @param modeBlackBox (in) at pk_true when the graph is without clusters, "false" otherwise
*/
_VIRTUAL_ void cGenerator::genStepFinalizeNotTrain(ptlINode ldelays, boolean modeBlackBox)
{
  genMicroStepDef();

  TextForm->printcallentrypointstepfinalize(progName, SEPARATE);

  genCodeEndochronousWithInternalClockRootEnd();
  StepFinalize(modeBlackBox);

  TextForm->printentrypointstepfinalize(progName, SEPARATE);
  genOrderedNodes(ldelays, tick, modeBlackBox, pk_false);
  genUpdateStateVariables();
  printCallEntrypointStepInitialize();
}

_VIRTUAL_ void cGenerator::genCodeEndochronousWithInternalClockRootBegin()
{
  if (bEndochronousWithInternalClockRoot)
    {
      TextForm->printBeginBlockMicroStep(getprefix(), getMicroStepName());
      genComment(comment_endointernalclock);
    }
}

/**
   It finalizes the generation of the step code when the graph is an endochronous with an internal clock root.
*/
_VIRTUAL_ void cGenerator::genCodeEndochronousWithInternalClockRootEnd()
{
  if (bEndochronousWithInternalClockRoot)
    TextForm->printEndBlockMicroStep(getprefix(), getMicroStepName());
}

_VIRTUAL_ void cGenerator::genSchedulerBegin()
{
  TextForm->printiterfunchead();
}

_VIRTUAL_ void cGenerator::genSchedulerEnd()
{
  TextForm->printiterfuncEnd();
}

/** 
    Generation of the entry point foo_step() that is a SEQUENTIAL scheduler of the program (for a use as main program).
*/
_VIRTUAL_ void cGenerator::genScheduler()
{
  genSchedulerBegin();

  genCodeEndochronousWithInternalClockRootBegin();
  if (nbSets == 1)
    {
      /* SubGraphTable[0]:: scheduler */
      genOrderedNodes(SubGraphTable[0]->ln, tick, pk_false, pk_false);
      genStepFinalizeNotTrain(SubGraphTable[0]->ldelays, pk_false);
    }
  else
    {				// Set of clusters + scheduler (in SubGraphTable[0])
      genOrderedNodes(SubGraphTable[0]->ln, tick, pk_false, pk_true);
      genOrderedNodes(SubGraphTable[0]->ldelays, tick, pk_false, pk_true);
      printCallEntrypointStepInitialize();
      genCodeEndochronousWithInternalClockRootEnd();
    }

  genSchedulerEnd();
}

/**
   Generation of the declaration of the entry point foo_step_Black_Box().
   @param pv (in) pk_true if the declaration mut be completed with the end declaration separator.
*/
_VIRTUAL_ void cGenerator::genEntryPointDeclarationSchedulerBlackBox(boolean pv)
{
  ptlinstances lins, louts;
  ptexpcal lInputs, lOutputs;
  ptident name;

  //  name = makeIdent_ident_ident(progName, idIterBlackBox);
  name = makeIdent_ident_string(progName, STRING_ITERATE);
  lins = Inputs;
  louts = Outputs;
  lInputs = EM->convertLinstances2ListExpcal(lins);
  lOutputs = EM->convertLinstances2ListExpcal(louts);
  genEntryPointDeclaration(name, lInputs, lOutputs, pv, pk_true);
}

/**
   Generation of the code of entry point foo_step_Black_Box(<parameters>) ; 
   it is a sequential scheduler of the program, whitout Clusters defined in the file foo_body_separate.c
*/
_VIRTUAL_ void cGenerator::genBlackBoxCode()
{
  if (nbSets == 1)
    {
      /* Without clusters. */
      genEntryPointDeclarationSchedulerBlackBox(pk_false);
      TextForm->blockIndentOpening();
      genCodeEndochronousWithInternalClockRootBegin();
      genOrderedNodes(SubGraphTable[0]->ln, tick, pk_true, pk_true);	/* 0 -->scheduler */
      // genCodeEndochronousWithInternalClockRootEnd ();

      genStepFinalizeNotTrain(SubGraphTable[0]->ldelays, pk_true);
      TextForm->printtruefuncend();
    }
}

/**
   @returns the number of auxiliary indexes to be declared for the generation of the nodes of the cluster (Cluster), starting with nb as this number.
*/
int cGenerator::getNbAuxRequiredIndexesIn(ptlINode ln, int nb, int nbblock)
{
  int m;
  ptlinstances ls;
  ptINode n;
  ptexpcal s;
  TypeType TT;
  ptlINode lnodes;
  ptlinsth lh;

  while (ptlINode_get_elem(&ln, &n))
    {
      if (getNodeTravelIndex(n, &ls))
	{
	  m = ptlinstances_get_size(ls);
	  if (m > nb)
	    nb = m;
	}
      if (NM->get_array_bodyDefinitions(n, &lnodes, &lh))
	{
	  nbblock++;
	  nb = getNbAuxRequiredIndexesIn(lnodes, nb, nbblock);
	}
      else
	{
	  ls = NM->getInodelOutputFlows(n);
	  while (ptlinstances_get_next(&ls, &s))
	    {
	      TT = EM->getType(s);
	      if (TypeTestArrayClass(TT))
		{
		  m = getNbDimFlattenedType(TT);
		  if (m - nbblock > nb)
		    nb = m;
		}
	    }
	}
    }
  return nb;
}

/**
   Generation of code for each cluster.
   In SEPARATE mode, entry points are generated with parameters, otherwise without parameters ; in this case, 
   the i/o variables are updated by side effect.
*/
_VIRTUAL_ void cGenerator::genSubGraphs()
{
  int i;
  ptINode aSubGraph;
  ptinsth h;
  ptexpcal lins, louts;
  ptlINode ld;

  if (nbSets == 1)
    return;			/*  Without train clusters */

  for (i = 1; i < nbSets; i++)
    {
      aSubGraph = SubGraphTable[i]->aSubGraph;
      genEntryPointSubGraphHeader(aSubGraph, pk_false);
      NM->get_inOutGraphRef(aSubGraph, &lins, &louts);
      TextForm->blockIndentOpening();

      genLocalIndexLstDeclaration(InternalIndexLst, getNbAuxRequiredIndexesIn(SubGraphTable[i]->ln, 0, 0));

      h = NM->get_nodeClock(aSubGraph);
      if (SEPARATE)
	genAssignIOSeparate(lins, h, pk_true);

      // setUndefinedValueForBooleanList(h, NM->getInodelLocalFlows(aSubGraph)); removed Sep 2014 LBesnard: already done when the "step_initialize" entrypoint is generated.
      genOrderedNodes(SubGraphTable[i]->ln, h, pk_false, pk_false);
      if (existPredefinedPragma(NM->getPragmas(aSubGraph), PRAGMA_DELAYCLUSTER))
	{
	  genMicroStepDef();
	  ld = SubGraphTable[i]->ldelays;
	  genOrderedNodes(ld, h, pk_false, pk_false);
	  genUpdateStateVariables();
	  if (SEPARATE)
	    genAssignIOSeparate(louts, h, pk_false);
	  printCallEntrypointStepInitialize();
	}
      else if (SEPARATE)
	genAssignIOSeparate(louts, h, pk_false);
      TextForm->printdclfuncend();
    }
}

_VIRTUAL_ void cGenerator::genFinalyze()
{
  if (TM->getMode())		/* mode ON */
    TextForm->printRefSimulator(progName);
}

/** 
    Code generation of the file foo_body.c
    - the declarations (includes, variables);
    - the entries: foo_initialize(), foo_STEP_initialize(),
    - an entry point for each cluster,
    - then the sequential scheduler (entry point foo_step ).
*/
_VIRTUAL_ void cGenerator::generateBodyCode()
{
  TextForm->printbodyfileIncludes(progName, pk_false);
  gendeclarationsBody();
  genInitialisations();
  genSubGraphs();
  genEntryPointStepInitialize();
  genScheduler();
  genFinalyze();
}

/** 
    Code generation of the file foo_body.c
    - the declarations (includes, variables);
    - the entries: foo_initialize(), foo_STEP_initialize(),
    - an entry point for each cluster,
    - then the sequential scheduler (entry point foo_step ).
*/
_VIRTUAL_ void cGenerator::genBodyCode()
{
  TextForm->printbodyfilebegin(progName, pk_false);	/* opening the file foo_body.c */
  generateBodyCode();
  TextForm->closefile();	/* closing the file foo_body.c */
}

/**
   Code generation of the files foo_body.c and foo_body.h (headers) for "main" program.
*/
_VIRTUAL_ void cGenerator::genBody()
{
  genBodyHeader();		/* file <P>_body.h */
  genBodyCode();		/* file <P>_body.c  */
}

/**
   It generates the file foo_main.c, where foo is the name of the current program.
*/
_VIRTUAL_ void cGenerator::genMain()
{
  TextForm->printmainfile(progName);
}

/** 
    Entry point of the code generation, with or without clusters, in non-separated mode.
    @param mainFile (in) pk_true if the file foo_main.c must be generated
    @param bodyFile (in) pk_true if the file foo_boby.c must be generated
    @param ioFile (in)  pk_true if the file foo_io.c must be generated
    @param commFile (in) not used
    \see{generate}
    
*/
_VIRTUAL_ void cGenerator::gencode(boolean mainFile, boolean bodyFile, boolean ioFile, boolean commFile)
{
  genExternalDeclarationsFile();	/* file foo_externals.h */
  dclTypes();			/* file foo_types.h  except the externals */
  if (mainFile)
    genMain();
  if (bodyFile)
    genBody();
  if (ioFile)
    geniomodule();
  // printf("DEBUG  exiting cGenerator::gencode\n"); fflush(stdout);
}

/**
   Assignment (attribute) identifiers for previous value of state variables.
*/
_VIRTUAL_ void cGenerator::stateVariablesAttribute()
{
  ptlinstances lexpr;
  ptexpcal exp, stateVar;
  ptident id, idPrevious;

  if (EM->getExprList(opvvarin, &lexpr))	// lexpr = Sets of ?x expressions
    while (ptlinstances_get_next(&lexpr, &exp))
      {
	stateVar = EM->getUnaryArg(exp);
	if (mustBeGenerated(stateVar))
	  {
	    id = EM->getName(stateVar);
	    idPrevious = nameManager->mkIdentNameState(id);
	    EM->addAttribute(exp, AttributeNameGencode, (attributeValue) idPrevious);
	  }
      }
}

/**
   It loads a predefined table of definitions of some operations 
   coded using pragmas (_exception, assert, and operations on matrices) for a language.

   \see sysGetFileCodeGenPredef(), loadTableOfPragmas()
*/
_VIRTUAL_ void cGenerator::loadOperatorsDef(char *fileName)
{
  ntrees p;
  boolean berr;

  tableOpCode = NILLPRAGMAS;
  berr = pk_false;
  if (fileName != NIL_String)
    {
      berr = !sysparser(ph_ol_prag, fileName, &p);
      if (!berr)
	berr = !ntreesPragmas2pragmas(p, &tableOpCode);
      if (berr)
	pK_messenger->warningident(warFileNotFound, string2ident(fileName));
    }
}

/**
   When the current graph is an internal graph of a module (module), 
   it imports the types, the constants, the parameters and the indexes of the module.
*/
_VIRTUAL_ void cGenerator::addImportsFrom(ptDCG_sch module)
{
  ptlinstances l;

  if (module->getUserTypesOneLevel(&l))
    Types = EM->mergeTypesCst(l, Types);

  l = module->getConstants();
  if (l != NILLINSTANCES)
    {
      Constants = ptlinstances_copy(Constants);
      Constants = ptlinstances_sortedfusion(l, Constants);
    }
  l = module->getParameters();
  if (l != NILLINSTANCES)
    {
      Parameters = ptlinstances_copy(Parameters);
      Parameters = ptlinstances_sortedfusion(l, Parameters);
    }
  l = module->getListIndexes();
  if (l != NILLINSTANCES)
    {
      Indexes = ptlinstances_copy(Indexes);
      Indexes = ptlinstances_sortedfusion(l, Indexes);
    }
}

/**
   It initializes the code generator.
   For more details, see the definition of the generator class.
   @param vGraph (in) graph for which the sequential code is asked.
   @param genMode (in) boolean at pk_true for separated generation code.
   @param vprefix (in) string used to prefix each reference to a data.
   For example: with data-> as vprefix, each reference to a signal x, will be generated as data->x. 
   It is useful for separating compiling to have several instances of the same model in an application (in C language).
*/
_VIRTUAL_ void cGenerator::initialize(ptDCG_sch vGraph, boolean genMode, ptrepstr vprefix)
{
  SEPARATE = genMode;

  AGRAPH = vGraph;
  EM = AGRAPH->getExpressionManager();
  CKM = AGRAPH->getClockManager();
  NM = AGRAPH->getNodeManager();
  nameManager = AGRAPH->getNameManager();
  TM = AGRAPH->getTraceabilityManager();	// TRACEABILITY

  progName = AGRAPH->getName();
  Inputs = AGRAPH->getInputs();
  Outputs = AGRAPH->getOutputs();
  Locals = AGRAPH->getLocalSignals();
  Types = NILLINSTANCES;
  if (AGRAPH->getUserTypesOneLevel(&Types)) 
    Types = ptlinstances_copy(Types);
  Constants = AGRAPH->getConstants();
  Parameters = AGRAPH->getParameters();
  Indexes = AGRAPH->getListIndexes();
  ExternalsProcs = AGRAPH->getListExternalProcedure(&ExternalsFuncs);
  bEndochronousWithInternalClockRoot = AGRAPH->isEndochronousWithInternalClockRoot(&vhComplete);

  // -----------------temporary attributes
  AttributeNameGencode = getMark();
  NOTGREYBOXBLACKBOXREFERENCE = getMark();
  travelIndexAttribute = getMark();
  IteratorAttribute = getMark();
  OPTIMIZED_INDEX = getMark();
  INIT_DONE = getMark();
  affineSampleAttribute = getMark();
  alreadyVisisted = getMark();
  vcdmark = getMark();
  // -----------------
  lAffineSample = NILLINODE;
  tick = CKM->getTickClock();
  setPrefix(vprefix);
  functionDeclaration = pk_false;
  someProblem = pk_false;
  sameLine = pk_false;
  TextForm->initialize(ident2repstr(progName));
  AGRAPH->getDCGSchInfos(&nbSets, &SubGraphTable, &lNodesHinit, &lClocksNotToBeInitiliazed);
  PragmaGencodeLanguage = getPredefinedPragma(PRAGMA_CODEGEN_C);
  PragmaImportLanguage = getPredefinedPragma(PRAGMA_CODEGEN_C_IMPORT);
  PragmaDeclareLanguage = getPredefinedPragma(PRAGMA_CODEGEN_C_DECLARE);
  loadOperatorsDef(sysGetFileCodeGenPredef("C"));
  lIncrDelay = NILLINSTANCES;
  stateVariablesAttribute();
  unexpandedModelInitialize();
  bCallEntrypointStepInitialize = pk_false;

  nbMaxIndex = 0;
  InternalIndexLst = NILLINSTANCES;
  indexStack = NILLINSTANCES;
  INTERNAL_BLOCK = pk_false;
  contextOp = NULLOP;

  Zero = EM->mkIntegerConstant(0);
  One = EM->mkIntegerConstant(1);

  ARRAY_CONTEXT = pk_false;
  undefinedType = TypeTypeGetUndefinedType();
  EM->setModeTypeChecking(pk_false);
  extensionInhibit = pk_false;
  externalObjects = pk_false;
  okForElse = pk_false;
  modeelsif = pk_false;
  visComment = pk_true;
}

/**
   It finalyzes the code generation. The graph is cleaned (internal attributes/marks are removed).
   @return "true" if the generating is complete, "false" otherwise (structure not implemented,...).
*/
_VIRTUAL_ boolean cGenerator::finalize()
{
  if (someProblem)
    pK_messenger->error(erropnotimpl);
  EM->resetMarkAll(AttributeNameGencode);
  EM->resetMarkAll(NOTGREYBOXBLACKBOXREFERENCE);
  EM->resetMarkAll(travelIndexAttribute);
  EM->resetMarkAll(IteratorAttribute);
  EM->resetMarkAll(INIT_DONE);
  EM->resetMarkAll(OPTIMIZED_INDEX);
  EM->resetMarkAll(affineSampleAttribute);
  EM->resetMarkAll(vcdmark);

  unexpandedModelFinalize();
  return !someProblem;
}

/**
   Separated compiling.\n
   Code generation of the file foo_body_separate.c for separated compiling.

   It calls 
   - printbodyfilebegin()
   - genInitialisations()
   - genSubGraphs()
   - genBlackBoxCode()
   - closefile()
*/
_VIRTUAL_ void cGenerator::genBodyCodeSeparate()
{
  TextForm->printbodyfilebegin(progName, pk_true);
  TextForm->printbodyfileIncludes(progName, pk_true);
  genInitialisations();
  genSubGraphs();
  genEntryPointStepInitialize();
  genBlackBoxCode();
  TextForm->closefile();
}

/**
   Separated compiling.\n
   It calls 
   - genBodyCodeSeparate() for the code generation of the file FOO_body_separate.c
   - genBodyHeader() for the code generation of the file  FOO_body_separate.h 
*/
_VIRTUAL_ void cGenerator::genBodySeparate()
{
  genBodyHeader();		/* file <P>_separate_body.h */
  genBodyCodeSeparate();	/* file <P>_body_separate.c  */
}

/**
   Separated compiling.\n
   Entry point of the code generation, with or without clusters, in separated compiling mode.
   \see{generateSeparate}
*/
_VIRTUAL_ void cGenerator::genCodeSeparate()
{
  genExternalDeclarationsFile();	/* foo_externals.h  file */
  dclTypes();			/* except the external ones */
  genBodySeparate();
}

/**
   Separated compiling.\n
   Used for separate compiling with clusters. 
   It generates for each signal X of the list (list)
   the assigment X = _X_ for inputs or *_X_ = X for outputs. The name _X_ is a name of a formal parameter of a cluster.
   @param list (in) list of i/o signals.
   @param bInputs (in) must be "true" when the signals of the list are input signals.
   @param sgClock (in) is the clock of the cluster. 

   The generation begins/terminates with an initialisation of the clock context to sgClock.
*/
_VIRTUAL_ void cGenerator::genAssignIOSeparate(ptexpcal list, ptinsth sgClock, boolean bInputs)
{
  ptexpcal exp;
  ptinsth h;

  if (list != NILEXPCAL)
    {
      initContextGenerator(sgClock);
      do
	{
	  exp = EM->getNaryArgs(&list);
	  h = EM->getClock(exp);
	  if (bInputs)
	    genReadEnvSeparate(exp, h);
	  else
	    genWriteEnvSeparate(exp, h);
	}
      while (list != NILEXPCAL);
      genBlockOpeningOf(sgClock);
    }
}

/**
   Misc. For internal tests. Not documented.\n
   Generation of code for node annotated with PRAGMA_UNEXPANDED pragma.
*/
_VIRTUAL_ void cGenerator::unexpandedModelInitialize()
{
  ptlINode lNodes;
  ptINode node;
  ptident name;
  ntrees statement;
  ptDCG aGraph;
  ptlpragmas prags;
  int i;

  if (NM->getNodeCalls(&lNodes))
    {
      i = 0;
      while (ptlINode_get_elem(&lNodes, &node))
	{
	  aGraph = NM->get_nodeCall_graph(node);
	  prags = aGraph->getPragmas();
	  if ( /* existPredefinedPragma(prags, PRAGMA_SEPARATE) ||*/ existPredefinedPragma(prags, PRAGMA_UNEXPANDED))
	    {
	      i++;
	      name = aGraph->getName();
	      statement = repstr2PragmaStatement(ident2repstr(name));
	      statement = addEnonceValue(statement, integer2repstr(i));
	      NM->addPragma(node, newPredefinedPragma(PRAGMA_BLACK_BOX, NILOBJPRAGMA, statement));
	    }
	}
    }
}

/**
   Misc. For internal tests. Not documented.\n
   Generation of code with PRAGMA_UNEXPANDED pragmas.
   It removes the informations previously set in unexpandedModelInitialize() for the 
   nodes annotated with PRAGMA_UNEXPANDED pragma.
*/
_VIRTUAL_ void cGenerator::unexpandedModelFinalize()
{
  ptlINode lNodes;
  ptINode node;
  ptDCG aGraph;
  ptlpragmas prags;

  // for TESTS (L. Besnard) 
  if (NM->getNodeCalls(&lNodes))
    while (ptlINode_get_elem(&lNodes, &node))
      {
	aGraph = NM->get_nodeCall_graph(node);
	prags = aGraph->getPragmas();
	if ( /* existPredefinedPragma(prags, PRAGMA_SEPARATE) ||*/ existPredefinedPragma(prags, PRAGMA_UNEXPANDED))
	  NM->removePragma(node, getPredefinedPragma(PRAGMA_BLACK_BOX));
      }
}

/**
   It generates an unparametrized comment given by a number (num).
*/
_VIRTUAL_ void cGenerator::genComment(int num)
{
  if (visComment)
    TextForm->printComment(pK_messenger->recupmessStr(num));
}
