/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
// --------------------------------------
#define BDEBUG pk_false
#define DEBUG(s)
// --------------------------------------
#define tab "    "
#define genBeginComment() ofile << ";" ; inComment=true;
#define genBeginAssign() ofile << "(= "
#define genSpace() ofile << " "
#define genAssert() genString("assert ");
// ite : if then else
#define genITEClock(vclk)  ofile << "(ite "; genClockId(vclk);
// The string Mapping of  is used in an external script. Don't modify this.
#define mapvarmess string("; ---- Mapping of variables ")

#include <string.h>
#include <algorithm>
#include <math.h> 
#include <sstream>
#define EOS std::string::npos


void SMTPrinter::setErrorInTranslation()
{
  if (! inComment) _ERROR_ = pk_true;
}

ptident SMTPrinter::error_identifier()
{
  return string2ident("_ERROR_IN_TRANSLATION_");
}

void SMTPrinter::NotYetImplemented(string message)
{
  genString("Not_yet_implemented : ");
  genString(message);
  setErrorInTranslation();
}

/**
   Open the ofile stream on the file FOO_<vsuffix>.smt or (FOO.smt is the vsuffix is empty).
   FOO is the name of the current graph.
 */
boolean SMTPrinter::openFile(const char *vsuffix)
{
  ptrepstr filename;
  char *vfile;
  ptrepstr idfile;

  if (strlen(vsuffix) > 0)
    idfile = ident2repstr(makeIdent_ident_string(idGraph, vsuffix));
  else
    idfile = ident2repstr(idGraph);
  if (!sys_suffixedFilePathInUserOutputDir(idfile, "smt", &filename))
    {
      setErrorInTranslation();
      return pk_false;
    }
  vfile = repstr2string(filename);
  ofile.open(vfile);
  pK_messenger->globalMessageStrStr(messGeneratorFile, string2repstr("SMT "), filename);
  return pk_true;
}

/**
   Close the ofile stream.
 */
void SMTPrinter::closeFile()
{
  ofile.close();
}

/**
   Open a SMT block.
 */
void SMTPrinter::genOpenBlock()
{
  ofile << "(";
}

/**
   Close  a SMT block.
*/
void SMTPrinter::genCloseBlock()
{
  ofile << ")";
}

/**
   Generates a SMT comment (vcomment prefixed by ';'.)
*/
void SMTPrinter::genComment(string vcomment)
{
  genBeginComment();
  genString(vcomment);
  genNewLine();
}

/**
   Generates a SMT comment in debug mode.
*/
void SMTPrinter::genCommentDebug(string vcomment)
{
  if (BDEBUG)
    genComment(vcomment);
}

void SMTPrinter::genNewLine()
{
  ofile << endl;
  if (indent) ofile << tab;
  if (inComment) inComment = false;
}

void SMTPrinter::genFormat()
{
  switch (vformat)
    {
    case DCG_BASIC:
      genString("BASIC");
      break;
    case DCG_POLY:
      genString("POLY");
      break;
    case DCG_ENDO:
      genString("ENDO");
      break;
    case DCG_BOOL:
      genString("BOOL");
      break;
    case DCG_FLAT:
      genString("FLAT");
      break;
    case DCG_SCH:
      genString("SCH");
      break;
    default:
      genString("UNKNOWN");
    }
}

/**
   @return the name of a clock (vh).
 */
ptident SMTPrinter::getClockId(ptinsth vh)
{
  ptexpcal vsig;
  ptident id;
  ptrepstr repstrResult;

  id = CKM->getName(vh);
  if (!CKM->isClockRepresentedBySignal(vh, &vsig) && !CKM->isClockUnifiedWithSignal(vh, &vsig))
    return id;
  if (EM->IsEvent(vsig)) return EM->getName(vsig); // id;
  makeRepstr_string_ident("^", id, &repstrResult);
  return repstr2ident(repstrResult);
}

/**
   @return the name of an identifier (e).
   e is assumed to be a signal.
*/
ptident SMTPrinter::getSigId(ptexpcal e)
{
  if (EM->getOperatorClass(e) == variabl) return EM->getName(e);
  setErrorInTranslation();
  return error_identifier();
}

ptident SMTPrinter::getSigIdMemory(ptexpcal e, ptINode node)
{
  if (EM->getOperatorClass(e) == variabl)
    return EM->getName(e);
  // constant ; e = (cste when H),
  // example:  zx := (false when ^ZX) $1 init true
  return makeIdent_ident_int(idMemCste, NM->get_num(node));
}

/**
   @return the name of an expression (e).
   e event it returns the true identifier
   e a signal it returns the name of the signal and signals the e must be generated (mustBeGenerated()).
   otherwise the "error" identifier is returned (error_identifier()).
 */
ptident SMTPrinter::getIdentCondition(ptexpcal e)
{
  if (EM->IsEvent(e)) return id_true;
  if (EM->IsSignal(e))
    {
      mustBeGenerated(e);
      lsig_CurrentDCG = ptlinstances_insert( lsig_CurrentDCG, e );
      return getSigId(e);
    }
  return error_identifier();
}


void SMTPrinter::genInteger(long int val)
{
  if (val < 0) 
    {
      genOpenBlock();
      genOperator(opmoin);
      genInteger(-val);
      genCloseBlock();
    }
  else
    ofile << val << ' ';
}

/**
   Generates the vstr string in the opened file.
 */
void SMTPrinter::genChar(char vstr)
{
  ofile << vstr;
}

/**
   Generates the vstr string in the opened file.
 */
void SMTPrinter::genString(string vstr)
{
  ofile << vstr.c_str();
}

/**
   Generates the id identifier in the opened file.
 */
void SMTPrinter::genIdent(ptident id)
{
  genString(ident2string(id));
}

/**
   Generates the identifier assigned to the vsig expression in the opened file.
 */
void SMTPrinter::genIdentifier(ptexpcal vsig)
{
  genIdent(getSigId(vsig));
}

/**
   Generated the SMT operator of the vop SIGNAL operator.
 */
void SMTPrinter::genOperator(noperators vop)
{
  // if (vop != opvvarin)

  if (vop == opfalse)
    genString("false");
  else if (vop == optrue)
    genString("true");
  else if (vop == opplus)
    genString("+");
  else if (vop == opdivi)
    genString("/");
  else if (vop == opmod)
    genString("mod");
  else if (vop == opmult)
    genString("*");
  else if (vop == opmoin)
    genString("-");
  else if (vop == oppuis)
    genString("pow");
  else if (vop == opdiff)
    genString("/=");
  else if (vop == opegal)
    genString("=");
  else if (vop == opegeg)
    genString("=");
  else if (vop == opiegal)
    genString("<=");
  else if (vop == opinfe)
    genString("<");
  else if (vop == opsegal)
    genString(">=");
  else if (vop == opsupe)
    genString(">");
  else if (vop == opiieg)
    genString("<=");
  else if (vop == opet)
    genString("and");
  else if (vop == opou)
    genString("or");
  else if (vop == opouex)
    genString("xor");
  else if (vop == opnega)
    genString("-");
  else if (vop == opnon)
    genString("not");
  else if (vop == opleft_tt)
    genString("left_tt");
  else
    NotYetImplemented("OPERATOR");
  genSpace();
}

/**
   @return true if a clock (vh) is an element of the current graph.
*/
boolean SMTPrinter::isInNewSystem(ptinsth vh)
{
  if (ptlinsth_isin_elem(Clocks_CurrentDCG, vh)) return pk_true;
  return ptlinsth_isin_elem(lAddedClocks, vh);
}

/**
   It specifies that a clock (vh) must be generated.
 */
void SMTPrinter::mustBeGenerated(ptinsth vh)
{
  if (!isInNewSystem(vh))
    {
      ClocksToBeGenerated = ptlinsth_sorted_insert_elem(ClocksToBeGenerated, vh);
      lAddedClocks = ptlinsth_sorted_insert_elem(lAddedClocks, vh);
    }
}

/**
   Generates a 'when expression' (exp).
   vexp is assumed to be such an expression.
   The contribution clock is marked to be generated.
*/
void SMTPrinter::genWhen(ptexpcal vexp, boolean inDefault)
{
 ptinsth hc = EM->getContributionClock(vexp);
 mustBeGenerated(hc);
 if (inDefault)
    {
      genITEClock(hc);
      genSpace();
    }
  genExpression(EM->getWhenOrCellArg(vexp));
  genSpace();
}

/** @return the "fixed" representation of a real (exp).
    If exp is not represented using mantissa and exponent, exp is returned.
    Otherwise the conversion to the fixed representation is returned using 30 as precision.
    Example : for 1.0e-16, the returned string is 0.000000000000000100000000000000

    NB: SMT-LIBv2 does not accept the (mantissa L exponent) representation (L in {e,d,E,D}).
*/
string SMTPrinter::getDecimalRepresentation(string exp)
{
   size_t index; 
   index= exp.find_first_of("eEdD", 0);
   if (index == EOS) return exp;
   string mantissa = exp.substr(0, index);
   // long double m = stold(mantissa, NULL);
   long double m = atof(mantissa.c_str());

   string exponent = exp.substr(index+1);

   // long double e = std::stod ( exponent, NULL);
   long double e = atof(exponent.c_str());

   long double result = m * pow(10, e); 

   ostringstream vfixedRepr;
   vfixedRepr.precision(30);
   vfixedRepr << std::fixed << result ;
   return vfixedRepr.str();
}

/**
   Generates a constant expression (vexp).
   vexp is assumed to be such an expression.
*/
void SMTPrinter::genConstantExpression(ptexpcal vexp)
{
  noperators vop;
  size_t idx=0;

  vop = EM->getOperator(vexp);
  if ((vop == optrue) || (vop == opfalse))
    genOperator(vop);
  else
    {
      ptrepstr v =  EM->getConstant(vexp);
      string vstr = repstr2string(v);
      if ((vop == opcstreel) || (vop == opcstdp) )
	vstr = getDecimalRepresentation(vstr);
      genString(vstr);
    }
}

/**
   Generates an unary expression (vexp).
   vexp is assumed to be such an expression.
*/
void SMTPrinter::genUnaryExpression(ptexpcal vexp)
{
  noperators vop;

  vop = EM->getOperator(vexp);
  genOpenBlock();
  genOperator(vop);
  genExpression(EM->getUnaryArg(vexp));
  genCloseBlock();
}

/**
   Generated a left_tt expression: arg1 is defined by arg2 assumed to be a left_tt expression
*/
void SMTPrinter::genLeft_tt(ptexpcal arg1, ptexpcal arg2)
{
  ptinsth vhexp1;

  if (EM->getSynchronousExprHorl(arg1, &vhexp1))
    {
      genString("ite");
      genSpace();
      genClockId(vhexp1);
      genSpace();
      genExpression(arg1);
      genSpace();
      genString("false");
    }
  else
    {
      genString("iteERROR");
      genSpace();
      genExpression(arg2);
      genSpace();
      genExpression(arg1);
      genSpace();
      genString("false");
    }
}

/**
  Generates a binary expression (vexp).
  vexp is assumed to be such an expression.
*/
void SMTPrinter::genBinaryExpression(ptexpcal vexp)
{
  ptexpcal arg1, arg2;
  noperators vop;
  ptinsth vh;
  boolean b;
  vop = EM->getOperator(vexp);
  arg1 = EM->getBinaryArgstAt(vexp, 1);
  arg2 = EM->getBinaryArgstAt(vexp, 2);
  genOpenBlock();
  if (vop == opleft_tt)
    genLeft_tt(arg1, arg2);
  else
    {
      b =  (vop == opdiff);
      if (b) { genString(" not ( "); vop = opegal;}
      genOperator(vop);
      genExpression(arg1);
      genSpace();
      genExpression(arg2);
      if (b) genString(" )");
    }
  genCloseBlock();
}

/**
    Generates a default expression or a when expression
    vexp is assumed to be such an expression.
    @return the number of sub expressions has been generated (number of '(')
*/
int SMTPrinter::genDefaultExpression(ptexpcal vexp)
{
  int n;

  if (EM->getOperator(vexp) == opmele)
    {
      n = genDefaultExpression(EM->getDefaultArgstAt(vexp, 1));
      genWhen(EM->getDefaultArgstAt(vexp, 2), pk_true);
      n++;
    }
  else
    {
      genWhen(vexp, pk_true);
      n = 1;
    }
  return n;
}

/**
   Generates a default expression (e), and finalizes the generation of the expression.
   e is assumed to be such an expression.
*/
void SMTPrinter::genDefault(ptexpcal e)
{
  int nbblocks = genDefaultExpression(EM->getDefaultArgstAt(e, 1));
  // The last one is generated without the clock condition.
  genExpression(EM->getWhenOrCellArg(EM->getDefaultArgstAt(e, 2)));
  for (int i = 1; i <= nbblocks; i++)
    genCloseBlock();
}

/**
   Generates a ternary expression (vexp).
   vexp is assumed to be such an expression.
*/
void SMTPrinter::genTernaryExpression(ptexpcal vexp)
{
  genString("(genTernaryExpression xxxx)");
  /*
     ptexpcal arg1, arg2, arg3;
     arg1 = EM->getTernaryArgstAt(vexp, 1);
     arg2 = EM->getTernaryArgstAt(vexp, 2);
     arg3 = EM->getTernaryArgstAt(vexp, 3);

     genExpression(arg1);
     genSpace();
     genExpression(arg2);
     genSpace();
     genExpression(arg3);
   */
}

/**
   Generates a function call expression (vexp).
    vexp is assumed to be such an expression.
*/
void SMTPrinter::genFunctionCallExpression(ptexpcal vexp)
{
  ptident id;
  ptexpcal vinput, vexpcal1, vexpcal2;
  string vstr;
  // vexpcal2 : parameters; vexpcal1: inputs
  EM->getFunctionCall(vexp, &id, &vexpcal2, &vexpcal1);	
  genString("("); 
  if (existPredefinedPragma(EM->getPragmas(vexp), PRAGMA_CONVERSION))
    {
      noperators ope = identType2SignalType(ident2repstr(id));
      switch ((int)ope) 
	{
	case oplogical:
	  vstr= "to_bool";
	  break;
	case opshort:
	case opinteger:
	case oplong:
	  vstr="to_int";
	  break;
	  
	case opreal:
	case opdpreal:
	  vstr="to_real";
	  break;

	case opevent:
	case opchar:
	case opstring:
	case opcomplex:
	case opdcomplex:
	   vstr= "UNDEFINED_CONVERSION";
	  break;
	default:
	   vstr= "UNDEFINED_CONVERSION";
	}
      genString(vstr);
    }
  else 
    {
      genIdent(id);
    }
  genSpace();
  genListExpression(vexpcal1, " ");
  genString(")");  
}

/**
   Generates a list expression (vexp) with sep as element separator.
   vexp is assumed to be such an expression.
*/
void SMTPrinter::genListExpression(ptexpcal vexp, string sep)
{
  ptexpcal elem;
  while (EM->getNaryArgs(&vexp, &elem))
    {
      genExpression(elem);
      if (vexp != NILEXPCAL) genString(sep);
    }
}
/**
   Generate a list of types of the expressions of vexp with sep as list separator.
   It is used to declare external function.
 */
void SMTPrinter::genTypesListExpression(ptexpcal vexp, string sep)
{
   ptexpcal elem;
  while (EM->getNaryArgs(&vexp, &elem))
    {
      genType(EM->getUnamedType(elem));
      if (vexp != NILEXPCAL) genString(sep);
    }
}


/**
   Generates a list expression (vexp).
   vexp is assumed to be such an expression.
*/
void SMTPrinter::genListExpression(ptexpcal vexp)
{
  genString("(");
  genListExpression(vexp, ",");
  genString(")");
}

/**
   
 */
void SMTPrinter::mustBeGenerated(ptexpcal vexp)
{
  if (modeCert && EM->IsSignal(vexp)) 
    {
      EM->addMark(vexp, vMarkSeen);
    }
}

boolean SMTPrinter::genValue (ptexpcal exp)
{
  TypeType TT;
  long int val;
  if (EM->IsIndex (exp)) return pk_false;
  if (!EM->IsAssigned (exp)) return pk_false;
  TT = EM->getUnamedType (exp);
  if (!TypeTestPredef (TT)) return pk_false;
  val = EM->getValue (exp);
  if (TypeTestLogical (TT))
    {
      genOperator( (val != 0) ? optrue : opfalse);
      return pk_true;
    }

  if (TypeTestIntNumber (TT))
    {
      genInteger(val);
      return pk_true;
    }
  return pk_false;
}

/**
   It generates a constant signal (vexp).
   \see genEqu()
*/
void SMTPrinter::genVariable(ptexpcal vexp)
{
  genIdentifier(vexp);
  mustBeGenerated(vexp);
}


void SMTPrinter::genExpression(ptexpcal vexp)
{
  ptexpcal nexp;

  if (vexp != NILEXPCAL)
    {
      switch ((int)(EM->getOperatorClass(vexp)))
	{
	case variabl:
	  genVariable(vexp);
	  break;

	  // reference to a constant expression.
	case constant:
	  genConstantExpression(vexp);
	  break;

	  // Unary expression : not , -
	case oper1:
	  genUnaryExpression(vexp);
	  break;

	  // Binary Expression : +,*,...
	case oper2:
	  if (EM->satisfiesList(vexp))
	    genListExpression(vexp);
	  else
	    genBinaryExpression(vexp);
	  break;

	  // Delay/Window (perhaps filtered before)
	case oper3:
	  genTernaryExpression(vexp);
	  break;

	  // function call id( i1,...., in)
	case call:
	  genFunctionCallExpression(vexp);
	  break;

	  // E1 when clk ( E1 cell clk has been filtered before)
	case opereh:
	  genWhen(vexp, pk_false);
	  break;

	case oper2tp:		// default operator
	  genDefault(vexp);
	  break;

	  // no occurrence
	case oper1tp:
	case oper3tp:
	case opertyp:
	  break;

	default:
	  pK_messenger->Caseerror(currentLine, currentFile);
	}
    }
}

/**
   Generates the identifier of a clock (vclk).
   The clock is marked to be generated.
 */
void SMTPrinter::genClockId(ptinsth vclk)
{
  genIdent(getClockId(vclk));
  mustBeGenerated(vclk);
}

/**
   Generated a SMT if for a clock (vclk)
   For example: (=> H_5_H 
 */
void SMTPrinter::genIfClock(ptinsth vclk)
{
  genOpenBlock();
  ofile << "=> ";
  genClockId(vclk);
}

/*
  Generates the SMT definition of "if vclk then id=e": (=> vclk (= id e)).
*/
void SMTPrinter::genDefinition(ptinsth vclk, ptident id, ptexpcal e)
{
  boolean b;

  b = indent;
  indent = true;
  genIfClock(vclk);
  genOpenBlock();
  genString("= ");
  genString(ident2string(id));
  genSpace();
  genExpression(e);
  genCloseBlock();
  indent = b;
}

/**
   Generates the assignement of a clock (vh) is SMT: "(= ID" where ID is the name of the operand clock (vh).
   The identifier is marked as generated: it is stored in the global vector generatedIdentifiers.
 */
void SMTPrinter::genAssignClockId(ptinsth vh)
{
  ptident id;

  genBeginAssign();
  id = getClockId(vh);
  genIdent(id);
  genSpace();
  generatedIdentifiers.push_back(id);
}

/**
   Generates the definition of an extraction clock (vh).
   vh = hcond[:idcond] (bsub is true) or vh = hcond[/:idcond] (bsub is false).
 */
void SMTPrinter::genSub(ptinsth vh, ptinsth hcond, ptident idcond, boolean bsub)
{
  genAssignClockId(vh);
  genOpenBlock();
  genString("and");
  genSpace();
  genClockId(hcond);
  genSpace();
  if (!bsub)
    genString("(not ");
  genIdent(idcond);
  genString("))");
  if (!bsub)
    genString(")");
  genNewLine();
  genClockDef(hcond);
}

/**
   Generates the definition of a clock formula (vh), and recursively its operands.
 */
boolean SMTPrinter::genFormula(ptinsth vh)
{
  ptinsth vharg1, vharg2, hfather;

  if (CKM->satisfiesMarked(vh, vMarkSeen))
    return pk_false;

  CKM->addMark(vh, vMarkSeen);
  switch ((int)(CKM->getOperator(vh)))
    {
    case hcomp:
      CKM->getUnaryArg(vh, &vharg1);
      genAssignClockId(vh);

      genOpenBlock();
      genString("and");
      genSpace();
      hfather = CKM->getFather(vh);
      genClockId(hfather);
      genSpace();
      genString("(not ");
      genClockId(vharg1);
      genString(")))");
      genNewLine();
      genClockDef(hfather);
      break;

    case hnot:
      genComment(" hnot detected !!!!");
      break;

    case sub:
      if (isSelectedClock(vh))
	{
	  ptinsth hcond = CKM->getClockSub(vh);
	  ptexpcal vcond = CKM->getConditionSub(vh);
	  ptident idcond = getIdentCondition(vcond);
	  genSub(vh, hcond, idcond, pk_true);
	}
      break;

    case nosub:
      if (isSelectedClock(vh) && isEventClockedFormat)
	{
	  genSub(vh, CKM->getClockSub(vh), getIdentCondition(CKM->getConditionSub(vh)), pk_false);
	}
      break;

    case infer:
      genAssignClockId(vh);
      genOpenBlock();
      genString("and");
      genSpace();

      if (CKM->getMinusArgs(vh, &vharg1, &vharg2))
	{
	  vharg1 = CKM->getRepresentativeFormula(vharg1);
	  vharg2 = CKM->getRepresentativeFormula(vharg2);
	  genClockId(vharg2);
	  genSpace();
	  genString("(not ");
	  genClockId(vharg1);
	  genString(")))");
	}
      else
	{
	  CKM->getInfArgs(vh, &vharg1, &vharg2);
	  vharg1 = CKM->getRepresentativeFormula(vharg1);
	  vharg2 = CKM->getRepresentativeFormula(vharg2);
	  genClockId(vharg1);
	  genSpace();
	  genClockId(vharg2);
	  genString("))");
	}
      genNewLine();
      genClockDef(vharg1);
      genClockDef(vharg2);
      break;

    case super:
      CKM->getSupArgs(vh, &vharg1, &vharg2);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      vharg2 = CKM->getRepresentativeFormula(vharg2);

      genAssignClockId(vh);
      genOpenBlock();
      genString("or");
      genSpace();
      genClockId(vharg1);
      genSpace();
      genClockId(vharg2);
      genString("))");
      genNewLine();
      genClockDef(vharg1);
      genClockDef(vharg2);

      break;

    default:;
    }
  return pk_true;
}

/**
   Generates the definition of a clock (ie all the formula of the class).
 */
void SMTPrinter::genClockDef(ptinsth clk)
{
  ptinsth clk1;

  clk = CKM->getRepresentativeFormula(clk);
  if (genFormula(clk))
    {
      clk1 = CKM->getNextFormula(clk, pk_true);
      while (clk1 != clk)
	{
	  genFormula(clk1);
	  addConstraint(clk, clk1);
	  clk1 = CKM->getNextFormula(clk1, pk_true);
	}

      clk1 = clk;
      while (CKM->getNextAssertion(clk1, &clk1))
	{
	  addAssertion(clk1, clk);
	}
    }
}

/** 
    Generates the begining of a definition of a SMT function with id as name.
    It generates:  (define-fun id () Bool (and 
 */
void SMTPrinter::genSMTFunctionInit(ptident id)
{
  // (define-fun id () Bool (and 
  genNewLine();
  genOpenBlock();
  genString("define-fun");
  genSpace();
  genIdent(id);
  genSpace();
  genString("()");
  genType(BooleanType);
  genNewLine();
  genOpenBlock();
  genString("and");
  indent = true;
  genNewLine();
}

/** 
    Generates the ending of a SMT function opened by genSMTFunctionInit().
 */
void SMTPrinter::genSMTFunctionFinal()
{
  indent = false;
  genString(mapvarmess);
  genFormat();
  genNewLine();

  genCloseBlock();		// closing "( and"
  genNewLine();
  genSpace();
  genCloseBlock();		// closing define-fun
  genNewLine();
}

/**
   Generates the SMT code for a SIGNAL type (TT).
 */
void SMTPrinter::genType(TypeType TT)
{
  // TODO : to be completed
  genSpace();
  if (TT == BooleanType)
    genString("Bool");
  else if (TT == PureType)
    genString("Bool");
  else if (TT == ShortType)
    genString("Int");
  else if (TT == IntegerType)
    genString("Int");
  else if (TT == LongType)
    genString("Int");
  else if (TT == RealType)
      genString("Real");
  else if (TT == DRealType)
    genString("Real");
  else
    NotYetImplemented("Type");
}


void SMTPrinter::genDeclFuncBegin(string id)
{
  genSpace();
  genOpenBlock();
  genString("declare-fun");
  genSpace();
  genString(id);
  genSpace();
}

void SMTPrinter::genDeclFuncEnd(TypeType TT)
{
  genType(TT);
  genCloseBlock();
  genNewLine();
}

/**
   Generate the declaration of a variable, named id, of SIGNAL type TT.
   Example: It generates "(declare-fun id () Bool)"  for TT the boolean type.
 */
void SMTPrinter::genIdentDecl(string id, TypeType TT)
{
  genDeclFuncBegin(id);
  genString("()");
  genDeclFuncEnd(TT);
}

/**
   Generate the declaration of a clock (vh).
   The clock is added to the clocks to be generated.
 */
void SMTPrinter::genClockIdentDecl(ptinsth vh)
{
  if (BDEBUG) printSignals(vh); // print the synchronous signals with vh.
  genIdentDecl(ident2string(getClockId(vh)), BooleanType);

  if (!ptlinsth_isin_elem(Clocks_CurrentDCG, vh))
    lAddedClocks = ptlinsth_sorted_insert_elem(lAddedClocks, vh);
}

/**
   Generate the declaration of formula (vh) and recursively its operands.
   @return "true" if it is done, "false" when the formula has been before generated.
 */
boolean SMTPrinter::declareFormula(ptinsth vh)
{
  ptinsth vharg1, vharg2;

  if (CKM->satisfiesMarked(vh, vMarkDeclared))
    return false;

  CKM->addMark(vh, vMarkDeclared);
  // Filtering the Tick clock/signal for avoiding a double declaration.
  if (! modeCert && ! isEventClockedFormat && CKM->satisfiesTickClock(vh) ) return pk_false;

  genClockIdentDecl(vh);
  switch ((int)(CKM->getOperator(vh)))
    {
    case sub:
      CKM->getComplInFather(vh, &vharg1);
      if (isSelectedClock(vharg1) && isEventClockedFormat)
	genClockDecl(vharg1);	// nosub
      genClockDecl(CKM->getClockSub(vh));
      break;

    case infer:
      if (!CKM->getMinusArgs(vh, &vharg1, &vharg2))
	CKM->getInfArgs(vh, &vharg1, &vharg2);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      vharg2 = CKM->getRepresentativeFormula(vharg2);
      genClockDecl(vharg1);
      genClockDecl(vharg2);
      break;

    case super:
      CKM->getSupArgs(vh, &vharg1, &vharg2);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      vharg2 = CKM->getRepresentativeFormula(vharg2);
      genClockDecl(vharg1);
      genClockDecl(vharg2);
      break;
    case hcomp:
      vharg1 = CKM->getFather(vh);
      vharg1 = CKM->getRepresentativeFormula(vharg1);
      genClockDef(vharg1);
      break;

    default:;
    }
  return true;
}

/**
   Generates as comment the signals with vclk as clock.
 */
void SMTPrinter::printSignals(ptinsth vclk)
{
  ptexpcal vsig;
  ptlinstances lsig;

  lsig = CKM->getSynchronousSignals(vclk);
  if (lsig != NILLINSTANCES)
    {
      genString("; Clock of ");
      while (ptlinstances_get_next(&lsig, &vsig))
	{
	  genIdent(getSigId(vsig));
	  if (lsig != NILLINSTANCES)
	    genString(", ");
	}
      genNewLine();
    }
}

/**
   Generates the declaration of the variables assigned to a clock (ie all the formula of the class, the associated assertions).
 */
void SMTPrinter::genClockDecl(ptinsth clk)
{
  ptinsth clk1;

  clk = CKM->getRepresentativeFormula(clk);
  declareFormula(clk);

  clk1 = CKM->getNextFormula(clk, pk_true);
  while (clk1 != clk)
    {
      declareFormula(clk1);
      clk1 = CKM->getNextFormula(clk1, pk_true);
    }

  clk1 = clk;
  while (CKM->getNextAssertion(clk1, &clk1))
    {
      declareFormula(clk1);
    }
}

/**
   Generates the declarations of a list of clocks (list).
 */
void SMTPrinter::genClockDecls(ptlinsth list)
{
  ptinsth vh;

  while (ptlinsth_get_next(&list, &vh))
    if (forceGeneration || isSelectedClock(vh))
     
	genClockDecl(vh);
}

/**
   Generates the definitions of a list of clocks (list).
 */
void SMTPrinter::genClockDefs(ptlinsth list)
{
  ptinsth vh;

  while (ptlinsth_get_next(&list, &vh))
    if (forceGeneration || isSelectedClock(vh))
      genClockDef(vh);
}

/**
   Generates the constraint vh1 = vh2.
 */
void SMTPrinter::addConstraint(ptinsth vh1, ptinsth vh2)
{
  if (vh1 != vh2)
    {
      genBeginAssign();
      genClockId(vh1);
      genSpace();
      genClockId(vh2);
      genCloseBlock();
      genNewLine();
    }
}

/**
   Generates the constraints on the placed clocks.
 */
void SMTPrinter::genConstraintsOnPlacedClocks()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  genComment(" >>> CLOCK CONSTRAINTS -- PLACED CLOCKS");
  CKM->getConstraintsOnPlacedClocks(&lh1, &lh2);
  while (ptlinsth_get_next(&lh1, &vh1))
    {
      ptlinsth_get_next(&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula(vh1);
      vh2 = CKM->getRepresentativeFormula(vh2);
      if (vh1 != vh2)
	addConstraint(vh1, vh2);
    }
  genComment(" <<< CLOCK CONSTRAINTS -- PLACED CLOCKS");
}

/**
   Generates the assertion vh1 = vh2.
*/
void SMTPrinter::addAssertion(ptinsth vh1, ptinsth vh2)
{
  ofile << " Assertion: not yet implemented ";
  genNewLine();
  setErrorInTranslation();

  /*
     genBeginAssign();
     genClockId(vh1);
     genSpace();
     genClockId(vh2);
     genCloseBlock();
     genNewLine(); */
}

/**
   Generates the assertions on placed clocks.
 */
void SMTPrinter::genAssertionsOnPlacedClocks()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  genComment(" >>> ASSERTIONS ");
  CKM->getAssertionsOnPlacedClocks(&lh1, &lh2);
  while (ptlinsth_get_next(&lh1, &vh1))
    {
      ptlinsth_get_next(&lh2, &vh2);
      vh1 = CKM->getRepresentativeFormula(vh1);
      vh2 = CKM->getRepresentativeFormula(vh2);
      if (vh1 != vh2)
	addAssertion(vh1, vh2);
    }
  genComment(" <<< ASSERTIONS");
}

/**
   Memory management ($, cell).
   @return the identifier Mem.idSig.
 */
ptident SMTPrinter::getMemoryId(ptident idSig, int num)
{
  return makeIdent_ident_ident(idMem, idSig);
}

/**
   Memory management ($, cell).
   @return the identifier id_!_0
 */
ptident SMTPrinter::getMemoryInitId(ptident id)
{
  id = makeIdent_ident_string(id, "!");
  return makeIdent_ident_int(id, 0);
}

/**
   Memory management ($, cell)
   @return the identifier id_!_next
*/
ptident SMTPrinter::getMemoryNextId(ptident id)
{
  id = makeIdent_ident_string(id, "!");
  return makeIdent_ident_string(id, "next");
}

void SMTPrinter::genSigDeclDelay(ptexpcal vsig)
{
  ptINode node;
  ptident id, idInit, idNext;
  ptexpcal vmemo;

  if (PRINTMEMORY)
    {
      node = EM->getNode(vsig);
      if (NM->getMemorizedSignal(node, &vmemo))
	{
	 
	  id = getMemoryId(getSigIdMemory(vmemo, node), -1);
	  idInit = getMemoryInitId(id);
	  idNext = getMemoryNextId(id);
	  TypeType TT = EM->getType(vmemo);
	  genIdentDecl(ident2string(id), TT);	// Mem.x
	  genIdentDecl(ident2string(idInit), TT); // Mem.x0
	  genIdentDecl(ident2string(idNext), TT); // Mem.x_next
	}
    }
}

void SMTPrinter::genSigDecl(ptexpcal vsig)
{
  // ptinsth h;  if (! EM->isSignalUnifiedWithClock (vsig, &h))
  
    if (! EM->IsLabel(vsig))
    {
      if (forceGeneration || isSelectedType(vsig))
	{
	  genIdentDecl(ident2string(getSigId(vsig)), EM->getType(vsig));
	  genSigDeclDelay(vsig);
	  EM->addMark(vsig, vMarkDeclared);
	}
    }
}

/**
   Generates the declarations of the signals of a list (lsig).
 */
void SMTPrinter::genSigDecls(ptlinstances lsig)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&lsig, &vsig))
      genSigDecl(vsig);
}

/**
   @return "true" if a node is selected for the generation, "false" otherwise.
   It is selected when it is not in the unused nodes of the current graph, or when it is in the selected list (lnodestogenerate).
 */
boolean SMTPrinter::isSelectedNode(ptINode n)
{
  if (ptlINode_isIn(unusedNodes, n))
    return pk_false;
  return ptlINode_isIn(lnodestogenerate, n);
}

/**
   @return "true" if a clock is selected for the generation(ie when the global member forceGeneration is true, or when the clock is not in the unused clocks of the current graph), "false" otherwise.
   A clock is selected when the global member forceGeneration is true, or when the clock is not in the unused clocks of the current graph
*/
boolean SMTPrinter::isSelectedClock(ptinsth vh)
{
  if (forceGeneration)
    return pk_true;
  return !ptlinsth_isin_elem(unusedClocks, vh);
}

/**
   @return "true" if the type of visg is selected in the current printing, "false" otherwise.
   Note that it exists a global member restrictedToBooleans to restrict to the boolean type.
 */
boolean SMTPrinter::isSelectedType(ptexpcal vsig)
{
  /* TypeType TT = EM->getType(vsig);
     return TypeTestPredef(TT); */

  if ( selectedType == ErrorType) return pk_true;
  TypeType TT = EM->getType(vsig);
  if (! TypeTestPredef(TT))
    return pk_false;
  return (selectedType <= TT );
}

void SMTPrinter::genEqu(ptexpcal leftPart, ptexpcal rightPart)
{
  genSpace();
  genBeginAssign();
  genExpression(leftPart);
  genSpace();
  if (! genValue(leftPart))
    genExpression(rightPart);
  genCloseBlock();
}

/**
   Generates the SMT code of an equation with vh as clock.
 */
void SMTPrinter::genEqu(ptinsth vh, ptINode aNode)
{
  ptexpcal leftPart, rightPart, basicExpr;

  //(=> vh (X EXP))

  if (NM->is_cellNode(aNode))
    genCell(vh, aNode);
  else if (NM->get_rightOperand(aNode, &rightPart))
    {
      leftPart = NM->get_leftOperand(aNode);
      basicExpr = EM->getBasicExpr(rightPart);
      if (EM->IsConstant(leftPart))
	genEqu(leftPart, basicExpr);
      else
	{
	  genIfClock(vh);
	  genEqu(leftPart, basicExpr);
	  genCloseBlock();
	}
      genNewLine();
    }
}

/**
   Generates the initilisation of a memory ($, cell)
   (= M.<idMemory>0 SMT(vinit))
*/
void SMTPrinter::genInitDelay(ptident idMemory, ptexpcal vinit)
{
  // initial value for a delay or cell
  ptident idInitialValue = getMemoryInitId(idMemory);
  genBeginAssign();
  genIdent(idInitialValue);
  genSpace();
  genExpression(vinit);
  genCloseBlock();
  genNewLine();
}

void SMTPrinter::genNextValue(ptinsth vh, ptident idMemory, ptexpcal sigret)
{
  // next 
  ptident idNextValue = getMemoryNextId(idMemory);
  genIfClock(vh);
  genSpace();
  genBeginAssign();
  genIdent(idNextValue);
  genSpace();
  genIdentifier(sigret);
  genCloseBlock();
  genCloseBlock();
  genNewLine();
}

void  SMTPrinter::setAttributes( ptINode node, ptexpcal vmemo, ptexpcal vinit)
{
  // Keeping some infos for nex step
  NM->addAttribute(node, SMT_MEMORY_ID, (attributeValue) vmemo);
  NM->addAttribute(node, SMT_MEMORY_INIT, (attributeValue) vinit);
}

/**
   Generates a delay (aNode) with vh as clock when PRINTMEMORY is true, no effect otherwise.
*/
void SMTPrinter::genDelay(ptinsth vh, ptINode node)
{
  ptident idMemory, idNextValue;
  ptexpcal resret, vmemo, vret, vfen, vinit;

  if (PRINTMEMORY)
    {

      NM->get_delay_arg(node, &resret, &vmemo, &vret, &vfen, &vinit);
      setAttributes(node, vmemo, vinit);
      idMemory = getMemoryId(getSigIdMemory(vmemo, node), -1);
      
      genInitDelay(idMemory, vinit);
      
      // zx := mem(x)
      genIfClock(vh);
      genSpace();
      genBeginAssign();
      genIdentifier(resret);
      genSpace();
      genIdent(idMemory);
      genCloseBlock();
      genCloseBlock();
      genNewLine();
      
      genNextValue(vh, idMemory, resret);
      
    }
}

/**
   Generates a cell (aNode) with vh as clock when PRINTMEMORY is true, no effect otherwise.
 */
void SMTPrinter::genCell(ptinsth vh, ptINode aNode)
{
  ptexpcal leftPart, rightPart, vinit;
  ptinsth h;
  ptident idMemory;

  if (PRINTMEMORY)
    {
      leftPart = NM->get_leftOperand(aNode);
      NM->get_rightOperand(aNode, &rightPart);

      // initial value.
      idMemory = getMemoryId(getSigId(leftPart), -1);
      vinit = EM->getInitialValue(leftPart);
      genInitDelay(idMemory, vinit);

      genIfClock(vh);
      genSpace();
      genBeginAssign();
      genExpression(leftPart);
      genSpace();

      genString("(ite");
      ptinsth hcontr = NM->get_contributionClockEquation(aNode);
      genSpace();
      genClockId(hcontr);
      genSpace();
      rightPart = EM->getWhenOrCellArg(rightPart);
      rightPart = EM->getBasicExpr(rightPart);
      genExpression(rightPart);
      genSpace();
      genIdent(idMemory);
      genString("))");

      genCloseBlock();
      genNewLine();

      genNextValue(vh, idMemory, leftPart);
      setAttributes(aNode, leftPart,vinit);
    }
}

void SMTPrinter::genNotImpl()
{
  ofile << "A Definition is not yet implemented";
  setErrorInTranslation();
  genNewLine();
}


boolean SMTPrinter::genNode(ptINode aNode)
{
  ptinsth vh;

  if (NM->isMarked(aNode, vMarkGenerated)) return pk_false;
  NM->addMark(aNode, vMarkGenerated);

  vh = NM->get_nodeClock(aNode);
  mustBeGenerated(vh);

  if ( ! isSelectedNode(aNode))  return pk_true;
  switch ((int)(NM->get_oper(aNode)))
    {
    case opdefsig:
      genEqu(vh, aNode);
      break;
      
    case opret:
      genDelay(vh, aNode);
      break;
      
    case opldefpart:
      genNotImpl();
      break;
      
    case opdefpart:
      genNotImpl();
      break;
      
    case opsubgraph:
      genNotImpl();
      break;
      
    case opmoti:
      genNotImpl();
      break;
      
    case opbund:
      genNotImpl();
      break;
      
    case opappe:
    case opnodecall:
    case opnil:
      break;
      
    default:
      pK_messenger->Caseerror(currentLine, currentFile);
    }
  return pk_true;
}

/**
   Generates the definition of a signal (vsig) when it is not yet done.
   @return true if the operation succeeds, false otherwise (already before generated).
 */
boolean SMTPrinter::genSigDef(ptexpcal vsig)
{
  ptINode aNode;

  if (EM->isMarked(vsig, vMarkGenerated))
    return pk_false;

  EM->addMark(vsig, vMarkGenerated);
  aNode = EM->getNode(vsig);
  if (forceGeneration || isSelectedNode(aNode))
    if (isSelectedType(vsig))
      {
	genNode(aNode);
	return pk_true;      
      }
  return pk_false;
}

/**
   Generates the definitions of the signals of the operand list (lsig).
 */
void SMTPrinter::genSigDefs(ptlinstances lsig)
{
  ptexpcal vsig;

  while (ptlinstances_get_next(&lsig, &vsig))
    genSigDef(vsig);
}


void SMTPrinter::genConstant(ptexpcal vconst)
{
  ptexpcal exp, leftPart,rightPart;
  ptINode n;

  if (isSelectedType(vconst))
    {
      n = EM->getNode (vconst);
      NM->get_equation_arg (n, &leftPart, &rightPart);
      genOpenBlock();
      genAssert();
      genEqu(leftPart, EM->getBasicExpr(rightPart));
      genCloseBlock();
      genNewLine();
    }
}

/**
   Generates the definitions of the signals of the operand list (lsig).
 */
void SMTPrinter::genConstants()
{
  ptexpcal vsig;
  ptlinstances lsig = Constants;
  while (ptlinstances_get_next(&lsig, &vsig))
    genConstant(vsig);
}


/**
   Generates the declaration of the clocks that has been selected in the Clocks_CurrentDCG global clocks and the 
   added ones in the lAddedClocks global list.
*/
void SMTPrinter::genClockDecls()
{
  indent = pk_false;
  genComment(" ---------------- Clock Declarations");
  genClockDecls(Clocks_CurrentDCG);
  genClockDecls(lAddedClocks);
}

/**
   Generates the definition of the clocks that has been selected in the Clocks_CurrentDCG global clocks.
*/
void SMTPrinter::genClockDefs()
{
  genClockDefs(Clocks_CurrentDCG);
}

/**
   Generates the declarations of the signals selected in the lsig_CurrentDCG global list.
 */
void SMTPrinter::genSigDecls()
{
  if (lsig_CurrentDCG != NILLINSTANCES)
    {
      genComment(" ---------------- Signal Declarations");
      genSigDecls(lsig_CurrentDCG);
    }
}

/**
   Generates the definition of the signals that has been marked to be generated (ie they have been referenced in an expression).
 */
void SMTPrinter::genSigDefs()
{
  ptexpcal vsig;
  ptlinstances lsig;
  boolean oncontinue;

  oncontinue= pk_true;
  while(oncontinue)
    {
      oncontinue= pk_false;
      lsig = lsig_CurrentDCG;
      while (ptlinstances_get_next(&lsig, &vsig))
	if (EM->isMarked(vsig, vMarkSeen))
	  if (genSigDef(vsig)) oncontinue= pk_true;
    }
}

/**
   @return the list of the generated signals.
   Useful in certification mode, to have this information at the next step.
 */
ptlinstances SMTPrinter::getGeneratedSignals()
{
  ptexpcal vsig;
  ptlinstances ls, lsig;

  ls = NILLINSTANCES;
  lsig = lsig_CurrentDCG;
  while (ptlinstances_get_next(&lsig, &vsig))
    if (EM->isMarked(vsig, vMarkGenerated))
      ls = ptlinstances_insert(ls, vsig);

  return ls;
}

void SMTPrinter::genExternalFunctions()
{
  ptlINode lextref;
  ptlinstances lfuncs;
  ptexpcal aCall, params, inputs;
  ptident id;
  ptlPointer ls;

  genComment(" ---------------- External functions");
  ls = NILLPOINTER;
  lextref = CurrentDCG->getListExternalProcedure(&lfuncs);
  while (ptlinstances_get_next(&lfuncs, &aCall))
    { 
      if (!EM->isMarked(aCall, INTRINSIC) && !EM->isMarked(aCall, DONTDECLARE))
	{
	  EM->getFunctionCall(aCall, &id, &params, &inputs);
	  if (! ptlPointer_IsIn(ls, id))
	    {
	      ls = ptlPointer_add(ls, id);
	      genDeclFuncBegin(ident2string(id));
	      genChar('(');
	      genTypesListExpression(inputs, " ");
	      genChar(')');
	      genDeclFuncEnd(EM->getType(aCall));
	    }
	}
    }
  ptlPointer_free(ls);	 
}


/**
   Generates the declarations (signals + clocks).
*/
void SMTPrinter::genDeclarations()
{
  genExternalFunctions();
  genClockDecls();
  genSigDecls();
}


void SMTPrinter::generate()
{
  genSigDefs();
  boolean oncontinue = (ClocksToBeGenerated != NILLINSTH);
  while (oncontinue)
    {
      genClockDefs(ClocksToBeGenerated);
      ClocksToBeGenerated = NILLINSTH;
      genSigDefs();
      oncontinue = (ClocksToBeGenerated != NILLINSTH);
    }
}

void SMTPrinter::genSigDefsExternals()
{
  ptlINode ln;
  ptINode node;
  ptexpcal lparam, lent, lsort;
  ptDCG vg;
  ptident id;

  if (NM->getProcCalls(&ln))
    while (ptlINode_get_elem(&ln, &node))
      {
	  if (! NM->is_readNode(node))
	    if (! NM->is_writeNode(node))
	      {
		NM->get_procCall_arg(node, &id, &lparam, &lent, &lsort); 

		genBeginComment();
		genIdent(id);
		genListExpression(lent);
		genNewLine();
	      }
	}
  if (NM->getNodeCalls(&ln))
    while (ptlINode_get_elem(&ln, &node))
	{
	  NM->get_nodeCall_arg(node, &vg, &lparam, &lent, &lsort);
	  genBeginComment(); 
	  genIdent(vg->getName ());
	  genListExpression(lent);
	  genNewLine();
	}
}

/**
   Generates the definitions in a function such that:

   (define-fun P () Bool
    (and  eq1 eq2 ...eqn)
    )
 */
void SMTPrinter::genDefinitions()
{
  //  genSMTFunctionInit(idGraph);
  genComment(" >>> ---------------- Clocks/Signals Definitions");
  genClockDefs();
  genConstraintsOnPlacedClocks();
  genAssertionsOnPlacedClocks();
  genSigDefs(CurrentDCG->getOutputs());
  genSigDefsExternals();
  generate();
  genComment(" <<< ---------------- Clocks/Signals Definitions");
  SignalsForTheNextLevel = getGeneratedSignals();
  ClocksForTheNextLevel = ptlinsth_duplicate(lAddedClocks);
  //  genSMTFunctionFinal();  indent = true;
}

void SMTPrinter::genDefinitions(ptlINode ln)
{
  ptlINode lnodes;
  ptINode node;
  
  lnodes = ln;
  while (ptlINode_get_elem(&lnodes,&node))
    {
      genNode(node);
    }
}


/** 
    Initializes the printer for a graph (vDCG).
 */
void SMTPrinter::initialize(ptDCG vDCG)
{
  ptlinstances laux;
  // Current graph
  CurrentDCG = vDCG;
  CKM = CurrentDCG->getClockManager();
  EM = CurrentDCG->getExpressionManager();
  NM = CurrentDCG->getNodeManager();
  NAMES = CurrentDCG->getNameManager();
  vformat = CurrentDCG->getFormatLevel();
  isEventClockedFormat = (vformat == DCG_BASIC) || (vformat == DCG_POLY) || (vformat == DCG_ENDO);
  unusedNodes = CurrentDCG->getUnused(&unusedClocks);
  idGraph = CurrentDCG->getLevelName();
  if (EM->getExprList(opident, &laux))
    lsig_CurrentDCG = ptlinstances_copy(laux);
  else
    lsig_CurrentDCG = NILLINSTANCES; 
  // marks
  vMarkSeen = getMark();
  vMarkDeclared = getMark();
  vMarkGenerated = getMark();

  Constants = CurrentDCG->getConstants();
  Parameters = CurrentDCG->getParameters();
  Indexes = CurrentDCG->getListIndexes();
  Types = NILLINSTANCES;
  // if (CurrentDCG->getUserTypesOneLevel(&Types)) Types = ptlinstances_copy(Types);
  
  // Internal parameters.
  _ERROR_ = pk_false;
  forceGeneration = false;
  indent = false;
  inComment = false;
  idMem = string2ident("Mem.");	// for delay($) and Cell.
  idMemCste = string2ident("?"); // for delay with constant.
  // local lists
  lhAlreadyGenerated = NILLINSTH;
  lSigAddedOptimized = NILLINSTANCES;
  lAddedClocks = NILLINSTH;
  ClocksToBeGenerated = NILLINSTH;
  ClocksForTheNextLevel = NILLINSTH;
  SignalsForTheNextLevel = NILLINSTANCES;
  
}

/** 
    Finalizes the current SMT print.
    It cleans the graph by removing added marks to objects.
 */
void SMTPrinter::finalize()
{
  CKM->resetMarkAll(vMarkDeclared);
  EM->resetMarkAll(vMarkDeclared);

  CKM->resetMarkAll(vMarkSeen);
  EM->resetMarkAll(vMarkSeen);

  CKM->resetMarkAll(vMarkGenerated);
  EM->resetMarkAll(vMarkGenerated);
  NM->resetMarkAll(vMarkGenerated);
}

/* Unused now: SMT_LOGIC shell variable is assumed set to specifiy the logic
   void SMTPrinter::printLogic(string vlogic)
   {
   genOpenBlock();
   genString("set-logic ");
   genString(vlogic);      
   genCloseBlock();
   genNewLine();
   }
   
   void SMTPrinter::printLogicNewFile(string vlogic)
   { if (openFile("Logic")) { printLogic(vlogic); closeFile();} }
*/
