/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_


/** Constructor.*/
SMTService::SMTService(ptDCG_basic g)
{
  vDCG = g;
  BERROR = pk_false;
  certificationMode = pk_false;
  lNodesToExport = NILLINODE;
  lhTovisit = NILLINSTH;
  idMem = string2ident("Mem.");
  selectedType= BooleanType;
}

/** Destructor.*/
SMTService::~SMTService()
{
}

/**
   @return "true" if the operation succeeds, "false" otherwise (operator not yet implemented,...).
   It prepares the graph for the export in SMT format.
   - When certMode is true (certification mode), the traceability manager is used to know the transformations of the graph since the previous call to this method. 
     The traces of the transformations will be exported.
   - When certMode is false, the abstraction, defined in the buildSMTAbstraction() method, of the current graph for SMT is generated. 
   Currently, the abstraction is composed of the boolean part of the graph.

   To print the SMT abstraction requires to call the print() method. 

*/
boolean SMTService::translation(boolean certMode, boolean bprint)
{
  ptDCG_poly gPoly;
  ptDCG_basic G;
  ptTraceabilityManager tm;
  ptlINode unusedNodes;
  ptlinsth unusedClocks;
  boolean COMPLETEGRAPH = pk_true;  // when false : only the Clock system is printed.

  if ( isProgramWithSMTRestrictions())
    {
      pK_messenger->printString(" >>> SMT printer: some constructions are not yet implemented (loops, type,...)");
      pK_messenger->printString("     *** SMT CODE RESTRICTED TO THE CLOCK SYSTEM.***");
      COMPLETEGRAPH = pk_false;
    }

  selectedType = setSelectedTypeAndLogic(); // SMT_LOGIC shell variable is assumed set.
  
  certificationMode = certMode;
  BERROR = pk_false;
  if ( certificationMode )
    {
      tm = vDCG->getTraceabilityManager ();
      vDCG-> undersamplingOnSignal(pk_false, pk_false);
      exportedDCG = vDCG-> getDCG();
      unusedNodes = vDCG->getUnused (&unusedClocks);
      if (COMPLETEGRAPH)
	{
	  ptNodeManager NM = vDCG->getNodeManager();
	  lNodesToExport = NILLINODE;
	  NM->getAllNodes(&lNodesToExport);
	}
      else
	{
	  lNodesToExport =  vDCG-> getNodesSMTAbstraction(selectedType);
	}
      lhTovisit = getClocksToExport(COMPLETEGRAPH);
      tm->ProofOfCompilerCorrectness_begin(vDCG->getFormatLevel (), lhTovisit, unusedClocks, unusedNodes);
    }
  else
    {
      if (COMPLETEGRAPH)
	{
	  vDCG->undersamplingOnSignal(pk_false, pk_false);
	  exportedDCG = vDCG-> getDCG();
	  vDCG->undersamplingOnSignal(pk_false, pk_false);
	  lhTovisit = NILLINSTH;
	}
      else 
	{
	  GraphLevelType vformat = vDCG->getFormatLevel ();
	  // vDCG->groupPartialDefinitions();
	  vDCG->undersamplingOnSignal(pk_false, pk_false);
	  if (vDCG->buildSMTAbstraction(&G, selectedType))
	    {
	      exportedDCG = G -> getDCG();
	      gPoly = DCG_poly::new_prePoly (exportedDCG);
	      ptClockManager CKM_gPoly = gPoly->getClockManager ();
	      CKM_gPoly->systemSolving_postlude();
	    
	      gPoly-> setIntrinsicClocks ();
	      exportedDCG->setFormatLevel(vformat);
	      // lhTovisit = getClocksToExport(pk_true);
	      lhTovisit = NILLINSTH;
	      if (bprint) gPoly -> unparse ();
	    }
	  else
	    {
	      pK_messenger->printString("ERROR during buildSMTAbstraction");
	      pK_messenger->printString("SMTService: Program not exported to SMT.");
	      BERROR = pk_true;
	    }
	}
    }
  return (! BERROR);
}

/**
   @return "true" if the operation succeeds, "false" otherwise (error);
   Generate the SMT code for the certification if the translation() method has been called 
   with the parameter, otherwise the SMT code of the clock system.
   In the certification mode, several files are generated: (LEVEL stands for BASIC | POLY | ENDO | BOOL | SCH )
   - FOO_<LEVEL>_Defs.smt: contains the definitions of the current graph
   - FOO_<LEVEL>_Decls.smt: contains the declarations  of the current graph
   - FOO_<LEVEL>_OPTIMIZED_Defs.smt: contains the definitions that have been optimized since the previous level.
   - FOO_<LEVEL>_OPTIMIZED_Decls.smt:contains the declarations that have been optimized since the previous level.
   - FOO_<LEVEL>_MappedVariables.smt: contains the mapping of variables: a mapping (= h1 h2) specify that h1 of the previous level is h2 in the current level.

   In the standard mode, one file is generated  
   - FOO_SMT2_ <LEVEL>.smt that contains the declarations and the definitions of the current graph
*/
boolean SMTService::print()
{
  boolean b;
  ptlinsth lClockForNextStep, lhUnused, lhPreviousSystem, lhModifiedNames;
  ptlPointer lprevNames;
  ptTraceabilityManager tm;
  ptlinstances lSigForNextStep, PrevSignals;
  ptlINode PrevMemoizedNodes;

  ptldependence lIODependences, lprevIODependences;

  ptSMTPrinter smt_printer = new SMTPrinter();

  if (certificationMode) 
    {
      tm = vDCG->getTraceabilityManager ();
      lhPreviousSystem = tm -> Certification_getClocksPreviousSystem();
      PrevSignals = tm -> Certification_getPreviousSignals();
      PrevMemoizedNodes = tm -> Certification_getPreviousMemorizedNodes();
      tm -> Certification_getClocksWithNewName(lhPreviousSystem, &lhModifiedNames, &lprevNames);
      lprevIODependences = tm -> Certification_getPreviousIODependences();
      b = smt_printer->print(exportedDCG, selectedType, lhTovisit, lhPreviousSystem, lhModifiedNames, lprevNames, lprevIODependences, lNodesToExport,
			     PrevSignals, PrevMemoizedNodes, &lClockForNextStep, &lSigForNextStep, &lIODependences);
      tm->Certification_end(lClockForNextStep,lSigForNextStep, lIODependences);
    }
  else
    {
      b = smt_printer->print(exportedDCG, selectedType);
    }
  delete smt_printer;

  return b;
}


