/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** Constructor.*/
  SigaliService::SigaliService (ptDCG_basic /*ptDCG_poly */ g)
{
  vDCG = g;
  vDCGSigali = NILDCG_SIGALI;
  BERROR = pk_false;
}

/** Destructor.*/
SigaliService::~SigaliService ()
{
}

/**
   @return "true" if the operation succeeds, "false" otherwise (operator not yet implemented,...).

   <em> This  method can be applied to a graph of a process, NOT on a graph of a module.</em>\n
   This is the entry point of the code generation for SIGALI prover.
  
  The code is generated in the file \e foo.z3z, where  \e foo is the name of the process.
  It is produced by two steps:
      - \ref translateProgram() "first step", a "left to right topdown" traversal of the hierarchy is done. 
      It expresses the synchronisations and the booleans expressions of the program  in the Z/3Z finite field.
      For such an object, a polynomial is created in a specific internal representation. 
      The objects are classified into sub-classes in this representation: declarations, events, states, initialisations, 
      evolutions, constraints, controllables, free conditions. Attributes are used to link graph objects to polynomials.

    - \ref unparse_all() "second step": the sigali textual form is produced from the result of the firts step.
    Moreover, if specific intrinsic functions of SIGALI (see library SIGALI.LIB in the Polychrony library directory),
    are referenced in the program, a command file (see \ref printPropertiesToProve()) is generated, with foo_CMD.z3z as name. 
    It contains:
       - some initialization commands of SIGALI tool,
       - the set of properties that have to be proved and/or synthesized.

  The graph is then cleaned (\ref finalyze()).
*/
boolean SigaliService::translation (boolean trace)
{
  ptDCG_poly gPoly;
  ptDCG_basic G;

  BERROR = pk_false;
  vDCGSigali = NILDCG_SIGALI;

  // vDCG->groupPartialDefinitions();
  vDCG->undersamplingOnSignal (pk_false, pk_false);
  if (vDCG->buildF3Abstraction (&G))
    {
      /* OLD OLD 
	 gPoly = DCG_poly::DCG_basic2poly (G);
	 gPoly->getDCG()->setClocksName(false); // The clocks are not substituted.
	 gPoly->setIntrinsicClocks ();
      */

      gPoly = DCG_poly::new_prePoly (G->getDCG ());
      gPoly->getDCG()->setClocksName(false); // The clocks are not substituted.
      gPoly->setIntrinsicClocks ();

      // Printing the graph G in Signal textual form. (to be removed)
      if (trace) gPoly->unparse ();
      // restriction: delay 1,  no array...
      if (DCG_Sigali::mkDCGSigali (gPoly, &vDCGSigali))
	{
	  // The SigaliGenerator should be cleaned/simplified (all the objects of the graph (G) must be/are printed)
	  initialize ();
	  markSignalsProperties (lproperties);
	  translateProgram ();
	  finalyze ();
	  return pk_true;
	}


    }
  else
    pK_messenger->printString ("ERROR during buildF3Abstraction");
  pK_messenger->printString ("SigaliService: Program not exported to Sigali.");
  BERROR = pk_true;
  return pk_false;
}

/**
   @return "true" if the operation succeeds, "false" otherwise (error);
   if there is no detected before this calling, it prints 
   - the properties to prove (foo_CMD.z3z file, see printPropertiesToProve())
   - the expression of the synchronization in z3z (foo.z3z file, see unparse_all())
*/
boolean SigaliService::print ()
{
  ptrepstr
    fname;
  if (vDCGSigali == NILDCG_SIGALI)
    {
      if (BERROR)
	{
	  pK_messenger->
	    printString ("SigaliService::print: the graph cannot be printed; you must take into account the result of SigaliService::translation()");
	  pK_messenger->error (errcompilerbug);
	}
      else
	pK_messenger->printString ("SigaliService::print: the graph cannot be printed; you must call SigaliService::translation() before.");
      return pk_false;
    }

  shaper->initprintpoly ();
  if (!sys_add_typeSuffix2repstr (ProcessNameStr, Z3ZSUFF, &fname)) return pk_false;
  if (!sys_filePathInUserOutputDir (fname, &fname)) return pk_false;
  pK_messenger->globalMessageStr (infoz3z, fname);
  printPropertiesToProve (lproperties);	// FOO_CMD.z3z file.
  unparse (fname);		// FOO.z3z file

  if (numetat > 0) pK_messenger->message (infoz3zState);	// state variables have been created
  if (notStandard) pK_messenger->message (infoz3zNotStandard);
  return pk_true;
}
