
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
#define ptDCG_OriginalGraphType ptDCG_flat
#include "Service_Macros.c"

ptDCG_flat SyndexService::getGraph()
{
  return vDCG;			// ->getDCG();
}

ptDCG_Syndex SyndexService::getCurrentGraph()
{
  NM = gw->getNodeManager();
  EM = gw->getExpressionManager();
  CKM = gw->getClockManager();
  return gw;
}

void SyndexService::setCurrentDCG(ptDCG g)
{
  gw->setDCG(g);
}

_PROTECTED_
/**
   @return the operand expression (exp) in which the last contribution clock of a default is removed.
   \verbatim
   For (E1 when h1) default (E2 when h2), 
   it returns (E1 when h1) default (E2).
   \endverbatim
*/
ptexpcal SyndexService::removeLastContributionClockDefault(ptexpcal exp)
{
  ptexpcal exp_save, e2, e1, e;
  boolean b;
  ptinsth hc;

  exp_save = exp;
  if (!EM->IsWhenOrCellExpr(exp))
    b = pk_false;
  else
    {
      b = pk_true;
      hc = EM->getContributionClock(exp);
      exp = EM->getBasicExpr(exp);
    }

  if (!EM->isDefaultExpression(exp))
    return exp_save;
  e2 = EM->getDefaultArgstAt(exp, 2);
  if (!EM->IsWhenOrCellExpr(e2))
    return exp_save;
  e1 = EM->getDefaultArgstAt(exp, 1);
  e2 = EM->getWhenOrCellArg(e2);
  e = EM->mkBinaryExpTp(opmele, e1, e2, EM->getBackAnnotation(exp));
  EM->setContributionClock(e, EM->getContributionClock(exp));
  if (!b)
    return e;
  return EM->mkWhenOrCellExp(EM->getOperator(exp_save), e, hc);
}

/**
   For a equation or a partial definition (node), it removes the last contribution clock of the default expression.
*/
void SyndexService::removeLastContributionClockDefault(ptINode node)
{
  ptexpcal vdedef, e;

  if (NM->is_partialDefinition(node) || NM->is_equation(node))
    if (NM->get_rightOperand(node, &vdedef))
      {
	e = removeLastContributionClockDefault(vdedef);
	if (e != vdedef)
	  NM->update(node, vdedef, e);
      }
}

/**
   It applies removeLastContributionClockDefault(ptINode) for each node of lnodes.
*/
void SyndexService::removeLastContributionClockDefault(ptlINode lnodes)
{
  ptINode node;
  while (ptlINode_get_elem(&lnodes, &node))
    removeLastContributionClockDefault(node);
}

/**
   It applies removeLastContributionClockDefault(ptlINode) for the set 
   of equation and the set of partial definitions of the node manager.
*/
void SyndexService::removeLastContributionClockDefault()
{
  ptlINode lnodes;
  ptINode node;

  if (NM->getEquations(&lnodes))
    removeLastContributionClockDefault(lnodes);
  if (NM->getGroupsOfPartialEquations(&lnodes))
    while (ptlINode_get_elem(&lnodes, &node))
      removeLastContributionClockDefault(NM->get_setOfPartialDefinitions_arg(node));
}

boolean SyndexService::verifDelayOfOne(ptlINode ln)
{
  ptINode node;
  ptexpcal e, x, vret, vfen, vinit;
  compilerInteger val;

  while (ptlINode_get_elem(&ln, &node))
    {
      NM->get_delay_arg(node, &x, &e, &vret, &vfen, &vinit);
      // Restriction: delay 1 assumed.
      if (!EM->getIntegerValue(vret, &val))
	return pk_false;
      if (val != 1)
	return pk_false;
    }
  return pk_true;
}

boolean SyndexService::isTranslatable()
{
  ptlINode lNodes;

  if (NM->getDelays(&lNodes))
    if (!verifDelayOfOne(lNodes))
      {
	pK_messenger->printString("SyndexService::translation: program with delay > 1.");
	return pk_false;
      }

  if (NM->getIterates(&lNodes))
    {
      pK_messenger->printString("SyndexService::translation: program with array/iterate definitions.");
      return pk_false;
    }

  if (NM->getBundles(&lNodes))
    {
      pK_messenger->printString("SyndexService::translation: program with bundle/structure.");
      return pk_false;
    }

  return pk_true;
}

/** 
    Signal->Syndex tranlation.

    Assume : the graph is a STS one.
    WARNING : x := y $ n (n > 1), windows N, ARE NOT TAKEN INTO ACCOUNT.
    @param mainProg (in) must be pk_true for main program, "false" otherwise.
    @return "true" if the operation succeeds, false otherwise.
    
    \ref SyndexGenerator "For more details..."
*/
boolean SyndexService::translationOneLevel()
{
  ptstatesVariables stateBuilder;
  ptinsth tick;
  boolean b;
  ptDCG_Syndex gw;

  gw = getCurrentGraph();

  if (gw->isGraphOfExternalProcess())
    return pk_true;
  if (!isTranslatable())
    {
      pK_messenger->printString("SyndexService:: (translation) the graph cannot be translated to a Syndex graph.");
      return pk_false;
    }
  b = pk_true;
  // SubGraphs2Graph(gp);  //Removed Feb 2011 The subGraphs are isolated as Graphs.???
  gw->lefttts2defaults();
  removeLastContributionClockDefault();

  // The left_tt are rewritten in default expression for the interface calculus
  // for c := leftt(C1, C2), only C1 is referenced in Syndex.
  gw->nonProductiveSignalElimination();
  gw->explicitConversions();

  // All Delays are defined at master clock. 
  // State Variable are replaced by status intrinsic function.
  stateBuilder = new statesVariables(gw->getDCG());
  tick = gw->getClockManager()->getTickClock();
  if (!stateBuilder->extendDelayFromClockToClock(tick, tick))
    b = pk_false;
  if (!stateBuilder->extendCellToCellFromClockToClock(tick, tick))
    b = pk_false;
  stateBuilder->stateVariables2Status();
  delete stateBuilder;

  gw->setInitialValues();
  if (mainProg && (vDCGSyndex->getDCG() == gw->getDCG()))
    gw->makeCallsForInputsOutputs();	// ???? a faire differemment. (encapsuler).
  return b;
}

/**
   It produces the Abstract Syndex tree of the current graph.
*/
boolean SyndexService::produceOneLevel()
{
  boolean bres;
  ntrees vt;
  ptDCG vdcg;
  ptDCG2Syndex gen;
  ptDCG_Syndex gw;

  gw = getCurrentGraph();
  vdcg = gw->getDCG();
  gen = new DCG2Syndex(vdcg);
  bres = gen->translate(mainProg, (vdcg != vDCG->getDCG()), &vt);
  if (bres)
    {
      if (IS_NILTREE(ASTSyndex))
	ASTSyndex = LL_make_ldecl_emptyList();
      ASTSyndex = ntreeList_appendLast(ASTSyndex, vt);
      nmakeattrntrees(vt, SyndexService_ATTRIBUTENUMBER, copy_ident(vdcg->getName()));
    }
  else
    pK_messenger->printString("SyndexService::produce: incomplete translation ");
  delete gen;
  return bres;
}
