/* 
   Authors: Loïc Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Constructor. It is linked with a graph (g).
 */
  intemkIGraph::intemkIGraph (ptDCG g)
{
  AGRAPH = g;
  EM = AGRAPH->getExpressionManager ();
  CKM = AGRAPH->getClockManager ();
  NM = AGRAPH->getNodeManager ();
  NAMEM = AGRAPH->getNameManager ();
  /* predefined clocks */
  hinit = CKM->getInitClock ();
  the_nullClock_GB_copy = CKM->getNullClock ();
  hundef = CKM->getUndefClock ();
  UndefinedType = TypeTypeGetUndefinedType ();

  labelsStack = NILSTACK;
  clkLabelsStack = NILSTACK;
  vRunOn = NILPRAGMA;
  vGreyBox = NILPRAGMA;
  vBlackBox = NILPRAGMA;
  CounterOfExternalProcessReference = 0;
  WaitConstantExpr = pk_false;
  AcceptStateVar = pk_false;
  splittingObject = new splitting (g);
  myMark = getMark ();
  caseContextClock = hundef;
  // Automaton management
  automaton_current_state = NILEXPCAL;
  automatonClock = hundef;
  automaton_is_clocked = false;
  InAutomatonIntrinsicContext = false;
}

/**
   Destructor.
*/
intemkIGraph::~intemkIGraph ()
{
}


/** 
    @defgroup exprGroup intemkIGraph: expression managing 
 */
/** @ingroup exprGroup
   It links an expression (assumed to be of the form x[?]),
   to the current index of an array of processes (See getStackArrayNode()). No effect, if there is no array block.

   Note: It is done by an attribute called \e LinkToArrayIndexAttribute.
*/
void
intemkIGraph::linkToCurrentIndexArray (ptexpcal e)
{
  ptINode arrayNode, n;
  ptexpcal s;

  if (getStackArrayNode (&arrayNode))
    {
      n = NM->get_indexNode (arrayNode);
      s = NM->get_index (n);
      EM->addAttribute (e, LinkToArrayIndexAttribute, (attributeValue) s);
    }
}


/** @ingroup exprGroup
   @return "true" if a signal (aSig) respects the single assignment, "false" otherwise.
   (Several partial defitions are allowed).
   When an error is detected, the operand tree (vtree) is annotated.
*/
boolean intemkIGraph::verifyDefSig (ptexpcal aSig, BackAnnotationType vtree)
{
  ptINode n;
  ptlinstances lsig;
  ptexpcal s;
  boolean berr;

  mkFieldAccess (aSig, vtree);
  n = EM->getNode (aSig);
  if (NM->is_setOfPartialDefinitions (n))
    return pk_true;		// Set of part-definitions 
  if (isRecUndefined (n))
    return pk_true;
  if (NM->is_lfields (n))	// structure/bundle
    {
      lsig = NM->get_lfields_arg (n);
      berr = pk_false;
      while ((!berr) && ptlinstances_get_next (&lsig, &s))
	if (!verifyDefSig (s, vtree))
	  berr = pk_true;
      if (!berr)
	return pk_true;
    }

  pK_messenger->errorTreeIdent (errMultipleDefinition, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), EM->getOriginalName (aSig));
  return pk_false;
}

/** @ingroup exprGroup
   @return the initial value for a signal (vsig).
   If vinit is 
   - defined, it is the initial value of the signal (ie if there is yet an initial expression associated with the signbal it is replaced by vinit),
   - otherwise, the initial returned value is the current initial value associated with it.
*/
ptexpcal
intemkIGraph::verifAssignInit (ptexpcal vsig, ptexpcal vinit)
{
  ptexpcal vinitSig;

  vinitSig = EM->getInitialValue (vsig);
  if (vinit != NILEXPCAL)
    {
      if (vinitSig != NILEXPCAL)
	pK_messenger->warningacc (warIgnoreInit, EM->getTreeOfSignalDeclaration (vsig));
      EM->setInitialValue (vsig, vinit);
      return vinit;
    }
  if (vinitSig == NILEXPCAL)
    {
      pK_messenger->warningaccnom (warctinit2, EM->getTreeOfSignalDeclaration (vsig), EM->getName (vsig));
      pK_messenger->warningstr (warctinit2, ident2repstr (EM->getName (vsig)));
    }
  return vinitSig;
}

/** @ingroup exprGroup
  @return a new signal with undefined (free) type.
  @param annotTree (in) back attribute to parser abtract tree for error annotations. It can be undefined.
  @param instent (in/out) the created signal is inserted into this list.
*/
ptexpcal
intemkIGraph::mkVariable (BackAnnotationType annotTree, ptlinstances * instent)
{
  return mkVariable (annotTree, UndefinedType, instent);
}

/** @ingroup exprGroup
  @return a new signal with TT as type.
  @param TT (in) the type of the created signal. It can be undefined, in this case, a free type is associated with the created signal.
  @param annotTree (in) back annotation to parser information for error managing.
  @param instent (in/out) the created signal is inserted into this list.
*/
ptexpcal
intemkIGraph::mkVariable (BackAnnotationType annotTree, TypeType TT, ptlinstances * instent)
{
  ptexpcal vexpcal;
  typinst t;
  ptident id;

  if (UndefinedType == TT)
    TT = AnyTypeCreate ();
  if (WaitConstantExpr)
    {
      t = tconstant;
      id = NAMEM->getConstantId ();
    }
  else
    {
      t = tsignal;
      id = NAMEM->getContextSubExprOperator ();
    }
  vexpcal = EM->mkIdentExp (id, t, TT, NM->mk_node_undefined (), hundef, annotTree);
  EM->setAuxiliarySignal (vexpcal);
  if (t == tsignal)
    *instent = ptlinstances_sortedfusion (ptlinstances_push_front (NILLINSTANCES, vexpcal), *instent);
  return vexpcal;

}

/** @ingroup exprGroup. Structure/bundle managing.
    @return the operand expression (exp).
    If the expression (exp) is
      - an access of a field of a structure (x.Field), it associates (see getSetField() method), if it is not done, a "composed node" associated 
       with the referenced signal (x): the nodes associated with the fields { x.f1, ...,x.fn} are the elements of the "composed node".

      - a list of expressions, this method is applied for each element of the list
    A back attribute to parser abstract tree (vt) can be provided for error annotations.
*/
ptexpcal
intemkIGraph::mkFieldAccess (ptexpcal exp, BackAnnotationType vt)
{
  cloper clop;
  noperators op;
  ptexpcal exp1, e, r;
  boolean b;

  if (EM->isNILEXPCAL (exp)) return exp;
  if (EM->getSignalAssignedTo (exp, &exp1)) return exp;
  clop = EM->getOperatorClass (exp);
  if (clop != oper2) return exp;

  op = EM->getOperator (exp);
  if (op == opcham)
    // Defining a field of a struct/bundle
    {
      if (!mkExpressionWithoutReferenceToStateVar (exp, &exp1))
	{
	  b = pk_false;
	  exp1 = exp;
	}
      else
	b = pk_true;
      r = getSetField (exp1, vt);
      if (b)
	EM->setSignalAssignedTo (exp, r);
    }
  else
    {				// list of expressions (call)
      exp1 = exp;
      while (EM->getNaryArgs (&exp1, &e))
	mkFieldAccess (e, vt);
    }
  return exp;
}

/** 
    @defgroup nodeGroup intemkIGraph: node managing.
 */
/** @ingroup nodeGroup
   @return "true" if the operand signal (vinst) is a "bundle" one, "false" otherwise.
   - for a bundle an input node is created for each field without modification of the interface (see fieldsInputs() method) of the graph.
   in this case this method is called for each element and nRef parameter if the node associated with the "bundle".
   - Otherwise, nRef must be undefined.
   In all the cases, 
    - the returned node (m) is the "input node" associated with the operand signal.
    - it takes into account the differents contexts (see takeIntoAccountContexts()).
*/
boolean
intemkIGraph::addInNode (ptexpcal vinst, ptINode nRef, ptINode * m)
{
  ptinsth vh;
  ptINode n;

  if (fieldsInputs (vinst))
    return pk_false;
  n = NM->mk_node_input (vinst, nRef, EM->getBackAnnotation(vinst));
  vh = setClockVariable (vinst);
  NM->insertNode (n, vh);
  takeIntoAccountContexts (n);
  *m = n;
  return pk_true;
}



boolean
intemkIGraph::addInNode (ptexpcal vinst, ptINode nRef, ptinsth vh, ptINode * m)
{
  //  ptinsth vh;
  ptINode n;

  if (fieldsInputs (vinst)) return pk_false;
  n = NM->mk_node_input (vinst, nRef, EM->getBackAnnotation(vinst));
  //  vh = setClockVariable (vinst);
  NM->insertNode (n, vh);
  takeIntoAccountContexts (n);
  *m = n;
  return pk_true;
}



/** @ingroup nodeGroup
   It adds a "write" node in the graph for the signal (vinst).
   For a bundle/structured type, it must be called for each sub-signal (see fieldsOutputs()).
   The parameter nRef unused currenly.
*/
ptINode
intemkIGraph::addOutNode (ptexpcal vinst, ptINode nRef)
{
  ptinsth vh, h;
  ptINode n, n1;
  ptpragma aPragma;

  n = NM->mk_node_output (vinst, EM->getBackAnnotation(vinst));
  vh = setClockVariable (vinst);
  NM->insertNode (n, vh);
  takeIntoAccountContexts (n);
  // For external node.
  n1 = EM->getNode (vinst);
  h = NM->get_nodeClock (n1);
  if (h == hundef)
    NM->insertNode (n1, vh);
  // explicit dependences are propagated.
  NM->ExplicitdependencesPasteOn (n1, n);
  // Partitionning pragma moved.
  if (getPragmaKeyword (NM->getPragmas (n1), PRAGMA_BINDINGIO, &aPragma))
    {
      NM->removePragma (n1, getPredefinedPragma (PRAGMA_BINDINGIO));
      NM->addPragma (n, aPragma);
    }
  return n;
}


/** @ingroup nodeGroup 
    Construction of the input nodes for the signals of the list (list). 
    Assumes these signals are inputs signals of the graph.
    This operation can induce splitting up for bundles.
    Note that the interface of the program is not modified.
    @return none.
*/
void
intemkIGraph::addInNodes (ptlinstances list)
{
  ptexpcal vinst;
  ptINode m;
  while (ptlinstances_get_next (&list, &vinst))
    addInNode (vinst, NILINODE, &m);
}


/** @ingroup nodeGroup 
    Construction of the output nodes for the signals of the list (list).
    This operation can induce splitting up for bundles.
    Note that the interface of the program is not modified.
    @return none
*/
void
intemkIGraph::addOutNodes (ptlinstances list)
{
  ptexpcal vinst;
  while (ptlinstances_get_next (&list, &vinst))
    if (!fieldsOutputs (vinst))
      addOutNode (vinst, NILINODE);
}



/** @ingroup nodeGroup 
   It associates to a node (node) the current pragmas (see pragmasContext() method)
   and when the current label node is 
   - a sub graph, it is inserted in the subgraph.
   - an array, it is inserted in the body  sub-graph  or the inital  sub-graph of the array (depends on the bodyGraph boolean context).
   The delay are not inserted in the array node (see addInputArrayNodes()).
*/
void
intemkIGraph::takeIntoAccountContexts (ptINode node)
{
  ptElemStack nodeLabel;
  ptINode n;

  if (stackTop (labelsStack, &nodeLabel) && (nodeLabel != NILINODE))	// nodeLabel == NILINODE when the label is associated with a state of an automaton.
    {
      n = (ptINode) nodeLabel;
      if (NM->is_subgraph (n))
	NM->subGraphInsertNode (n, node);
      else if (NM->is_array (n))
	if (!NM->is_delay (node))
	  {
	    if (bodyGraph)
	      NM->arrayBodyInsertNode (n, node);
	    else
	      NM->arrayInitInsertNode (n, node);
	    if (NM->is_array (node))
	      NM->set_inComposedNode (node, n);	// A revoir avec les lignées
	  }
	else
	  //  added as inputs of the array;
	  addInputArrayNodes (NM->get_leftOperand (node));
      else;			// error
    }
  pragmasContext (node);
}


/** @ingroup nodeGroup
   It creates the definition of a constant (vexpcal) by the expression Edef.
   The associated clock is the predefined "initial" clock.
   A back attribute to parser abstract tree (refTree) for error annotations can be provided.
*/
void
intemkIGraph::createConstantDefinition (ptexpcal vexpcal, ptexpcal Edef, BackAnnotationType refTree)
{
  if (!EM->isNILEXPCAL (Edef))
    {
      NM->mk_node_equation (mkFieldAccess (vexpcal, refTree), Edef, hinit, NILLINSTANCES, refTree);
      EM->setValuesEq (vexpcal, Edef, opnil);
    }
  AGRAPH->addConstant (vexpcal);
}


/** @ingroup nodeGroup
   It associates with the node that defined an identifier (expcal1) an element composed of: 
  - expcal2 as expression of definition.
  - vhcal as clock at which the expression expcal2 is used to define the signal expcal1.

  The list of signals (instent) are assumed to be the signals referenced in expcal2: it is used to install dependences.
  A back attribute to parser abstract tree (vtree) for error annotations can be provided.\n
 
  After some verifications about the clock of assigment it calls mkEquation() method.\n
  Some particular cases:
     - the constant definitions are filtered (see createConstantDefinition() method). 
     - the cartesian product operation is processed in mkCartesianProductEquation() method.

  @return the synchronized clock of expcal1 with vhvar.
*/
ptinsth
intemkIGraph::createEquation (ptexpcal expcal1, ptexpcal expcal2, ptinsth vhcal, ptinsth vhvar, ptlinstances instent, BackAnnotationType vtree)
{
  ptinsth vh, vh1;
  boolean b;

  if (!EM->isNILEXPCAL (expcal2) && EM->getOperator (expcal2) == oppcar)
    {
      vh = mkCartesianProductEquation (expcal1, expcal2, vhcal, instent, vhvar, vtree);
    }
  else
    {
      if (EM->IsConstant (expcal1))
	{
	  createConstantDefinition (expcal1, expcal2, vtree);
	  return vhvar;
	}
      if ((vhcal == hinit) && ! EM->IsConstantExpr (expcal2)) vhcal = vhvar;	/* state variables */
      b = EM->isStateVar (expcal1);
      if (isPartialDefinitionOf (expcal1))
	{
	  vh1 = instContextClock (vhvar, vtree);
	}
      else
	{
	  if (b) pK_messenger->errorTreeIdent (errStateVar, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), EM->getOriginalName (expcal1));
	  vh1 = vhcal;
	}
      if (!b) vh = getSignalClockSynchronizedWith (expcal1, vhvar); else vh = vh1;
      mkEquation (expcal1, expcal2, vh1, instent, vh, vtree);
    }

  return vh;
}

/** @ingroup nodeGroup
    Add a partial equation. Similar to createEquation().
 */
ptinsth 
intemkIGraph::createPartialEquation (ptexpcal expcal1, ptexpcal expcal2, ptinsth vhcal, ptinsth vhvar, ptlinstances instent, BackAnnotationType vtree)
{
  ptinsth vh;

  EM->addMark (expcal1, PART_DEFINITION);
  vh = createEquation (expcal1, expcal2, vhcal, vhcal, instent, vtree);
  EM->removeMark (expcal1, PART_DEFINITION);
  return vh;
}

/** @ingroup nodeGroup
   It associates with the node that defined an identifier (expcal1) an element composed of: 
   - expcal2 as expression of definition.
   - vhcal as clock at which the expression expcal2 is used to define the signal expcal1.

   The list of signals (instent) are assumed to be the signals referenced in expcal2: it is used to install dependences.
   A back attribute to parser abstract tree (vt) for error annotations can be provided.\n
   Here, nothing is said about the clock of the defined signal. 

   The node is not inserted in the graph: it can be used when a signal is defined by a default-like expression (see mkEquation()).
   Then, createEquation() method must be called for the last expression to insert the node in the graph. 
*/
void
intemkIGraph::addSubEquation (ptexpcal expcal1, ptexpcal expcal2, ptinsth vhcal, ptlinstances instent, BackAnnotationType vt)
{
  vhcal = instContextClock (vhcal, vt);
  mkEquation (expcal1, expcal2, vhcal, instent, hundef, vt);
}

/** 
    @ingroup nodeGroup
    It creates a delay node (\e zexp := \e exp $ \e delayExp window \e winExp  init \e vinit)
    with vh as clock.
    Note : when the value (d) of the delay and the value of the window (w) are available, and, d=0 and w=1
    an equation node is created.
*/
void
intemkIGraph::mkDelayWindowNode (ptexpcal zexp, ptexpcal exp, ptexpcal delayExp, ptexpcal winExp, ptexpcal vinit, ptinsth vh, BackAnnotationType vtree)
{
  ptINode vinstr;
  integer v1, v2;
  boolean b;

  exp = EM->getCloneClassHead (exp);
  b = pk_true;
  if (getExpressionValue (delayExp, &v1) && getExpressionValue (winExp, &v2))
    if ((v1 == 0) && (v2 == 1))
      {
	createEquation (zexp, exp, vh, vh, ptlinstances_push_front (NILLINSTANCES, exp), vtree);
	b = pk_false;		// it is not a delay.
      }
  if (b)
    if (verifyDefSig (zexp, vtree))
      {
	if (EM->isStateVar (zexp) && !isPartialDefinitionOf (zexp)) 
	  pK_messenger->errorTreeIdent (errStateVar, BACK_ANNOT_GET_ABSTRACT_TREE(vtree), EM->getOriginalName (zexp));
	vinit = verifAssignInit (zexp, vinit);
	if (adjustInitialValue (&vinit, delayExp, winExp, vtree)) EM->setInitialValue (zexp, vinit);

	vinstr = NM->mk_node_delay (mkFieldAccess (zexp, vtree), exp, delayExp, winExp, vinit, vtree);
	NM->insertNode (vinstr, vh);
	takeIntoAccountContexts (vinstr);
      }
}

/** @ingroup nodeGroup
   @return "false" if the model (modelName) is the "assert" predefined model, "true" otherwise.
   For the assertion, see AssertionCallNode().
   
   It adds a procedure call node in the graph \e louts := \e modelName { \e lparam } ( \e lins ).
   @param modelName (in)  name of the model,
   @param lparam (in) sets of parameters,
   @param  lins (in)   sets of referenced signals,
   @param louts  (in) sets of produced signals,
   @param vh (in) clock of the node,
   @param code  (in) safety property tag,
   @param vtree  (in)  back annotation to the reference of the model,
   @param modelDecl  (in)  back annotation to the declaration of the model,
   @param  bintrinsic (in) intrinsic model when bintrinsic, 
   @param vn (out) the inserted node (available only when it returns "true").
*/
boolean
  intemkIGraph::mkProcCallNode (ptident modelName, ptexpcal lparam, ptexpcal lins, ptexpcal louts, 
				ptinsth vh, safetyPropertyType code, BackAnnotationType vtree, BackAnnotationType modelDecl, boolean bintrinsic, ptINode * vn)
{
  ptINode vinstr;
  ptexpcal ls;

  if (AssertionCallNode (modelName, lins, vh))
    return pk_false;

  ls = mkFieldAccess (louts, vtree);
  vinstr = NM->mk_procCall (modelName, lparam, lins, ls, modelDecl);
  NM->set_safetyProperty (vinstr, code);
  if (bintrinsic)
    {
      NM->set_intrinsicfunc (vinstr);
      EM->TypeCheckingIntrinsic (modelName, lparam, lins, ls, vtree);
    }
  NM->insertNode (vinstr, vh);
  takeIntoAccountContexts (vinstr);
  *vn = vinstr;
  return pk_true;
}



/** @defgroup pragmaGroup intemkIGraph: pragma managing 
    Pragma managing methods: some pragmas (grey box, black box, run on) play a special feature
    during the building of the graph.
*/
/**
   It sets vr as the current "Run On" pragma.
*/
void
intemkIGraph::setRunOnPragma (ptpragma vr)
{
  // to be rewritten using StackPragmaTables
  vRunOn = vr;
}

/** @ingroup pragmaGroup
   @return the current "Run On" pragma.
*/
ptpragma intemkIGraph::getRunOnPragma ()
{
  // to be rewritten using StackPragmaTables
  return vRunOn;
}

/** @ingroup pragmaGroup
    It pushes a pragma (vr) in the StackPragmaTables[pragma name of (vr)].
*/
void
intemkIGraph::pushPragma (ptpragma vr)
{
  ptrepstr name;

  name = getPragmaName (vr);
  if (StackPragmaTables.find (name) == StackPragmaTables.end ())
    StackPragmaTables[name] = NILSTACK;
  StackPragmaTables[name] = stackPush (StackPragmaTables[name], (ptElemStack) vr);
}

/** @ingroup pragmaGroup
    It pops a pragma(vr) in the StackPragmaTables[pragma name of (vr)].
*/
void
intemkIGraph::popPragma (ptpragma vr)
{
  ptrepstr name;

  name = getPragmaName (vr);
  if (StackPragmaTables.find (name) != StackPragmaTables.end ())
    StackPragmaTables[name] = stackPop (StackPragmaTables[name]);
}


/**@ingroup pragmaGroup
   vgb becomes the current "Grey box" pragma, and the 
   counter of external processes (CounterOfExternalProcessReference attribute) is increadsed.
*/
void
intemkIGraph::newGreyBoxPragma (ptpragma vgb)
{
  // to be rewritten using StackPragmaTables
  CounterOfExternalProcessReference++;
  setGreyBoxPragma (vgb);
}

/** @ingroup pragmaGroup
   vgb becomes the current "Grey box" pragma.
*/
void
intemkIGraph::setGreyBoxPragma (ptpragma vgb)
{
  // to be rewritten using StackPragmaTables
  vGreyBox = vgb;
}

/** *
    @ingroup pragmaGroup
    @return the current "Grey box" pragma.
*/
ptpragma intemkIGraph::getGreyBoxPragma ()
{
  // to be rewritten using StackPragmaTables
  return vGreyBox;
}


/**
   @ingroup pragmaGroup
   vgb becomes the current "Black box" pragma.
*/
void
intemkIGraph::setBlackBoxPragma (ptpragma vgb)
{
  // to be rewritten using StackPragmaTables
  vBlackBox = vgb;
}

/**
   @ingroup pragmaGroup
   @return the current "Black box" pragma.
*/
ptpragma intemkIGraph::getBlackBoxPragma ()
{
  return vBlackBox;
}

/**
   @defgroup labelGroup intemkIGraph: sub graph managing (labels).
   Labels are used to group nodes of the graph. This construction is 
   provided by the expression E: (| set of equations |). The array of processes
   is also characterized by two sub graphs; one for the initial part, another for the body.
   So, some methods are provided to manage this context block (sub-graph): it is a classical stack managing.
*/
/**
   vls becomes the current label stack (labelsStack attribute).
 */
void
intemkIGraph::setLabelStack (ptStack vls)
{
  labelsStack = vls;
}

/** @ingroup labelGroup
   @return the current label stack (labelsStack attribute).
 */
ptStack intemkIGraph::getLabelStack ()
{
  return labelsStack;
}

/**
   @ingroup labelGroup
   It pushes the element (vClk) to  the top of the clock label stack (clkLabelsStack attribute).
*/
void
intemkIGraph::pushContextClock (ptElemStack vClk)
{
  clkLabelsStack = stackPush (clkLabelsStack, vClk);
}

/**
   @ingroup labelGroup
   It pops the clock label stack (clkLabelsStack attribute).
*/
void
intemkIGraph::popContextClock ()
{
  clkLabelsStack = stackPop (clkLabelsStack);
}

/**
   @ingroup labelGroup
   @return "true" if the clock label stack is not empty, "false" otherwise.
   @param el (out) the top of the stack.
*/
boolean intemkIGraph::topContextClock (ptinsth * el)
{
 pointer x;
 
 if (!stackTop (clkLabelsStack, &x)) return pk_false;
  (*el) = (ptinsth) x;
  return pk_true;
}

/**
   @ingroup labelGroup
   @return the clock, if it exists, that is at the top of the clock label stack, or a new one otherwise.
*/
ptinsth
intemkIGraph::getLabelClock (ntrees absTree)
{
  ptinsth el;
  if (topContextClock (&el)) return el;
  pK_messenger->warningacc (warLabelAction, absTree);
  pK_messenger->warning (warLabelAction);
  return CKM->new_variFormulaRootClock ();
}


/**
   @ingroup labelGroup
   It pushes 
   - the node (el) associated to a label to the top of the label stack. el is NILINODE for a state of an automaton.
   - and the clock (vClk) to the top of the clock label stack.
*/
void
intemkIGraph::pushLabel (ptINode el, ptinsth vClk)
{
  labelsStack = stackPush (labelsStack, (ptElemStack) el);
  pushContextClock ((ptElemStack) vClk);
}

/**
   @ingroup labelGroup
   It pops the label stack (labelsStack attribute) and the clock label stack (clkLabelsStack attribute).
*/
void
intemkIGraph::popLabel ()
{
  labelsStack = stackPop (labelsStack);
  popContextClock ();
}


/**
   @ingroup labelGroup
   @return the top of the label stack.
*/
boolean intemkIGraph::topLabel (ptElemStack * el)
{
  return stackTop (labelsStack, el);
}

/** @ingroup labelGroup
    It pushes an array node (arrayNode) in the "label stack" context.
    and sets the bodyGraph to "false": here, we suppose that the compiler begins to 
    compile the "initial" part of an array of processes.
 */
void
intemkIGraph::pushArray (ptINode arrayNode)
{
  // clk context not modified;
  labelsStack = stackPush (labelsStack, arrayNode);
  bodyGraph = pk_false;		// default the init. (non empile? a revoir)
}

/** @ingroup labelGroup
    Array managing.
    vbody must be at "true" value for the compiling of the body of the array; "false" for the with part. */
void
intemkIGraph::setCurrentSubgraph (boolean vbody)
{
  bodyGraph = vbody;
}

/** @ingroup labelGroup
    Pop array/label stack.
*/
void
intemkIGraph::popArray ()
{
  labelsStack = stackPop (labelsStack);
  // attention au bodyGraph (non empile??)
}


/** @ingroup pragmaGroup
    It associates the current pragmas (RunOn, Black Box, Grey Box,...) to a node (node).
*/
void
intemkIGraph::pragmasContext (ptINode node)
{
  PragmaPropagateOn (vRunOn, node);

  if ((NM->isProcessCall (node)) || NM->isNodeCall (node))
    {
      if (vBlackBox != NILPRAGMA)
	if (vGreyBox == NILPRAGMA)
	  NM->addPragma (node, addIndexForPragma (vBlackBox, pk_true));
	else
	  {
	    // IT IS A BLACK BOX CALL in the Grey Box specification.
	    NM->addPragma (node, vBlackBox);
	    NM->addPragma (node, addIndexForPragma (vGreyBox, pk_false));
	  }
      else if (vGreyBox != NILPRAGMA)
	NM->addPragma (node, addIndexForPragma (vGreyBox, pk_false));
    }
  else
    {
      if (vGreyBox != NILPRAGMA)
	NM->addPragma (node, vGreyBox);
      if (vBlackBox != NILPRAGMA)
	NM->addPragma (node, vBlackBox);
    }
}

/** @ingroup pragmaGroup
    @return a new pragma built from another one (aPragma) and with a new "value" given by the current 
    value (or increased if bincr) of the internal counter (CounterOfExternalProcessReference).
    It is used to mark the set of clusters (balck boxes) of a same reference to an external process (seen as a grey box).
*/
ptpragma
intemkIGraph::addIndexForPragma (ptpragma aPragma, boolean bincr)
{
  ptObjPragma obj;
  ptrepstr name, extRefOcc;
  ntrees enonce, newEnonce;

  if (bincr) CounterOfExternalProcessReference++;

  extRefOcc = integer2repstr (CounterOfExternalProcessReference);
  getPragma (aPragma, &name, &obj, &enonce);
  newEnonce = copyEnonce (enonce);
  newEnonce = addEnonceValue (newEnonce, extRefOcc);
  return newpragma (name, obj, newEnonce);
}

/**
   @defgroup clockGroup intemkIGraph:clock managing.
   clock managing.
*/

/**
   @ingroup labelGroup
   clk becomes the case context clock. It is used to fix the clock of the constants in the case expressions.
*/
ptinsth intemkIGraph::pushCaseContextClock (ptinsth clk)
{
  ptinsth pclk;
  pclk = caseContextClock;
  caseContextClock = clk;
  return pclk;
}

/**
   @ingroup labelGroup
   clk becomes the case context clock. It is used to fix the clock of the constants in the case expressions.
*/
ptinsth intemkIGraph::getCaseContextClock ()
{
  return caseContextClock;
}

/**
   @ingroup clockGroup
   @return the clock obtained by the synchronisation of
   - the clock with a free part vh1
   - the clock vh2 with a free part if locfree is pk_true.
*/
ptinsth
intemkIGraph::instop2syncfree (ptinsth vhcste, ptinsth vh1, ptinsth vh2, boolean locfree)
{
  ptinsth vharg1, vharg2;

  vh1 = CKM->makeSynchro (vh1, vh2);
  if (!locfree)
    return vh1;
  if (CKM->satisfiesUndefClock (vhcste))
    return vh1;
  if (!CKM->getSupArgs (vh1, &vharg1, &vharg2))
    return vh1;
  if (vharg1 == vh1)
    return CKM->makeSynchro (vh1, CKM->makeSup (vharg2, vhcste));
  if (vharg2 == vh1)
    return CKM->makeSynchro (vh1, CKM->makeSup (vharg1, vhcste));
  return vh1;
}

/** @ingroup clockGroup
    @return the clock of a signal (vsig). This clock is created if it does not exist.
 */
ptinsth
intemkIGraph::mkClockInstance (ptexpcal vsig)
{
  ptinsth vh;

  vh = EM->getClock (vsig);
  if (!CKM->satisfiesUndefClock (vh))
    return vh;
  vh = CKM->new_variFormulaRootClock ();
  EM->setClock (vsig, vh);
  return vh;
}

/** @ingroup clockGroup
   If the clock of the signal (vinstance) is undefined, a new clock is associated with it, otherwise it is unchanged.
   The clock of a index, parameter, constant is undefined.
   @return the clock of the operand identifier.
*/
ptinsth
intemkIGraph::setClockVariable (ptexpcal vinstance)
{
  ptexpcal vx;

  if (EM->isStateVar (vinstance))
    return CKM->getFictitiousRoot ();
  if (!isPartialDefinitionOf (vinstance))
    {
      if (EM->getOperatorClass (vinstance) != variabl)
	{
	  // field of a bundle/structure: to be completed for an access to an array x[].
	  // The following line added Jan 2012 LBesnard.
	  if ((EM->getOperator (vinstance) == opcham) && EM->getSignalAssignedTo (vinstance, &vx))
	    return setClockVariable (vx);
	  return hundef;
	}
      if (!EM->IsSignal (vinstance) && !EM->IsLabel (vinstance) && !EM->IsField (vinstance))
	return hundef;
    }
  return mkClockInstance (vinstance);
}

/** 
    @ingroup clockGroup
    It associates to a clock (vh) the current pragmas.
*/
void
intemkIGraph::takeIntoAccountContexts (ptinsth vh)
{
}


/**
   @ingroup clockGroup
   @return the clock of an identifier (vexpcal) assumed to be 
       - a signal, parameter, index, label, state variable
       - or a reference to a field of a structure,
       - or a reference to an element of an array.
  using a context clock (vhcste) combined with context mode (isfree) for fixing this clock for state variable and index.
  The list of signals (l) is completed with the identifier when it is a signal or a label.
*/
ptinsth
intemkIGraph::getSetIdClock (ptexpcal vexpcal, boolean * isfree, ptinsth vhcste, BackAnnotationType vt, ptlinstances * l)
{
  noperators op;
  ptinsth vh;

  op = EM->getOperator (vexpcal);
  if (op == opcham)
    {
      *isfree = !EM->IsSignal (EM->getDeclaredSignal (vexpcal));
      return fieldAccess (vexpcal, vhcste, l, vt);
    }
  if (op == opelem)
    {
      do
	{
	  vexpcal = EM->getBinaryArgstAt (vexpcal, 1);
	  op = EM->getOperator (vexpcal);
	}
      while (op == opelem);
      return getSetIdClock (vexpcal, isfree, vhcste, vt, l);
    }
  // opident

  if (!EM->IsSignal (vexpcal))
    {
      if (EM->IsLabel (vexpcal))
	{
	  *l = ptlinstances_insert (*l, vexpcal);
	  if (InAutomatonIntrinsicContext) return hundef; // Added Nov 2016
	  *isfree = pk_false;
	  return setClockVariable (vexpcal);
	}
      if (EM->IsIndex (vexpcal))
	{
	  vh = setClockVariable (vexpcal);
	  if (!CKM->satisfiesUndefClock (vh))
	    (*isfree) = pk_false;
	  return vh;
	}
      // constant, parameter.
      return instsiginvhcste (isfree, vhcste, vt);
    }
  // Signal.
  if (isInConstantContext ()) pK_messenger->errorTree (errNotConstantExpr,  BACK_ANNOT_GET_ABSTRACT_TREE(vt));
  if (EM->isStateVar (vexpcal))
    {
      if (!isInStateVarContext ()) pK_messenger->errorTree (errcsteint,  BACK_ANNOT_GET_ABSTRACT_TREE(vt));
      vh = instsiginvhcste (isfree, vhcste, vt);
    }
  else
    {
      *isfree = pk_false;
      vh = setClockVariable (vexpcal);
    }
  if (EM->IsEvent (vexpcal))
    vexpcal = EM->getTrue ();
  else
    *l = ptlinstances_insert (*l, vexpcal);
  return vh;
}

/** 
    @ingroup clockGroup
    It associates a clock (vh) to a signal (vinstance).
    @return the clock of the signal.
    If the identifier (vinstance) is
    - a constant, the clock is the predefined "initial" clock.
    - a state variable, the clock is the predefined fictitious root.
    - otherwise, if the clock (vh) is 
       - undefined, or the signal (vinstance) is currently defined by a partial expression (isPartialDefinitionOf() method), 
       it is defined by setClockVariable() method.
       - otherwise, a synchronisation is added between the clock of the signal and the operand clock (vh).
*/
ptinsth
intemkIGraph::getSignalClockSynchronizedWith (ptexpcal vinstance, ptinsth vh)
{
  ptinsth vh1;

  if (EM->IsConstant (vinstance))
    return hinit;
  if (EM->isStateVar (vinstance))
    return CKM->getFictitiousRoot ();
  if (CKM->satisfiesUndefClock (vh))
    return setClockVariable (vinstance);
  if (isPartialDefinitionOf (vinstance))
    return setClockVariable (vinstance);
  vh1 = EM->getClock (vinstance);
  if (CKM->satisfiesUndefClock (vh1))
    EM->setClock (vinstance, vh);
  else
    vh = CKM->makeSynchro (vh, vh1);
  return vh;
}

/**  @ingroup clockGroup
     @return the clock of the synchronisation of a list (l) of signals. */
ptinsth
intemkIGraph::synchronizeIdentifiers (ptlinstances l)
{
  ptinsth vh, vhres;
  ptexpcal vinst;
  ptlinstances liste;
  liste = EM->getSignalsReferencedInList (l);
  l = liste;
  if (!ptlinstances_get_next (&liste, &vinst))
    return CKM->getUndefClock ();
  vh = setClockVariable (vinst);
  while (((vh == hundef) || (vh == hinit)) && ptlinstances_get_next (&liste, &vinst))
    vh = setClockVariable (vinst);
  while (ptlinstances_get_next (&liste, &vinst))
    vh = getSignalClockSynchronizedWith (vinst, vh);
  if (vh != hinit)
    vhres = vh;
  else
    vhres = CKM->getUndefClock ();
  ptlinstances_free (l);
  return vhres;
}

/**  @ingroup clockGroup
     @return the upper bound of the clocks of the signals of a list (l). */
ptinsth
intemkIGraph::upperBoundClockOfIdentifiers (ptlinstances l)
{
  ptinsth vh;
  ptexpcal vsig;
  ptlinstances liste;

  liste = EM->getSignalsReferencedInList (l);
  l = liste;
  vh = CKM->getUndefClock ();
  while (ptlinstances_get_next (&liste, &vsig))
    if (!EM->isStateVar (vsig))
      vh = CKM->makeSup (vh, setClockVariable (vsig));
  ptlinstances_free (l);
  return vh;
}



/** @defgroup globalProcessinGroup intemkIGraph:methods for sets of objects
 */

/** @ingroup globalProcessinGroup
   For each node, element of the list of partial definitions (ie set of x::=E1,..., x::=En), it finalyzes 
   by fixing its clock and inserting it in the graph.
   Two cases are considered: the state variable (see statepartialDefinitionFinalize() method), and the signals (see signalpartialDefinitionFinalize()).
*/
void
intemkIGraph::partialDefinitionFinalize ()
{
  ptlINode lNodes;
  ptINode node;
  ptexpcal s;

  if (NM->getGroupsOfPartialEquations (&lNodes))
    // lNodes : set of nodes defined by a set of ::= nodes
    while (ptlINode_get_elem (&lNodes, &node))
      {
	s = NM->get_leftOperand (node);
	if (EM->isStateVar (s))
	  statepartialDefinitionFinalize (node);
	else
	  signalpartialDefinitionFinalize (node);
      }
}


/** @ingroup globalProcessinGroup
    @return "true" if all the fields of structure/bundle signal are defined, "false" otherwise. 
    For each "structure/bundle" node defining a signal (x),
      - it adds a dependence x.a_i --> x for each field of a structure (not for a bundle).
      - it inserts the node associated with the "structure/bundle" into the graph at the clock
            - the synchronized clock of the field x.a_1, ..., x.a_n.
            - upper bound of the signals x.a_1, ..., x.a_n for a bundle
*/
boolean
intemkIGraph::verifyStructBundleDefinitions ()
{
  ptlINode lNodes;
  ptINode node, node1;
  ptlinstances lsig;
  ptexpcal aSig, s, left;
  boolean err, bres, bstruct, binput, bstate;
  ptinsth h, h1;
  TypeType TT;

  if (AGRAPH->isGraphOfExternalProcess ())
    return pk_true;
  if (!NM->getBundles (&lNodes))
    return pk_true;

  bres = pk_true;
  while (ptlINode_get_elem (&lNodes, &node))
    {
      left = NM->get_leftOperand (node);
      s = left;
      if (NM->is_readNode (node))
	binput = pk_true;
      else
	binput = (EM->getOperator (left) == oplident);	// ??
      if (!binput)		// revoir 
	{
	  bstate = EM->isStateVar (s);
	  TT = EM->getUnamedType (s);
	  bstruct = TypeTestStructureClass (TT);
	  if (bstruct)
	    h = EM->getClock (s);
	  else
	    h = hundef;
	  lsig = NM->get_lfields_arg (node);
	  err = pk_false;
	  while (!err && ptlinstances_get_next (&lsig, &aSig))
	    {
	      node1 = EM->getNode (aSig);
	      if (NM->is_undefined (node1))
		err = !CKM->satisfiesNullClock (EM->getClock (aSig));
	      else if (bstruct)
		{
		  if (h == hundef)
		    h = NM->get_nodeClock (node1);
		  NM->addDependence (node1, node, h);
		}
	      else
		{
		  if (bstate)
		    h1 = NM->get_nodeClock (node1);
		  else
		    h1 = EM->getClock (aSig);
		  h = CKM->makeSup (h, h1);
		}
	    }
	  if (err)
	    {
	      bres = pk_false;
	      pK_messenger->errorTreeIdent (errFieldNotDefined, EM->getTreeOfSignalDeclaration (s), EM->getOriginalName (s));
	    }
	  else if (!bstruct)
	    if (h != hundef)
	      h = CKM->makeSynchro (EM->getClock (s), h);
	  NM->insertNode (node, h);
	}
    }
  return bres;
}

/**
   @ingroup globalProcessinGroup
   It finalyzes the construction  of the associated graph by:
    - \ref graphClassSpecificity "applying graph Class specificity",
    - \ref propagateTypes "propagating types",
    - \ref DCG::setInitialValues "verifying the initialisations",
    - \ref addDefinitionForUndefinedSignals "adding definitions (x:=x) for undefined signals that are not declared as inputs,
    - \ref enumNameConflicts "solving conflicts for enumerated type",
    - \ref verifyStructBundleDefinitions "finalyzing the structure/bundle definitions"
    - \ref partialDefinitionFinalize "finalyzing the partial definitions definitions"
    - adding the pragma PRAGMA_DEFINED_CLOCK_HIERARCHY to the clock systems when it is present in the graph (it specified that the hierarchy is available).
    @return "true" if the construction succeeds, "false" otherwise.
*/
boolean
intemkIGraph::graphFinalyze ()
{
  NM->removingSubGraphs ();
  graphClassSpecificity ();
  propagateTypes ();
  AGRAPH->setInitialValues ();
  AGRAPH->eliminateWhenEvent ();
  // LabelsEliminating(); NM->setDependencesLabels(); removed, LBesnard, Sep 2014
  if (addDefinitionForUndefinedSignals ())
    pK_messenger->warning (warDefForUndefinedSignals);
  // enumNameConflicts (); NOW done on DCG_SCHED ( for code generation.) removed, LBesnard, Sep 2015
  verifyStructBundleDefinitions ();
  delaysFinalizing ();
  if (AGRAPH->delaysRewriting ())
    AGRAPH->eliminateWhenEvent ();
  partialDefinitionFinalize ();
  CKM->setConstraintsMode (existPredefinedPragma (AGRAPH->getPragmas (), PRAGMA_DEFINED_CLOCK_HIERARCHY));
  return pK_messenger->getErrorCode ();
}


/** @defgroup particularContextGroup intemkIGraph: some particular contexts managing.
 */


/** @ingroup particularContextGroup
   Context Clock for constant expression.\n
   @return vhcste if it is not undefined, a new clock variable otherwise and the tree (vtree) is annotated with an error. 
*/
ptinsth
intemkIGraph::instContextClock (ptinsth vhcste, BackAnnotationType vtree)
{
  if (!CKM->satisfiesUndefClock (vhcste))
    {
      if (isInConstantContext ()) return vhcste;
      if (CKM->satisfiesUndefClock (caseContextClock)) return vhcste;
      return CKM->makeInf (caseContextClock, vhcste);
    }
  pK_messenger->errorTree (errcsteint, BACK_ANNOT_GET_ABSTRACT_TREE(vtree));
  return CKM->new_variFormulaRootClock ();
}

/** @ingroup particularContextGroup
   Context Clock for constant expression.\n
   @return vhcste if context mode (isfree) is "true", otherwise it calls instContextClock() method.
   note: isfree is not modified.
*/
ptinsth
intemkIGraph::instsiginvhcste (boolean * isfree, ptinsth vhcste, BackAnnotationType vtree)
{
  if (!(*isfree))
    return instContextClock (vhcste, vtree);
  return vhcste;
}

/**
   @ingroup particularContextGroup
   @return the current context for constant expression, before fixing it (WaitConstantExpr attribute) to b.
   when b, the expressions must be expressions on constants (constant value, parameters, constant identifiers,...)
   otherwise errors are signaled.
*/
boolean
intemkIGraph::getSetConstantContext (boolean b)
{
  boolean vb;
  vb = WaitConstantExpr;
  WaitConstantExpr = b;
  return vb;
}

/**
  @ingroup particularContextGroup
  @return "true" if the expressions must be expressions on constants (constant value, parameters, constant identifiers,...), "false" otherwise.
*/
boolean
intemkIGraph::isInConstantContext ()
{
  return WaitConstantExpr;
}

/**
   @ingroup particularContextGroup
   @return the current context for "state variable" in expression, before fixing it (AcceptStateVar attribute) to b.
   when b, the expressions can contain references to state variables otherwise errors are signaled.
*/
boolean
intemkIGraph::getSetStateVarContext (boolean b)
{
  boolean vb;
  vb = AcceptStateVar;
  AcceptStateVar = b;
  return vb;
}

/**
   @ingroup particularContextGroup
   @return "true" if expressions can contain references to some state variables, "false" otherwise.
*/
boolean
intemkIGraph::isInStateVarContext ()
{
  return AcceptStateVar;
}


/**
   @ingroup globalProcessinGroup
   It updates the names of the signals by 
     - building the list of the undefined and non referenced signals.
     - removing these signals from the graph,
     - modifying the signals with the same prefix than the removed ones.
   @return "true" if some updates are done, "false" otherwise.
*/
boolean
intemkIGraph::updateSignalsNames ()
{
  ptlinstances lident, curlist, ls, lin, lout;
  ptexpcal vsig;
  ptINode node;
  ptinsth h;
  ptident id;
  boolean b;

  if (!EM->getExprList (opident, &lident))
    return pk_false;
  // building the list of the undefined and non referenced signals(ls) : the signals are removed from the graph.
  markSignalsReferencedInExtractions ();
  lin = AGRAPH->getInputs ();
  lout = AGRAPH->getOutputs ();
  ls = NILLINSTANCES;

  while (ptlinstances_get_next (&lident, &vsig))
    if (EM->IsSignal (vsig))
      if (!EM->isMarked (vsig, myMark))
	if (!ptlinstances_isin_exp (lin, vsig))
	  if (!ptlinstances_isin_exp (lout, vsig))
	    {
	      node = EM->getNode (vsig);
	      if (NM->is_undefined (node))
		{
		  b = EM->getExtractionLink (vsig, &h) || NM->isReferenced (node) || (EM->getLuse (vsig) != NILLEXPCAL);
		  if (!b)
		    {
		      h = EM->getClock (vsig);
		      if (!CKM->satisfiesNullClock (h))
			if (removeIdent (vsig, &id))
			  {
			    b = EM->IsEvent (vsig);
			    if (CKM->removeFromClock (h, vsig, b))
			      {
				EM->remove (vsig);
				if (!b)
				  {
				    NAMEM->remove (id);
				    ptlinstances_isin_insert_exp (&ls, &curlist, vsig);
				  }
			      }
			  }
		    }
		}
	    }
  // Cleaning the expressions (marked in markSignalsReferencedInExtractions() method)
  EM->resetMarkAll (myMark);
  // modifying the signals with the same prefix than the removed ones.
  if (ls == NILLINSTANCES)
    return pk_false;
  while (ptlinstances_get_next (&ls, &vsig))
    EM->changeNameSignalSamePrefix (vsig);
  return pk_true;
}

/**
   \copydoc  splitting::mkFieldsDescription().
*/
ptlinstances
intemkIGraph::mkFieldsDescription (ptexpcal exp)
{
  ptlinstances ls;
  ptINode n;

  if (!EM->getFieldsOf (exp, &ls))
    {
      ls = splittingObject->mkFieldsDescription (exp);
      n = EM->getNode (exp);
      n = NM->mk_node_lfields (exp, n, ls, EM->getBackAnnotation (exp));
    }
  return ls;
}


void
intemkIGraph::setAuxiliarySignalName (ptexpcal aSig, ptexpcal exp, noperators vop)
{
  ptident idp, id;

  if (EM->IsAuxiliarySignal (aSig))
    if (getNameAssociatedWithOperator (vop, &idp))
      if (getOriginalName (exp, &id))
	{
	  id = newIdent_concIdentIdent (id, idp);
	  NAMEM->insertInIdentNameTable (id);	// to always add a suffix for each identifier.
	  EM->mkSignalName (aSig, id, pk_true);
	}
}


/**
   @defgroup automatonContextGroup intemkIGraph: automaton managing.
*/

/**
   @ingroup automatonContextGroup
   clk becomes the automaton context clock. 
   It is used to fix the clock of the conditions built on the current state variable associated with an automaton.
   @return the previous one. The user must managed the stack.
*/
ptinsth intemkIGraph::setClockAutomaton (ptinsth clk)
{
  ptinsth pclk;
  pclk = automatonClock;
  automatonClock = clk;
  return pclk;
}

/**
   @ingroup automatonContextGroup

   @return the current automaton context clock.
   It is used to fix the clock of the conditions built on the current state variable associated with an automaton.
*/
ptinsth intemkIGraph::getClockAutomaton ()
{
  return automatonClock;
}


/**
   @ingroup automatonContextGroup
   Sets the current state variable (of an automaton) to vstatevar
*/
void
intemkIGraph::setAutomatonCurrentStateVar (ptexpcal vstatevar)
{
  automaton_current_state = vstatevar;
}

/**
   @ingroup automatonContextGroup
   @return the current state variable (associated with an automaton) previously set by setAutomatonCurrentStateVar().
*/
ptexpcal
intemkIGraph::getAutomatonCurrentStateVar ()
{
  return automaton_current_state;
}

/**
   @ingroup automatonContextGroup
   @return true i the clock automaton has been fixed using the Automaton_clock () intrinsic, "false" otherwise.
*/
boolean
intemkIGraph::isClockedAutomaton()
{
  return automaton_is_clocked;
}

void intemkIGraph::setClockedAutomaton(boolean b)
{
  automaton_is_clocked = b;
}


boolean
intemkIGraph::setInAutomatonIntrinsicContext(boolean b)
{
  boolean bprev=InAutomatonIntrinsicContext;
  InAutomatonIntrinsicContext = b;
  return bprev;
}
