/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** @return the clock composed of a new variable. */
  insthClock ClockSystem::new_variFormulaRootClock (void)
{
  insthClock result;

  result = ptinsth_newClock (vari);
  register_variClockRoot (result);
  return result;
}

/** 
   @return a new condition with c as undersampling expression. 
   It creates the undersampling clock [c] by this expression and its negation [not c].
   The expression (c) 
      - is reduced to a signal if isasignal is "true";
      - its clock is vhorl, 
      - references the signals of the list linst.

   Moreover, it is declared an equivalent formula if isAnequivalent is pk_true (used for bDC+ sub-format).
*/
ptcondition
ClockSystem::new_conditionExtractions (insthClock vhorl, ptexpcal c, ptlinstances linst, boolean isasignal, boolean isAnequivalent)
{
  insthClock vinsth, vinsthns;
  ptcondition aux;

  vinsth = ptinsth_newClock (sub);
  EM->setExtractionLink (c, vinsth);
  vinsthns = ptinsth_newClock (nosub);

  aux = new condition (vinsth, vinsthns, vhorl, c, linst);
  nbConditions++;
  setArgSub (vinsth, aux);
  setReferenceToFrom (vinsth, vhorl);
  setArgSub (vinsthns, aux);
  setReferenceToFrom (vinsthns, vhorl);

  if (isAnequivalent)
    {
      setIsAnEquivalentForm (vinsth, pk_true);
      setIsAnEquivalentForm (vinsthns, pk_true);
    }
  else
    {
      insertlhroots (vinsth);
      insertlhroots (vinsthns);
      lssub = insertlcond (lssub, vinsth, isasignal);
    }
  setComplementary (vinsth, vinsthns);
  return aux;
}

/** - misc -
    @return the clock composed of the h ^- h1 formula where h denotes the mother clock of h1 in
    the hierarchy. It assumes that such a clock exists: h1 is placed in a tree.
*/
insthClock
ClockSystem::new_compClockInTree (insthClock vharg)
{
  insthClock result;
  // it is used only in recur_subterms().
  result = ptinsth_newClock (hcomp);
  setUnaryArg (result, vharg);
  return result;
}

/** @return the clock composed of the lower-bound vh1 ^* vh2 formula. */
ptinsth
ClockSystem::makeInf (insthFormula vh1, insthFormula vh2)
{
  int vres;
  ptinsth vhres;

  vh1 = getRepresentativeFormula (vh1);
  vh2 = getRepresentativeFormula (vh2);

  if (ptinsth_is_hundef (vh1)) return vh2;
  if (ptinsth_is_hundef (vh2)) return vh1;
  if (IsInSimplifyMode)
    {
      if (hredinfsup (vh1, vh2, &vres))
	{
	  if (vres == 0) return vh1;
	  return the_nullClock_CKS_copy;
	}
      if (hredupb (vh1, vh2, &vhres)) return vhres;
      if (hredhnot (vh1, vh2)) return vh2;
      if (hredhnot (vh2, vh1)) return vh1;
    }
 return rewriteInf (getUndefClock (), vh1, vh2);
}

/**
   Rewriting the clock of the expression of the sub-formula vfsub in vhfrom.
   vfsub is assumed to be a sub-formula.
*/
void
ClockSystem::rewriteSub (ptinsth vfsub, ptinsth vhfrom)
{
  /* fourni par Irina mais pas appliqué => recalcul pour des formules déja placées. 
     ptAffineRelation x;
     if ( getAffineRelation(vfsub, &x))
     rewriteAffineSub(vfsub, vhfrom); 
     else
   */
  if (IsInSimplifyMode) replaceArgSub (vfsub, vhfrom);
}

/** 
    For an assumed sub-formula vfsub [e], e is replaced by nexp. Nexp references the signals of the list l.
*/
void
ClockSystem::rewriteCondSub (ptinsth vfsub, ptexpcal nexp, ptlinstances l)
{
  ptexpcal e;
  e = getConditionSub (vfsub);
  EM->setExtractionLink (e, NILINSTH);
  getArgSub (vfsub)->linkExpression (nexp, l);
  EM->setExtractionLink (nexp, vfsub);
}

/** 
    Rewriting in vhfrom of the clock of the negate-formula vfnot.
*/
void
ClockSystem::rewriteNot (ptinsth vfnot, ptinsth vhfrom)
{
  ptinsth otherocc;
  boolean isavar;

  if (IsInSimplifyMode)
   {
     otherocc = removeNot (vfnot, &isavar);
     vfnot = otherocc;
     if (findForce_notFormula (&otherocc, vhfrom))
       hcladdsynchro (otherocc, vfnot);
   }
}

/**
   @return the class of the formula (vharg1 ^* vharg2) build if vfinf=hundef
   rebuild otherwise, and vfinf is removed from its class and the equivalence are announced.
   vfinf is assumed to be a inf-formula.
*/
ptinsth
ClockSystem::rewriteInf (ptinsth vfinf, ptinsth vharg1, ptinsth vharg2)
{
  ptinsth otherterm, vform;
  boolean isavar, bassert;

  if (ptinsth_is_hundef (vfinf))
    {
      findForce_inferFormula (&vfinf, vharg1, vharg2, &vform);
      return vfinf;
    }

  if (IsInSimplifyMode)
    {
      bassert = satisfiesIsAnAssumption (vfinf);
      otherterm = removeInf (vfinf, &isavar);
      if (!bassert)
	{
	  vfinf = otherterm;
	  if (findForce_inferFormula (&vfinf, vharg1, vharg2, &vform))
	    hcladdsynchro (vfinf, otherterm);
	  return vfinf;
	}
      vfinf = hundef;
      findForce_inferFormula (&vfinf, vharg1, vharg2, &vform);
      makeAssumption (vfinf, otherterm);
      return otherterm;
    }
  return vfinf;
}

/**
   @return the class of the formula (vharg1 ^+ vharg2) build if vfsup=hundef
   rebuild otherwise, and vfsup is removed from its class and the equivalence are announced.
   vfsup is assumed to be a sup-formula.
*/
ptinsth
ClockSystem::rewriteSup (ptinsth vfsup, ptinsth vharg1, ptinsth vharg2)
{
  ptinsth otherterm, vform;
  boolean isavar, bassert;

  if (ptinsth_is_hundef (vfsup))
    {
      findForce_superFormula (&vfsup, vharg1, vharg2, &vform);
      return vfsup;
    }
  if (IsInSimplifyMode)
    {
      bassert = satisfiesIsAnAssumption (vfsup);
      otherterm = removeSup (vfsup, &isavar);
      if (!bassert)
	{
	  vfsup = otherterm;
	  if (findForce_superFormula (&vfsup, vharg1, vharg2, &vform))
	    hcladdsynchro (vfsup, otherterm);
	  return vfsup;
	}
      vfsup = hundef;
      findForce_superFormula (&vfsup, vharg1, vharg2, &vform);
      makeAssumption (vfsup, otherterm);
      return otherterm;
    }
  return vfsup;
}
/** 
    It replaces the clock vhsub by vhnew in the system and the equivalences are announced.
    When keep is pk_true, the formula are kept.
*/
void
ClockSystem::replaceSub (ptinsth vhsub, ptinsth vhnew, boolean keep)
{
  boolean isavar, bkill, b;
  ptinsth vhsubn;

  if (vhsub == vhnew) return;
  b = satisfiesIsAnAssumption (vhsub);

  bkill = !keep || satisfiesNullClock (vhnew) || b;
  vhsubn = removeSub (vhsub, bkill, &isavar);
  if (vhsubn == vhnew)
    hcladdmodified (vhnew);
  else if (b)
    makeAssumption (vhnew, vhsubn);
  else
    hcladdsynchro (vhnew, vhsubn);
}

/** 
    It replaces the clock vhsub by vhnew in the system and the equivalences are announced.
    When keep is pk_true, the formula are kept.
*/
void
ClockSystem::replaceSub_null (ptinsth vhsub)
{
  boolean isavar, b;
  ptinsth vhsubn;

  if (vhsub != the_nullClock_CKS_copy)
    {
      b = satisfiesIsAnAssumption (vhsub);
      
      vhsubn = removeSub (vhsub, pk_true, &isavar);
      if (vhsubn == the_nullClock_CKS_copy)
	hcladdmodified (the_nullClock_CKS_copy);
      else if (b)
	makeAssumption (the_nullClock_CKS_copy, vhsubn);
      else
	hcladdsynchro (the_nullClock_CKS_copy, vhsubn);
    }
}

/**
   @return the class vhnew.
   If vfsup,  an assumed sup-formula, is not the "undefined clock", 
   this formula is removed from its class and the equivalence with vhnew is announced.
*/
ptinsth
ClockSystem::replaceSup (ptinsth vfsup, ptinsth vhnew)
{
  boolean isavar, b;
  ptinsth vh;

  if (!ptinsth_is_hundef (vfsup))
    {
      b = satisfiesIsAnAssumption (vfsup);
      vh = removeSup (vfsup, &isavar);
      if (vh == vhnew)
	hcladdmodified (vfsup);
      else if (b)
	makeAssumption (vhnew, vh);
      else
	hcladdsynchro (vhnew, vh);
    }
  return vhnew;
}

/**

   @return the class vhnew.
   If vfinf, an assumed inf-formula, is not the "undefined clock", 
   this formula is removed from its class and the equivalence with vhnew is announced.
*/
ptinsth
ClockSystem::replaceInf (ptinsth vfinf, ptinsth vhnew)
{
  boolean isavar, b;
  ptinsth vh;

  if (vfinf != vhnew)
    if (!ptinsth_is_hundef (vfinf))
      {
	b = satisfiesIsAnAssumption (vfinf);
	vh = removeInf (vfinf, &isavar);
	if (!satisfiesNullClock (vhnew))
	  if (vh == vhnew)
	    hcladdmodified (vfinf);
	  else if (b)
	    makeAssumption (vhnew, vh);
	  else
	    hcladdsynchro (vhnew, vh);
	else
	  hclunionwithhnull (vfinf);
      }
  return vhnew;
}

/** It merges a clock (vh1) with the null clock and realizes induced merges.
 */
void
ClockSystem::hclunionwithhnull (ptinsth vh1)
{
  ptlinsth lh, lhs;
  ptinsth vh;

  // Added Nov 2014. h1 ^+ h2 ^= ^0 => h1 ^=  ^0  and h2 ^= ^0.
  lh = getTermsUpperBound (vh1, NILLINSTH);
  if (lh != NILLINSTH)
    {
      lhs = lh;
      while (ptlinsth_get_next (&lh, &vh))
	if (vh != vh1)
	  unionwithhnull (vh);
      ptlinsth_free (lhs);
    }				// --end  Added Nov 2014
  unionwithhnull (vh1);
}

/**
   @return 
     - false if it does not exists some equivalences to do, 
     - true otherwise, in this case it builds this equivalences and the induced equivalences.
*/
boolean
ClockSystem::hclmklequiv0 ()
{
  ptinsth vh1, vh2;

  if (!ModeSynchroSolving)
    return pk_false;
  if (!hclsomesynchro (&vh1, &vh2))
    return pk_false;
  do
    {
      vh1 = hclmkunion0 (vh1, vh2);
    }
  while (hclsomesynchro (&vh1, &vh2));
  return pk_true;
}

/** Formula.
    The representative formula of the class {clk, bclk, ...} becomes bclk.
    This clock is a child of father clock.
*/
void
ClockSystem::switchFormula (ptinsth clk, ptinsth bclk, ptinsth father)
{
  substitute_clhead (clk, bclk);
  setIsAnEquivalentForm (clk, pk_true);

  setIsAnEquivalentForm (bclk, pk_false);
  removeChild (father, clk);
  addChild (father, bclk);
  fusionInfosGraph (clk, bclk);
  fusion_pragmas (clk, bclk);
}

/**
   It replaces a formula (vh1) by another (vh2) for each formula of a list (luse).
*/
void
ClockSystem::replaceForm (ptinsth vh1, ptinsth vh2, ptlinsth luse)
{
  ptinsth use, nf, h1, h2;

  while (ptlinsth_get_next (&luse, &use))
    {
      if (getDepth (use) == 0)

	switch ((int) (getOperator (use)))
	{
	case hnot:
	  rewriteNot (use, vh2);
	  break;
	case infer:
	  getBinaryArgs (use, &h1, &h2);
	  h1 = getRepresentativeFormula (h1);
	  h2 = getRepresentativeFormula (h2);
	  if (h1 == vh1)
	    replaceInf (use, makeInf (vh2, h2));
	  else
	    if (h2 == vh1) 
	      replaceInf (use, makeInf (h1, vh2));
	  break;
	case super:
	  getBinaryArgs (use, &h1, &h2);
	  h1 = getRepresentativeFormula (h1);
	  h2 = getRepresentativeFormula (h2);
	  if (h1 == vh1)
	    replaceSup (use, makeSup (vh2, h2));
	  else
	    if (h2 == vh1)
	      replaceSup (use, makeSup (h1, vh2));
	  break;
	  
	case sub:
	case nosub:
	case hcomp:
	case vari:
	case predef:
	  break;
	default:
	  pK_messenger->Caseerror (__LINE__, __FILE__);
	}
    }
}

/**
   It realizes the synchronisations that are not marked by m.
   (Currently used for assertions managing).
*/
void
ClockSystem::makeSynchrosUnmarked (typeMark m)
{
  ptinsth vh1, vh2;
  ptlinsth lh1, lh2, l1, l2;

  if (lterm1equ != NILLINSTH)
    {
      lh1 = lterm1equ;
      lh2 = lterm2equ;
      lterm1equ = NILLINSTH;
      lterm2equ = NILLINSTH;
      l1 = NILLINSTH;
      l2 = NILLINSTH;
      while (ptlinsth_get_next (&lh1, &vh1))
	{
	  ptlinsth_get_next (&lh2, &vh2);
	  vh1 = getRepresentativeFormula (vh1);
	  vh2 = getRepresentativeFormula (vh2);
	  if (!satisfiesMarked (vh1, m) && !satisfiesMarked (vh2, m))
	    makeSynchro (vh1, vh2);
	  else
	    {
	      l1 = ptlinsth_insert_elem (l1, vh1);
	      l2 = ptlinsth_insert_elem (l2, vh2);
	    }
	}
      lterm1equ = l1;
      lterm2equ = l2;
    }
}

/**
   It returns the synchrozisations not yet solved (when ModeSynchroSolving is "false" for example) in lh1, lh2.
   We have lh1 = { h1,...., hN,...,hv}, lh2 = { k1,...., kN,...kv}, and hi is equivalent to ki.
   (Currently used for assertions managing).
*/
void
ClockSystem::getSuspendedSynchros (ptlinsth * lh1, ptlinsth * lh2)
{
  *lh1 = lterm1equ;
  *lh2 = lterm2equ;
}

/**
   It removes a clock (clk) from the system by :
     - updating root clocks or placed clocks sub-set,
     - updating the formula sub-set.
     Informations are kept for tracability.
*/
void
ClockSystem::removeClock (ptinsth clk)
{
  pkClockType vop;
  boolean b;
  ptcondition vcond;
  ptinsth f;

  if (!removelhroots (clk))
    if (!satisfiesRoot (clk))
      removelhplaced (clk, pk_false);
  // But the BDD manager is not updated.
  f = getFather (clk);
  if (f != clk)
    removeChild (f, clk);
  setHierarchy (clk, clk, clk, 0);	// Added Sept 2013 (LBesnard)
  vop = getOperator (clk);
  switch (vop)
    {
    case nosub:
    case sub:
      if (!dlcond (&lssub, clk, &vcond))
	dlcond (&lsextract, clk, &vcond);
      removeSubLink (clk, vcond);
      break;
    case hcomp:
      /* ?????? */
      break;
    case infer:
      lsinfer = ptlinsth_remove_elem (lsinfer, clk, &b);
      if (!b)
	removeSortedList (clk, &lsinfintree, &lsinfintreeEnd);
      removeReferencesBinaryFormula (clk);
      break;
    case super:
      lssuper = ptlinsth_remove_elem (lssuper, clk, &b);
      if (!b)
	removeSortedList (clk, &lssupintree, &lssupintreeEnd);
      removeReferencesBinaryFormula (clk);
      break;
    case vari:
      removeVari (clk);
      break;
    default:			//predef, troot, hnot, equ
      break;
    }
}

/** Removing an EXTRACTION formula.
   It cleans the use of its arguments and the internal lists.
   @param vform an assumed sub-formula to remove.
   @param killFormula (in) the formula vform is removed of the class, otherwise it is 
   kept as an equivalent.
   @param becomesvari (out) at pk_true when the clock becomes an empty set of formula.
   @return the class of vform.
*/
ptinsth
ClockSystem::removeSub (ptinsth vform, boolean killFormula, boolean * becomesvari)
{
  ptcondition vcond;
  ptinsth vh;

  removelhroots (vform);
  if (!dlcond (&lssub, vform, &vcond))
    dlcond (&lsextract, vform, &vcond);
  if (killFormula)
    {
      vh = removeSubFormulaFromClass (vform, vcond, becomesvari);
      EM->setExtractionLink (vcond->get_condsub (), NILINSTH);
      if (*becomesvari)
	{
	  setOperator (vh, vari);
	  register_variClockRoot (vh);
	}
      return vh;
    }
  setIsAnEquivalentForm (vform, pk_true);
  return vform;
}

/** 
    @return
    - false if the clock (vh) of a signal (sig) is the null clock. In this case, the link signal/clock is preserved.
    - true otherwise and the link is destroyed : the signal is removed from the set of the signals assigned to vh.
      If there is no references to this clock, it is removed from the system.
      
      if bk, the signal (sig) is kept in a specific set; its name could be used for the name of the clock (vh).
*/
boolean
ClockSystem::removeFromClock (ptinsth vh, ptexpcal sig, boolean bk)
{
  ptlinstances lsvh;

  if (satisfiesNullClock (vh))
    return pk_false;

  removeSignalFromClass (vh, sig);
  lsvh = getSynchronousSignals (vh);

  // Removing the clock if it is not referenced.
  if (lsvh == NILLINSTANCES)
    if (satisfiesVari (vh))
      if (getClockReferencesTo (vh) == NILLINSTH)
	removeVari (vh);

  if (bk && EM->IsSignal (sig))
    {
      // saving the signal for tracability.
      lsvh = getSynchronousSignalsKept (vh);
      lsvh = ptlinstances_insert (lsvh, sig);
      setSynchronousSignalsKept (vh, lsvh);
    }
  return pk_true;
}

/** @return true if a clock (vh) is a free clock (i.e a variable without any constraints)
    false otherwise.*/
boolean
ClockSystem::isFreeClock (ptinsth vh)
{
  boolean bcontrainst;
  ptlinsth luse;
  ptinsth use;

  if (!satisfiesVari (vh))
    return pk_false;
  luse = getClockReferencesTo (vh);
  bcontrainst = pk_false;
  while (!bcontrainst && ptlinsth_get_next (&luse, &use))
    bcontrainst = isRecursiveFormula (use);
  return !bcontrainst;
}

_PUBLIC_
/**
  Undersampling by a positive condition.
  @return the clock in which the sub-formula that represents the set of the value "true" of the 
  condition c with c ^= vh, is an element. It assumes that c is a positive expression which references the signals given in the list pred.
  This sub-formula is noted [c] and the following properties hold:
    - [c] ^+ [not c] = ^c
    - [c] ^* [not c] = ^0
    \see makeNoSub
*/
  ptinsth ClockSystem::makeSub (insthFormula vh, ptexpcal c, ptlinstances pred)
{
  ptinsth
    vhaux,
    vh1;
  ptcondition
    vcond;

  vh = getRepresentativeFormula (vh);
  if (satisfiesNullClock (vh))
    return vh;
  if (rchlsub (pk_true, c, &vhaux))
    {
      if ((pred != NILLINSTANCES))
	return vhaux;
      /* Une constante est disponible a la borne sup de ses utilisations */
      vh1 = getClockSub (vhaux);
      if (ptinsth_is_hundef (vh1))
	rewriteSub (vhaux, vh);
      else
	rewriteSub (vhaux, makeSup (vh1, vh));
      return vhaux;
    }
  vcond = new_conditionExtractions (vh, c, pred, EM->IsSignal (c), pk_false);	// Il faudrait supprimer EM->IsSignal(c)
  return vcond->get_clksub ();
}

/**
  Undersampling by a negative condition.
  @return the clock in which the sub-formula, that represents the set of the value "false" of the 
  condition c with c ^= vh, is an element. It assumes that c is a positive expression.
  This sub-formula is noted [not c].
  \see makeSub
*/

ptinsth
ClockSystem::makeNoSub (ptexpcal c)
{
  ptinsth vhaux;

  if (rchlsub (pk_false, c, &vhaux))
    return vhaux;
  return the_nullClock_CKS_copy;
}

/** @return the clock composed of the upper-bound vh1 ^+ vh2 formula. */
ptinsth
ClockSystem::makeSup (ptinsth vh1, ptinsth vh2)
{
  int vres;
  ptinsth vh11, vh12;

  vh1 = getRepresentativeFormula (vh1);
  vh2 = getRepresentativeFormula (vh2);
  if (ptinsth_is_hundef (vh1)) return vh2;
  if (ptinsth_is_hundef (vh2)) return vh1;
  if (satisfiesInitClock (vh1)) return vh2;
  if (satisfiesInitClock (vh2)) return vh1;

  if (IsInSimplifyMode)
    {
      if (hredinfsup (vh1, vh2, &vres))
	{
	  if (vres < 0)
	    return vh2;
	  return vh1;
	}
      if (hredupb (vh1, vh2, &vh11))
	return getTickClock ();
      if (getMinusArgs (vh1, &vh11, &vh12))
	{
	  if (vh11 == vh2)
	    return makeSup (vh12, vh2);
	  return rewriteSup (getUndefClock (), vh1, vh2);
	}
      if (getMinusArgs (vh2, &vh11, &vh12))
	{
	  if (vh11 == vh1)
	    return makeSup (vh12, vh1);
	  return rewriteSup (getUndefClock (), vh1, vh2);
	}
    }
  return rewriteSup (getUndefClock (), vh1, vh2);
}

/** @return the class of -(vh). */
ptinsth
ClockSystem::getForce_notClock (insthFormula vh)
{
  ptinsth vhaux;

  vh = getRepresentativeFormula (vh);
  vhaux = hundef;
  findForce_notFormula (&vhaux, vh);
  return vhaux;
}

/** @return the clock composed of the vh1 ^- vh2 formula.
    Applied elementary reduction:
       - vh1 ^- vh1 -> ^0
*/
ptinsth
ClockSystem::makeMinus (ptinsth vh1, ptinsth vh2)
{
  vh1 = getRepresentativeFormula (vh1);
  vh2 = getRepresentativeFormula (vh2);
  if (vh1 == vh2)
    return the_nullClock_CKS_copy;
  return makeInf (vh1, getForce_notClock (vh2));
}

/** @return the clock obtained by the merge of vh1 and vh2.
    a representative formula is chosen.
 */
ptinsth
ClockSystem::makeSynchro (ptinsth vh1, ptinsth vh2)
{
  if (ptinsth_is_hundef (vh1)) return vh2;
  if (ptinsth_is_hundef (vh2)) return vh1;
  vh1 = hclmkunion0 (vh1, vh2);
  hclmklequiv0 ();
  return getRepresentativeFormula (vh1);
}

/** @return the clock obtained by the merge of vh1 and vh2.
    This constraint is assumed satisfied (assertion).
*/
ptinsth
ClockSystem::makeAssumption (ptinsth vh1, ptinsth vh2)
{
  if (vh1 == vh2) return vh1;
  if (ptinsth_is_hundef (vh1)) return vh2;
  if (ptinsth_is_hundef (vh2)) return vh1;
  // if (ptinsth_satisfies_predef (vh1)) return vh1; removed Apr 2016 LBesnard.

  if (getDepth (vh1) > 0)
    {
      setAssumption (vh2, pk_true);
      return makeSynchro (vh2, vh1);
    }
  else
    {
      setAssumption (vh1, pk_true);
      return makeSynchro (vh1, vh2);
    }
}

/**
   Adds in the clock system the assertions:
   - (| when vcond ^= vh | when not vcond ^=0 |) when bpos, 
   - (| when not vcond ^= vh | when vcond ^=0 |) when not bpos.

   vh is assumed to be the clock of the condition (vcond).
*/
void
ClockSystem::makeAssert (ptinsth vh, ptexpcal vcond, boolean bpos, ptinsth * vhsub, ptinsth * vhnosub)
{
  *vhsub = makeSub (vh, vcond, EM->getPredecessors (vcond));
  *vhnosub = makeNoSub (vcond);
  if (bpos)
    {
      makeAssumption (*vhsub, vh);
      makeAssumption (*vhnosub, the_nullClock_CKS_copy);
    }
  else
    {
      makeAssumption (*vhsub, the_nullClock_CKS_copy);
      makeAssumption (*vhnosub, vh);
    }
}

/** @return the class obtained by the merge of vh1 and vh2.
    vh1 is a proved equivalent formula of vh2.
*/
ptinsth
ClockSystem::makeEquivForm (ptinsth vh1, ptinsth vh2)
{
  removelhroots (vh1);
  setIsAnEquivalentForm (vh1, pk_true);
  return makeSynchro (vh1, vh2);
}

/** @return the class obtained by the merge of vh1 and [c], with 
      - c a positive condition.
      - hc ^= c
      - [c] is a proved equivalent formula of vh1.
      - vlist list of  the signals referenced in the condition c.
    \note these kept formula are used for interformat translations.
*/
ptinsth
ClockSystem::makeEquivForm (ptinsth vh1, ptexpcal c, ptlinstances vlist, ptinsth hc)
{
  ptinsth vhresult, vh;
  ptcondition vcond;

  hc = getRepresentativeFormula (hc);
  if (satisfiesNullClock (hc))
    return hc;
  if (!rchlsub (pk_true, c, &vhresult))
    {
      vcond = new_conditionExtractions (hc, c, vlist, pk_true, pk_true);
      vhresult = vcond->get_clksub ();
      vh = makeSynchro (vh1, vhresult);
    }
  else
    vhresult = makeEquivForm (vh1, vhresult);
  return vhresult;
}

/**
   It places as roots the set of unplaced clocks.
 */
void
ClockSystem::setClocksAsRoots ()
{
  ptinsth vh, vh1;
  ptlinsth list;

  list = getFormulaList (sub);
  while (ptlinsth_get_next (&list, &vh1))
    {
      addClockAsRoot (vh1);
      getComplInFather (vh1, &vh);
      addClockAsRoot (vh);
    }

  list = getFormulaList (vari);
  while (ptlinsth_get_next (&list, &vh))
    {
      vh1 = getRepresentativeFormula (vh);
      if (vh == vh1)
	addClockAsRoot (vh);
    }

  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
    addClockAsRoot (vh);

  list = getFormulaList (super);
  while (ptlinsth_get_next (&list, &vh))
    addClockAsRoot (vh);

}

/**
   It realizes the fusion of two clocks vhold and vhnew. This induces the fusion of:
   the formula, the lists of synchronous signals, the sub-graphs of nodes assigned to each clock.
*/
void
ClockSystem::hclfusion (ptinsth vhold, ptinsth vhnew)
{
  removelhroots (vhold);
  if (!satisfiesRoot (vhold))
    {
      if (getDepth (vhold) != 0)
	removelhplaced (vhold, pk_false);
      addClockAsRoot (vhnew);
    }
  fusionclass (vhnew, vhold);
  fusionInfosGraph (vhold, vhnew);
  fusion_pragmas (vhold, vhnew);
}

/**
  @return true if the formula VH2 has not the priority on vh1 for the choice of a representative element.
  remark : a placed formula has always the priority on a non-placed one.
*/
boolean
ClockSystem::isFormulaBeforeFormula (ptinsth vh1, ptinsth vh2)
{
  if (getDepth (vh1) > 0)
    {
      if (getDepth (vh2) > 0)
	return satisfies_order_sametree (vh1, vh2);
      return pk_true;
    }
  if (getDepth (vh2) > 0)
    return pk_false;
  if (isFictitiousRootFormula (vh1))
    return pk_true;
  if (isFictitiousRootFormula (vh2))
    return pk_false;
  return satisfies_order_sametree (vh1, vh2);
}


boolean
ClockSystem::free_formula (ptinsth vh)
{
  pkClockType op;

  op = getOperator (vh);
  if (op == predef)
    return pk_false;
  if (op == sub)
    delete getArgSub (vh);
  ptinsth_free (vh);
  return pk_true;
}
