
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Sets to a clock (clk) its hierarchy informations: its father clock (hfather), its root (hroot) and its depth (depth).
*/
  void
ClockSystem::setHierarchy (ptinsth clk, ptinsth hfather, ptinsth hroot, int depth)
{
  setDepth (clk, depth);
  setFather (clk, hfather);
  setRoot (clk, hroot);
}

/**
   Ordering formulas (vh1, vh2) by increasing number.
 */
void
ClockSystem::setorder_from_num (ptinsth * vh1, ptinsth * vh2)
{
  ptinsth vh;
  if (!satisfies_order_num (*vh1, *vh2))
    {
      vh = *vh1;
      *vh1 = *vh2;
      *vh2 = vh;
    }
}

/** @return 
    - pk_true if the clocks vh1 and vh2 are placed.
    - pk_false, not placed and they are re-ordered (null clock first)
*/
boolean
ClockSystem::sortFormula (ptinsth * vh1, ptinsth * vh2)
{
  if (satisfiesNullClock (*vh2))
    {
      *vh2 = *vh1;
      *vh1 = the_nullClock_CKS_copy;
      return pk_false;
    }
  if (satisfiesNullClock (*vh1))
    return pk_false;
  if (satisfiesRoot (*vh1))
    {
      setorder_from_num (vh1, vh2);
      return pk_false;
    }
  if (satisfiesRoot (*vh2))
    {
      setorder_from_num (vh1, vh2);
      return pk_false;
    }
  return pk_true;
}

/**
   It removes a condition from the system.
   @return true if the condition of the clock (assumed to be an extraction) is in the list vlcond, false otherwise.
   When it is present, the condition is removed from the list and returned in output vcondition parameter.
*/
boolean
ClockSystem::dlcond (ptlinsth * vlcond, ptinsth velem, ptcondition * vcondition)
{
  *vcondition = getArgSub (velem);
  return ptlinsth_remove_condition (vlcond, *vcondition);
}

/**
   List managing when the beginning of the list and the end of the list are known (added for optimizing clock calculus).
   It adds a formula (vh) in such a list (lbegin, lend).
*/
void
ClockSystem::insertSortedList (ptinsth vh, ptlinsth * lbegin, ptlinsth * lend)
{
  boolean b;
  int n, m;
  ptinsth x;

  if ((*lend) != NILLINSTH)
    {
      x = ptlinsth_get_elem (*lend);
      n = getNum (vh);
      m = getNum (x);
      if (n != m)
	if (n > m)
	  *lend = ptlinsth_insert_post (*lend, vh);
	else
	  *lbegin = ptlinsth_sorted_insert_elem_bool (*lbegin, vh, &b);
    }
  else
    {
      *lbegin = ptlinsth_sorted_insert_elem_bool (*lbegin, vh, &b);
      *lend = *lbegin;
    }
}

/**
   List managing when the beginning of the list and the end of the list are known (added for optimizing clock calculus).
   It removes a formula (vh) from such a list (lbegin, lend).
*/
boolean
ClockSystem::removeSortedList (ptinsth vh, ptlinsth * lbegin, ptlinsth * lend)
{
  boolean b;
  int n, m;
  ptinsth x;

  if ((*lend) != NILLINSTH)
    {
      x = ptlinsth_get_elem (*lend);
      n = getNum (vh);
      m = getNum (x);
      if (n == m)
	{
	  *lbegin = ptlinsth_remove_last_elem (*lbegin, lend);
	  return pk_true;
	}
    }
  *lbegin = ptlinsth_remove_elem (*lbegin, vh, &b);
  return b;
}

void
ClockSystem::clean_variableClockList ()
{
  ptlinsth list;
  ptinsth vh;
  list = lvariables;
  while (ptlinsth_get_next (&list, &vh))
    if (vh != getRepresentativeFormula (vh))
      removeVari (vh);
}

/**
   System management.
   @return 
   - pk_true if the formula -vharg is yet registerd. Then, vhresult (out) is the class of the found formula.
   - false otherwise.
*/
boolean
ClockSystem::find_notClockFormula (ptinsth * vhresult, ptinsth vharg)
{
  ptlinsth aux;
  ptinsth velem, vh1;
  aux = lsnot;
  while (ptlinsth_get_next (&aux, &velem))
    {
      (void) ptinsth_get_1arg (velem, &vh1);
      if (vharg == vh1)
	{
	  *vhresult = getRepresentativeFormula (velem);
	  return pk_true;
	}
    }
  return pk_false;
}

/**
   System management.
   @return 
   - pk_true if the formula op(vh1, vh2) is in the list VLHINFSUP. Then, vhresult (out) is the class of the found formula.
   - false otherwise.
*/
boolean
ClockSystem::find_binaryClockFormula (ptlinsth vlhinfsup, ptinsth vh1, ptinsth vh2, ptinsth * vhresult)
{
  ptinsth velem, vharg1, vharg2;

  while (ptlinsth_get_next (&vlhinfsup, &velem))
    {
      getBinaryArgs (velem, &vharg1, &vharg2);
      if ((vh1 == vharg1) && (vh2 == vharg2))
	{
	  *vhresult = getRepresentativeFormula (velem);
	  return pk_true;
	}
    }
  return pk_false;
}

/**
   System management.
   @return true if the formula op(vh1, vh2) exists using the references to vh1.
   @return false if the formula op(vh1, vh2) does not exist using the references to vh1.
   @param(vhresult) is assigned the existing formula if any; it remains unchanged otherwise
*/
boolean
ClockSystem::rchFormula (pkClockType op, ptinsth vh1, ptinsth vh2, ptinsth * vhresult)
{
  ptinsth velem, vharg1, vharg2;
  ptlinsth vlhinfsup;

  vlhinfsup = getClockReferencesTo (vh1);
  while (ptlinsth_get_next (&vlhinfsup, &velem))
    if (getOperator (velem) == op)
      {
	getBinaryArgs (velem, &vharg1, &vharg2);
	if ((vh1 == vharg1) && (vh2 == vharg2))
	  {
	    *vhresult = getRepresentativeFormula (velem);
	    return pk_true;
	  }
      }
  return pk_false;
}

/** 
    System management.

    @return true if an expression (cond) is argument of an extraction clock on condition (velem), 
    of a list of extraction clocks (list).
    Assumed: list is an extraction clocks list.
*/
boolean
ClockSystem::recherche (ptlinsth list, ptexpcal cond, ptcondition * velem)
{
  ptlinsth aux;
  ptinsth vh;

  aux = list;
  while (ptlinsth_get_next (&aux, &vh))
    if (EM->isEquivalent (cond, getConditionSub (vh), pk_false))
      {
	*velem = getArgSub (vh);
	return pk_true;
      }
  return pk_false;
}

/**
   It inserts an extraction  clock (vhsub) in the list vlsub.
   Assumed : vhsub is an extraction clock.
*/
void
ClockSystem::register_variClockRoot (ptinsth vh)
{
  insertSortedList (vh, &lvariables, &lvariablesEnd);
  insertlhroots (vh);
}

// lssub, lsinfer, lssuper, lsnot

/**
   It inserts an extraction  clock (vhsub) in the list vlsub.
   Assumed : vhsub is an extraction clock.
   @return the modified list.
*/
ptlinsth
ClockSystem::insertlcond (ptlinsth vlsub, ptinsth vhsub)
{
  if (vlsub == NILLINSTH)
    return ptlinsth_insert_elem (NILLINSTH, vhsub);
  return ptlinsth_insert_elem (vlsub, vhsub);
}

/**
   It inserts an extraction  clock (vhsub) at the end of the list (vlsub) if isasignal or at the beginning of the list otherwise.
   Assumed : vhsub is an extraction clock.
   @return the modified list.
*/
ptlinsth
ClockSystem::insertlcond (ptlinsth vlsub, ptinsth vhsub, boolean isasignal)
{
  if (vlsub == NILLINSTH)
    return ptlinsth_insert_elem (NILLINSTH, vhsub);
  if (!isasignal)
    return ptlinsth_insert_elem (vlsub, vhsub);
  ptlinsth_append (ptlinsth_get_last (vlsub), ptlinsth_insert_elem (NILLINSTH, vhsub));
  return vlsub;
}

/**
   It removes a condition from the system.
   @return true if the condition vcondition is in the list vlsub, 
   in this case it is removed from it, false otherwise.
*/
boolean
ClockSystem::ptlinsth_remove_condition (ptlinsth * vlsub, ptcondition vcondition)
{
  boolean found;
  ptlinsth pred, liste, aux;
  ptinsth vh;

  if ((*vlsub) == NILLINSTH)
    return pk_false;
  liste = *vlsub;
  aux = *vlsub;
  pred = NILLINSTH;
  found = pk_false;
  while (!found)
    {
      if (!ptlinsth_get_next (&aux, &vh))
	return pk_false;
      if (getArgSub (vh) == vcondition)
	found = pk_true;
      else
	{
	  pred = liste;
	  liste = aux;
	}
    }
  if (pred == NILLINSTH)
    *vlsub = aux;
  else
    ptlinsth_append (pred, aux);
  return found;
}

/** 
    When 
      - b is pk_true, a new clock is created with a formula with vop as type operator.
      - b is pk_false, it adds a formula, with vop as type operator, in the class of vhcl. The type operator of vhcl is 
      modified if it is a free clock.

      @return the class of the added formula.
*/

/** creates a new formula of type (operator) vop; 
                   this formula is unstable since arguments are not set
    if the formula of (* vhcl) is undefined the newly created formula is made a clock returned in (* vhcl)
    otherwise if the formula of (* vhcl) is a vari, it is replaced by the new formula
            provided that (* vhcl) is a root and is not the meta root
    @return the unstable  created formula.
*/
insthFormula
ClockSystem::insert_newFormula (insthClock * vhcl, pkClockType vop)
{
  insthFormula vform;

  if (ptinsth_is_hundef (*vhcl))
    {
      *vhcl = ptinsth_newClock (vop);
      return (insthFormula) (*vhcl);
    }
  if (getOperator (*vhcl) == vari)
    if (satisfiesNullDepth (*vhcl) && !isFictitiousRootFormula (*vhcl))
      {
	setOperator (*vhcl, vop);
	if (vop != vari) removeVari (*vhcl);
	return (insthFormula) (*vhcl);
      }
  vform = ptinsth_newClock (vop);
  ptinsth_insert_form (*vhcl, vform);
  return vform;
}


/** @return True if vh is a recursive form in vform, false otherwise.
    Local to \ref ishodefrec. */
boolean
ClockSystem::ishodefrecin (ptinsth vh, ptinsth vform)
{
  int p;
  ptinsth vh1;

  if (vh == vform) return pk_true;
  vh1 = getRoot (vh, &p);
  if (!satisfiesFictitiousRoot (vh1)) vh = vh1;
  if (satisfiesMarked (vh, mark3)) return pk_false;
  addMark (vh, mark3);
  return ishodefrec (vh, vform);
}

/** @return true if vh is a recursive form in vform, false otherwise.
 */
boolean
ClockSystem::ishodefrec (ptinsth vh, ptinsth vform)
{
  ptinsth vh1, vh2;

  switch (getOperator (vh))
    {
    case sub:
    case nosub:
      return ishodefrecin (getClockSub (vh), vform);

    case hcomp:
      return ishodefrecin (getFather (vh), vform);

    case hnot:
      getUnaryArg (vh, &vh1);
      return ishodefrecin (getRepresentativeFormula (vh1), vform);

    case infer:
    case super:
      getBinaryArgs (vh, &vh1, &vh2);
      if (ishodefrecin (getRepresentativeFormula (vh1), vform))
	return pk_true;
      return ishodefrecin (getRepresentativeFormula (vh2), vform);

    default:;
    }
  return pk_false;
}

/** 
   It rewrites the formula of the list (luse) by replacing the references to vh1 by the null clock.
*/
void
ClockSystem::setlusetohnull (ptinsth vh1, ptlinsth luse)
{
  ptinsth use, otherocc, vhs1, vhs2;
  ptlinsth lusenot;
  boolean b;

  // ptinsth_DEBUG_traceClock("entering ClockSystem::setlusetohnull vh1", vh1);
  while (ptlinsth_get_next (&luse, &use))
    {
      // ptinsth_DEBUG_traceClock("     in  ClockSystem::setlusetohnull while use", use);
      switch ((int) (getOperator (use)))
	{
	case sub:
	case nosub:
	  if (!satisfiesNullDepth (use))
	    ptlinsth_sorted_remove_elem (&lsnullpl, use);
	  vhs1 = removeSub (use, pk_true, &b);
	  // ptinsth_DEBUG_traceClock("     in  ClockSystem::setlusetohnull sub, nosub vhs1", vhs1);
	  hcladdhnull (vhs1, b);
	  break;
	case hnot:
	  lusenot = getClockReferencesTo (use);
	  otherocc = removeNot (use, &b);
	  while (lusenot != NILLINSTH)
	    {
	      ptlinsth_get_next (&lusenot, &otherocc);
	      vhs1 = otherArgOf (otherocc, use);
	      vhs2 = removeInf (otherocc, &b);
	      hcladdsynchro (vhs1, vhs2);
	    }
	  break;
	case infer:
	  if (!satisfiesNullDepth (use))
	    ptlinsth_sorted_remove_elem (&lsnullpl, use);
	  vhs1 = removeInf (use, &b);
	  // ptinsth_DEBUG_traceClock("     in  ClockSystem::setlusetohnull infer vhs1", vhs1);
	  hcladdhnull (vhs1, b);
	  break;
	case super:
	  vhs1 = otherArgOf (use, vh1);
	  vhs2 = removeSup (use, &b);
	  hcladdsynchro (vhs1, vhs2);
	  break;

	case hcomp:
	case vari:
	case predef:
	  break;
	default:
	  pK_messenger->Caseerror (__LINE__, __FILE__);
	}
    }
}

/**
   @return the class result of the merge of two classes (x1 and x2).
   It realizes recursively the induced merging.
   Warning: the boolean reductions are not applied. It is used during the creation of the clock system (only elementary reductions are appled).
   \see {hclmkunion(), mergingClasses0(), hclunionwithhnull()}
*/
ptinsth
ClockSystem::hclmkunion0 (ptinsth x1, ptinsth x2)
{
  ptinsth vhs1, vh1, vh2;

  vh1 = getRepresentativeFormula (x1);
  vh2 = getRepresentativeFormula (x2);
  if (vh1 == vh2)
    {
      return vh1;
    }

  if (satisfiesIsAnAssumption (x1))
    {
      setAssumption (x2, pk_true);
      setAssumption (vh2, pk_true);
    }
  else if (satisfiesIsAnAssumption (x2))
    {
      setAssumption (x1, pk_true);
      setAssumption (vh1, pk_true);
    }

  if (satisfiesNullClock (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, the_nullClock_CKS_copy, pk_false);
      hclunionwithhnull (vh2);
      return the_nullClock_CKS_copy;
    }
  if (satisfiesNullClock (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, the_nullClock_CKS_copy, pk_false);
      hclunionwithhnull (vh1);
      return the_nullClock_CKS_copy;
    }

  if (satisfiesIsAnEquivalentForm (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      hclfusion (vh1, vh2);
      return vh2;
    }

  if (satisfiesIsAnEquivalentForm (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      hclfusion (vh2, vh1);
      return vh1;
    }
  // if [C] ^= [not C] then ^0 ^= ^C
  if (satisfiesSub (vh1))
    if (getComplExtraction (vh1, &vhs1))
      if (vh2 == vhs1)
	{
	  hclunionwithhnull (getClockSub (vh1));
	  return the_nullClock_CKS_copy;
	}
  // ------
  if (existHierarchyConstraintsOn (vh2, vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_true);
      addConstraintPlacedClocks (vh1, vh2);
      return vh1;
    }
  // Assumption managing.
  if (satisfiesIsAnAssumption (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      if (!satisfiesNullDepth (vh1))
	{
	  mergingClasses0 (vh1, vh2);
	  {
	    return vh1;
	  }
	}
      mergingClasses0 (vh2, vh1);
      return vh2;
    }

  if (satisfiesIsAnAssumption (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      if (!satisfiesNullDepth (vh2))
	{
	  mergingClasses0 (vh2, vh1);
	  {
	    return vh2;
	  }
	}
      mergingClasses0 (vh1, vh2);
      return vh1;
    }

  if (isFictitiousRootFormula (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      mergingClasses0 (vh1, vh2);
      return vh1;
    }

  if (isFictitiousRootFormula (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      mergingClasses0 (vh2, vh1);
      return vh2;
    }
  if (satisfiesVari (vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      mergingClasses0 (vh1, vh2);
      removeVari (vh2);
      return vh1;
    }
  if (satisfiesVari (vh1))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
      mergingClasses0 (vh2, vh1);
      removeVari (vh1);
      return vh2;
    }

  if (!ModeSynchroSolving)
    {
      hcladdsynchro (vh1, vh2);
      return vh1;
    }

  if (isFormulaBeforeFormula (vh1, vh2))
    {
      PRINTCLOCKSYNCHRO (vh1, vh2, vh1, pk_false);
      mergingClasses0 (vh1, vh2);
      return vh1;
    }
  PRINTCLOCKSYNCHRO (vh1, vh2, vh2, pk_false);
  mergingClasses0 (vh2, vh1);
  return vh2;
}

ptinsth ClockSystem::rewriteInf0(ptinsth use, ptinsth vh1, ptinsth vh2)
{
  boolean isavar;
  if ((vh1 == vh2) &&  (vh1 == use)) return removeInf(use, &isavar);
  return rewriteInf(use, vh1, vh2);  
}

ptinsth ClockSystem::rewriteSup0(ptinsth use, ptinsth vh1, ptinsth vh2)
{
  boolean isavar;
  if ((vh1 == vh2) &&  (vh1 == use)) return removeSup(use, &isavar);
  return rewriteSup(use, vh1, vh2);  
}

/**
   @return the result of the merge of two clocks (classes of formula), vh1 andf vh2.
   It realizes recursively the induced merges.
   \see hclmkunion0.
*/
void
ClockSystem::mergingClasses0 (ptinsth vh1, ptinsth vh2)
{
  ptlinsth luse;
  ptinsth use;

  luse = getClockReferencesTo (vh2);
  hclfusion (vh2, vh1);
  while (luse != NILLINSTH)
    {
      ptlinsth_get_next (&luse, &use);
      switch ((int) (getOperator (use)))
	{
	case sub:
	  rewriteSub (use, vh1);
	  break;
	case nosub:
	  break;
	case hcomp:
	  break;
	case vari:
	case predef:
	  break;

	case hnot:
	  rewriteNot (use, vh1);
	  break;
	case infer:
	  use = rewriteInf0(use, getRepresentativeFormula (vh1), getRepresentativeFormula (otherArgOf (use, vh2)));
	  break;
	case super:
	  use = rewriteSup0(use, getRepresentativeFormula (vh1), getRepresentativeFormula (otherArgOf (use, vh2)));
	  break;

	default:
	  pK_messenger->Caseerror (__LINE__, __FILE__);
	}
    }
}

/**
   The pragmas associated with a clock (vhold) are inserted to those of an another one (vhnew).
*/
void
ClockSystem::fusion_pragmas (ptinsth vhold, ptinsth vhnew)
{
  ptlpragmas l_old;
  ptpragma aPragma;

  l_old = getPragmas (vhold);
  while (ptlpragmas_get_next (&l_old, &aPragma))
    addPragma (vhnew, aPragma);
}

/** 
    Assumed : the formula vh and vform have the same clock father.
    It substitutes the current representative formula (vh) by another one (vform).
 */
void
ClockSystem::substitute_clhead (ptinsth vh, ptinsth vform)
{
  int d;
  ptinsth hr, hf;
  ptlinsth lh;

  hf = getFather (vh);
  hr = getRoot (vh, &d);
  lh = getChildren (vh);

  setIsRepresentativeFormula (vform, pk_true);
  moveReferences (vh, vform);
  setNodeReferencesTo (vform, getNodeReferencesTo (vh));
  setNodeReferencesTo (vh, NILLINODE);
  belongsNotclhead (vh);
  setHierarchy (vform, hf, hr, d);
  setChildren (vform, lh);
  setChildren (vh, NILLINSTH);
}

/** It substitutes the current representative formula (vh) by another one (vform). */
void
ClockSystem::modify_clhead (ptinsth vh, ptinsth vform)
{
  substitute_clhead (vh, vform);
  setGraphInfos (vform, getGraphInfos (vh));
  setHierarchy (vform, vform, vform, 0);
}

void
ClockSystem::hclexclusiverules (ptinsth vh)
{
  ptinsth vh12, vh11;

  if (getOperator (vh) == infer)
    if (getMinusArgs (vh, &vh11, &vh12))	// vh = vh12 ^- vh11
      {				// a ^- b= 0, a ^< b
	replaceInf (makeInf (vh11, vh12), vh12);
	replaceSup (makeSup (vh11, vh12), vh11);
      }
    else if (getInfArgs (vh, &vh11, &vh12))
      {				// a ^* b= 0, a ^- b = a, b ^- a = b.
	replaceInf (makeMinus (vh11, vh12), vh11);
	replaceInf (makeMinus (vh12, vh11), vh12);
      }
}

/** It merges a clock (vh1) with the null clock and realizes induced merges.
 */
void
ClockSystem::unionwithhnull (ptinsth vh1)
{
  ptlinsth luse;
  ptinsth vhroot;
  int p;
  boolean b;

  do
    {
      vh1 = getRepresentativeFormula (vh1);
      if (!satisfiesNullClock (vh1))
	{
	  luse = getClockReferencesTo (vh1);
	  b = pk_true;
	  if (satisfiesNullDepth (vh1))
	    {
	      hclfusion (vh1, the_nullClock_CKS_copy);
	      hclexclusiverules (vh1);
	    }
	  else
	    {
	      vhroot = getRoot (vh1, &p);
	      if (vhroot == the_nullClock_CKS_copy)
		hclfusion (vh1, the_nullClock_CKS_copy);
	      else
		{
		  b = pk_false;
		  addPlacedNullClocks (vh1);
		}
	    }
	  if (b)
	    setlusetohnull (vh1, luse);
	}
    }
  while (hclsomehnull (&vh1));
}

/**
   It merges a clock (vh) with the null clock and realizes induced merges, for vh a upper bound.
 */
ptlinsth
ClockSystem::getTermsUpperBound (ptinsth vh, ptlinsth lh)
{
  ptinsth vharg1, vharg2;

  if (ptlinsth_isin_elem (lh, vh))
    return lh;
  if (getSupArgs (vh, &vharg1, &vharg2))
    {
      lh = getTermsUpperBound (vharg1, lh);
      lh = getTermsUpperBound (vharg2, lh);
    }
  else
    lh = ptlinsth_insert_elem (lh, vh);
  return lh;
}

_PROTECTED_

boolean ClockSystem::TestExclusiveConditions(ptinsth vharg1, ptinsth vharg2)
{ 
  ptinsth vh1;
  getComplExtraction (vharg2, &vh1);
  if (getRepresentativeFormula (vh1) == vharg1) return pk_true;
  return EM->TestExclusiveConditions(getConditionSub(vharg1), getOperator (vharg1) == sub, getConditionSub(vharg2), getOperator (vharg2) == sub);
}


/**
   if (vh1,vh3) is in the set of placed constraints, (vh2,vh3) is added to the set.
 */
void ClockSystem::ConstraintPlacedClocksTransitivity(ptinsth vh1, ptinsth vh2)
{
  ptinsth vh11, vh22;
  ptlinsth l1, l2;
  
  l1 = lterm1equpl;
  l2 = lterm2equpl;
  while (ptlinsth_get_next (&l1, &vh11))
    {
      vh11 = getRepresentativeFormula (vh11);
      ptlinsth_get_next (&l2, &vh22);
      vh22 = getRepresentativeFormula (vh22);
      if (vh1 == vh11)
	if (! ptlinsth_search_couple (vh2, vh22, lterm1equpl, lterm2equpl)) {
	  lterm1equpl = ptlinsth_insert_elem (lterm1equpl, vh22);
	  lterm2equpl = ptlinsth_insert_elem (lterm2equpl, vh2);
	}
      else
	if (vh1 == vh22)
	  if (! ptlinsth_search_couple (vh2, vh11, lterm1equpl, lterm2equpl)) {
	    lterm1equpl = ptlinsth_insert_elem (lterm1equpl, vh11);
	    lterm2equpl = ptlinsth_insert_elem (lterm2equpl, vh2);
	  }
    }
}

void
ClockSystem::addConstraintPlacedClocksNotAssumptions (ptinsth vh1, ptinsth vh2)
{
  ptinsth vhaux;

  vh1 = getRepresentativeFormula (vh1);
  vh2 = getRepresentativeFormula (vh2);
  if (vh1 != vh2)
    {
      if (getNum (vh1) > getNum (vh2))
	{
	  vhaux = vh2;
	  vh2 = vh1;
	  vh1 = vhaux;
	};
      // affine clocks or not, they are added in the same sets (lterm1equpl, lterm2equpl)
      if (! ptlinsth_search_couple (vh1, vh2, lterm1equpl, lterm2equpl))
	{
	  ConstraintPlacedClocksTransitivity(vh1, vh2);
	  ConstraintPlacedClocksTransitivity(vh2, vh1);
	  lterm1equpl = ptlinsth_insert_elem (lterm1equpl, vh1);
	  lterm2equpl = ptlinsth_insert_elem (lterm2equpl, vh2);
	}
    }
}

/**
   @return "true" if a formula (vh) depends on the value of a signal (s), "false" otherwise.
*/
boolean
ClockSystem::__isaFormulaDependOnValueOf (ptinsth vh, ptexpcal s)
{
  //UNUSED
  ptinsth vh1, vh2;
  boolean b;
  ptlinstances ls;

  if (satisfiesMarked (vh, mark3))
    return pk_false;
  addMark (vh, mark3);

  switch ((int) (getOperator (vh)))
    {
    case super:
      getSupArgs (vh, &vh1, &vh2);
      if (__isaFormulaDependOnValueOf (vh1, s))
	b = pk_true;
      else
	b = __isaFormulaDependOnValueOf (vh2, s);

      break;
    case infer:
      getInfArgs (vh, &vh1, &vh2);
      if (__isaFormulaDependOnValueOf (vh1, s))
	b = pk_true;
      else
	b = __isaFormulaDependOnValueOf (vh2, s);
      break;
    case sub:
    case nosub:
      ls = getSignalsOfCondition (vh);
      if (ptlinstances_isin_exp (ls, s))
	b = pk_true;		// else xxxxxx(); // visit the definitions of the signals(ls)!!
      break;

    case hcomp:
    case hnot:
      if (getUnaryArg (vh, &vh1))
	b = __isaFormulaDependOnValueOf (vh1, s);
      break;
    default:
      b = pk_false;
    }

  removeMark (vh, mark3);
  return b;
}

