/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Building an elementary tree for extractions [c], [not c] that are assertions.
 */
  boolean ClockManager::mkTreeAssumption (ptinsth vhcond, ptinsth aSub, ptinsth aNoSub)
{
  ptinsth vh1head, vh2head;

  vh1head = getRepresentativeFormula (aSub);
  vh2head = getRepresentativeFormula (aNoSub);
  removelssub (aSub);
  removeFormulaFromClass (vh1head, aSub);
  removeFormulaFromClass (vh2head, aNoSub);
  aNoSub = getRepresentativeFormula (aNoSub);
  aSub = getRepresentativeFormula (aSub);
  addClockAsRoot (vhcond);
  buildElementaryTreeFromSub (vhcond, aSub, aNoSub);
  if (!satisfiesNullClock (vh1head))
    {
      hclmkunion (0, aSub, vh1head);
    }
  if (!satisfiesNullClock (vh2head))
    {
      hclmkunion (0, aNoSub, vh2head);
    }
  return pk_true;
}

/** Assumed : aSub is a positive extraction [C], where C is assumed to be a free condition.
    @return 
    - pk_true if the tree ^C,[C], [not C]  with father([C]) =  ^C, and  father([not C]) =  ^C is built;
    - false otherwise: it exists (not proved) constraints on free conditions.
*/
boolean
ClockManager::buildExtractionTree (ptinsth aSub)
{
  ptinsth vhcond, aNoSub;

  vhcond = getRepresentativeFormula (getClockSub (aSub));
  getComplInFather (aSub, &aNoSub);
  if (!satisfiesIsAnAssumption (aSub))
    return buildElementaryTreeFromSub (vhcond, aSub, aNoSub);
  return mkTreeAssumption (vhcond, aSub, aNoSub);
}

/**
   Placing clocks build on free conditions (provided by the when/cell  signal-expression).
   @return true if it exists (not proved) constraints on free conditions, false otherwise.
*/
boolean
ClockManager::buildExtractionTree ()
{
  boolean someproblems;
  int vtimedebop;
  ptlinsth list;
  ptinsth vh;

  vtimedebop = pK_messenger->messageBeginOperVerbose (infofreeform);
  someproblems = pk_false;
  list = getFormulaList (sub);
  while (ptlinsth_get_next (&list, &vh))
    {
      if (!buildExtractionTree (vh))
	someproblems = pk_true;
      solving (1);
      solvingAffine ();
    }
  pK_messenger->messageEndOperVerbose (vtimedebop);
  return someproblems;
}

/**
   @return true if the placing of the formula vh (assumed to be a ^* formula) is 
   performed (ie the 2 operands are in the same tree),
   false otherwise.
 */
boolean
ClockManager::infNormalFormTreePlacing (ptinsth vh)
{
  ptinsth vharg1, vharg2, otherterm;

  if (getInfArgs (vh, &vharg1, &vharg2))
    if (treeReduceInf (vharg1, vharg2, &otherterm))
      {
	otherterm = replaceInf (vh, otherterm);
	return pk_true;
      }
  return pk_false;
}

/**
   Placing of the unplaced ^* formulas when the two operands are in the same tree.
   @return true if one placing is perfomed, false otherwise.
 */
boolean
ClockManager::infNormalFormTreePlacing ()
{
  ptinsth vh;
  boolean qqcplace;
  ptlinsth list;

  qqcplace = pk_false;
  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
    if (!satisfiesIsCanonicalRepr (vh))
      if (infNormalFormTreePlacing (vh))
	qqcplace = pk_true;
  return qqcplace;
}

/**
   @return true if the placing of the formula vh (assumed to be a ^+ formula) is 
   performed (ie the 2 operands are in the same tree),
   false otherwise.
 */
boolean
ClockManager::supNormalFormTreePlacing (ptinsth vh)
{
  ptinsth vharg1, vharg2, otherterm;

  if (getSupArgs (vh, &vharg1, &vharg2))
    if (treeReduceSup (vharg1, vharg2, &otherterm))
      {
	otherterm = replaceSup (vh, otherterm);
	return pk_true;
      }
  return pk_false;
}

/**
   Placing of the unplaced ^+ formulas when the two operands are in the same tree.
   @return true if one placing is perfomed, false otherwise.
 */
boolean
ClockManager::supNormalFormTreePlacing ()
{
  ptinsth vh;
  boolean qqcplace;
  ptlinsth list;

  qqcplace = pk_false;
  list = getFormulaList (super);
  while (ptlinsth_get_next (&list, &vh))
    if (!satisfiesIsCanonicalRepr (vh))
      if (supNormalFormTreePlacing (vh))
	qqcplace = pk_true;
  return qqcplace;
}

/**
   @return true if the placing of the formula vh (assumed to be a ^- formula) is 
   performed (ie the 2 operands are in the same tree),
   false otherwise.
 */
boolean
ClockManager::minusNormalFormTreePlacing (ptinsth vh)
{
  ptinsth vharg1, vharg2, otherterm;

  if (getMinusArgs (vh, &vharg1, &vharg2))
    if (!satisfiesIsCanonicalRepr (vh))
      if (treeReduceInfhnot (vh, vharg1, vharg2, &otherterm))
	{
	  otherterm = replaceInf (vh, otherterm);
	  return pk_true;
	}
  return pk_false;
}

/**
   Placing of the unplaced ^- formulas when the two operands are in the same tree.
   @return true if one placing is perfomed, false otherwise.
 */
boolean
ClockManager::minusNormalFormTreePlacing ()
{
  ptinsth vh;
  boolean qqcplace;
  ptlinsth list;

  qqcplace = pk_false;
  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
    if (minusNormalFormTreePlacing (vh))
      qqcplace = pk_true;
  return qqcplace;
}

/**
   Placing of the unplaced (^*, ^*,^-) formulas when the two operands are in the same tree.
   @return true if one placing is perfomed, false otherwise.
 */
boolean
ClockManager::NormalFormTreePlacing ()
{
  int vtimedebop;
  boolean qqcplace, hplace;

  vtimedebop = pK_messenger->messageBeginOperVerbose (infonormform);
  hplace = pk_false;
  qqcplace = pk_true;
  while (qqcplace)
    {
      qqcplace = infNormalFormTreePlacing ();
      if (!qqcplace)
	{
	  qqcplace = supNormalFormTreePlacing ();
	  if (!qqcplace)
	    qqcplace = minusNormalFormTreePlacing ();
	}
      if (qqcplace)
	{
	  hplace = pk_true;
	  solving (0);
	}
    }
  if (hplace)
    solvingAffine ();
  pK_messenger->messageEndOperVerbose (vtimedebop);
  return hplace;
}

/**
   Applying boolean reductions to the ^* formula not placed with 
   holoph as the depth of resolution.
   @return true if such formulas exist, false otherwise.
*/
boolean
ClockManager::infBooleanReducing (int holoph)
{
  boolean notempty;
  ptinsth vhres, vharg1, vharg2, vh;
  ptlinsth list;

  notempty = pk_false;
  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
    if (!getMinusArgs (vh, &vharg1, &vharg2))
      {
	notempty = pk_true;
	getInfArgs (vh, &vharg1, &vharg2);
	vhres = booleanReducedInf (holoph, vh, vharg1, vharg2, pk_true);
      }
  return notempty;
}

/**
   Applying boolean reductions to the ^+ formula not placed with 
   holoph as the depth of resolution.
   @return true if such formulas exist, false otherwise.
*/
boolean
ClockManager::supBooleanReducing (int holoph)
{
  boolean notempty;
  ptinsth vhres, vharg1, vharg2, vh;
  ptlinsth list;

  notempty = pk_false;
  list = getFormulaList (super);
  while (ptlinsth_get_next (&list, &vh))
    {
      notempty = pk_true;
      getSupArgs (vh, &vharg1, &vharg2);
      vhres = booleanReducedSup (holoph, vh, vharg1, vharg2, pk_true);
    }
  return notempty;
}

/**
   Applying boolean reductions to the ^- formula not placed with 
   holoph as the depth of resolution.
   @return true if such formulas exist, false otherwise.
*/
boolean
ClockManager::minusBooleanReducing (int holoph)
{
  boolean notempty;
  ptinsth vhres, vharg1, vharg2, vh;
  ptlinsth list;

  notempty = pk_false;
  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
    if (getMinusArgs (vh, &vharg1, &vharg2))
      if (!satisfiesIsCanonicalRepr (vh))
	{
	  notempty = pk_true;
	  getInfArgs (vh, &vharg1, &vharg2);
	  vhres = booleanReducedInf (holoph, vh, vharg1, vharg2, pk_true);
	}
  return notempty;
}

/**
   Applying boolean reductions to  not placed formulas with 
   holoph as the depth of resolution : the expressions are visited at most holoph-level for reductions.
*/
void
ClockManager::BooleanReducing (int holophmax)
{
  int vtimedebop;
  boolean qqplace, empty;
  int holoph;

  if (holophmax > 0)
    {
      holoph = 1;
      vtimedebop = pK_messenger->messageBeginOperVerbose (infosolveroots);
      empty = pk_false;
      while ((holoph <= holophmax) && !empty)
	{
	  pK_messenger->messageIntVerbose (infodepth, holoph);
	  qqplace = pk_false;
	  empty = !infBooleanReducing (holoph);
	  if (!empty)
	    {
	      solving (0);
	      qqplace = NormalFormTreePlacing ();
	    }
	  if (!qqplace)
	    {
	      empty = !supBooleanReducing (holoph);
	      if (!empty)
		{
		  solving (0);
		  qqplace = NormalFormTreePlacing ();
		}
	    }
	  if (!qqplace)
	    {
	      empty = !minusBooleanReducing (holoph);
	      if (!empty)
		{
		  solving (holoph);
		  qqplace = NormalFormTreePlacing ();
		}
	    }
	  solvingAffine ();

	  holoph = holoph + 1;

	  pK_messenger->messageEndOperVerbose (0);
	}
      pK_messenger->messageEndOperVerbose (vtimedebop);
    }
}

/**
   Solving by modifying the representative formula for root classes.
   @return true if the application of this strategy has induced the placing of some formula,
   false otherwise.
*/
boolean
ClockManager::solvingModifyRepr ()
{
  int vtimedebop;
  boolean qqmodif, qqplace, b;
  ptinsth vh;
  ptlinsth list;

  /* ATTENTION : La méthode tryToChangRepresentativeFormula(vh) */
  /* insere des contraintes (gerees dans mksethorl) sur les formules des horloges racines. */
  /* Il faut imperativement reconstruire les classes.(methode   rebuiltRootClasses();) */
  vtimedebop = pK_messenger->messageBeginOperVerbose (infochrepr);
  b = strongInclusionTest;
  strongInclusionTest = pk_true;
  qqplace = pk_false;
  qqmodif = pk_true;
  while (qqmodif)
    {
      qqmodif = pk_false;
      list = getFormulaList (infer);
      while (!qqmodif && ptlinsth_get_next (&list, &vh))
	qqmodif = tryToChangRepresentativeFormula (vh);
      if (qqmodif)
	{
	  solving (1);
	  if (NormalFormTreePlacing ())
	    qqplace = pk_true;
	}
      else
	{
	  list = getFormulaList (super);
	  while (!qqmodif && ptlinsth_get_next (&list, &vh))
	    qqmodif = tryToChangRepresentativeFormula (vh);
	  if (qqmodif)
	    {
	      solving (1);
	      if (NormalFormTreePlacing ())
		qqplace = pk_true;
	    }
	}
    }
  solving (0);
  solvingAffine ();
  pK_messenger->messageEndOperVerbose (vtimedebop);
  strongInclusionTest = b;
  return qqplace;
}

/** Verifying equivalence of formulas for root classes. */
void
ClockManager::verifyConstraintsOnRoots ()
{
  int vtimedebop;
  boolean b;
  ptinsth vh;
  ptlinsth list;

  vtimedebop = pK_messenger->messageBeginOperVerbose (infocontrroots);
  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
    if (isOtherSyntacticFormOfRepresentativeFormula (vh))
      vh = removeInf (vh, &b);

  list = getFormulaList (super);
  while (ptlinsth_get_next (&list, &vh))
    if (isOtherSyntacticFormOfRepresentativeFormula (vh))
      vh = removeSup (vh, &b);

  pK_messenger->messageEndOperVerbose (vtimedebop);
}

/**
   Rewriting  not-formula : the operator h ^- g is implemented as h * not(g)
*/
void
ClockManager::rewriteHnotFormulas ()
{
  ptlinsth lh;
  ptinsth vh, vh1, vh2;

  lh = getFormulaList (hnot);
  while (ptlinsth_get_next (&lh, &vh))
    {
      getUnaryArg (vh, &vh1);
      vh2 = getRepresentativeFormula (vh1);
      if (vh1 != vh2)
	rewriteNot (vh, vh2);
    }
  solving (0);
}

/**
   The operator h ^- g is implemented as u = (h * k) with k = not(g).
   For solving, the reference "k referenced by u" is installed
   and the reference "g referenced by u" is removed.  
*/
void
ClockManager::setNotFormulaReferences ()
{
  ptlinsth lh;
  ptinsth vh;

  rewriteHnotFormulas ();
  lh = getlhroots ();
  while (ptlinsth_get_next (&lh, &vh))
    switchReferencesNotFormula (vh);

  lh = getlhplaced ();
  while (ptlinsth_get_next (&lh, &vh))
    switchReferencesNotFormula (vh);
}

boolean
ClockManager::rewritingUsingPlacedConstrainst (ptinsth vh1, ptinsth vh2)
{
  ptlinsth luse1, luse2;
  ptinsth vh;

  if (vh1 == vh2) return pk_false;
  
  luse1 = getClockReferencesTo (vh1);
  luse2 = getClockReferencesTo (vh2);
  if (luse1 == NILLINSTH)
    {
      if (luse2 != NILLINSTH)
	{
	  replaceForm (vh2, vh1, luse2);
	  return pk_true;
	}
      return pk_false; 
    }
  if (luse2 == NILLINSTH)
    {
      replaceForm (vh1, vh2, luse1);
      return pk_true;
    }
  return pk_false; 
}


/**
   Solving using the placed constraints. Applied only if the global parameter UsingPlacedConstrainstAlreadyApplied is false.
   For two formula (h1, h2) of the same class, if h1 and h2 are placed during the clock calculus and have not the
   same canonical arborescent form, the constraint is kept and a choice is done to continue the clock calculus.
   The switch on placed formula is done, then the solving is applied.
*/
void
ClockManager::rewritingUsingPlacedConstrainst (int vholophmax)
{
  ptlinsth lh1, lh2, luse1, luse2;
  ptinsth vh1, vh2;
  boolean bret;
  int s, sn, snbVari, nbVari;


  if (! UsingPlacedConstrainstAlreadyApplied)
    {
      bret = pk_false;
      sn = getNbRoots ();
      nbVari = ptlinsth_get_size(lvariables);
      do
	{
	  s = sn; 
	  snbVari = nbVari;
	  getConstraintsOnPlacedClocks (&lh1, &lh2);
	  while (ptlinsth_get_next (&lh1, &vh1))
	    {
	      ptlinsth_get_next (&lh2, &vh2);
	      
	      luse1 = getClockReferencesTo (vh1);
	      replaceForm (vh1, vh2, luse1);
	      NormalFormTreePlacing ();
	      BooleanReducing (vholophmax);
	      
	      luse2 = getClockReferencesTo (vh2);
	      replaceForm (vh2, vh1, luse2);
	      NormalFormTreePlacing ();
	      BooleanReducing (vholophmax);
	      bret = pk_true;
	    }
	  sn = getNbRoots ();
	  nbVari = ptlinsth_get_size(lvariables);
	}
      while (sn + nbVari < s + snbVari);
      
      if (bret) UsingPlacedConstrainstAlreadyApplied = pk_true;
    }
}

/**
   It solves the equivalences on affine clocks. 
   For each couple of clocks (vh1, vh2) of the lists lterm1equpl/lterm2equpl, 
   it applies isEquivAffineFormTree() on vh1 and vh2 if the branching clock is defined.
   The constraint is removed form the lists if it is proved.
*/
void
ClockManager::solvingAffine ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2, vhemb;

  lh1 = lterm1equpl;
  lh2 = lterm2equpl;
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      vh1 = getRepresentativeFormula (vh1);
      vh2 = getRepresentativeFormula (vh2);
      vhemb = getBranching (vh1, vh2);
      if (!ptinsth_is_hundef (vhemb))
	if (isEquivAffineFormTree (vh1, vh2, vhemb))
	  ptlinsth_remove_pair (&lterm1equpl, &lterm2equpl, vh1, vh2);
    }
}

/**
   @return true if the current representative formula of a class (vh) is a recursive expression
   and if another formula of the class has been chosen as the new representative element.
*/
boolean
ClockManager::tryToChangRepresentativeFormula (ptinsth vh)
{
  ptinsth vh1, vhrepr, first;

  if (satisfiesSingleton (vh))
    return pk_false;
  vhrepr = getRepresentativeFormula (vh);
  if (vhrepr != vh)
    return pk_false;

  if (isFictitiousRootFormula (vhrepr))
    return pk_false;
  if (satisfiesIsCanonicalRepr (vhrepr))
    return pk_false;
  if (!isRecursiveFormula (vhrepr))
    return pk_false;
  first = vhrepr;
  vh1 = getNextFormula (vhrepr, pk_true);
  while (vh1 != first)
    {
      hclchangform (vhrepr, vh1);
      hcladdmodified (vh1);
      if (!isRecursiveFormula (vh1))
	return pk_true;
      vhrepr = vh1;
      vh1 = getNextFormula (vh1, pk_true);
    }
  solving (0);
  return pk_false;
}

/**
   It solves (recursively) all announced equivalences and the formula whose one of argument 
   has been modified.
*/
void
ClockManager::solving (int holoph)
{
  boolean arreter;
  ptinsth use, vh, h1, h2;
  ptlinsth luse;

  if (ModeSynchroSolving)
    {
      arreter = pk_false;
      while (!arreter)
	{
	  if (hclsomehmodified (&vh))
	    {
	      luse = getClockReferencesTo (vh);
	      while (ptlinsth_get_next (&luse, &use))
		switch ((int) (getOperator (use)))
		  {
		  case sub:
		    rewriteSub (use, vh);
		    break;
		  case nosub:
		    break;
		  case hcomp:
		    break;
		  case hnot:
		    rewriteNot (use, vh);
		    break;
		  case infer:
		    getBinaryArgs (use, &h1, &h2);
		    use = booleanReducedInf (holoph, use, h1, h2, pk_true);
		    break;
		  case super:
		    getBinaryArgs (use, &h1, &h2);
		    use = booleanReducedSup (holoph, use, h1, h2, pk_true);
		    break;
		  case vari:
		  case predef:
		    break;
		  default:
		    pK_messenger->Caseerror (__LINE__, __FILE__);
		  }
	      arreter = pk_false;
	    }
	  else
	    arreter = pk_true;
	  if (hclmklequiv (holoph))
	    arreter = pk_false;
	}
    }
}


/**
   It frees the clocks, elements of the operand list (lh).

   CURRENTLY WITHOUT EFFECTS: Side effect in the BDD package.
*/
void
ClockManager::free_clocks (ptlinsth lh)
{
  ptinsth vh;

  while (ptlinsth_get_next (&lh, &vh))
    {
      if (free_formula (vh))
	removeBDDClock (vh);
    }
}


void ClockManager::rewClocksUsingPlacedConstraints()
{
  // Added for PolyEndo.
  ptlinsth luse1, luse2, lh1, lh2;
  ptinsth vh1, vh2;

  if (getConstraintsOnPlacedClocks(&lh1, &lh2))
    {
      while (ptlinsth_get_next (&lh1, &vh1))
	{
	  ptlinsth_get_next (&lh2, &vh2);
	  luse1 = getClockReferencesTo (vh1);
	  luse2 = getClockReferencesTo (vh2);
	  replaceForm (vh2, vh1, luse2);
	  replaceForm (vh1, vh1, luse1);
	  solving(0);
	}
    }
  NormalFormTreePlacing ();
  BooleanReducing(0);
}


boolean
ClockManager::rewMinusClocksUsingPlacedConstraints (ptinsth vh)
{
  ptinsth velem, vharg1, vharg2, vhemb;
  ptlinsth lh_save, lhEq;
  boolean found = pk_false;

  // Added for PolyEndo.
  if (getMinusArgs (vh, &vharg1, &vharg2)) 
    {
      // vh = vharg2 ^- vharg1
      found = pk_false;
      lhEq = getClocksPlacedEquivTo(vharg1);
      lh_save = lhEq;
      while (!found && ptlinsth_get_next (&lhEq, &velem))
	{
	  vhemb = getBranching (vharg2, velem);
	  if (!ptinsth_is_hundef (vhemb))
	    {
	      found = rewritingUsingPlacedConstrainst(vharg1, velem);
	    }
	}
      ptlinsth_free(lh_save);
    }
  return found;
}


boolean ClockManager::testExclusiveTerms(ptinsth vh1, ptinsth vh2, ptlinsth lhNullClockAssertions)
{
  ptinsth vharg1, vharg2, vh;

  // Added for PolyEndo.
  while (ptlinsth_get_next (&lhNullClockAssertions, &vh))
    {
      if (!getMinusArgs (vh, &vharg1, &vharg2))
	{
	  getInfArgs (vh, &vharg1, &vharg2);
	  if ((vh1 == vharg1) && (vh2 == vharg2)) return true;
	  if ((vh1 == vharg2) && (vh2 == vharg1)) return true;
	}
    }
  return pk_false;
}

boolean ClockManager::testExclusive(ptinsth vh1, ptinsth vh2, ptlinsth lhNullClockAssertions)
{
  ptinsth vharg1, vharg2, vh;
  
  // Added for PolyEndo.
  if (testExclusiveTerms(vh1, vh2, lhNullClockAssertions)) return pk_true;
  // excluding ^- sub terms
  if (getMinusArgs (vh1, &vharg1, &vharg2)) return pk_false;
  if (getMinusArgs (vh2, &vharg1, &vharg2)) return pk_false;
  if (getInfArgs (vh1, &vharg1, &vharg2))
    {
      if (testExclusive(vh2, vharg1,lhNullClockAssertions)) return pk_true;
      if (testExclusive(vh2, vharg2,lhNullClockAssertions)) return pk_true;
    }
  // for the sup ... later
  return pk_false;
}

/**
   Simplification of H ^- K.
   If H = h1^*...^*hn, K = k1^* ..^*.kp and it exists i,j such that hi*kj=^0
   then H ^- K is reduced to H.
 */
boolean
ClockManager::rewMinusClocksUsingExcluInfos()
{
  ptlinsth lhNullAssert, list;
  ptinsth vharg1, vharg2, vh;
  boolean modif = pk_false;

  // Added for PolyEndo.
  if (getNullClockAssertions(&lhNullAssert))
    {
      list = getFormulaList (infer);
      while (ptlinsth_get_next (&list, &vh))
	{
	  if (getMinusArgs (vh, &vharg1, &vharg2))
	    {
	      // vh = vharg2 ^- vharg1
	      if (testExclusive(vharg1, vharg2, lhNullAssert))
		{
		  replaceInf (vh, vharg2);
		  modif = pk_true;
		}
	    }
	}
    }
  if (modif) 
    {
      solving(0);
      NormalFormTreePlacing (); 
      BooleanReducing(0);
    }
 return modif;
}

int ClockManager::getNumberOfMinusFormula()
{
  ptlinsth list;
  ptinsth vharg1, vharg2, vh;
  int n;

  // Added for PolyEndo.
  n = 0;
  list = getFormulaList (infer);
  while (ptlinsth_get_next (&list, &vh))
     if (getMinusArgs (vh, &vharg1, &vharg2)) n++;
  // printf("getNumberOfMinusFormula() = %d\n", n);
  return n;
}


void ClockManager::rewMinusClocksUsingPlacedConstraints()
{
  ptlinsth list;
  ptinsth vh;
  boolean qqrew, bcont;
  
  // Added for PolyEndo.
  bcont= pk_true;
  while (bcont)
    {
      bcont= pk_false;
      qqrew = pk_false;
      list = getFormulaList (infer);
      while ( !qqrew && ptlinsth_get_next (&list, &vh))
	if (rewMinusClocksUsingPlacedConstraints (vh))
	  {
	    qqrew = pk_true;
	    solving(0);
	    bcont= pk_true;
	  }
    }
}


void ClockManager::rewMinusClocksUsingPlacedConstraints(int holoph)
{
  
  // Added for PolyEndo.
  rewMinusClocksUsingExcluInfos();
  rewClocksUsingPlacedConstraints();
  rewMinusClocksUsingPlacedConstraints();
  NormalFormTreePlacing (); 
  BooleanReducing(holoph);
}


void ClockManager::reduceMinusFormulae (int holoph)
{
  // Added for PolyEndo.
  int n, m;

  rewMinusClocksUsingPlacedConstraints(holoph);
  n = getNumberOfMinusFormula();
  if (n > 0) 
    {
      m = n-1;
      while ( m < n )
	{
	  n = m;
	  rewMinusClocksUsingPlacedConstraints(holoph);
	  m = getNumberOfMinusFormula();
	}
    }
}
