/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

/**
 @return a new subgraph.
 */
extern ptgfd
ptgfd_create (void)
{
  ptgfd vgfd;

  vgfd = new gfd ();
  return ptgfd_init (vgfd);
}

/**
   @return a copy of the subgraph vgfdref. If vgfdmod is undefined a new descriptor is created, 
   otherwise vgfdmod is used to defined the copy.
*/
extern ptgfd
ptgfd_copy (ptgfd vgfdmod, ptgfd vgfdref)
{
  if (vgfdref == NILGFD)
    return vgfdmod;
  if (vgfdmod == NILGFD)
    vgfdmod = ptgfd_create ();
  vgfdmod->interface = ptinterface_duplicate (vgfdref->interface);
  vgfdmod->lnodes = vgfdref->lnodes;
  vgfdmod->lclocks = vgfdref->lclocks;
  return vgfdmod;
}



/**
   It initializes a subgraph (vgfd) by empty lists.
   @return the subgraph.
*/
extern ptgfd
ptgfd_init (ptgfd vgfd)
{
  vgfd->interface = NILINTERFACE;
  vgfd->lnodes = NILLINODE;
  vgfd->lclocks = NILLINSTH;
  vgfd->iostep = NILINSTH;
  vgfd->projectedSystem = NILLINSTH;
  return vgfd;
}


/**
   @return the merging of two subgraphs (vgfdmod, vgfdres) with result in vgfdres.
   \warning vgfdres may be modified.
*/
extern ptgfd
ptgfd_fusion (ptgfd vgfdmod, ptgfd vgfdres)
{
  ptlINode ln;

  if (vgfdres == NILGFD)
    return vgfdmod;
  if (vgfdmod == NILGFD)
    return vgfdres;
  vgfdres->interface = ptinterface_union (vgfdres->interface, vgfdmod->interface);
  vgfdmod->interface = NILINTERFACE;
  ln = ptlINode_fusion (ptgfd_get_nodes (vgfdmod), ptgfd_get_nodes (vgfdres));
  ptgfd_set_nodes (vgfdres, ln);
  return vgfdres;
}

/**
   @return true if the operation suceeds, "false" otherwise (vnode is not in the subgraph).
   It removes a node (vnode) from a subgraph (vgfd).
*/
extern boolean
ptgfd_remove_node (ptgfd vgfd, ptINode vnode)
{
  ptlINode li;

  li = ptgfd_get_nodes (vgfd);
  if (!ptlINode_remove_elem (&li, vnode))
    return pk_false;
  ptgfd_set_nodes (vgfd, li);
  return pk_true;
}

/**
   It inserts a node (vnode) in a subgraph (vgfd).
*/
extern void
ptgfd_insert_node (ptgfd vgfd, ptINode vnode)
{
  ptlINode li;

  li = ptgfd_get_nodes (vgfd);
  if (ptlINode_insert_elem (&li, vnode))
    ptgfd_set_nodes (vgfd, li);
}


/**
   It removes a clock (vclock) from a subgraph (vgfd).
 */
extern void
ptgfd_remove_clock (ptgfd vgfd, ptinsth vclock)
{
  ptlinsth li;

  li = ptgfd_get_clocks (vgfd);
  if (ptlinsth_sorted_remove_elem (&li, vclock))
    ptgfd_set_clocks (vgfd, li);
}

/**
   It inserts a clock (vclock) in a subgraph (vgfd).
*/
extern void
ptgfd_insert_clock (ptgfd vgfd, ptinsth vclock)
{
  ptlinsth li;
  boolean b;

  li = ptgfd_get_clocks (vgfd);
  li = ptlinsth_sorted_insert_elem_bool (li, vclock, &b);
  ptgfd_set_clocks (vgfd, li);
}



/**
   It assigns the nodes (l) to a subgraph (vgfd).
*/
extern void
ptgfd_set_nodes (ptgfd vgfd, ptlINode l)
{
  vgfd->lnodes = l;
}

/**
   It assigns the clocks (l) to a subgraph (vgfd).
*/
extern void
ptgfd_set_clocks (ptgfd vgfd, ptlinsth l)
{
  vgfd->lclocks = l;
}

/**
   It assigns the projection information of a subgraph (vgfd):
   -  the input/output step clock (vclock)
   -  the set of clocks of the projected system (l)
*/
extern void
ptgfd_set_projectionInfos (ptgfd vgfd, ptinsth vclock, ptlinsth l)
{
  vgfd->iostep = vclock;
  vgfd->projectedSystem = l;
}


/**
   It assigns an interface (vinter) to a subgraph (vgfd).
*/
extern void
ptgfd_set_interface (ptgfd vgfd, ptinterface vinter)
{
  vgfd->interface = vinter;
}


/**
   It assigns the signals (l) as local signals in the interface of a subgraph (vgfd).
*/
extern void
ptgfd_set_lloc (ptgfd vgfd, ptlinstances l)
{
  vgfd->interface = ptinterface_set_localSignals (vgfd->interface, l);
}

/**
   It assigns the signals (l) as input signals in the interface of a subgraph (vgfd).
*/
extern void
ptgfd_set_inputSignals (ptgfd vgfd, ptlinstances l)
{
  vgfd->interface = ptinterface_set_inputSignals (vgfd->interface, l);
}


/**
   It assigns the signals (l) as output signals in the interface of a subgraph (vgfd).
*/
extern void
ptgfd_set_outputSignals (ptgfd vgfd, ptlinstances l)
{
  vgfd->interface = ptinterface_set_outputSignals (vgfd->interface, l);
}

/**
   It assigns the signals (l) as input signals in the interface of a subgraph (vgfd).
*/
extern void
ptgfd_set_inOutSignals (ptgfd vgfd, ptlinstances l)
{
  vgfd->interface = ptinterface_set_inOutSignals (vgfd->interface, l);
}


/**
   It assigns the clocks (l) as input clocks in the interface of a subgraph (vgfd).
*/
extern void
ptgfd_set_inputClocks (ptgfd vgfd, ptlinsth l)
{
  vgfd->interface = ptinterface_set_inputClocks (vgfd->interface, l);
}


/**
   It resets (the lists become empty) the interface of a subgraph (vgfd). 
*/
extern void
ptgfd_reset_interface (ptgfd vgfd)
{
  interface_free (vgfd->interface);
}



/**
   @return the list of the nodes assigned to a subgraph (vgfd).
*/
extern ptlINode
ptgfd_get_nodes (ptgfd vgfd)
{
  if (vgfd == NILGFD)
    return NILLINODE;
  return vgfd->lnodes;
}

/**
   @return the list of the clocks assigned to a subgraph (vgfd).
*/
extern ptlinsth
ptgfd_get_clocks (ptgfd vgfd)
{
  if (vgfd == NILGFD)
    return NILLINSTH;
  return vgfd->lclocks;
}

/**
   @return "true" if the subgraph is empty (no nodes, no clocks), "false" otherwise.
*/
extern boolean
ptgfd_satisfies_empty (ptgfd vgfd)
{
  if (vgfd == NILGFD)
    return pk_true;
  return (vgfd->lnodes == NILLINODE) && (vgfd->lclocks == NILLINSTH);
}


/**
   @return the interface assigned to a subgraph (vgfd).
*/
extern ptinterface
ptgfd_get_interface (ptgfd vgfd)
{
  return vgfd->interface;
}


/**
   @return the list of the clocks of the interface of a subgraph (vgfd).
*/
extern ptlinsth
ptgfd_get_inputClocks (ptgfd vgfd)
{
  return ptinterface_get_inputClocks (vgfd->interface);
}

/**
   @return the list of the input signals of the interface of a subgraph (vgfd).
*/
extern ptlinstances
ptgfd_get_inputSignals (ptgfd vgfd)
{
  return ptinterface_get_inputSignals (vgfd->interface);
}

/**
   @return the list of the output signals  of the interface of a subgraph (vgfd).
*/
extern ptlinstances
ptgfd_get_outputSignals (ptgfd vgfd)
{
  return ptinterface_get_outputSignals (vgfd->interface);
}


/**
   @return the list of the local signals  of the interface of a subgraph (vgfd).
*/
extern ptlinstances
ptgfd_get_localSignals (ptgfd vgfd)
{
  return ptinterface_get_localSignals (vgfd->interface);
}

/**
   @return the list of the in-out signals  of the interface of a subgraph (vgfd).
*/
extern ptlinstances
ptgfd_get_inOutSignals (ptgfd vgfd)
{
  return ptinterface_get_inOutSignals (vgfd->interface);
}

/**
   @return the clock assigned as the input-output step clock
   and the list of the clocks (lh) of the projected system, assigned to a subgraph (vgfd).
*/
extern ptinsth
ptgfd_get_projectionInfos (ptgfd vgfd, ptlinsth * lh)
{
  *lh = vgfd->projectedSystem;
  return vgfd->iostep;
}
