/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** It adds vinsth as a successor of vinstr.
    @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph). 
*/
    boolean NodeManager::addDependence(ptINode vinstr, ptinsth vinsth)
{
  ptlinsth vlinsth;
  boolean b;

  if ( CKM->getRepresentativeFormula(vinsth) != vinsth) return pk_false;
  if ( CKM->satisfiesIsAnEquivalentForm(vinsth)) return pk_false;

  if (existPredefinedPragma(getPragmas(vinstr), PRAGMA_DELAYCLUSTER))
    return pk_false;
  vlinsth = get_clockSuccessors(vinstr);
  vlinsth = ptlinsth_sorted_insert_elem_bool(vlinsth, vinsth, &b);
  set_clocksSuccessors(vinstr, vlinsth);
  return b;
}

/**  It adds vhto as a successor of vhfrom.
     @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph). 
*/
boolean NodeManager::addDependence(ptinsth vhfrom, ptinsth vhto)
{
  ptlinsth vlinsth;
  boolean b;

  vlinsth = CKM->getClockReferencesTo(vhfrom);
  vlinsth = ptlinsth_sorted_insert_elem_bool(vlinsth, vhto, &b);
  if (b)
    CKM->setReferencesAt(vhfrom, vlinsth);
  return b;
}

/**  It adds vinstr as a successor of vh.
     @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph or vh is the null clock). 
*/
boolean NodeManager::addDependence(ptinsth vh, ptINode vinstr)
{
  ptlINode ln;
  boolean b;
  ptlinsth lh;

  if (CKM->satisfiesNullClock(vh))
    return pk_false;
  ln = CKM->getNodeReferencesTo(vh);
  if (ptlINode_insert_elem(&ln, vinstr))
    CKM->setNodeReferencesTo(vh, ln);
  lh = get_clockPredecessors(vinstr);
  set_clockPredecessors(vinstr, ptlinsth_sorted_insert_elem_bool(lh, vh, &b));
  return b;
}

/**
   It adds the dependence vinstrFrom -> vinstrTo when h with tdep as kind of dependence.
   @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph).
*/
boolean NodeManager::addDependence(ptINode vinstrFrom, ptINode vinstrTo, ptinsth h, TypeDependenceKind tdep)
{
  ptldependence li;
  ptdependence aDep;
  ptinsth h1, h2;

  if (CKM->satisfiesNullClock(h))
    return pk_false;
  if (CKM->satisfiesUndefClock(h))
    {
      h1 = get_nodeClock(vinstrFrom);
      h2 = get_nodeClock(vinstrTo);
      if (!CKM->satisfiesUndefClock(h1) || !CKM->satisfiesUndefClock(h2))
	h = CKM->makeInf(h1, h2);
    }

  li = get_successors(vinstrFrom);
  if (!li->exist_dependence(vinstrFrom, vinstrTo, &aDep))
    {
      aDep = new Idependence(vinstrFrom, vinstrTo, h);
      GdependencesList = GdependencesList->insert(aDep);
      aDep->setDependenceKind(tdep);
      li = li->insert(aDep);
      set_successors(vinstrFrom, li);

      li = get_predecessors(vinstrTo);
      li = li->insert(aDep);
      set_predecessors(vinstrTo, li);
      return pk_true;
    }
  aDep->setDependenceKind(tdep);
  h1 = aDep->getClockDependence();
  aDep->setClockDependence(CKM->makeSup(h, CKM->getRepresentativeFormula(h1)));
  return pk_false;
}

/**  
     It adds the data-flow dependence vinstrFrom -> vinstrTo when h.
     @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph). 
*/
boolean NodeManager::addDependence(ptINode vinstrFrom, ptINode vinstrTo, ptinsth h)
{
  return addDependence(vinstrFrom, vinstrTo, h, dataFlowDep);
}

/**
   It adds the explicit dependence vinstrFrom -> vinstrTo when h.
   @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph). 
*/
boolean NodeManager::addExplicitDependence(ptINode vinstrFrom, ptINode vinstrTo, ptinsth h)
{
  return addDependence(vinstrFrom, vinstrTo, h, explicitDep);
}

/**
   It adds the explicit dependence vinstrFrom -> vinstrTo when h if this dependence does not exist.
   @return "true" if the operation succeeds, "false" otherwise (it already exists in the graph). 
*/
boolean NodeManager::addReinforcingDependence(ptINode vinstrFrom, ptINode vinstrTo, ptinsth h)
{
  ptdependence dep;
  if (getDependence(vinstrFrom, vinstrTo, &dep))
    return pk_false;
  return addDependence(vinstrFrom, vinstrTo, h, reinforcementDep);
}

/** 
    It adds the (data-flow) dependences n -> m when h, for n in lnodesFrom, and m element of lnodeTo.
    It calls addDependence(ptINode, ptINode , ptinsth ).
*/
void NodeManager::addDependences(ptlINode lnodesFrom, ptlINode lnodeTo, ptinsth h)
{
  ptINode nodeFrom, nodeTo;
  ptlINode list;

  while (ptlINode_get_elem(&lnodesFrom, &nodeFrom))
    {
      list = lnodeTo;
      while (ptlINode_get_elem(&list, &nodeTo))
	addDependence(nodeFrom, nodeTo, h);
    }
}

/**
   It adds the dependence n -> m when (h ^* x1 ^* x2), with
   - n the node of aSig
   - m the node of sigTo
   - x1 he clock of aSig
   - x2 he clock of sigTo

    It is an explicit dependence when bexplicit, a data-flow one otherwise   
*/
boolean NodeManager::addDependence(ptexpcal aSig, ptexpcal sigTo, ptinsth h, boolean bexplicit)
{
  ptINode nodeFrom, nodeTo;
  ptinsth hdep, vh1, vh2;
  TypeDependenceKind tdep;

  nodeFrom = EM->getNode(aSig);
  vh1 = EM->getClock(aSig);

  nodeTo = EM->getNode(sigTo);
  vh2 = EM->getClock(sigTo);
  hdep = CKM->makeInf(vh1, vh2);
  if (h != hundef)
    hdep = CKM->makeInf(hdep, h);
  if (bexplicit)
    tdep = explicitDep;
  else
    tdep = dataFlowDep;
  return addDependence(nodeFrom, nodeTo, hdep, tdep);
}

/** It sets the dependences for downsampling clocks (when Condition) of a list (lh). */
void NodeManager::setDependencesINodeToClocks(ptlinsth lh)
{
  ptinsth vh1;
  while (ptlinsth_get_next(&lh, &vh1))
    {
      vh1 = CKM->getRepresentativeFormula(vh1);
      setDependencesDownSampling(vh1);
    }
}

/**
   For each clock of the list lh, it adds (addDependence(ptinsth, ptINOde)) a dependence from its clock to each
   node computed at this clock.
*/
void NodeManager::setDependencesClocksToINodes(ptlinsth lh)
{
  ptinsth vh1;
  ptgfd vgfd;
  ptINode Node;
  ptlINode linstr;

  while (ptlinsth_get_next(&lh, &vh1))
    {
      vh1 = CKM->getRepresentativeFormula(vh1);
      vgfd = CKM->getClockSubGraph(vh1);
      linstr = ptgfd_get_nodes(vgfd);
      while (ptlINode_get_elem(&linstr, &Node))
	addDependence(vh1, Node);
    }
}

/** 
    It sets a dependence from a node (anode) to a subgraph node (aSubGraph) 
    iff it exists a dependence from anode to one of the node of the subgraph.
*/
boolean NodeManager::setDependencesToSubGraph(ptINode anode, ptINode aSubGraph)
{
  ptexpcal OrderedOutputs, aSig;
  ptINode n;
  boolean b;

  //Currently, only referenced in TrainCluster.

  if (anode == aSubGraph) return pk_false;
  // if (! ptINode_satisfies_subGraph(aSubGraph)) return pk_false;  removed Aug 2016 (dependence on inputs!)

  if (existDependenceFromToOneNodeOf(anode, aSubGraph))
    {
      addExplicitDependence(anode, aSubGraph, hundef);
      addDependenceBooleanClockTo(anode, aSubGraph);
      return pk_true;
    }
  if (isNodeCall(anode))
    {
      b = pk_false;
      OrderedOutputs = get_leftOperand(anode);
      while (EM->getNaryArgs(&OrderedOutputs, &aSig))
	{
	  n = EM->getNode(aSig);
	  if (existDependenceFromToOneNodeOf(n, aSubGraph))
	    {
	      addExplicitDependence(n, aSubGraph, hundef);
	      addDependenceBooleanClockTo(n, aSubGraph);
	      addDependenceBooleanClockTo(anode, aSubGraph);
	      b = pk_true;
	    }
	}
      return b;
    }
  return pk_false;
}

/** 
    It adds the dependences n -> m when h, with
    - n the node of one signal of lFrom,
    - m the node of one signal of lTo,
    
    It is an explicit dependence when bexplicit, a data-flow one otherwise
    It calls addDependence(ptexpcal , ptexpcal, ptinsth, boolean).
*/
void NodeManager::addDependences(ptlinstances lFrom, ptlinstances lTo, ptinsth h, boolean bexplicit)
{
  ptlinstances list;
  ptexpcal aSig, sigTo;

  while (ptlinstances_get_next(&lFrom, &aSig))
    {
      list = lTo;
      while (ptlinstances_get_next(&list, &sigTo))
	addDependence(aSig, sigTo, h, bexplicit);
    }
}

/**
   It Propagates the dependences associated with a node to its sucesssors/predecessors.
*/
void NodeManager::dependencesPropagate(ptINode node)
{
  ptldependence lsucc, lpred, ls;
  ptdependence aDep1, aDep2;
  ptinsth hdep, hdep1, hdep2;
  boolean bexplicit;
  ptINode n11, n12, n21, n22;
  TypeDependenceKind tdep;

  lsucc = get_successors(node);
  lpred = get_predecessors(node);
  while (lpred->get_next(&aDep1, &lpred))
    {
      hdep1 = aDep1->getDependence(&n11, &n12);
      ls = lsucc;
      while (ls->get_next(&aDep2, &ls))
	{
	  hdep2 = aDep2->getDependence(&n21, &n22); // n21 = n12 = node 
	  hdep = CKM->makeInf(hdep1, hdep2);
	  bexplicit = aDep1->satisfiesExplicitDependence() || aDep2->satisfiesExplicitDependence();
	  if (bexplicit) tdep = explicitDep; else tdep = dataFlowDep;
	  addDependence(n11, n22, hdep, tdep);
	}
    }
}

/**
   It propagates the explicit dependences associated with a node (node1) to another one (node2).
*/
void NodeManager::ExplicitdependencesPasteOn(ptINode node1, ptINode node2)
{
  ptldependence ldep;
  ptdependence aDep;
  ptINode n1, n2;
  ptinsth hdep;

  ldep = get_successors(node1);
  while (ldep->get_next(&aDep, &ldep))
    {
      hdep = aDep->getDependence(&n1, &n2);
      if (aDep->satisfiesExplicitDependence())
	if (! is_writeNode(node2))
	  addDependence(node2, n2, hdep, explicitDep);
    }
}

/**
   It propagates the dependences associated with a node (node1) to another one (node2).
   Only the successors of node1 are considered.
*/
void NodeManager::dependencesPasteOn(ptINode node1, ptINode node2)
{
  ptldependence ldep;
  ptdependence aDep;
  ptINode n1, n2;
  ptinsth hnodeTo, hdep;

  hnodeTo = get_nodeClock(node2);
  ldep = get_successors(node1);
  while (ldep->get_next(&aDep, &ldep))
    {
      hdep = aDep->getDependence(&n1, &n2);
      hdep = CKM->makeInf(hdep, hnodeTo);
      addDependence(node2, n2, hdep, aDep->getDependenceKind());
    }
}

/**
   It propagates the dependences associated with a node (nodeFrom) to another one (nodeTo).
   Predecessors/successors of nodeFrom are considered.
*/
void NodeManager::dependencesMove(ptINode nodeFrom, ptINode nodeTo)
{
  ptldependence ldep;
  ptdependence aDep;
  ptlinsth lh;
  ptINode n1, n2;
  ptinsth hnodeTo, hdep, vh1;
  if (nodeFrom != nodeTo)
    {
      hnodeTo = get_nodeClock(nodeTo);
      ldep = get_predecessors(nodeFrom);
      while (ldep->get_next(&aDep, &ldep))
	{
	  hdep = aDep->getDependence(&n1, &n2);
	  hdep = CKM->makeInf(hdep, hnodeTo);
	  addDependence(n1, nodeTo, hdep, aDep->getDependenceKind());
	}
      lh = get_clockPredecessors(nodeFrom);
      while (ptlinsth_get_next(&lh, &vh1))
	{
	  vh1 = CKM->getRepresentativeFormula(vh1);
	  addDependence(vh1, nodeTo);
	}
      
      dependencesPasteOn(nodeFrom, nodeTo);
      lh = get_clockSuccessors(nodeFrom);
      while (ptlinsth_get_next(&lh, &vh1))
	{
	  vh1 = CKM->getRepresentativeFormula(vh1);
	  addDependence(nodeTo, vh1);
	}
    }
}

/**
   It removes the dependences node -> obj, when obj (a clock, or a node) is marked (mark).
*/
void NodeManager::removeDependenceMarkedSuccessors(ptINode node, typeMark mark)
{
  ptldependence list, nlist;
  ptlinsth lhuse;
  ptINode n1, n2;
  ptinsth h;
  ptdependence el;

  lhuse = get_clockSuccessors(node);
  lhuse = ptlinsth_removeMarked(lhuse, mark);
  set_clocksSuccessors(node, lhuse);
  list = get_successors(node);
  nlist = list;
  while (list->get_next(&el, &list))
    {
      h = el->getDependence(&n1, &n2);
      if (isMarked(n2, mark))
	nlist->remove(el);
    }
  set_successors(node, nlist);
}

/**
   @return "true" if a node (node) is referenced in the graph.
*/
boolean NodeManager::isReferenced(ptINode node)
{
  ptlinsth lh;
  ptldependence lsucc;
  ptdependence aDep;
  ptinsth h1;

  lsucc = get_successors(node);
  if (lsucc->get_next(&aDep, &lsucc))
    return pk_true;
  lh = get_clockSuccessors(node);
  return ptlinsth_get_next(&lh, &h1);
}

/**
   @return the number of explicit dependence it exists from a node (node).
*/
int NodeManager::getCardExplicitPredecessors(ptINode node)
{
  ptldependence lpred;
  ptdependence aDep;
  int n;

  lpred = get_predecessors(node);
  n = 0;
  while (lpred->get_next(&aDep, &lpred))
    if (aDep->satisfiesExplicitDependence())
      n++;
  // n = n + ptlinsth_size(node->get_clockPredecessors());
  return n;
}

/** It removes the dependences reaching a node (node). */
void NodeManager::removeDependenceTo(ptINode node)
{
  removeNodeDependencesReaching(node);
  removeClockDependencesReaching(node);
}

/** It removes the dependences outgoing from a node (node). */
void NodeManager::removeDependenceFrom(ptINode node)
{
  ptldependence lsucc;
  ptdependence aDep;
  ptlinsth lh;
  ptinsth h1;
  ptINode in;

  // TO BE COMPLETED
  lsucc = get_successors(node);
  while (lsucc->get_next(&aDep, &lsucc))
    removeDependence(aDep);

  lh = get_clockSuccessors(node);
  while (ptlinsth_get_next(&lh, &h1)) ;	// ????????

  if (get_subGraphInputNode(node, &in))
    removeDependenceFrom(in);	// added Sep 2014
}

/** It removes the dependences outgoing from a clock (vh). */
void NodeManager::removeDependenceFrom(ptinsth vh)
{
  ptlsucc vlsucci;
  ptINode vinstr;
  ptldependence li;
  ptdependence aDep;
  ptlinsth lh;

  vlsucci = CKM->getNodeReferencesTo(vh);
  while (ptlINode_get_elem(&vlsucci, &vinstr))
    {
      li = get_predecessors(vinstr);
      while (li->get_next(&aDep, &li))
	if (CKM->getRepresentativeFormula(aDep->getClockDependence()) == vh)
	  removeDependence(aDep);

      lh = get_clockPredecessors(vinstr);
      if (ptlinsth_sorted_remove_elem(&lh, vh))
	set_clockPredecessors(vinstr, lh);
    }
  CKM->setNodeReferencesTo(vh, NILLINODE);
}

/** It removes the dependences reaching a clock (vh). */
void NodeManager::removeDependenceTo(ptinsth vh)
{
  ptexpcal sig;
  ptINode node;
  ptlinstances lsig, l;

  CKM->removeLinksTo(vh);
  if (CKM->satisfiesSub(vh))
    {
      // lsig = CKM->getSignalsOfCondition(vh); ?? a revoir.
      lsig = EM->getSignalsReferencedIn(CKM->getConditionSub(vh));
      l = lsig;
      while (ptlinstances_get_next(&lsig, &sig))
	{
	  node = EM->getNode(sig);
	  removeDependence(node, vh);
	}
      ptlinstances_free(l);
    }
}

/** It removes the dependences associated with the predefined null clock. */
void NodeManager::removeDependenceNullClock()
{
  ptinsth vh;

  vh = CKM->getNullClock();
  removeDependenceFrom(vh);
  removeDependenceWithLabel(vh);
}

/** It removes a dependence (aDep). */
void NodeManager::removeDependence(ptdependence aDep)
{
  ptinsth hdep;
  ptldependence ls;
  ptINode n1, n2;

  // TO BE COMPLETED FOR THE EXPLICIT DEPENDENCES.
  hdep = aDep->getDependence(&n1, &n2);
  ls = get_successors(n1);
  ls = ls->remove(aDep);
  set_successors(n1, ls);
  ls = get_predecessors(n2);
  ls = ls->remove(aDep);
  set_predecessors(n2, ls);

  GdependencesList = GdependencesList->remove(aDep);
}

/** It removes the dependence vh -> vNodeTo. */
void NodeManager::removeDependence(ptinsth vh, ptINode vNodeTo)
{
  ptlsucc lsuc;
  ptlinsth lh;

  lsuc = CKM->getNodeReferencesTo(vh);
  if (ptlINode_remove_elem(&lsuc, vNodeTo))
    CKM->setNodeReferencesTo(vh, lsuc);
  lh = get_clockPredecessors(vNodeTo);
  if (ptlinsth_sorted_remove_elem(&lh, vh))
    set_clockPredecessors(vNodeTo, lh);
}

/** It removes the dependence vinstr -> vinsth. */
void NodeManager::removeDependence(ptINode vinstr, ptinsth vinsth)
{
  ptlinsth vlinsth;

  vlinsth = get_clockSuccessors(vinstr);
  if (ptlinsth_sorted_remove_elem(&vlinsth, vinsth))
    set_clocksSuccessors(vinstr, vlinsth);
}

/** It removes the dependence vNodeFrom -> vNodeTo. */
void NodeManager::removeDependence(ptINode vNodeFrom, ptINode vNodeTo)
{
  ptldependence li;
  ptdependence aDep;

  li = get_successors(vNodeFrom);
  if (li->exist_dependence(vNodeFrom, vNodeTo, &aDep))
    removeDependence(aDep);
}

/**
   @return
   - "true" if it exists a dependence vFrom -> vTo, returned in aDep
   - "false" otherwise, aDep is not modified.
*/
boolean NodeManager::getDependence(ptINode vFrom, ptINode vTo, ptdependence * aDep)
{
  ptldependence li;
  li = get_successors(vFrom);
  return li->exist_dependence(vFrom, vTo, aDep);
}

/**
   It sets the dependences for each node call of the graph.
*/
void NodeManager::setDependencesNodeCall()
{
  ptlINode ln, lReadsProducedByNodeCall, laux;
  ptINode aNodeCall, n, nout;
  ptexpcal nodeCallInputs, e;
  ptinsth hdep;
  boolean b;
  ptldependence lsucc;
  ptdependence aDep;

  if (getNodeCalls(&ln))
    while (ptlINode_get_elem(&ln, &aNodeCall))
      {
	lReadsProducedByNodeCall = ptINode_get_nodecallProducedNodes(aNodeCall);
	hdep = get_nodeClock(aNodeCall);	// hdep = hundef;

	get_rightOperand(aNodeCall, &nodeCallInputs);
	while (EM->getNaryArgs(&nodeCallInputs, &e))
	  if (addDependenceSig2Node(e, aNodeCall, hdep))
	    {
	      n = EM->getNode(e);
	      lsucc = get_successors(n);
	      addDependence(n, aNodeCall, hdep);
	      b = pk_false;
	      laux = lReadsProducedByNodeCall;
	      while (ptlINode_get_elem(&laux, &nout) && !b)
		if (lsucc->exist_dependence(n, nout, &aDep))
		  {
		    addDependence(aNodeCall, nout, hdep);
		    b = pk_true;
		  }
	    }
	// For Other dependences see setsbDCplusDependences().
      }
}

/**
   It sets the dependences induced by labelled objects:
   the internal nodes of a subgraph depends on a "fictitious" node (internal to the representation).
*/
void NodeManager::setDependencesLabels()
{
  ptINode InternalInputNode, aNode, aSubGraph;
  ptlINode lSubGraphs, ln;
  ptlinsth lh;
  ptinsth vhemb;
  ptldependence lsucc;

  if (getSubGraphs(&lSubGraphs))
    while (ptlINode_get_elem(&lSubGraphs, &aSubGraph))
      {
	lsucc = get_successors(aSubGraph);
	// lhsucc = aSubGraph->get_clockSuccessors(); TBC
	get_subGraphInputNode(aSubGraph, &InternalInputNode);
	get_subGraphDefinitions(aSubGraph, &ln, &lh);
	vhemb = hundef;
	addInternalDependences(InternalInputNode, ln, lh, vhemb);
	addComposedNodesDependences(aSubGraph, NILINODE, InternalInputNode, ln, lh);
	// while (ptlINode_get_elem(&ln, &aNode)) addDependenceFromTo(aNode, lsucc);  // removed Oct 2014 (LB)
      }
}

/** 
    It sets the dependences for array of processes.
    Internal nodes of an array depends on the index of the array.
*/
void NodeManager::setDependencesArrays()
{
  ptINode vinternalArrayNode, anArrayNode;
  ptlINode larrays, ln;
  ptlinsth lh;

  if (getIterates(&larrays))
    while (ptlINode_get_elem(&larrays, &anArrayNode))
      {
	arrayLinkNodes(anArrayNode, pk_false);
	get_array_inputNode(anArrayNode, &vinternalArrayNode);
	get_array_bodyDefinitions(anArrayNode, &ln, &lh);
	addComposedNodesDependences(anArrayNode, get_indexNode(anArrayNode), vinternalArrayNode, ln, lh);
	addInternalDependences(vinternalArrayNode, ln, lh, hundef);
      }
}

/** 
    Assumes: node is a list of partial defintions node.
    It adds a dependence from each component to node. For
    - N1 defined by x::=E1, 
    - ... 
    - Nn defined by x:=En
    - N the node (node) associated to x
    .
    it adds Ni->N
*/
void NodeManager::setDependencesPartDefinition(ptINode node)
{
  ptlINode ln;
  ptINode n;
  ptinsth hdep;

  ln = get_setOfPartialDefinitions_arg(node);
  while (ptlINode_get_elem(&ln, &n))
    {
      hdep = get_nodeClock(n);
      addDependence(n, node, hdep);
    }
}

/** 
    It sets the dependences for nodes defined by a set of partial definitions. 
    The node depends on all part definition nodes. 
*/
void NodeManager::setDependencesPartDefinitions()
{
  ptlINode lnodes;
  ptINode node;
  if (getGroupsOfPartialEquations(&lnodes))
    while (ptlINode_get_elem(&lnodes, &node))
      setDependencesPartDefinition(node);
}

/**
   It reenforces the dependences for node calls: it adds a dependence from each input to each output of such a node.
   It is used for sequential code  generators.
   \note it may add cycles in the graph.
   @return "true" if some dependences has been added, "false" otherwise.
*/
boolean NodeManager::setsbDCplusDependencesForNodeCalls()
{
  ptinsth h, hdep;
  ptlINode ln, lReadsProducedByNodeCall;
  ptINode aNodeCall, aNodeCall1, node;
  boolean b;
  ptldependence lsucc;
  ptINode n1, n2;
  ptdependence aDep;

  if (!getNodeCalls(&ln))
    return pk_false;
  hdep = hundef;
  b = pk_false;
  while (ptlINode_get_elem(&ln, &aNodeCall))
    {
      lReadsProducedByNodeCall = ptINode_get_nodecallProducedNodes(aNodeCall);
      // hdep = get_nodeClock(aNodeCall );
      while (ptlINode_get_elem(&lReadsProducedByNodeCall, &node))
	{
	  addDependence(aNodeCall, node, hdep);
	  b = pk_true;
	  // if node--> n1 and n1 is produced by an another node call aNodeCall1, then aNodeCall -> aNodeCall1
	  lsucc = get_successors(node);
	  while (lsucc->get_next(&aDep, &lsucc))
	    {
	      h = aDep->getDependence(&n1, &n2);
	      if (is_outputNodeCall(n2, &aNodeCall1))
		addDependence(aNodeCall, aNodeCall1, hdep);
	    }
	}
    }
  return b;
}

/**
   Assumed: the operand expression is a normalized default or a when expression.
   It adds a dependence for the signals of the expression (ne) to a node (node).
*/
void NodeManager::addDependences(ptINode node, ptexpcal ne)
{
  ptexpcal arg;
  ptinsth h;

  if (EM->getOperator(ne) == opmele)
    {
      addDependences(node, EM->getDefaultArgstAt(ne, 1));
      addDependences(node, EM->getDefaultArgstAt(ne, 2));
    }
  else
    {
      arg = EM->getWhenOrCellArg(ne);
      h = EM->getContributionClock(ne);
      addDependenceSigList2NodeAt(EM->getSignalsReferencedIn(arg), node, h);
    }
}

/**
   Replace the dependence (oNode --> vTo) by the dependence (nNode  --> vTo). The clock of the dependence is kept.
   @return true if the operation succeeds, "false" otherwie (no dependence (oNode  --> vTo))
*/
boolean NodeManager::replaceDependence(ptINode oNode, ptINode nTo, ptINode nNode)
{
  ptinsth hdep;
  ptdependence aDep;

  if (!getDependence(oNode, nTo, &aDep))
    return pk_false;
  hdep = aDep->getClockDependence();
  removeDependence(aDep);
  addDependence(nNode, nTo, hdep);
  return pk_true;
}

void NodeManager::free_dependences(ptldependence ldep)
{
  ptdependence aDep;
  while (ldep->get_next(&aDep, &ldep))
    delete aDep;
}



/**
   Reinforcing the dependencies for the "rendez-vous".
   The send,receive are done before all actions at the clock of the rendez-vous.
   A new node pK_RendezVous(rdv1, rdv2,...,rdvn) is added to the graph:
    - rdv1 is the symbol output of the pk_send() node, 
    - the other ones are the symbols of the pk_receive() nodes.
    @return the createdx  node  pK_RendezVous(rdv1, rdv2,...,rdvn).
*/
ptINode NodeManager::setDependencesRendezVous(ptinsth vh, ptlinstances outputs)
{
  ptlINode ln, lnodes2, lnodes1, lnodes;
  ptINode RendezVous,SendNode, aReceiveNode, node1, node2;
  ptlinsth lh;
  ptinsth masterClock,vh1;
  ptexpcal linputs, idCond, vinputs,vsig, voutput, vinput, vlab;
  boolean addRdv;
  ptlinstances lids,l;

  // ADDED FOR POLYENDO
  lnodes = CKM->getNodesAssignedTo(vh);
  lnodes1= lnodes;
  addRdv = pk_false;
  linputs = NILEXPCAL;
  while (ptlINode_get_elem(&lnodes1, &node1))
    {
      if (isReceiveNode(node1))
	{
	  addRdv = pk_true;
	  addDependencesPolyEndo(node1, lnodes, vh);
	  lh = CKM->getClockReferencesTo(vh);
	  while (ptlinsth_get_next(&lh, &vh1))
	    addDependence(node1, vh1);
	  aReceiveNode = node1;
	  // list of the signals of the rendezvous.
	  vinputs = get_rightOperand(node1);
	  vsig = EM->getNaryArgs(&vinputs);
	  linputs = EM->mkNaryExp(oplesign, vsig, linputs, NILBACKANNOTATION);
	}
      else
	if (isSendNode(node1))
	  {
	    lnodes2= lnodes;
	    while (ptlINode_get_elem(&lnodes2, &node2))
	      {
		if (! isSendNode(node2))
		    addExplicitDependence(node1, node2, vh);
	      }
	    SendNode = node1;
	  }
    }
  if (!addRdv) return NILINODE; 

  /* A new node pK_RendezVous(rdv1, rdv2,...) is added to the graph, 
     rdv1 is the symbol output of the pk_send() node, 
     the other ones are the input symbols of the pk_receive() nodes.
  */
  vinputs = get_leftOperand(SendNode);
  vsig = EM->getNaryArgs(&vinputs);
  linputs = EM->mkNaryExp(oplesign, vsig, linputs, NILBACKANNOTATION);
  RendezVous = mkIntrinsicUnsafeAt(getRendezVousId(), NILEXPCAL, linputs, NILEXPCAL, vh, NILBACKANNOTATION);
  vlab = EM->mkLabelIdent (getLabelId (), mk_node_undefined (), vh, NILBACKANNOTATION, pk_false);
  set_label (RendezVous, vlab);
  EM->setNode(vlab, RendezVous);
  ExplicitdependencesPasteOn(aReceiveNode, RendezVous);
  lnodes1= lnodes;
  while (ptlINode_get_elem(&lnodes1, &node1))
    if (isReceiveNode(node1))
      addExplicitDependence(node1, RendezVous, vh);
  
  return RendezVous;
}

boolean NodeManager::isSelectedDependenceForAbstraction(ptdependence aDep, ptlINode lnodes, ptlinstances ls)
{
  ptINode vFrom, vTo;
  ptexpcal sig;
  ptinsth hdep;

  hdep = aDep->getDependence(&vFrom, &vTo);
  hdep = CKM->getRepresentativeFormula(hdep);
  if (CKM->satisfiesNullClock(hdep)) return pk_false;
  if ( ! ptlINode_isIn(lnodes, vFrom) && ! ptlINode_isIn(lnodes, vTo))
    {
      if (! getDependenceNodeLabel(vFrom, &sig)) return pk_false;
      if (! (EM->IsInputFlow(sig) || EM->IsOutputFlow(sig))) return pk_false;
      if (! getDependenceNodeLabel(vTo, &sig)) return pk_false;
      if (!(EM->IsInputFlow(sig) || EM->IsOutputFlow(sig))) return pk_false;
    }
  return pk_true;
}

boolean NodeManager::isSelectedDependenceForAbstraction(ptdependence aDep, ptlINode lnodes)
{
  ptINode vFrom, vTo;
  ptexpcal sig;
  ptinsth hdep;

  hdep = aDep->getDependence(&vFrom, &vTo);
  hdep = CKM->getRepresentativeFormula(hdep);
  if (CKM->satisfiesNullClock(hdep)) return pk_false;
  if ( ! ptlINode_isIn(lnodes, vFrom) && ! ptlINode_isIn(lnodes, vTo))
    {
      if (! getDependenceNodeLabel(vFrom, &sig)) return pk_false;
      if (! (EM->IsInputFlow(sig) || EM->IsOutputFlow(sig))) return pk_false;
      if (! getDependenceNodeLabel(vTo, &sig)) return pk_false;
      if (!(EM->IsInputFlow(sig) || EM->IsOutputFlow(sig))) return pk_false;
    }
  return pk_true;
}


/**
   @return true it exists (vsig) a label or an output for the operand node, false otherwise.
*/
boolean NodeManager::getDependenceNodeLabel(ptINode vnode, ptexpcal * vsig)
{
  ptexpcal l;

  if (get_label(vnode, vsig)) return pk_true;
  if (is_ioNode(vnode))
    {
      *vsig = get_io_arg(vnode);
      return pk_true;
    }

  *vsig = get_leftOperand(vnode);
  if ((*vsig) == NILEXPCAL)
    return pk_false;
  if (!EM->satisfiesList(*vsig))
    return pk_true;
  // several outputs ? an i/o is selected if it exists such a signal
  l = *vsig;
  while (EM->getNaryArgs(&l, vsig))
    {
      if (EM->IsInputFlow(*vsig))
	return pk_true;
      if (EM->IsOutputFlow(*vsig))
	return pk_true;
    }
  return pk_true;
}

boolean NodeManager::getClockDependence(ptldependence lDep, ptINode vFrom, ptINode vTo, ptinsth *hDep)
{
  ptdependence aDep;
  ptINode vCurrentFrom, vCurrentTo;

  while (lDep->get_next(&aDep, &lDep))
    {
      *hDep = CKM->getRepresentativeFormula(aDep->getDependence(&vCurrentFrom, &vCurrentTo));
      if (( vCurrentFrom == vFrom) && ( vCurrentTo == vTo )) return pk_true;
    }
  return pk_false;
}
