/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   Added to take into account the substitutable clocks.
   An input/output clock of the interface(vinterface) that is substitutable and reduced to a signal, is replaced by this signal.
*/
    void
 NodeManager::updateInterfaceSubstitutableClocks(ptinterface vinterface)
{
  ptlinsth lh;
  ptinsth vh;
  ptexpcal cond;
  ptlinstances lsig;

  lh = ptinterface_get_inputClocks(vinterface);
  if (lh != NILLINSTH)
    while (ptlinsth_get_next(&lh, &vh))
      if (CKM->isSubstitutable(vh))
	{
	  cond = CKM->getConditionSub(vh);
	  if (EM->getOperatorClass(cond) == variabl)
	    {
	      ptinterface_remove_inputClock(vinterface, vh);
	      ptinterface_insert_inputSignal(vinterface, cond);
	    }
	}

  lh = ptinterface_get_outputClocks(vinterface);
  if (lh != NILLINSTH)
    while (ptlinsth_get_next(&lh, &vh))
      if (CKM->isSubstitutable(vh))
	{
	  cond = CKM->getConditionSub(vh);
	  if (EM->getOperatorClass(cond) == variabl)
	    {
	      ptinterface_remove_outputClock(vinterface, vh);
	      ptinterface_insert_outputSignal(vinterface, cond);
	    }
	}
}

/**
   @return the interface of a list of nodes (ln). All the signals defined in the list are declared as output
*/
ptinterface NodeManager::interfaceListINodes(ptlINode ln)
{
  ptINode aNode;
  ptinterface vinterface;

  vinterface = NILINTERFACE;
  while (ptlINode_get_elem(&ln, &aNode))
    if (!is_writeNode(aNode))
      vinterface = ptinterface_union(vinterface, getSet_interface(aNode));
  return vinterface;
}

/**
   For an extraction (vh), it inserts,in an interface (vinterface), as inputs the 
   signals referenced in the condition associated with vh. 
*/
void NodeManager::addInputSignalsForExtraction(ptinterface vinterface, ptinsth vh)
{
  ptlinstances lsig, l;
  ptexpcal aSig;
  if (!CKM->satisfiesSub(vh))
    return;
  //  lsig = CKM->getSignalsOfCondition(vh);
  lsig = EM->getSignalsReferencedIn(CKM->getConditionSub(vh));
  l = lsig;
  while (ptlinstances_get_next(&lsig, &aSig))
    vinterface = ptinterface_insert_inputSignal(vinterface, aSig);
  ptlinstances_free(l);
}

/**
   Adding vh to lh if there is no equivalent clock to vh in lhDefinedInSubGraph.
*/
ptlinsth NodeManager::addClockSignalsFiltered(ptlinsth lh, ptinsth vh, ptlinsth lhDefinedInSubGraph)
{
  // added for polyendochrony
  ptinsth vhaux;
  boolean b;
  
  if ( ! CKM->hasAnEquivalenceIn(vh,lhDefinedInSubGraph , &vhaux))
    lh = ptlinsth_sorted_insert_elem_bool (lh, vh, &b);
  return lh;
}

/**
   @return the list composed of the clock of each signal of a list of signals (lsig) added in lh.
   Such a clock is not added if it exists an equivelent clock in the list lh (polyendochrony).
*/
ptlinsth NodeManager::addClockSignalsFiltered (ptlinstances lsig, ptlinsth lh, ptlinsth lhDefinedInSubGraph)
{
  ptexpcal vinst;

  while (ptlinstances_get_next (&lsig, &vinst))
    if (! EM->isStateVar (vinst))
      {
	lh = addClockSignalsFiltered(lh,  EM->getClock (vinst), lhDefinedInSubGraph);
      }
  return lh;
}


/**
   @return the branching clock of the clocks referenced in the interface after the completion.
   It completes an interface (vinterface) by adding as input, the clocks not in lh1, of
   - the input signals of the interface,
   - the nodes of the list ln.
*/
ptinsth NodeManager::SubGraphSetClock(ptinterface vinterface, ptlINode ln, ptlinsth lhDefinedInSubGraph)
{
  ptlinsth lh;
  ptlinstances ls;
  ptinsth vhemb, vh;
  ptINode anode;

  lh = ptlinsth_duplicate(ptinterface_get_inputClocks(vinterface));
  lh = ptlinsth_sorted_fusion(lh, ptinterface_get_localClocks(vinterface));
  lh = ptlinsth_sorted_fusion(lh, ptinterface_get_outputClocks(vinterface));

  ls = ptinterface_get_inputSignals(vinterface);
  lh = addClockSignalsFiltered(ls, lh, lhDefinedInSubGraph);

  ls = ptinterface_get_outputSignals(vinterface);
  lh = addClockSignalsFiltered(ls, lh, lhDefinedInSubGraph);

  while (ptlINode_get_elem(&ln, &anode))
    {
      lh = addClockSignalsFiltered(lh, get_nodeClock(anode), lhDefinedInSubGraph);
    }

  vhemb = CKM->getBranchingClocks(lh);
  if (CKM->satisfiesUndefClock(vhemb))
    {
      vhemb = CKM->getTickClock();
      if (CKM->satisfiesUndefClock(vhemb))
	vhemb = CKM->getFictitiousRoot();
    }
  while (ptlinsth_get_next(&lh, &vh))
    if (!ptlinsth_isin_elem(lhDefinedInSubGraph, vh))
      ptinterface_insert_inputClock(vinterface, vh);
  return vhemb;
}

/**
   For each clock of lh:
   - it inserts it as a local object in the interface (vinterface)
   - it inserts as inputs, the signals referenced in the condition when the clock is an extraction (addInputSignalsForExtraction).
*/
ptinterface NodeManager::SubGraphInterfaceEnd(ptinterface vinterface, ptlinsth lh)
{
  ptinsth vh;

  while (ptlinsth_get_next(&lh, &vh))
    {
      vinterface = ptinterface_insert_localClock(vinterface, vh);
      addInputSignalsForExtraction(vinterface, vh);
    }
  return vinterface;
}



/**
   @return the interface of a graph (vgfd) and the branching clock of its nodes/clocks (vhemb).
   It calls interfaceListINodes(),SubGraphSetClock(),SubGraphInterfaceEnd().
*/
ptinterface NodeManager::buildSubGraphInterface(ptgfd vgfd, ptinsth * vhemb)
{
  ptinterface vinterface;
  ptlINode ln;
  ptlinsth lh;

  if (ptgfd_satisfies_empty(vgfd))
    {
      (*vhemb) = CKM->getUndefClock();
      return NILINTERFACE;
    }
  vinterface = ptgfd_get_interface(vgfd);
  if (vinterface != NILINTERFACE)
    return vinterface;
  ln = ptgfd_get_nodes(vgfd);
  lh = ptgfd_get_clocks(vgfd);
  vinterface = interfaceListINodes(ln);
  // pour eviter bouclage
  if (vinterface == NILINTERFACE)
    vinterface = ptinterface_create();
  ptgfd_set_interface(vgfd, vinterface);
  *vhemb = SubGraphSetClock(vinterface, ln, lh);
  vinterface = SubGraphInterfaceEnd(vinterface, lh);
  return vinterface;
}

/**
   @return the interface of the sub-graph of the body of an assumed array node (aNode).
   It calls buildSubGraphInterface().
*/
ptinterface NodeManager::bodyArrayInterface(ptINode aNode)
{
  ptinsth vhemb;
  ptinterface vinterface;

  vinterface = buildSubGraphInterface(get_array_bodyGraph(aNode), &vhemb);
  return vinterface;
}

/**
   @return the interface of the sub-graph of the initial part (part with) of an assumed array node (aNode).
   It calls buildSubGraphInterface().
*/
ptinterface NodeManager::initArrayInterface(ptINode aNode)
{
  ptinsth vhemb;
  ptinterface vinterface;

  vinterface = buildSubGraphInterface(get_array_initGraph(aNode), &vhemb);
  return vinterface;
}

ptinterface NodeManager::addInputClock(ptinterface vinterface, ptinsth vh)
{
  ptlinstances lsig;
  ptexpcal vsig;

  if (CKM->isSubstitutable(vh))	// resticted to extraction.
    {
      lsig = EM->getSignalsReferencedIn(CKM->getConditionSub(vh));
      while (ptlinstances_get_next(&lsig, &vsig))
	vinterface = ptinterface_insert_inputSignal(vinterface, vsig);
    }
  else if (CKM->isClockRepresentedBySignal(vh, &vsig))
    vinterface = ptinterface_insert_inputSignal(vinterface, vsig);
  else
    vinterface = ptinterface_insert_inputClock(vinterface, vh);
  return vinterface;
}

/**
   Adds the clocks of the operand list (lh) to the operand interface (vinterface) taking into account 
   the substitutable ones.
*/
ptinterface NodeManager::addInputClocks(ptinterface vinterface, ptlinsth lh)
{
  ptinsth vh;

  // vinterface = ptinterface_set_inputClocks(vinterface,lh); replaced by (substituable clocks)
  while (ptlinsth_get_next(&lh, &vh))
    vinterface = addInputClock(vinterface, vh);
  return vinterface;
}

/** @return the interface of a node (aNode). 
    When the interface is undefined, it is computed.
*/
ptinterface NodeManager::buildinterface(ptINode aNode)
{
  ptexpcal vopg, vexpsigres, vexpsigret, vexpret, vexpfen, lp, lin, lout, vdedef, le, ls, vinit;
  ptlinstances lsort, lent, lpar;
  ptinterface vinterface, vinter1, vinter2;
  ptinsth vhemb;
  ptDCG vgraph;
  ptlinsth lh;
  boolean bdummy;
  ptident id;

  vinterface = get_interface(aNode);
  if (vinterface != NILINTERFACE)
    return vinterface;
  switch (get_oper(aNode))
    {
    case opret:
      get_delay_arg(aNode, &vexpsigres, &vexpsigret, &vexpret, &vexpfen, &vinit);
      lent = ptlinstances_sortedfusion(EM->getIdsReferencedIn(vexpret), EM->getIdsReferencedIn(vexpfen));
      lent = ptlinstances_sortedfusion(EM->getIdsReferencedIn(vexpsigret), lent);
      lent = ptlinstances_sortedfusion(EM->getIdsReferencedIn(vinit), lent);
      vinterface = inSplipListToInterface(vinterface, lent);
      vinterface = outSplipListToInterface(vinterface, EM->getSignalsReferencedIn(vexpsigres));
      break;

    case opdefsig:		// :=
    case opdefpart:		// ::= 
      get_equation_arg(aNode, &vopg, &vdedef);
      if (get_rightOperand(aNode, &vdedef))
	{
	  if (EM->IsCellExpr(vdedef))
	    vdedef = EM->getWhenOrCellArg(vdedef);
	  lent = ptlinstances_sortedfusion(EM->getIdsReferencedIn(vdedef), EM->getIdsReferencedIn(EM->getInitialValue(vopg)));
	  vinterface = inSplipListToInterface(vinterface, lent);
	  lh = EM->getClocksReferencedIn(vdedef);
	  lh = ptlinsth_remove_elem(lh, get_nodeClock(aNode), &bdummy);
	  lh = ptlinsth_remove_elem(lh, CKM->getInitClock(), &bdummy);
	  vinterface = addInputClocks(vinterface, lh);
	}
      vinterface = outSplipListToInterface(vinterface, EM->getSignalsReferencedIn(vopg));
      break;

    case opappe:
      get_procCall_arg(aNode, &id, &lp, &le, &ls);
      lent = EM->getIdsReferencedIn(le);
      lsort = EM->getSignalsReferencedIn(ls);
      lh = EM->getClocksReferencedIn(le);
      vinterface = inSplipListToInterface(vinterface, lent);
      vinterface = outSplipListToInterface(vinterface, lsort);
      vinterface = addInputClocks(vinterface, lh);
      break;

    case opmoti:
      get_array_io(aNode, &le, &ls);
      lent = EM->getIdsReferencedIn(le);
      lsort = EM->getSignalsReferencedIn(ls);
      vinterface = inSplipListToInterface(vinterface, lent);
      vinterface = outSplipListToInterface(vinterface, lsort);
      vinter1 = bodyArrayInterface(aNode);
      vinter2 = initArrayInterface(aNode);
      vinterface = ptinterface_union(vinterface, vinter1);
      vinterface = ptinterface_union(vinterface, vinter2);
      break;

    case opnodecall:
      get_nodeCall_arg(aNode, &vgraph, &lp, &lin, &lout);
      lpar = EM->getIdsReferencedIn(lp);
      lent = EM->getIdsReferencedIn(lin);
      lsort = EM->getSignalsReferencedIn(lout);
      vinterface = inSplipListToInterface(vinterface, lent);
      vinterface = outSplipListToInterface(vinterface, lsort);
      vinterface = ptinterface_set_parameters(vinterface, lpar);
      break;

    case opsubgraph:
      vinterface = buildSubGraphInterface(ptINode_get_subGraph(aNode), &vhemb);
      set_subGraphProjectedSystemInfos(aNode, vhemb, NILLINSTH);
      break;

    case opnil:
      vinterface = NILINTERFACE;
      break;

    case opbund:		// fields (struct, bundle)
      vinterface = NILINTERFACE;
      break;
    case opldefpart:		// set of ::= nodes
      vinterface = interfaceListINodes(get_setOfPartialDefinitions_arg(aNode));
      break;
    default:;
    }
  updateInterfaceSubstitutableClocks(vinterface);
  set_interface(aNode, vinterface);
  return vinterface;
}

/**
   It switches the outputs (clocks and signals) to the locals excepts some signals (lexcept) and some clocks (lhexcept).
*/
void NodeManager::moveOutputs2localsExceptClocks(ptinterface vinterface, ptlinsth lhexcept)
{
  ptlinsth lhout;
  ptinsth h;

  lhout = ptinterface_get_outputClocks(vinterface);
  while (ptlinsth_get_next(&lhout, &h))
    if (!ptlinsth_isin_elem(lhexcept, h))
      {
	ptinterface_remove_outputClock(vinterface, h);
	ptinterface_insert_localClock(vinterface, h);
      }
}

/**
   It switches the output signals to the locals excepts some signals (lexcept).
*/
void NodeManager::moveOutputs2localsExceptSignals(ptinterface vinterface, ptlinstances lexcept)
{
  ptlinstances lout;
  ptexpcal vsig;

  lout = ptinterface_get_outputSignals(vinterface);
  while (ptlinstances_get_next(&lout, &vsig))
    if (!ptlinstances_isInSortedList(lexcept, vsig) && !EM->isSharedSignalBySeveralGraphs(vsig))
      {
	ptinterface_remove_outputSignal(vinterface, vsig);
	ptinterface_insert_localSignal(vinterface, vsig);
      }
}
