/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    Modifying a definition.

    It replaces the definition of the signal vexp by "vexp := nexp at vhcal", 
    the new definition is inserted at the vhnode clock.
    If the new expression (nexp) is
    - a delay expression, the node of vexp becomes a memorization node,
    - a function call, it is replaced by a procCall node
    - otherwise, it is an equation.
 */
    void
 NodeManager::modifiedDefinition(ptexpcal vexp, ptexpcal nexp, ptinsth vhcal, ptinsth vhnode)
{
  ptINode aNode;
  noperators op;
  ptexpcal parameters, inputs, outputs;
  ptident id;

  if (vexp == NILEXPCAL)
    {
      printf("##INTERNAL ERROR##: in NodeManager::modifiedDefinition, vexp is nil\n");
      fflush(stdout);
      return;
    }
  if (vexp == NILEXPCAL)
    {
      printf("##INTERNAL ERROR##: in NodeManager::modifiedDefinition, nexp is nil\n");
      fflush(stdout);
      return;
    }
  if (vhcal == NILINSTH)
    {
      printf("##INTERNAL ERROR##: in NodeManager::modifiedDefinition, vhcal is nil\n");
      fflush(stdout);
      return;
    }
  if (vhnode == NILINSTH)
    {
      printf("##INTERNAL ERROR##: in NodeManager::modifiedDefinition, vhnode is nil\n");
      fflush(stdout);
      return;
    }
  if (vexp != nexp)
    {
      aNode = EM->getNode(vexp);
      removeDependenceTo(aNode);
      removeInSetCase(aNode);
      ptINode_set_interface(aNode, NILINTERFACE);

      op = EM->getOperator(nexp);
      if (op == opret)
	ptINode_becomes_delay(aNode, vexp, nexp);
      else if (op == opappe && (!EM->isMarked(nexp, DONTDECLARE)))
	{
	  outputs = EM->mkNaryExp(oplesign, vexp, NILEXPCAL, NILBACKANNOTATION);
	  EM->getFunctionCall(nexp, &id, &parameters, &inputs);
	  ptINode_becomes_extern(aNode, id, parameters, inputs, outputs);
	}
      else
	{
	  ptINode_resetAsEquation(aNode);
	  addaff(vexp, nexp, vhcal, EM->getPredecessors(nexp));
	}
      insertNode(aNode, vhnode);
      getSet_interface(aNode);
    }
}

/** 
    Switching dÃ©finitions. 

    For two equations/delays n1, x1 := E1 and n2, x2:= E2, it produces the nodes x1 := E2 and x2:= E1.
    It may be used for preparing the renaming: x1 is an output defined by x1:= x2, x2 := E (x2 is not an output).
    It is rewriten in x1:=E and x2:=x1, then x2 may be replaced by x1 and deleted.
    
    It is assumed that there is no action on clocks.    
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean NodeManager::switchingNodes(ptINode n1, ptINode n2)
{
  ptlinsth lpredClockn1, lpredClockn2;
  ptldependence lpredn1, lpredn2;
  ptexpcal e1, e2;

  return pk_false;
  if (!(is_equation(n1) || is_delay(n1)))
    return pk_false;
  if (!(is_equation(n2) || is_delay(n2)))
    return pk_false;

  lpredn1 = get_predecessors(n1);
  lpredClockn1 = get_clockPredecessors(n1);
  removeDependenceTo(n1);
  removeInSetCase(n1);
  ptINode_set_interface(n1, NILINTERFACE);

  lpredn2 = get_predecessors(n2);
  lpredClockn2 = get_clockPredecessors(n2);
  removeDependenceTo(n2);
  removeInSetCase(n2);
  ptINode_set_interface(n2, NILINTERFACE);

  addDependenceFromTo(lpredn2, n1);
  addDependenceFromTo(lpredClockn2, n1);
  addDependenceFromTo(lpredn1, n2);
  addDependenceFromTo(lpredClockn1, n2);
  if (is_equation(n1))
    insertTable(n1, &ldefsig, &ldefsigEnd);
  else
    insertTable(n1, &ldelay, &ldelayEnd);
  if (is_equation(n2))
    insertTable(n2, &ldefsig, &ldefsigEnd);
  else
    insertTable(n2, &ldelay, &ldelayEnd);
  get_rightOperand(n1, &e1);
  get_rightOperand(n2, &e2);
  update(n1, ptINode_get_leftOperand(n1), e2);
  update(n2, ptINode_get_leftOperand(n2), e1);
  getSet_interface(n1);
  getSet_interface(n2);
  return pk_true;
}

/**
   It transforms a read node (node) into a intrinsic unsafe procedure call name "readS".
*/
void NodeManager::makeInputByInstrinsicRead(ptINode node)
{
  boolean b;
  ptINode NodeCall;

  if (is_readNode(node))
    {
      b = is_outputNodeCall(node, &NodeCall);
      if (b)
	b = isNodeCall(NodeCall);
      if (!b)
	{
	  removeTable(node, &lread, &lreadEnd);
	  ptINode_set_NameCall(node, string2ident("readS"));	// read simple (one arg)
	  ptINode_set_intrinsicfunc(node);
	  ptINode_set_safetyProperty(node, opunsafe);
	  insertTable(node, &listrefextern, &listrefexternEnd);
	}
    }
}

/**
   It transforms a write-node (node) into a intrinsic unsafe procedure call name "writeS".
*/
void NodeManager::makeOutputByInstrinsicWrite(ptINode node)
{
  if (is_writeNode(node))
    {
      removeTable(node, &lwrite, &lwriteEnd);
      ptINode_set_NameCall(node, string2ident("writeS"));	// write simple (one arg)
      ptINode_set_intrinsicfunc(node);
      ptINode_set_safetyProperty(node, opunsafe);
      insertTable(node, &listrefextern, &listrefexternEnd);
    }
}

/**
   @return "true" if node denotes a set of partial definition associated with a signal, "false" otherwise.
   When the signal is a state variable, it applies PartialDefinitionsMergingStateVar() method,
   otherwise it merges the partial definitions in an equation. The contribution clocks are computed.
*/
boolean NodeManager::PartialDefinitionsMerging(ptINode node)
{
  ptexpcal e1;

  if (!is_setOfPartialDefinitions(node))
    return pk_false;
  e1 = ptINode_get_leftOperand(node);
  if (EM->isStateVar(e1))
    return PartialDefinitionsMergingStateVar(node);
  return PartialDefinitionsMergingSharedSignal(node);
}

/**
   It transforms a node (aSubGraph) assumed to be a node associated with a sub-graph, with vh as clock,
   into a node call that references a model (g).
*/
void NodeManager::subGraphNodeBecomesNodeCall(ptINode aSubGraph, ptDCG g, ptinsth vh)
{
  ptexpcal parameters, lin, lout;
  ptlinstances louts;

  removeInSetCase(aSubGraph);
  parameters = NILEXPCAL;
  lin = EM->convertLinstances2ListExpcal(getInodelInputFlows(aSubGraph));
  louts = getInodelOutputFlows(aSubGraph);
  lout = EM->convertLinstances2ListExpcal(louts);
  ptINode_becomes_nodecall(aSubGraph, g, parameters, lin, lout);
  insertTable(aSubGraph, &lNodeCall, &lNodeCallEnd);
  addInputNodesForOutputs(lout, aSubGraph, vh);
}

/**
   It evaluates the intrinsic functions: Bounds, Dim, Size.
*/
void NodeManager::spatialProcessingIntrinsicEvaluation(ptINode node)
{
  ptident id;
  ptexpcal lp, le, ls;
  ptexpcal e1, e2, e3, s1;
  compilerInteger n;

  if (is_intrinsicfunc(node) && !is_ioNode(node))
    {
      get_procCall_arg(node, &id, &lp, &le, &ls);
      if (isBoundsId(id))
	{
	  EM->getBinaryArgs(le, &e1, &le);
	  EM->getBinaryArgs(le, &e2, &le);
	  EM->getBinaryArgs(le, &e3, &le);
	  EM->getBinaryArgs(ls, &s1, &ls);
	  // fixer les bornes. (to be completed)
	}
      else
	// dim : array ; for Scalar-type, external type or enumerated type 
      if (isDimId(id))
	{
	  EM->getBinaryArgs(le, &e1, &le);
	  n = EM->getNumberOfDimensions(e1);
	  EM->getBinaryArgs(ls, &s1, &ls);
	  modifiedDefinition(s1, EM->mkIntegerConstant(n), CKM->getInitClock(), get_nodeClock(node));
	}
      else
	// size(T,i) defined for  1 <= i <= dim(T)
      if (isSizeId(id))
	{
	  EM->getBinaryArgs(le, &e1, &le);
	  // second arg must be a int positive value.
	  EM->getBinaryArgs(le, &e2, &le);
	  if (EM->getIntegerValue(e2, &n))
	    //  first arg.
	    if (EM->getIthDimension(e1, (int)n, &e3))
	      {
		EM->getBinaryArgs(ls, &s1, &ls);
		modifiedDefinition(s1, e3, CKM->getInitClock(), get_nodeClock(node));
	      }
	}
    }
}

/**
   It updates the definition of a node (node). The operator of the node is not modify.
   It replaces the expression e1 by  the expression e2 in the node.
   This operation is not implemented for sub-graph and array nodes.
*/
void NodeManager::update(ptINode node, ptexpcal e1, ptexpcal e2)
{
  ptexpcal x;

  ptINode_set_interface(node, NILINTERFACE);
  switch (ptINode_get_oper(node))
    {
    case opdefsig:		// equation x := e1 becomes x := e2
      set_def(node, e2);
      x = get_leftOperand(node);
      EM->TypeCheckingEquationFinalize(x, e2, getBackAnnotation(node));
      EM->setValuesEq(x, e2, opnil);
      break;

    case opdefpart:		// equation x ::= e1 becomes x ::= e2
      set_def(node, e2);
      break;

    case opret:		// delay
      ptINode_set_delayExpression(node, e2);
      EM->setValuesEq(get_leftOperand(node), e2, opret);
      break;

    case opappe:		// A revoir avec nodeCall
      if (ptINode_get_rightOperand(node, &x) && (e1 == x))
	ptINode_set_inputExpressions(node, e2);
      else
	ptINode_update_NodeCall(node, e1, e2);	// parameters, outputs
      break;

    case opnodecall:		/* node call */
      ptINode_update_NodeCall(node, e1, e2);
      break;

    case opsubgraph:		/* sub-graph */
      break;
    case opmoti:		/* Array */
      break;
    case opnil:
      break;
    }
}

/**
   @return "true" if the operation has been applied to some nodes, "false" otherwise.
   It merges the partial definitions associated with a signal to one node (PartialDefinitionsMerging(ptINode)).
*/
boolean NodeManager::PartialDefinitionsMerging()
{
  ptlINode lNodes;
  ptINode node;
  boolean b;

  b = pk_false;
  if (getGroupsOfPartialEquations(&lNodes))
    while (ptlINode_get_elem(&lNodes, &node))
      if (PartialDefinitionsMerging(node))
	b = pk_true;
  return b;
}

/**
   For each node of the graph, x:=E1 default ...default EN, 
   it produces (DefaultSpliting(ptINode)) a set of partial definition (x::=E1,..., x::=EN).
*/
boolean NodeManager::DefaultSpliting()
{
  ptlINode lNodes;
  ptINode node;
  boolean b;
  ptinsth undefClk;

  b = pk_false;
  if (getEquations(&lNodes))
    {
      undefClk = CKM->getUndefClock();
      while (ptlINode_get_elem(&lNodes, &node))
	if (DefaultSpliting(node, undefClk))
	  b = pk_true;
    }
  return b;
}

/**
   It expands the left_tt operator into its kernel definition:
   c := left_tt(c1, c2), ^c = ^c1 ^+ ^c2  is rewritten in c:= (true when c1) default (pk_false)
*/
boolean NodeManager::lefttts2defaults()
{
  ptlINode lNodes;
  ptINode aNode;
  ptexpcal vopg, vopd, nvopd;
  ptinsth clkinstr;
  boolean b;

  b = pk_false;
  if (getEquations(&lNodes))
    while (ptlINode_get_elem(&lNodes, &aNode))
      {
	vopg = ptINode_get_leftOperand(aNode);
	if (EM->IsBoolean(vopg))
	  if (get_rightOperand(aNode, &vopd))
	    {
	      clkinstr = get_nodeClock(aNode);
	      if (EM->lefttts2defaults(vopd, clkinstr, &nvopd))
		{
		  modifiedDefinition(vopg, nvopd, clkinstr, clkinstr);
		  b = pk_true;
		}
	    }
      }
  return b;
}

/**
   Evaluating of the intrinsic functions: Bounds, Dim, Size of the node manager.
   (It calls spatialProcessingIntrinsicEvaluation(ptINode))
 */
void NodeManager::spatialProcessingIntrinsicEvaluation()
{
  ptlINode lNodes;
  ptINode node;

  if (getProcCalls(&lNodes))
    while (ptlINode_get_elem(&lNodes, &node))
      spatialProcessingIntrinsicEvaluation(node);
}

/**
   It rewrites the definition of the operande node (node), assumed to be an equation (not a delay).
   The node is such that: x := (E1 when h1) default ...default (En when hn)

   When an expression Ei is reduced to a signal (S) and the signal is also defined by an equation (i.e.  S:=  F1 when k1 default ...Fm when km),
   then (Ei when hi) is replaced by ( F1 when k1 default ...Fm when km) when hi. The resulting expression is then normalized.
   The dependencies are updated.
   
   @return "true" if the node has been modified, "false" otherwise.
*/
boolean NodeManager::rewDefinition(ptINode node)
{
  ptexpcal vdedef, x, e, ne;
  ptinsth h;
  boolean b;
  noperators op;

  b = pk_false;
  if (get_rightOperand(node, &vdedef))
    {
      x = get_leftOperand(node);
      h = EM->getContributionClock(vdedef);
      if (!EM->IsCellExpr(vdedef))
	{
	  e = EM->getBasicExpr(vdedef);
	  op = EM->getOperator(e);
	  if (op == opmele)
	    {			//  x := (E1 when h1) default ...default (En when hn)
	      ne = EM->rewriteDefault(e, x);
	      b = (ne != e);
	    }
	  else
	    // x:= E when h
	  if (!CKM->satisfiesInitClock(h))
	    {
	      ne = EM->rewriteWhen(vdedef, x);
	      b = (ne != vdedef);
	    }
	  if (b)
	    {
	      resetAsEquation(node);
	      ptINode_set_interface(node, NILINTERFACE);
	      becomes_assign(node, x, ne);
	      addDependences(node, ne);
	      EM->setContributionClock(ne, h);
	      getSet_interface(node);
	    }
	}
    }
  return b;
}

/**
   The operand node is assumed to be a set of partial definitions (| x::= E1 when h1 | ... | x ::= En when hn |), 
   and the constraints for deterministic assigments are assumed to be already in the graph.

   It applies the rewDefinition() for each partial definition.
   @return "true" if one of the partial definition is modified, "false" otherwise.
*/
boolean NodeManager::rewDefinitionPartialDef(ptINode node)
{
  ptlINode ln;
  ptINode n;
  boolean b;

  b = pk_false;
  ln = get_setOfPartialDefinitions_arg(node);
  while (ptlINode_get_elem(&ln, &n))
    if (rewDefinition(n))
      b = pk_true;
  // if (b) SplitDefaultPartialDef(node);
  return b;
}

/**
   aNode is an equation (not a delay) or a partial definition.
   If the utility clock is computed (denoted UC(x) for a signal x)
   - if UC(x) is the null clock, the node is not modified (it is unused, it will be removed later).
   - otherwise,
      - an equation X:= E is replaced by (| X::= E when UC(x) | x ::= defaultvalue x$1 |)
      - a list of partial definition:X::= E1 .... X::=En is replaced by  X::= E1 when UC(c) .... X::=En when UC(c) and completed with the default value.
   
*/
boolean NodeManager::rewUsingProductiveClock(ptINode aNode)
{
  ptinsth hu;

  if (!(is_equation(aNode) || is_setOfPartialDefinitions(aNode)))
    return pk_false;
  if (!getUtilityClock(aNode, &hu))
    return pk_false;
  if (CKM->eqClock(hu, hundef))
    return pk_false;
  if (CKM->eqClock(hu, the_nullClock_NM_copy))
    return pk_false;		// Unused, it will be removed from the graph.
  if (CKM->eqClock(get_nodeClock(aNode), hu))
    return pk_false;
  if (is_equation(aNode))
    return rewEqUsingProductiveClock(aNode, hu);
  return rewPartialDefUsingProductiveClock(aNode, hu);
}

/**
   If the definition of the operand node (aNode) contains the expression (E when h) default ... default (E when k), it is aggregated in (E when h ^+ k).
   By construction  the default expressions are normalized in (F1 when k1) default ...(Fm when km) and the clocks ki, kj are exclusive.
   For a partial definition, the clocks are also exclusive by construction.
   @return "true" is the operand node is modified, "false" otherwise.
*/
boolean NodeManager::normalDefaults(ptINode aNode)
{
  if (is_equation(aNode))
    return normalDefaultsEquation(aNode);
  if (is_setOfPartialDefinitions(aNode))
    return normalDefaultsEquationPartialDef(aNode);
  return pk_false;
}

/**
   Substitute vsig1 by vsig2 in the operand node (node).
*/
void NodeManager::replace(ptexpcal vsig1, ptexpcal vsig2, ptINode node)
{
  ptlINode ln;
  ptINode n;
  ptexpcal nExprDef, ExprDef;
  ptlinsth lh;

  switch (get_oper(node))
    {
    case opnil:		// undefined
      break;

    case opdefpart:		// x ::= E
    case opdefsig:		// x:=E
    case opret:		// delay
      if (get_rightOperand(node, &ExprDef))
	{
	  nExprDef = EM->replace(vsig1, vsig2, ExprDef);
	  if (nExprDef != ExprDef)
	    {
	      set_def(node, nExprDef);
	      ptINode_set_interface(node, NILINTERFACE);
	      replaceDependence(EM->getNode(vsig1), node, EM->getNode(vsig2));
	    }
	}
      break;

    case opldefpart:		// { x::= e1, ... x::= eN}, 
      ln = get_setOfPartialDefinitions_arg(node);
      while (ptlINode_get_elem(&ln, &n))
	replace(vsig1, vsig2, n);
      break;

    case opsubgraph:		// sub-graph
      get_subGraphDefinitions(node, &ln, &lh);
      while (ptlINode_get_elem(&ln, &node))
	replace(vsig1, vsig2, node);
      break;

      // Must be completed for the followings
    case opmoti:
      break;

    case opnodecall:
      break;

    case opappe:		// procedure call
      break;

    case opbund:		// Set of the nodes of a struct/bundle (lfields)
      break;

    }
}

/**
   @return "true" if the operation has been applied to some nodes, "false" otherwise.
   It merges the partial definitions associated with a signal to one node (PartialDefinitionsMerging(ptINode)).
*/
boolean NodeManager::SharedVariableRemoving()
{
  ptlINode lNodes;
  ptINode node;
  boolean b;

  b = pk_false;
  if (getGroupsOfPartialEquations(&lNodes))
    while (ptlINode_get_elem(&lNodes, &node))
      if (SharedVariableRemoving(node))
	b = pk_true;
  return b;
}

boolean NodeManager::cell2kernel(ptINode n)
{
  ptexpcal vopg, vopd, zx, exp;
  ptinsth hCell;

  if (!is_cellNode(n))
    return pk_false;
  vopg = ptINode_get_leftOperand(n);
  ptINode_get_rightOperand(n, &vopd);
  hCell = EM->getContributionClock(vopd);
  exp = EM->getWhenOrCellArg(vopd);
  ptINode_set_rightOperand(n, exp);
  zx = mkDelayOf(vopg);
  addaff(vopg, zx, CKM->makeMinus(hCell, EM->getContributionClock(exp)), ptlinstances_insert(NILLINSTANCES, zx));
  return pk_true;
}



/**
  It replaces the references to vh1 by vh2 in the operand node.
 */
void NodeManager::replace(ptINode node, ptinsth vh1, ptinsth vh2)
{
  // Added for POLYENDO.

  ptexpcal nExprDef, ExprDef;

  switch (get_oper(node))
    {

    case opdefpart:		// x ::= E
    case opdefsig:		// x  := E

      if (get_rightOperand(node, &ExprDef))
	{
	  nExprDef = EM->replace(vh1, vh2, ExprDef);
	  if (nExprDef != ExprDef)
	    {
	      set_def(node, nExprDef);
	      ptINode_set_interface(node, NILINTERFACE);
	      removeDependence(vh1, node);
	      addDependence(vh2, node);
	      getSet_interface(node);
	    }
	}
      break;

    case opldefpart: // { x::= e1, ... x::= eN}, 
      /*
	ptINode n;
	ln = get_setOfPartialDefinitions_arg(node);
	while (ptlINode_get_elem(&ln, &n))
	replace(n, vh1, vh2);
      */
      break;

    case opsubgraph:		// sub-graph
      /*
	ptlINode ln;
	ptlinsth lh;
	get_subGraphDefinitions(node, &ln, &lh);
	while (ptlINode_get_elem(&ln, &node))
	replace(node, vh1, vh2);
      */
      break;

      // ---- Must be completed for
    case opmoti:
    case opnodecall:
    case opappe:
    case opbund: // Set of the nodes of a struct/bundle (lfields)
    case opnil:		// undefined
    case opret:		// delay
      break;
    }
}



