/* 
   Authors: Loic Besnard, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_

/**
   Enable/Disable the tracability of the rewritings when b is true/b is false.
*/
void
TraceabilityManager::setModeKeepSubstitutions (boolean b)
{
  modeKeepSubstitutions = b; 
}

/** 
    When the tracability of the rewritings is active (see setModeKeepSubstitutions),
    it pushes the current infos (Clocks and Signals) of the graph, without effect otherwise.
*/
void TraceabilityManager::ProofOfCompilerCorrectness_begin(GraphLevelType glevel, ptlinsth lhToVisit, ptlinsth lUnusedClocks, ptlINode lUnusedNodes)
{
  ptlinsth lh;
  ptlinstances lsig;
  ptClockNamesTable clocknametable;
  
  newLevel = pk_false;
  if (modeKeepSubstitutions)
    {
      if (! initialized) { newLevel= pk_true; initialized = pk_true; } else newLevel=(CurrentLevel != glevel);
      // a new one only when the graph level is modified since the previous call.
      if (newLevel) {
	CurrentLevel = glevel;
	clocknametable = new ClockNamesTable();
	lh = mkClockFormulaToKeep(lhToVisit, NILLINSTH, clocknametable, lUnusedClocks);
	if (!EM->getExprList(opident, &lsig)) lsig= NILLINSTANCES;
	lsig = ptlinstances_copy(lsig);
	lsig = removeUnused(lsig, lUnusedNodes);

	// push the clock system and the set of signals of the graph.
	ClockSystemsStack = stackPush (ClockSystemsStack, (ptElemStack) lh);
	SignalsStack = stackPush (SignalsStack, (ptElemStack) lsig);
	ClockNamesTableStack = stackPush (ClockNamesTableStack, (ptElemStack) clocknametable);
	StatesStack = stackPush (StatesStack, (ptElemStack) getMemorizedNodes(lsig));

	IODepStack = stackPush (IODepStack, (ptElemStack) NILLDEPENDENCE);

	printmodifiedSinceLastMarkedFunction(); // Debug.

	// Fixing the current 
	ClockSystemsStackRef = ClockSystemsStack;
	SignalsStackRef = SignalsStack;
	ClockNamesTableStackRef = ClockNamesTableStack;
	StatesStackRef = StatesStack;
	IODepStackRef = IODepStack;
      }
    }
}

/**
   @return true if it exists clocks of the  list lhpreviousSystem that have changed of name from the previous level of the graph (BASIC, POLY, ENDO,...).
   The old/new names of the renamed clocks are returned in the list lprevNames/lhWithNewName: 
   the i-th element of the lprevNames (old name) is associated with the i-th element of lhWithNewName (new name).
 */
boolean TraceabilityManager::Certification_getClocksWithNewName( ptlinsth lhpreviousSystem, ptlinsth *lhWithNewName, ptlPointer *lprevNames)
{
  ptClockNamesTable idtablePrevious;
  ptident curId, prevId;
  ptinsth vh;
  boolean b;

  *lhWithNewName = NILLINSTH;
  *lprevNames = NILLPOINTER;
  if ( ! getIdentifiersTablePrevious (&idtablePrevious)) return pk_false;

  b = pk_false;
  while (ptlinsth_get_next (&lhpreviousSystem, &vh))      
    if ( (*idtablePrevious).find (vh) != (*idtablePrevious).end ())
      {
	prevId = (*idtablePrevious)[vh];
	curId =  getClockId(vh);
	if (! eq_idents(curId, prevId))
	  {
	    *lhWithNewName = ptlinsth_insert_elem ( *lhWithNewName, vh);
	    *lprevNames = ptlPointer_add_front(*lprevNames, prevId);
	    b = pk_true;
	  }
      }
  return b;
}


/**
   @return the defined signals of the previous graph.
 */
ptlinstances TraceabilityManager::Certification_getPreviousSignals()
{
  ptStack S;
  pointer x;
  if (!stackPrevious(SignalsStackRef, &S)) return NILLINSTANCES;
  if (!stackTop(S, &x)) return NILLINSTANCES;
  return (ptlinstances) x;
}


/**
   @return the defined signals of the previous graph.
 */
ptlINode TraceabilityManager::Certification_getPreviousMemorizedNodes()
{
  ptStack S;
  pointer x;
  if (!stackPrevious(StatesStackRef, &S)) return NILLINODE;
  if (!stackTop(S, &x)) return NILLINODE;
  return (ptlINode) x;
}


/**
   @return the IO dependences of the previous graph.
 */
ptldependence TraceabilityManager::Certification_getPreviousIODependences()
{
  ptStack S;
  pointer x;
  if (!stackPrevious(IODepStackRef, &S)) return NILLDEPENDENCE;
  if (!stackTop(S, &x)) return NILLDEPENDENCE;
  return (ptldependence) x;
}


/**
   @return the list of the clocks of the previous system.
 */
ptlinsth TraceabilityManager::Certification_getClocksPreviousSystem()
{
  pointer x;
  ptlinsth lhprev;

  if (!modeKeepSubstitutions) return NILLINSTH;
  if (ClockSystemsStack == NILSTACK) return NILLINSTH; // Empty stack
  if (ClockSystemsStackRef == NILSTACK) return NILLINSTH; // No ClockSystem Reference
  if (ClockSystemsStack == ClockSystemsStackRef)
    {
      ptStack S;
      if (!stackPrevious(ClockSystemsStack, &S)) return NILLINSTH;
      if (!stackTop(S, &x)) return NILLINSTH;
      lhprev = (ptlinsth) x; 
    }
  else
    {
      if (!stackTop(ClockSystemsStackRef, &x)) return NILLINSTH;
      lhprev = (ptlinsth) x;
    }
  return ptlinsth_duplicate(lhprev);
}


void TraceabilityManager::Certification_end( ptlinsth lAddedClocks, ptlinstances lAddedSignals, ptldependence lIODependences)
{
  pointer x;

  if ( lAddedClocks != NILLINSTH )
    {
      if (stackTop(ClockSystemsStackRef, &x))
	{
	  ptlinsth lh = (ptlinsth) x;
	  lh = ptlinsth_sorted_fusion (lh, lAddedClocks);
	  StackReplaceValue(ClockSystemsStackRef, (pointer) lh);
	}
      if (stackTop(ClockNamesTableStackRef, &x))
	{
	  ptClockNamesTable clocknametable;
	  clocknametable = (ptClockNamesTable) x;
	  addIdentToTable(lAddedClocks, clocknametable);
	}
    }

  if (lAddedSignals != NILLINSTANCES)
    {
      if (stackTop(SignalsStackRef, &x))
	{
	  // ptlinstances lsig = (ptlinstances) x; lsig = ptlinstances_sortedfusion (lsig, lAddedSignals);
	  StackReplaceValue(SignalsStackRef, (pointer) lAddedSignals);
	  StatesStack = stackPush (StatesStack, (ptElemStack) getMemorizedNodes(lAddedSignals));
	}
    }
  
  if (lIODependences != NILLDEPENDENCE)
    {
      if (stackTop(IODepStackRef, &x))
	{
	  StackReplaceValue(IODepStackRef, (pointer) lIODependences);
	}
    }
}
