/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PROTECTED_
/**
    Abstract method.
    @return the translation of a clock (clk).
 */
  _VIRTUAL_ ptinsth InterFormatTranslator::formulaCoding (ptinsth clk)
{
  return clk;
}

/**
   Abstract method.
   Called after the coding of clk to visit its references.   
*/
_VIRTUAL_ void
InterFormatTranslator::propagateClock (ptinsth clk)
{
}

/**
   In debugging mode (DEBUG attribute member), it prints message following by the name of the clock.
*/
void
InterFormatTranslator::clk_trace (const char *message, ptinsth clk)
{
  if (DEBUG)
    {
      pK_messenger->printString (message);
      pK_messenger->printIdent (CKM->getName (clk));
    }
}

/**
   In debugging mode (DEBUG attribute member), it prints the constraints on placed formula.
*/
_VIRTUAL_ void
InterFormatTranslator::DEBUG_ClockConstraints ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2;

  if (DEBUG)
    {
      CKM->getConstraintsOnPlacedClocks (&lh1, &lh2);
      while (ptlinsth_get_next (&lh1, &vh1))
	{
	  vh1 = CKM->getRepresentativeFormula (vh1);
	  if (ptlinsth_get_next (&lh2, &vh2))
	    {
	      vh2 = CKM->getRepresentativeFormula (vh2);
	      if (vh1 != vh2)
		{
		  pK_messenger->printString ("CONSTRAINT : \n ");
		  clk_trace ("   --->Formula :", vh1);
		  clk_trace ("   --->Formula :", vh2);
		}
	    }
	}
    }
}

/** 
    It defines the boolean clock of clk at clkFather by the expression vexp.
    A signal is created when vexp is not a signal.
    @return the extraction clock assigned to clk.
*/
_VIRTUAL_ ptinsth
InterFormatTranslator::endDefinition (ptinsth clk, ptexpcal vexp, ptinsth clkFather)
{
  ptINode aNode;
  ptident name;
  ptINode vblockarray;
  ptinsth vh, vhres;
  ptlclosure lclos;

  if (!EM->IsSignal (vexp))
    {
      name = choiceNameBooleanClock (clk);
      vexp = NM->insertEquation (name, get_BooleanType (), vexp, clkFather, clkFather, &aNode);
      NM->addPragmas (aNode, CKM->getPragmas (clk));
      if (CKM->getInComposedNode (clk, &vblockarray))
	{
	  NM->arrayBodyInsertNode (vblockarray, aNode);
	  NM->set_inComposedNode (aNode, vblockarray);
	}
    }

  if (! EM->getAttribute (vexp, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
    {
      if (CKM->getAttribute (clk, CLOSUREATTRIBUTE, (attributeValue *) & lclos))
	EM->addAttribute (vexp, CLOSUREATTRIBUTE, (attributeValue) lclos);
    } 

  CKM->setClockRepresentedBySignal (clk, vexp);
  vhres = CKM->setClockAsBoolean (clk, vexp, EM->getPredecessors (vexp), clkFather);
  return vhres;
}

/** 
    The following actions are 
    - it applies a translation (formulaCoding()) to a clock (clk).
    - it marks the clock (CST_MYMARK internal attribute).
    - and it signals to the sucessors of clk (propagateClock()) that it has been translated.
    
    @return always "true"
*/
_VIRTUAL_ boolean
InterFormatTranslator::clockCoding (ptinsth clk)
{
  CKM->addMark (clk, CST_MYMARK);
  if (CKM->satisfiesPredefinedClock (clk))
    return pk_false;		// Armouring
  clk_trace ("Convert clock <begin> : ", clk);
  clk = formulaCoding (clk);
  clk_trace ("Convert clock <end> : ", clk);
  propagateClock (clk);
  CKM->addMark (clk, CST_MYMARK);
  return pk_true;
}

/** @return "true" if the interformat translation  can be applied to the operand clock (clk), "false" otherwise.
    It can be applied if the father clock has been translated and for upper bound, lower bound, 
    if the operands of the clock have been translated.
    
    It uses the CST_MYMARK mark to test if operands have been visisted; this mark is set in clockCoding() method.
*/
_VIRTUAL_ boolean
InterFormatTranslator::satisfiesCodingProperty (ptinsth clk)
{
  ptinsth vh1, vh2;

  if (CKM->satisfiesPredefinedClock (clk))
    return pk_false;		// Armouring
  if (CKM->satisfiesMarked (clk, CST_MYMARK))
    return pk_false;
  if (!CKM->satisfiesMarked (CKM->getFather (clk), CST_MYMARK))
    return pk_false;

  switch (CKM->getOperator (clk))
    {
    case infer:
      if (!CKM->getMinusArgs (clk, &vh1, &vh2))
	CKM->getBinaryArgs (clk, &vh1, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getRepresentativeFormula (vh2);
      return ((CKM->satisfiesMarked (vh1, CST_MYMARK)) && (CKM->satisfiesMarked (vh2, CST_MYMARK)));
      break;

    case super:
      CKM->getBinaryArgs (clk, &vh1, &vh2);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getRepresentativeFormula (vh2);
      return ((CKM->satisfiesMarked (vh1, CST_MYMARK)) && (CKM->satisfiesMarked (vh2, CST_MYMARK)));
      break;

    case hcomp:
      CKM->getUnaryArg (clk, &vh1);
      vh1 = CKM->getRepresentativeFormula (vh1);
      vh2 = CKM->getFather (clk);
      return ((CKM->satisfiesMarked (vh1, CST_MYMARK)) && (CKM->satisfiesMarked (vh2, CST_MYMARK)));
      break;

    case sub:
    case nosub:
      return pk_true;
      break;
    case vari:			// the tick of the program
      return pk_false;
      break;

    default:;
    }
  return pk_false;
}

/** 
    @return "true" if the interfomat translation can be applied to the operand clock (clk), "false" otherwise.
    It returns "true" when satisfiesCodingProperty() for clk and clk is a useful clock.
*/
_VIRTUAL_ boolean
InterFormatTranslator::canBeCoded (ptinsth clk)
{
  if (!satisfiesCodingProperty (clk))
    return pk_false;
  if (CKM->satisfiesSub (clk))
    return pk_true;
  return !ptlinsth_sorted_isin_elem (unusedClocks, clk);
}

/** 
    while it exists a placed clock than can be coded (canBeCoded()) 
    it translates it (clockCoding(ptinsth)).
    
    The mark CST_MYMARK is cleanned at the end of the method.
*/
_VIRTUAL_ void
InterFormatTranslator::clocksCoding ()
{
  boolean bcontinue;
  ptlinsth list;
  ptinsth clk;

  bcontinue = pk_true;
  while (bcontinue)
    {
      bcontinue = pk_false;
      list = CKM->getlhplaced ();
      while (ptlinsth_get_next (&list, &clk))
	{
	  clk = CKM->getRepresentativeFormula (clk);
	  if (canBeCoded (clk))
	    if (clockCoding (clk))
	      bcontinue = pk_true;
	}
    }
  CKM->resetMarkAll (CST_MYMARK);
  DEBUG_ClockConstraints ();
}

/** 
    @return a name for the boolean representation of a clock (clk).
*/
ptident InterFormatTranslator::choiceNameBooleanClock (ptinsth clk)
{
  ptlinstances
    lsig;
  ptident
    id;

  if (CKM->satisfiesTickClock (clk))
    return string2ident ("Tick");
  lsig = CKM->getSynchronousSignals (clk);
  if (!EM->getident (lsig, &id))
    id = CKM->getName (clk);
  if (id == NILIDENT)
    return getPrefixBooleanClock ();
  if (CKM->satisfiesSub (clk))
    return id;
  return namBooleanClockfrom (id);
}

boolean
InterFormatTranslator::satisfiesIOstepClock (ptinsth vh)
{
  return CKM->eqClock (vh, IOstepClock);
}
