/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PRIVATE_
/**
   @return "true" if the graph is not a DC+ and vh is an positive extraction, "false" otherwise.
*/
  boolean ProfileBuilder::isTypeGraphSelected (ptinsth h)
{
  return (CKM->getOperator (h) == sub);
}

/**
   @return "true" if a clock (vh) is representated by a signals in the inputs/outputs of the graph.
*/
boolean ProfileBuilder::isIOevent (ptinsth vh)
{
  ptexpcal aSig;

  if (is_output_clock (vh)) return pk_true;
  if (is_input_clock (vh)) return pk_true; 
  if (!CKM->isClockUnifiedWithSignal (vh, &aSig)) return pk_false;
  if (ptlinstances_isin_exp (inputs, aSig)) return pk_true;
  return ptlinstances_isin_exp (outputs, aSig);
}

/**
   @return "true" if a node (vinstr) is useful, "false" otherwise.
   useful = not in the unused nodes set associated with the graph.
*/
boolean ProfileBuilder::isSelectedNode (ptINode vinstr)
{
  return !ptlINode_isIn (unNodes, vinstr);
}

/**
   @return "true" if a clock (vh) is useful, "false" otherwise.
   useful = not in the unused clocks set associated with the graph.
*/
boolean ProfileBuilder::isSelectedClock (ptinsth vh)
{
  if (ptlinsth_sorted_isin_elem (unClocks, vh))
    return pk_false;
  if (vh != CKM->getRepresentativeFormula (vh))
    return pk_false;
  return pk_true;
}

/**
   @return "true" if the clock (vh) is declared as an input of the graph, "false" otherwise.
*/
boolean ProfileBuilder::is_input_clock (ptinsth vh)
{
  return (ptlinsth_sorted_isin_elem (stInterface_get_lhoin (rootghdc), vh));
}

/**
   @return "true" if the clock (vh) is declared as an output of the graph, "false" otherwise.
*/
boolean ProfileBuilder::is_output_clock (ptinsth vh)
{
  return (ptlinsth_sorted_isin_elem (stInterface_get_lhoout (rootghdc), vh));
}


/** Very strange: should be replaced by getFahter. */
ptinsth ProfileBuilder::getForcedFather (ptinsth vh)
{
  int p;

  if (CKM->getRoot (vh, &p) == NILINSTH) return rootghdc;
  return CKM->getFather (vh);
}

/** 
    @return the branching of vh1 and vh2 if it exists, or the fictitious root clock otherwise.
*/
ptinsth ProfileBuilder::getForcedBranching (ptinsth vh1, ptinsth vh2)
{
  ptinsth vhemb;

  if (CKM->getOperator (vh1) == hnot) return rootghdc;
  if (CKM->getOperator (vh2) == hnot) return rootghdc;
  vhemb = CKM->getBranching (vh1, vh2);
  if (CKM->satisfiesUndefClock (vhemb)) return rootghdc;
  return vhemb;
}


/**  
     It adds a signal (vexp) as an input for all the nodes of the hierarchy from vhf until vht.
     
     Assumed : vht is upper than vhf in the hierarchy.
*/
void
ProfileBuilder::insert_signal_input_to_except (ptinsth vhf, ptexpcal vexp, ptinsth vht)
{
  ptinsth vhemb, h;

  if (CKM->satisfiesPredefinedClock (vhf))
    return;
  if (EM->isSignalUnifiedWithClock (vexp, &h))
    insert_clock_input_to_except (vhf, CKM->getRepresentativeFormula (h), vht);
  else
    {
      vhemb = vhf;
      while (vhemb != vht)
	{
	  stInterface_insert_ent (vhemb, vexp);
	  vhemb = CKM->getFather (vhemb);
	}
    }
}

/**
   It adds a clock vhi as an input for all the nodes of the hierarchy from vhf until vht.
   
   Assumed: vht is upper than vhf in the hierarchy.
*/
void
ProfileBuilder::insert_clock_input_to_except (ptinsth vhf, ptinsth vhi, ptinsth vht)
{
  ptinsth vhemb;
  ptlinstances lsig;
  ptexpcal e;

  if (CKM->satisfiesPredefinedClock (vhf))
    return;
  if (CKM->satisfiesPredefinedClock (vhi))
    return;
  if (vhf == vht)
    return;

  if (isTypeGraphSelected (vhi))
    {
      lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vhi));
      while (ptlinstances_get_next (&lsig, &e))
	insert_signal_input_to_except (vhf, e, vht);
    }
  else
    {
      vhemb = vhf;
      while (vhemb != vht)
	{
	  stInterface_insert_hoin (vhemb, vhi);
	  vhemb = CKM->getFather (vhemb);
	}
    }
}



/** 
    It adds a signal (vexp) as an output for all the node of the hierarchy on the branch vhf until vht (excluded).
    Assumed : vht is upper than vhf in the hierarchy.
*/
void
ProfileBuilder::insert_signal_output_to_except (ptinsth vhf, ptexpcal vexp, ptinsth vht)
{
  ptinsth vhemb, h;

  if (CKM->satisfiesPredefinedClock (vhf))
    return;
  if (ptlinstances_isin_exp (outputs, vexp))
    return;
  if (EM->isSignalUnifiedWithClock (vexp, &h))
    insert_clock_output_to_except (CKM->getRepresentativeFormula (h), vht, pk_true);
  else
    {
      vhemb = vhf;
      
      while (vhemb != vht)
	{
	  stInterface_insert_sort (vhemb, vexp);
	  h = CKM->getFather (vhemb);
	  if (h == vht)
	    stInterface_insert_loc (h, vexp);
	  vhemb = h;
	}
    }
}



/** 
    It adds a signal (vexp) as an input-output for all the node of the hierarchy on the branch vhf until vht (excluded).
    Assumed : vht is upper than vhf in the hierarchy.
*/
void
ProfileBuilder::insert_signal_inout_to_except (ptinsth vhf, ptexpcal vexp, ptinsth vht)
{
  ptinsth vhemb, h;

  if (CKM->satisfiesPredefinedClock (vhf))
    return;
  if (EM->isSignalUnifiedWithClock (vexp, &h))
    insert_clock_output_to_except (CKM->getRepresentativeFormula (h), vht, pk_true);
  else
    {
      vhemb = vhf;
      while (vhemb != vht)
	{
	  ptinterface_remove_localSignal (CKM->getInterfaceSubHierarchy (vhemb), vexp);
	  vhemb = CKM->getFather (vhemb);
	}
    }
}

/**
   It adds a clock vh (dc+ graph) or the boolean signals referenced in the condition otherwise, as a local to mclock.
   Assumed: mclock is the mother clock of vh (otherwise hierarchy will be unconsistent).
*/
void
ProfileBuilder::insert_clock_local (ptinsth mclock, ptinsth vh)
{
  boolean b;
  ptlinstances lsig;
  ptexpcal e;

  if (!CKM->satisfiesPredefinedClock (vh))
    {
      if (isTypeGraphSelected (vh))
	{
	  lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vh));
	  while (ptlinstances_get_next (&lsig, &e))
	    stInterface_insert_loc (mclock, e);
	  if (CKM->isClockRepresentedBySignal (vh, &e))
	    stInterface_insert_loc (mclock, e);
	}
      else
	{
	  if (!DCplusGraph) b = pk_true; else b = !isIOevent (vh);
	  if (b) stInterface_insert_holoc (mclock, vh);
	}
    }
}


/**
   It adds a clock (vho) in the list of the outputs for all the nodes of the hierarchy
   from vho until vht (excluded). Then, if except, vho is becomes a local to vht, otherwise an output of vht.

   Assumed : vht is upper than vho in the hierarchy.
*/
void
ProfileBuilder::insert_clock_output_to_except (ptinsth vho, ptinsth vht, boolean except)
{
  ptinsth vhemb, mclock;
  ptexpcal exp;
  ptlinstances lsig;

  if (CKM->satisfiesPredefinedClock (vho)) return;
  if (vho == vht) return;

  if (isTypeGraphSelected (vho))
    {
      lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vho));
      mclock = CKM->getFather (vho);
      while (ptlinstances_get_next (&lsig, &exp))
	{
	  insert_signal_output_to_except (mclock, exp, vht);
	  stInterface_insert_loc (vht, exp);
	}
      if (CKM->isClockRepresentedBySignal (vho, &exp))
	stInterface_insert_loc (vht, exp);
    }
  else
    {
      vhemb = CKM->getFather (vho);
      while (vhemb != vht)
	{
	  stInterface_insert_hoout (vhemb, vho);
	  vhemb = CKM->getFather (vhemb);
	}
      if (!except)
	stInterface_insert_hoout (vht, vho);
      else if (!isIOevent (vho))
	stInterface_insert_holoc (vht, vho);
    }
}

/**
   It finalizes the interface of the sub graph.
     - the output signals are removed from the inputs
     - the boolean clocks are removed from the inputs.
*/
void
ProfileBuilder::setInterfaceSubGraphFinalize (ptgfd vgfd)
{
  ptlinstances vlent, vlsort;
  ptinsth vh;
  ptlinsth lh;
  ptexpcal vsig;

  vlent = ptgfd_get_inputSignals (vgfd);
  vlsort = ptgfd_get_outputSignals (vgfd);
  while (ptlinstances_get_next (&vlsort, &vsig))
    ptlinstances_isin_remove_exp (&vlent, vsig);

  lh = ptgfd_get_clocks (vgfd);
  while (ptlinsth_get_next (&lh, &vh))
    {
      vsig = CKM->getBooleanSignal (vh);
      ptlinstances_isin_remove_exp (&vlent, vsig);
    }
  ptgfd_set_inputSignals (vgfd, vlent);
}

/**
   @return the list obtained by the insertion of the signals of the list (l) into another 
   list (lmain). Moreover, the "declared" signals are also added to lmain (ie for the references X.Y, X[], X is also added).
   \warning  side effects on the lmain list.
*/
ptlinstances ProfileBuilder::addListAndDeclaredSignals (ptlinstances l, ptlinstances lmain)
{
  ptexpcal elem, x;
  while (ptlinstances_get_next (&l, &elem))
    {
      lmain = ptlinstances_insert (lmain, elem);
      x = EM->getDeclaredSignal (elem);
      if (x != elem)
	lmain = ptlinstances_insert (lmain, x);
    }
  return lmain;
}

/**
   It updates the visibilty of signals (lsig) referenced at a clock (vh).
   For a signal (x) of the list, if x is 
     - an input of the graph it is added as input for each node of the branch from vh to the root of the hierarchy. 
     - not an input of the graph, the branching (e) of h and ^x is computed
           - if (vh = ^x) and bup is pk_true, e is the mother clock of vh, otherwise (bup pk_false) e is undefined.
	   - otherwise the branching (e) of vh and ^x is computed.
	   .
       If e is defined:
        - x is added as output on the branch from ^x until e,
        - x is added as input on the branch from vh until e,
	- x is added as a local signal at e ((if it is not already an output of the hierachy with e as root)
*/
void
ProfileBuilder::visibilitySignalsAtClock (ptlinstances lsig, ptinsth vh, boolean bup)
{
  ptexpcal vsig;
  ptinsth vhsig, vhemb, hund;
  ptINode n, sg;

  hund = CKM->getUndefClock ();
  vh = CKM->getRepresentativeFormula (vh);
  while (ptlinstances_get_next (&lsig, &vsig))
    {
      if (ptlinstances_isin_exp (inputs, vsig))
	insert_signal_input_to_except (vh, vsig, rootghdc);
      else
	{
	  n = EM->getNode (vsig);
	  if (NM->get_inComposedNode (n, &sg))
	    vhsig = NM->get_nodeClock (sg);
	  else
	    vhsig = NM->get_nodeClock (n);

	  vhsig = CKM->getRepresentativeFormula (vhsig);
	  if (vh == vhsig)
	    {
	      if (bup) vhemb = CKM->getFather (vh); else vhemb = hund;
	    }
	  else
	    vhemb = getForcedBranching (vh, vhsig);

	  if (!CKM->satisfiesUndefClock (vhemb))
	    {
	      insert_signal_output_to_except (vhsig, vsig, vhemb);
	      insert_signal_input_to_except (vh, vsig, vhemb);
	      stInterface_insert_loc (vhemb, vsig);
	    }
	}
    }
}

/**
   For each node of a sub graph (vgfd)
     - it updates the interface of vgfd with the input/output signals and the clock referenced in the node
     - it updates the visibilty of the input signals referenced in the node (visibilitySignalsAtClock()).
*/
void
ProfileBuilder::setInterfaceSubGraphNodes (ptgfd vgfd)
{
  ptlinsth vlhorl;
  ptlinstances vlent, vlsort;
  ptINode vinstr;
  ptlINode linstr;

  linstr = ptgfd_get_nodes (vgfd);
  while (ptlINode_get_elem (&linstr, &vinstr))
    if (isSelectedNode (vinstr))
      {
	vlsort = NM->getInodelOutputFlows (vinstr);
	if (!NM->is_readNode (vinstr))
	  {
	    vlent = NM->getInodelInputFlows (vinstr);
	    vlhorl = NM->getInodelInputClocks (vinstr);
	    ptgfd_set_inputClocks (vgfd, ptlinsth_sorted_fusionexcept (ptgfd_get_inputClocks (vgfd), vlhorl, CKM->getInitClock ()));
	    ptgfd_set_outputSignals (vgfd, addListAndDeclaredSignals (vlsort, ptgfd_get_outputSignals (vgfd)));
	    if (vlent != NILLINSTANCES)
	      {
		ptgfd_set_inputSignals (vgfd, ptlinstances_sortedfusion (vlent, ptgfd_get_inputSignals (vgfd)));
		if ((!NM->is_subgraph (vinstr)) && (!NM->is_array (vinstr)))
		  visibilitySignalsAtClock (vlent, NM->get_nodeClock (vinstr), pk_false);
	      }
	  }
	else
	  ptgfd_set_inputSignals (vgfd, ptlinstances_sortedfusion (vlsort, ptgfd_get_inputSignals (vgfd)));

	vlsort = NM->getInodelInOutFlows (vinstr);
	ptgfd_set_inOutSignals (vgfd, ptlinstances_sortedfusion (vlsort, ptgfd_get_inOutSignals (vgfd)));
      }
}

/**
   Without effect!.
*/
void
ProfileBuilder::setInterfaceSubGraphClocks (ptgfd vgfd)
{
}


/** It builds the interface of a sub-graph (vgfd)-- Such an element is assigned to a clock--:
    - its nodes are visited (setInterfaceSubGraphNodes()),
    - its clocks are visited (setInterfaceSubGraphClocks()),
    - the calculus is then finalized (setInterfaceSubGraphFinalize())
*/
void
ProfileBuilder::setInterfaceSubGraph (ptgfd vgfd)
{
  setInterfaceSubGraphNodes (vgfd);
  setInterfaceSubGraphClocks (vgfd);
  setInterfaceSubGraphFinalize (vgfd);
}


/** 
    It updates the visibility of the signals referenced in the sub-graph associated with a clock (vh).
 */
void
ProfileBuilder::mkSubHierarchyInterface (ptinsth vh)
{
  boolean b;
  ptinsth vhemb, vhuse;
  ptlinsth lhorluse;
  ptlinstances vlsort, lsig;
  ptexpcal vexp, x;
  ptgfd vgfd;

  vgfd = CKM->getClockSubGraph (vh);
  if (!ptgfd_satisfies_empty (vgfd))
    {
      vlsort = ptgfd_get_outputSignals (vgfd);
      while (ptlinstances_get_next (&vlsort, &vexp))
	{
	  b = ptlinstances_isin_exp (stInterface_get_lsort (rootghdc), vexp);
	  if (b)
	    insert_signal_output_to_except (vh, vexp, rootghdc);
	  else
	    {
	      vhuse = NM->get_nodeClock (EM->getNode (vexp));
	      b = pk_true;
	      if (CKM->isClockRepresentedBySignal (vhuse, &x))
		if (!CKM->isSubstitutable (vhuse))
		  b = (vexp != x);
	      if (b)
		stInterface_insert_loc (vhuse, vexp);
	      else
		stInterface_insert_sort (vhuse, vexp);
	    }
	}
      lhorluse = ptgfd_get_inputClocks (vgfd);
      while (ptlinsth_get_next (&lhorluse, &vhuse))
	{
	  vhuse = CKM->getRepresentativeFormula (vhuse);
	  if (isSelectedClock (vhuse))
	    if (!CKM->satisfiesIsAnAssumption (vhuse))
	      {
		vhemb = getForcedBranching (getForcedFather (vhuse), vh);
		if (!CKM->satisfiesUndefClock (vhemb))
		  {
		    insert_clock_output_to_except (vhuse, vhemb, pk_true);
		    insert_clock_input_to_except (vh, vhuse, vhemb);
		  }
	      }
	}
      visibilityOf (ptgfd_get_inputSignals (vgfd), vh);
      visibilityOfSharedStateVar (ptgfd_get_inOutSignals (vgfd), vh);
    }
}

/**
   Actions:
   - vh is declared as a local at it mother clock, and an input of its own hierarchy,
   - it builds the interface (setInterfaceSubGraph()) of a sub-graph associated with vh,
   - it updates the visibility of the signals referenced in the sub graph associated with vh (mkSubHierarchyInterface()),
   - then, vh is added as output on the branch vh until e, where e is the branching of the references of vh.
*/
void
ProfileBuilder::visibilityOfClock (ptinsth vh)
{
  ptlinsth luse;
  ptinsth use, vhemb, vhembSave;

  setInterfaceSubGraph (CKM->getClockSubGraph (vh));
  mkSubHierarchyInterface (vh);
  stInterface_insert_hoin (vh, vh);

  luse = CKM->getClockReferencesTo (vh);
  vhemb = vh;
  while (ptlinsth_get_next (&luse, &use))
    {
      if (!CKM->satisfiesIsAnAssumption (use))
	if (!CKM->satisfiesIsAnEquivalentForm (use))
	  if (isSelectedClock (use))
	    if (CKM->getOperator (use) != hnot)
	      {
		vhembSave = vhemb;
		vhemb = getForcedBranching (vhemb, use);
		if (CKM->satisfiesUndefClock (vhemb))
		  {
		    vhemb = vhembSave;
		    NM->removeDependenceTo (use);	// WHY HERE ???
		  }
	      }
    }
  if (vh != vhemb)
    insert_clock_output_to_except (vh, vhemb, pk_true);
  else
    insert_clock_local (CKM->getFather (vh), vh);
}

/**
   It takes into account the formula/assertions of a clock (clk) 
   for the interface calculus. It applies insert_clock_local() for each element.
*/
void
ProfileBuilder::addConstraintsDeclaration (ptinsth clk)
{
  ptinsth clk1;

  clk = CKM->getRepresentativeFormula (clk);
  clk1 = CKM->getNextFormula (clk, pk_true);
  while (clk1 != clk)
    {
      insert_clock_local (rootghdc, clk1);
      clk1 = CKM->getNextFormula (clk1, pk_true);
    }
  while (CKM->getNextAssertion (clk1, &clk1))
    insert_clock_local (rootghdc, clk1);
}

/**
   It creates the interface of an assumed root clock (vh).
   It calls setInterfaceSubGraph(), mkSubHierarchyInterface() and adds the clock as an input of its own hierarchy.
*/
void
ProfileBuilder::rootProcessing (ptinsth vh)
{
  ptexpcal vsig;
  ptlinstances lsig;

  setInterfaceSubGraph (CKM->getClockSubGraph (vh));
  mkSubHierarchyInterface (vh);
  if (!CKM->satisfiesPredefinedClock (vh))
    {
      if (!is_input_clock (vh))
	{
	  insert_clock_local (rootghdc, vh);
	  if (CKM->isSubstitutable (vh))
	    if (CKM->satisfiesSub (vh))
	      {
		lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vh));
		while (ptlinstances_get_next (&lsig, &vsig))
		  insert_signal_output_to_except (NM->get_nodeClock (EM->getNode (vsig)), vsig, rootghdc);
	      }
	}
      stInterface_insert_hoin (vh, vh);
    }
}

/** It adds, as input, each signals of a list (vlent) on the branch form the clock of the node of the signal until the root of the hierarchy.
    This operation is not applied for a signal when it is marked "NO_DIFFUSION" (defined in CompilerAttributes.h).
*/
void
ProfileBuilder::InputsGraphVisibility (ptlinstances vlent)
{
  ptinsth h, vhroot;
  ptexpcal vexp;

  vhroot = CKM->getTickClock ();
  while (ptlinstances_get_next (&vlent, &vexp))
    if (!EM->isSignalUnifiedWithClock (vexp, &h))
      {
	stInterface_insert_ent (vhroot, vexp);
	// if (! EM->isMarked(vexp, NO_DIFFUSION))
	insert_signal_input_to_except (NM->get_nodeClock (EM->getNode (vexp)), vexp, vhroot);
      }
}


/**
   @return the clock of the node of a signal (vsig) taking into account the composed (array, sub graph):
   the clock is the clock of the upper level node.
*/
ptinsth ProfileBuilder::getUpperNodeClock (ptexpcal vsig)
{
  ptINode
    n,
    n1;

  n = EM->getNode (vsig);
  while (NM->get_inComposedNode (n, &n1))
    n = n1;
  return CKM->getRepresentativeFormula (NM->get_nodeClock (n));
}

/**
   The visibility of the input signals of a node (aNode) with vh as clock (assumed) is updated in the graph.
   For each input signal (x), and e the branching clock of vh with ^x, it adds
   - x as output on the branch from ^x until e,
   - x as input on the branch from vh until e,
   - x is added as a local signal of e (if it is not already an output of the hierachy with e as root).

   Obviously, it takes into account the case where vh and ^x are in the same branch.
*/
void
ProfileBuilder::visibilityOfInputs (ptINode aNode, ptinsth vh)
{
  ptlinstances lsig;
  ptinsth vhinput, vhemb;
  ptexpcal vsig;
  ptlinsth lh;

  lsig = NM->getInodelInputFlows (aNode);
  while (ptlinstances_get_next (&lsig, &vsig))
    {
      vhinput = getUpperNodeClock (vsig);
      vhemb = getForcedBranching (vh, vhinput);
      if (!CKM->satisfiesUndefClock (vhemb))
	{
	  insert_signal_output_to_except (vhinput, vsig, vhemb);
	  insert_signal_input_to_except (vh, vsig, vhemb);
	  if (!ptlinstances_isin_exp (stInterface_get_lent (rootghdc), vsig))
	    stInterface_insert_loc (vhemb, vsig);
	}
    }

  lh = NM->getInodelInputClocks (aNode);
  while (ptlinsth_get_next (&lh, &vhinput))
    {
      vhinput = CKM->getRepresentativeFormula (vhinput);
      vhemb = getForcedBranching (vh, vhinput);
      if (!CKM->satisfiesUndefClock (vhemb))
	if (!CKM->isSubstitutable (vh))
	  {
	    insert_clock_output_to_except (vhinput, vhemb, pk_true);
	    insert_clock_input_to_except (vh, vhinput, vhemb);
	  }
    }
}

/**
   The visibility of the local signals of a node (aNode) with vh as clock (assumed) is updated in the graph.
   Only the signals that are the ouputs of the graph are considered. For such a signal (x), 
   - x is removed from the locals of the node,
   - x is added as output on the branch from vh until the root of the hierarchy.
   \note the interface of the operand node can be modified.
*/
void
ProfileBuilder::visibilityOfLocals (ptINode aNode, ptinsth vh)
{
  ptlinstances lsig, l;
  ptinsth vhinput;
  ptexpcal vsig;
  boolean bout;
  ptinterface vinter;

  vinter = NM->get_interface (aNode);
  lsig = ptinterface_get_localSignals (vinter);
  l = lsig;
  while (ptlinstances_get_next (&lsig, &vsig))
    {
      bout = ptlinstances_isin_exp (outputs, vsig);
      if (bout)
	{
	  l = ptlinstances_remove_exp (l, vsig);
	  vhinput = EM->getClock (vsig);
	  insert_signal_output_to_except (vhinput, vsig, rootghdc);
	}
    }
  ptinterface_set_localSignals (vinter, l);
}

/**
   The visibility of the output signals of a node (aNode) with vh as clock (assumed) is updated in the graph.
   For such a signal (x), x is added as a local signal of vh, if it is not already an output of the hierachy 
   with vh as root.
*/
void
ProfileBuilder::visibilityOfOutputs (ptINode aNode, ptinsth vh)
{
  ptlinstances lsig;
  ptexpcal vsig;

  lsig = NM->getInodelOutputFlows (aNode);
  while (ptlinstances_get_next (&lsig, &vsig))
    stInterface_insert_loc (vh, vsig);
}

/**
   The visibility of a list of signals (lsig) referenced at a clock (vh) is updated in the graph.
   Only the signals that are the ouputs of the graph are considered. For such a signal (x), and e the branching
   of vh and ^x,
   - x is added as output on the branch ^x until e (when e = vh)
   - x is addded as a local signal of vh (if it is not already an output of the hierachy with vh as root).
*/
void
ProfileBuilder::visibilityOf (ptlinstances lsig, ptinsth vh)
{
  ptinsth vhinput, vhemb;
  ptexpcal vsig;
  boolean bout;

  while (ptlinstances_get_next (&lsig, &vsig))
    {
      bout = ptlinstances_isin_exp (stInterface_get_lsort (rootghdc), vsig);
      if (bout)
	{
	  vhinput = EM->getClock (vsig);
	  vhemb = getForcedBranching (vh, vhinput);
	  if (!CKM->satisfiesUndefClock (vhemb) && (vhemb == vh))
	    insert_signal_output_to_except (vhinput, vsig, vhemb);
	  stInterface_insert_loc (vh, vsig);
	}
    }
}


void
ProfileBuilder::visibilityOfSharedStateVar (ptlinstances lsig, ptinsth vh)
{
  ptexpcal vsig;
  ptinsth vhemb, localTo;
  ptinterface vstru;

  while (ptlinstances_get_next (&lsig, &vsig))
    {
      localTo = EM->getClock (vsig);
      vhemb = getForcedBranching (vh, localTo);
      insert_signal_inout_to_except (vh, vsig, vhemb);
      //  if (EM->isStateVar(vsig)) localTo = rootghdc; else localTo = EM->getClock(vsig); // il faut calculer l'embranchement des references. (comme pour les autres signaux)
      addLocal(vhemb, vsig);
    }
}


/**
   The visibility of the signals, referenced in a node (aNode):
   - as inputs (visibilityOfInputs(), 
   - as local (vibilityOfLocals()),
   - as output  (visibilityOfOutputs()),
   at the clock of the node is updated in the graph.
*/
void
ProfileBuilder::interfaceCalculus_composednode (ptINode aNode)
{
  ptinsth vh;

  vh = NM->get_nodeClock (aNode);
  visibilityOfInputs (aNode, vh);
  visibilityOfLocals (aNode, vh);
  visibilityOfOutputs (aNode, vh);
}

/**
   It applies interfaceCalculus_composednode() for each element of a list (lNodes) of nodes.
   (Recall: the interface of a single node is already known.)
*/
void
ProfileBuilder::interfaceCalculus_composednodes (ptlINode lNodes)
{
  ptINode vi;

  while (ptlINode_get_elem (&lNodes, &vi))
    interfaceCalculus_composednode (vi);
}

/**
   It takes into account the constraints on placed clocks for the calculus of the visibility of the clocks.
   For such an element (h1=h2) and e their branching,
   h1 (h2) is added as output in the interface of each node of the hierarchy from h1 (resp. h2) to e.
*/
void
ProfileBuilder::ConstrainstPlacedClocks ()
{
  ptlinsth lh1, lh2;
  ptinsth vh1, vh2, vhemb;

  CKM->getConstraintsOnPlacedClocks (&lh1, &lh2);
  while (ptlinsth_get_next (&lh1, &vh1))
    {
      ptlinsth_get_next (&lh2, &vh2);
      // vhemb = getForcedBranching(vh1, vh2);  
      // replaced (Dec 2013) by the root clock because the clock constraints are visible at the root clock-.
      vhemb = rootghdc;
      insert_clock_output_to_except (vh1, vhemb, pk_true);
      insert_clock_output_to_except (vh2, vhemb, pk_true);
    }
}

/**
   @return "true" if a node (vinstr) produces an output (sig), "false" otherwise.
   (local to mkInterfaceDependence())
*/
_VIRTUAL_ boolean
ProfileBuilder::getDependenceFlow (ptINode vinstr, ptexpcal * sig)
{
  ptlinstances lsort;

  switch (NM->get_oper (vinstr))
    {
    case opdefsig:
    case opdefpart:
    case opldefpart:
    case opmoti:
    case opret:
    case opappe:
    case opnodecall:
      lsort = NM->getInodelOutputFlows (vinstr);
      return ptlinstances_get_next (&lsort, sig);

    case opsubgraph:
    case opnil:
      return pk_false;
    default:
      pK_messenger->Caseerror (currentLine, currentFile);
    }
  return pk_false;
}


/** 
    For a dependence (n1 -> n2) when h, 
    s1 (resp. s2 and h) is added as output from its clock to the root of the hierarchy, where s1 (resp. s2) is the result of getDependenceFlow()
    with n1 (resp. n2) as parameter.
    \note dependencies are visible at the root of the process (Choice).
*/
void
ProfileBuilder::mkInterfaceDependence (ptdependence aDep)
{
  ptINode vFrom, vTo;
  ptexpcal s1, s2;
  ptinsth h, ckr;

  h = aDep->getDependence (&vFrom, &vTo);
  h = CKM->getRepresentativeFormula (h);
  if (! ptlinsth_sorted_isin_elem (unClocks, h)) {
  if (getDependenceFlow (vFrom, &s1))
    if (getDependenceFlow (vTo, &s2))
      {
	ckr = CKM->getTickClock ();
	if (CKM->satisfiesUndefClock (ckr))
	  ckr = rootghdc;
	if (!CKM->satisfiesUndefClock (h))
	  insert_clock_output_to_except (h, ckr, pk_true);
	insert_signal_output_to_except (NM->get_nodeClock (vFrom), s1, ckr);
	insert_signal_output_to_except (NM->get_nodeClock (vTo), s2, ckr);
      }
}
}

/** Taking into account the explicit dependencies for the interface of a process.
    Dependencies are visible at the root (choice) of the process.
    It applies mkInterfaceDependence() for each explicit dependence.
*/
void
ProfileBuilder::mkInterfaceDependences ()
{
  ptldependence lDep;
  ptdependence aDep;

  lDep = AGRAPH->getExplicitDependences ();
  while (lDep->get_next (&aDep, &lDep))
    mkInterfaceDependence (aDep);
  lDep = AGRAPH->getReinforcedDependences ();
  while (lDep->get_next (&aDep, &lDep))
    mkInterfaceDependence (aDep);
}

/**
   Interface calculus for the node calls, array, subgraphs
   It applies the interfaceCalculus_composednodes() for each of them.
*/
void
ProfileBuilder::interfaceComposedNodes ()
{
  ptlINode lNodes;

  if (NM->getNodeCalls (&lNodes))
    interfaceCalculus_composednodes (lNodes);
  if (NM->getSubGraphs (&lNodes))
    interfaceCalculus_composednodes (lNodes);
  if (NM->getIterates (&lNodes))
    interfaceCalculus_composednodes (lNodes);
}


/** This method realizes a left to right topdown traversal of the hierarchy with \e clock as root.
    The interface of the subgraph associated with the clock is produced then the xxx;
    For each \e useful node of the hierarchy, it applies:
    - rootProcessing() for a root of the hierarchy, visibilityOfClock() otherwise ,
    - addConstraintsDeclaration()
*/
void
ProfileBuilder::leftRightTopdownTraversal (ptinsth clock)
{
  ptinsth vh;
  ptlinsth lchild;

  lchild = CKM->getChildren (clock);
  while (ptlinsth_get_next (&lchild, &vh))
    {
      vh = CKM->getRepresentativeFormula (vh);
      if (isSelectedClock (vh))
	{
	  if (CKM->satisfiesRoot (vh))
	    rootProcessing (vh);
	  else
	    visibilityOfClock (vh);
	  leftRightTopdownTraversal (vh);
	  addConstraintsDeclaration (vh);
	}
    }
}

boolean ProfileBuilder::visibilitySignalsOfExtraction (ptinsth vh, ptinsth vhAt)
{
  ptlinstances lsig;

  if (!CKM->satisfiesSub (vh)) return pk_false;
  // printf( "visibilitySignalsOfExtraction = %d, at = %d \n", CKM->getNum(vh), CKM->getNum(vhAt));
  lsig = EM->getSignalsReferencedIn (CKM->getConditionSub (vh));
  visibilitySignalsAtClock (lsig, vhAt, pk_true);
  ptlinstances_free (lsig);
  return pk_true;
}

/**
   Method without effect when the formula (vh) is not an extraction.
   The visibility of the signals referenced in the expression associated with the formula is updated in the graph.
   It applies visibilitySignalsAtClock() for each signal.
*/
void
ProfileBuilder::extractionInterface (ptinsth vh)
{
  ptinsth h1, h2;
  boolean b1, b2;

  if (! visibilitySignalsOfExtraction (vh, vh))
    {
      // The operands may be substituable.
      b1 = pk_true;
      b2 = pk_true;
      if (!CKM->getMinusArgs (vh, &h1, &h2))
	if (!CKM->getSupArgs (vh, &h1, &h2))
	  if (!CKM->getInfArgs (vh, &h1, &h2))
	    {
	      b2 = pk_false;
	      b1 = CKM->getUnaryArg (vh, &h1);
	    }
      if (b1)
	visibilitySignalsOfExtraction (h1, vh);
      if (b2)
	visibilitySignalsOfExtraction (h2, vh);
    }
}

/** 
*/
void
ProfileBuilder::substituableClocksInterface (ptinsth clock)
{
  ptinsth vh;
  ptlinsth lchild;

  lchild = CKM->getChildren (clock);
  while (ptlinsth_get_next (&lchild, &vh))
    {
      vh = CKM->getRepresentativeFormula (vh);
      if (isSelectedClock (vh))
	{
	  extractionInterface (vh);
	  substituableClocksInterface (vh);
	}
    }
}


void ProfileBuilder::NullClockAssertions()
{
  ptlinsth lh;
  ptinsth vh, vhemb;

  if (DCplusGraph && CKM->getNullClockAssertions (&lh))
    while (ptlinsth_get_next (&lh, &vh))
      {
	vhemb = rootghdc;
	insert_clock_local (rootghdc, vh);
      }
}
