/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
  If the operand clock (vh) is not yet marked, it is marked, and recursively, marks its references.
*/
    void
 InputsTrainClusters::colouringClock(ptinsth vh, typeMark amark)
{
  vh = CKM->getRepresentativeFormula(vh);
  if (!ptlinsth_sorted_isin_elem(SelectedClocks, vh))
    if (ptinsthAttr_append_mark(vh, amark))
      colouringFrom(vh, amark);
}

/**
  If the operand node (aNode) is not yet marked, it is marked, and recursively, marks its references.
*/
void InputsTrainClusters::colouringNode(ptINode aNode, typeMark amark)
{
  if (ptINodeAttr_append_mark(aNode, amark))
    colouringFrom(aNode, amark);
}

/** 
    It marks by aMark each clock of a list (lh) of clocks (transitively).
*/
void InputsTrainClusters::colouring(ptlinsth lh, typeMark amark)
{
  ptinsth vh;
  while (ptlinsth_get_next(&lh, &vh))
    colouringClock(vh, amark);
}

/** 
    It marks by aMark each node of a list (ln) of nodes (transitively).
*/
void InputsTrainClusters::colouring(ptlINode ln, typeMark amark)
{
  ptINode aNode;

  while (ptlINode_get_elem(&ln, &aNode))
    colouringNode(aNode, amark);
}

/** 
    It marks by aMark the nodes n such that m->n is an element of ldep (transitively).
*/
void InputsTrainClusters::colouring(ptldependence ldep, typeMark amark)
{
  ptINode n1, aNode;
  ptdependence adep;

  while (ldep->get_next(&adep, &ldep))
    {
      adep->getDependence(&n1, &aNode);
      colouringNode(aNode, amark);
    }
}

/** 
    It marks by  aMark nodes/clocks referencing aNode.
*/
void InputsTrainClusters::colouringFrom(ptINode aNode, typeMark aMark)
{
  colouring(NM->get_clockSuccessors(aNode), aMark);
  colouring(NM->get_successors(aNode), aMark);
}

/** 
    It marks by  aMark nodes/clocks referencing vh.
*/
void InputsTrainClusters::colouringFrom(ptinsth vh, typeMark aMark)
{
  colouring(CKM->getNodeReferencesTo(vh), aMark);
  colouring(CKM->getClockReferencesTo(vh), aMark);
}

boolean InputsTrainClusters::colouringNodeCall(ptINode aNode, typeMark aMark)
{
  ptexpcal anInput, linputs;

  if (!NM->isNodeCall(aNode))
    return pk_false;
  linputs = NM->get_leftOperand(aNode);
  while (EM->getNaryArgs(&linputs, &anInput))
    {
      colouringFrom(EM->getNode(anInput), aMark);
    }

  return pk_true;
}

/** 
    It marks nodes/clocks for (input) clustering:
    - for each input node (element of SelectedNodes list), a new mark is taken,
    and (transitively) all the successors of the input are marked with this mark (colouringFrom(ptINode,typeMark)).

    - for each input clock (element of SelectedClocks list), a new mark is taken,
    and (transitively) all the successors of the input are marked with this mark (colouringFrom(ptlinsth,typeMark)).
*/
void InputsTrainClusters::colouring()
{
  ptlINode ln;
  ptlinsth lh;
  ptINode aNode;
  ptinsth vh;
  typeMark aMark;

  // markDelays();
  // if (NM->getIterates(&ln)) while(ptlINode_get_elem(&ln, &aNode)) colouringFrom(aNode, getNewMark());    
  ln = SelectedNodes;
  while (ptlINode_get_elem(&ln, &aNode))
    {
      aMark = getNewMark();
      if (!colouringNodeCall(aNode, aMark))
	colouringFrom(aNode, aMark);
      addTable(aNode, aMark);
    }
  lh = SelectedClocks;
  while (ptlinsth_get_next(&lh, &vh))
    colouringFrom(CKM->getRepresentativeFormula(vh), getNewMark());
}

/** It defines:
    - the set (SelectedNodes) of the selected nodes for the calculus (SelectedNodes is  a list declared in the TrainClusters class)
    - the set (SelectedClocks) of the selected clocks for the calculus (SelectedClocks is  a list declared in the TrainClusters class)
    A node is selected when 
    - it is an effective inputs of the graph
    - it is produced by a external call without inputs (example : x:=foo())
    A clock is selected when 
    - it is a root of the hierarchy.

    It initiliases also (InputsGraph/OutputsGraph are lists declared in the TrainClusters class)
        - the set (InputsGraph) of the selected signals for the calculus 
	- the set (OutputsGraph)  of the outputs signals of the graph
 */
void InputsTrainClusters::selectNodes()
{
  ptlinstances lsig;
  ptlinsth lh;
  ptexpcal vsig, le, ls, par;
  ptident nom;
  ptinsth vh;
  boolean bdummy;
  ptlINode ln;
  ptINode aNode;
  ptDCG vg;

  // Entry Points for the calculus.
  // inputs = inputs declared in the interface + reads done by external calls
  lsig = AGRAPH->getInputs();
  InputsGraph = ptlinstances_copy(lsig);
  OutputsGraph = AGRAPH->getOutputs();

  while (ptlinstances_get_next(&lsig, &vsig))
    {
      aNode = EM->getNode(vsig);
      ptlINode_insert_elem(&SelectedNodes, aNode);
    }

  if (NM->getNodeCalls(&ln))
    while (ptlINode_get_elem(&ln, &aNode))
      if (!NM->is_inComposedNode(aNode))
	{
	  NM->get_nodeCall_arg(aNode, &vg, &par, &le, &ls);
	  nom = vg->getName ();
	  if (le == NILEXPCAL)
	    {
	      ptlINode_insert_elem(&SelectedNodes, aNode);
	      InputsGraph = ptlinstances_sortedfusion(NM->getInodelOutputFlows(aNode), InputsGraph);
	    }
	}

  if (NM->getProcCalls(&ln))
    while (ptlINode_get_elem(&ln, &aNode))
      if (!NM->is_inComposedNode(aNode))
	{
	  NM->get_procCall_arg(aNode, &nom, &par, &le, &ls);
	  if (le == NILEXPCAL && ! isSendId(nom))
	    {
	      ptlINode_insert_elem(&SelectedNodes, aNode);
	      InputsGraph = ptlinstances_sortedfusion(NM->getInodelOutputFlows(aNode), InputsGraph);
	    }
	  else
	    if (isReceiveId(nom))
	      {
		ptlINode_insert_elem(&SelectedNodes, aNode);
		InputsGraph = ptlinstances_sortedfusion( NM->getInodelInputFlows(aNode), InputsGraph);
	      }
	}

  /* Clocks */
  lh = CKM->getlhroots();
  while (ptlinsth_get_next(&lh, &vh))
    {
      vh = CKM->getRepresentativeFormula(vh);
      if (CKM->satisfiesVari(vh))
	SelectedClocks = ptlinsth_sorted_insert_elem_bool(SelectedClocks, vh, &bdummy);
    }

}

/**
   It annotates  with the pragma PRAGMA_CLUSTER the nodes of lSubGraphs.
   Moreover, it adds to the clusters the pragma  PRAGMA_DEFINED_CLOCK_HIERARCHY when the current graph is annotated with.
*/
void InputsTrainClusters::attributesSetting()
{
  ptpragma vp;
  ptINode aNode, node;
  ptlMark lmCluster;
  typeMark vmark;
  ptlinstances vl;
  ptexpcal elem, outs;
  ptlINode lSubGraphs;

  // Clusters::attributesSetting();
  lSubGraphs = lSubGraph;
  while (ptlINode_get_elem(&lSubGraphs, &aNode))
    {
      lmCluster = ptINodeAttr_get_marks(aNode);
      vl = NILLINSTANCES;
      while (ptlMark_get_elemNext(&lmCluster, &vmark))
	{
	  if (getTableNode(vmark, &node))
	    {
	      outs = NM->get_leftOperand(node);
	      // adding the signals to the current pragma.
	      if (outs != NILEXPCAL)
		{
		  if (EM->satisfiesList(outs))
		    while (EM->getNaryArgs(&outs, &elem))
		      vl = ptlinstances_insert(vl, elem);
		  else
		    vl = ptlinstances_insert(vl, outs);
		}
	    }
	}
      vp = newPredefinedPragma(PRAGMA_CLUSTER, (ptObjPragma) vl, NILTREE);
      NM->addPragma(aNode, vp);
    }
}
