/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_
/**
   It assigns a clock (vh) to a partition (sub-graph).
*/
    void
 RunOnMark::assignClock(ptinsth vh)
{
  ptINode aSubGraph;
  typeMark aMark;
  ptlMark lm;

  lm = ptinsthAttr_get_marks(vh);
  while (ptlMark_get_elemNext(&lm, &aMark))
    {
      aSubGraph = SubGraphTable[aMark - firstMark];
      NM->subGraphInsertClock(aSubGraph, vh);
    }
}

/**
   It assigns a node (aNode) to a partition (sub-graph).
*/
void RunOnMark::assignNode(ptINode aNode)
{
  ptINode aSubGraph;
  ptlMark lm;
  typeMark aMark;

  lm = ptINodeAttr_get_marks(aNode);
  while (ptlMark_get_elemNext(&lm, &aMark))
    {
      aSubGraph = SubGraphTable[aMark - firstMark];
      NM->subGraphInsertNode(aSubGraph, aNode);
    }
}

_PROTECTED_
/**
   It inserts a sub-graph (aSubGraph) if it is useful in the global list (lSubGraph of the Class Cluster).
*/
_VIRTUAL_ void RunOnMark::selectSubGraph(ptINode aSubGraph)
{
  ptlINode ln;
  ptlinsth lh;

  if (NM->get_subGraphDefinitions(aSubGraph, &ln, &lh))
    if (ln != NILLINODE || lh != NILLINSTH)
      ptlINode_insert_elem(&lSubGraph, aSubGraph);
    else
      NM->removeNode(aSubGraph);
}

/**
   It initializes the partitionning functionality by allocating a 
   vector of Subgraphs. The size is internally computed (maxMark - firstMark +1, where 
   maxMark and firstMark are internal parameters).
   The vector of subgraphs will be replaced in finalizeUseInternalRepresentation() 
   by another structure to apply methods of Cluster class.
*/
_VIRTUAL_ void RunOnMark::partitionningInitialize()
{
  integer i;
  ptlMark lm;

  nbsets = maxMark - firstMark + 1;
  SubGraphTable = (ptINode *) malloc(nbsets * sizeof(ptINode));
  for (i = 0; i < nbsets; i++)
    {
      lm = NILLMARK;
      ptlMark_insert_elem(&lm, i + firstMark);
      SubGraphTable[i] = mkNodeCluster(lm);
    }
}

/** 
    The vector of Subgraphs is stored in the general list (lSubGraph of the class Cluster) to apply methods of Cluster class.
    This management is previously opened in partitionningInitialize() method.
*/
_VIRTUAL_ void RunOnMark::finalizeUseInternalRepresentation()
{
  integer i;
  ptINode aSubGraph;
  ptrepstr vstr;

  lSubGraph = NILLINODE;
  for (i = 0; i < nbsets; i++)
    {
      aSubGraph = SubGraphTable[i];
      selectSubGraph(aSubGraph);
      vstr = ExtractString(RunOnNameTable[i]);
      setpragmapartition(aSubGraph, vstr);
    }
}

/**
   It assigns the pragma "Processor" of the current graph (if it exists) to the partition (aSubGraph).
   
   Example:
   pragmas
      Processor "cpiom1" "CPIOM"
      Binding { } cpiom1
    end pragmas
*/
_VIRTUAL_ void RunOnMark::addProcessorPragma(ptINode aSubGraph, ptrepstr vstr)
{
  ptlpragmas lprag;
  ptpragma aPragma;
  boolean bfound;
  ptrepstr vstrProcessor;

  lprag = AGRAPH->getPragmas();
  bfound = pk_false;
  vstrProcessor = getPredefinedPragma(PRAGMA_PROCESSOR_NAME);
  while (!bfound && ptlpragmas_get_next(&lprag, &aPragma))
    if (getPragmaName(aPragma) == vstrProcessor)
      bfound = (strFirstStmtsValue(aPragma) == vstr);

  if (bfound)
    NM->addPragma(aSubGraph, aPragma);
}

/**
   It assigns to a partition (aSubGraph):
     - the pragma \b RunOn \b vstr
     - the pragma \b Processor 
   
*/
_VIRTUAL_ void RunOnMark::setpragmapartition(ptINode aSubGraph, ptrepstr vstr)
{
  NM->addPragma(aSubGraph, newPredefinedPragma(PRAGMA_RUN_ON, NILOBJPRAGMA, repstr2PragmaStatement(vstr)));
  addProcessorPragma(aSubGraph, vstr);
  // addBindingIO(aSubGraph);
}

/** It imports the clocks referenced in the assertions:
    For a clock (h) defined in a sub-graph (aSubGraph), if it exists an assertions h=h1 and h1 is not 
    defined in aSubGraph, h1 is added as input of the subgraph.
*/
_VIRTUAL_ void RunOnMark::importAssertions(ptINode aSubGraph)
{
  ptlINode ln;
  ptinsth vh, vh1;
  ptlinsth lh1, lh;
  ptinterface vinterface;

  if (NM->get_subGraphDefinitions(aSubGraph, &ln, &lh))
    {
      vinterface = NM->getSet_interface(aSubGraph);
      lh1 = lh;
      while (ptlinsth_get_next(&lh, &vh))
	// if (isEventClockedFormat(graph_type)) 
	while (CKM->getNextAssertion(vh, &vh1))
	  {
	    if (!ptlinsth_isin_elem(lh1, vh1))
	      ptinterface_insert_inputClock(vinterface, vh1);
	    vh = vh1;
	  }
    }
}

/**
   See Clusters::subGraphsBuildProfile ().
   The specialisation consists in adding some input for subgraphs: a signal x is added as input 
   when the clock is defined in the subgraph and it is represented by x.
 */
_VIRTUAL_ void RunOnMark::subGraphsBuildProfile()
{
  ptINode n, aSubGraph;
  ptlINode lnodes, ln;
  ptlinsth lhinputClocks, lh1, lh2;
  ptinsth rootClock, vh2;
  ptexpcal vsig;
  ptinterface vinterface;
  boolean b;

  Clusters::subGraphsBuildProfile();


  ln = lSubGraph;
  b = false;
  while (ptlINode_get_elem(&ln, &aSubGraph))
    {
      vinterface = NM->get_interface(aSubGraph);
      NM->get_subGraphDefinitions(aSubGraph, &lnodes, &lh1);
      NM->get_subGraphProjectedSystemInfos(aSubGraph, &rootClock, &lh2);
      while (ptlinsth_get_next(&lh2, &vh2))
	{
	  if (vh2 != rootClock)
	    if (CKM->isClockRepresentedBySignal(vh2, &vsig))
	      {
		n = EM->getNode(vsig);
		if (!ptlINode_isIn(lnodes, n))
		  {
		    ptinterface_insert_inputSignal(vinterface, vsig);
		    b = true;
		  }
	      }
	}
    }
  if (b)
    InterfaceCalculusSubGraphsUpdate();
}
