/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

static ptDCG_basic gw = NILDCG_BASIC;

#include "DCGs_Macros.c"

_PRIVATE_ void
DCG_basic::init (ptDCG g)
{
  if (gw == NILDCG_BASIC)
    gw = new DCG_basic ();
  gw->setDCG (g);
}

_PROTECTED_
/** Sub-Graph.
    It applies for all sub-graphs of the manager the method \ref subGraphUnlinkNodesFromClocks
 */
  void
DCG_basic::subGraphsUnlink ()
{
  ptINode aSubGraph;
  ptlINode lSubGraphs;

  require_node_manager (NM);
  if (NM->getSubGraphs (&lSubGraphs))
    while (ptlINode_get_elem (&lSubGraphs, &aSubGraph))
      NM->unlinkNodesFromClock (aSubGraph);

}

/**
   Recursively write each elementary process "PP" that has a subexpression "Ei" as "(| PP' | xi:=Ei|) where xi"
   such that xi is a new identifier and PP' is PP in which all occurences of "Ei" are replaced by "xi"
   @return the value "true" if the transformation succeeds, "false" otherwise.
   \ref Graph2Kernel "For more details..."
*/
boolean DCG_basic::expressionExpandOneLevel ()
{
  ptDCG2KernelTranslator
    o;
  boolean
    b;

  o = new DCG2KernelTranslator (this);
  b = o->apply ();
  delete
    o;
  return b;
}

/**
   It rewrites Boolean expressions of the undersamplings (i.e. expression such that "when E") 
   for logical and relational operators. It can be useful for proving constraints.
   @return the value "true" if an expression is rewritten, "false" otherwise.

   \ref Booleans2Events "For more details..."
 
*/
boolean DCG_basic::transformBooleanClocks2EventsOneLevel ()
{
  ptbooleans2events
    arew;
  boolean
    b;

  pK_messenger->messageIdent (inforewc, getName ());
  arew = new booleans2events (getDCG ());
  b = arew->apply (pk_true, pk_true, pk_true);
  delete
    arew;
  return b;
}


/** 
    It verifies that the graph can be translated into a sequential one.
    @return "true" if the current graph contains constraints, "false" otherwise.
*/
boolean DCG_basic::hasUnsolvedClockConstraintsOneLevel ()
{
  ptlinsth
    lh;
  boolean
    bc;

  // It is not applied to referenced sub-graphs.
  bc = getClockManager ()->isConstrainedSystem ();
  bc = bc || getNullClockConstraints (&lh);
  if (bc)
    pK_messenger->messageIdent (messlequnoprov, getName ());
  return bc;
}




/**
   Unifying operation.

   This unifying operation consists in the merge of nodes into one node when their definition expression
   are equal (syntacticly). It calls the following methods:
   - removeSignalEqualitiesOneLevel() for the elimination of the renamings (x:=y)
   - DCG::ConstantsUnifying() for the unification of the constant definitions                      
   - DCG::nodesUnification() the more general cases.
   @return "true" when some nodes have been unified, "false" otherwise.
*/
boolean DCG_basic::removeSignalClonesOneLevel ()
{
  boolean
    b,
    bres;
  ptDCG
    vdcg;

  pK_messenger->messageIdent (infoUnify, getName ());
  vdcg = getDCG ();
  bres = pk_false;
  b = pk_true;
  while (b)
    {
      b = pk_false;
      if (removeSignalEqualitiesOneLevel ())
	b = pk_true;
      if (vdcg->ConstantsUnifying ())
	b = pk_true;
      if (vdcg->nodesUnification ())
	b = pk_true;
      if (b)
	bres = pk_true;
    }
  return bres;
}

/**
   This operation consists in the substitution of a signal by its definition expression
   when the number of its references is less than a given value (maxRef).
   @param maxRef (in) int that specifies the bound used by the algorithm.
   @return "true" when substitutions have been applied, "false" otherwise.
*/
boolean
DCG_basic::substituteExpr2VarOneLevel (int maxRef)
{
  typeMark m;
  boolean b, find;

  m = getMark ();
  b = pk_true;
  find = pk_false;
  ptDCG vdcg = getDCG ();
  while (b)
    {
      b = pk_false;
      if (vdcg->markReferencedMaxTime (m, maxRef, pk_false))
	{
	  find = pk_true;
	  b = pk_true;
	  vdcg->substituteMarkedVariables (m);
	}
    }
  unmarkObjects (m);
  if (find)
    substituteEvaluatedExpression ();
  return find;
}


/** 
    Removing, at one level, unused nodes for the current graph.
    It realizes elementary optimizations by deleting objects of the graph which are 
    not used to produce (resp. consume) outputs (resp. inputs) of the program.

    @param bLastContributionClockDefaultUsed (in) must be "true" if the last contribution 
    clock of a default expression is considered, "false" otherwise.

    \ref Optimisations "For more details..."

*/
void
DCG_basic::nonProductiveSignalEliminationOneLevel (boolean bLastContributionClockDefaultUsed, boolean withClockDependence)
{
  int date;
  ptPruner aPruner;
  ptlinsth unusedClocks, lh;
  ptlINode unusedNodes, ln;

  date = pK_messenger->messageBeginOperVerbose (infomarksigho);
  aPruner = new Pruner (getDCG ());
  aPruner->apply (bLastContributionClockDefaultUsed, withClockDependence, &ln, &lh);
  delete aPruner;

  removeNodes (ln, lh);		// Cleaning

  unusedNodes = getUnused (&unusedClocks);
  unusedClocks = ptlinsth_sorted_fusion (unusedClocks, lh);
  unusedNodes = ptlINode_fusion (unusedNodes, ln);
  setUnused (unusedNodes, unusedClocks);
  updatePragmas ();
  pK_messenger->messageEndOperVerbose (date);
}


/** 
    It creates states variables at the clock \b hto from the tree of root \b hfrom.
    \warning it Assumes that \b hfrom is under \b hto in the hierarchy of clocks.
    @return "true" if the operation succeeds, "false" otherwise.
    \ref States "For more details..."

*/
boolean
DCG_basic::extendBooleanDelayClockOneLevel (ptinsth hfrom, ptinsth hto)
{
  ptbooleanStatesVariables stateBuilder;

  stateBuilder = new booleanStatesVariables (getDCG ());
  return stateBuilder->extendDelayFromClockToClock (hfrom, hto);
}

/** 
    It performs a shift register transformation.
    It rewrites synchronous function f such that 
    Y := f(X1 $ m1 init V1,...,Xn $ mn init Vm) 
    into 
    Y := y' $ j init f(V1',...,Vm') and y' := f(X1 $ m1' init V1",...., Xm $ mn' init Vm").

    \ref States "For more details..."
*/
void
DCG_basic::moveIn2OutDelaysOneLevel ()
{
  ptdownstreamRetiming obj;
  int date;

  //------ TODO for one node   
  date = pK_messenger->messageIdent (dretiming, getName ());
  obj = new downstreamRetiming (getDCG ());
  obj->applyRetiming ();
  delete obj;
  pK_messenger->messageEndOperVerbose (date);
}


/**
   It replaces a reference to a graph (node call) by an insertion of a copy of the 
   referenced graph in the context of the call.
   @return "true" if the operation succeeds, "false" otherwise (external node).

   \ref TheMorphisms "For more details..."
*/
boolean
DCG_basic::expandModelReferenceOneLevel (ptINode nodeCall, boolean binclude)
{
  ptInclusion obj;
  int date;
  ptDCG_basic g1;
  ptINode SubGraphNode, n;
  ptlINode ln;
  ptlinsth lh;

  //--TODO--new name expandModelReference (node call) is a DCG_basic function
  // new name expandModelReference (model reference) is a DCG_basic function
  // new name expandModelReference (boolean force) is a DCG_basic function

  require_node_manager (NM);
  g1 = (ptDCG_basic) NM->get_nodeCall_graph (nodeCall);
  if (g1->isGraphOfExternalProcess ())
    return pk_false;
  date = pK_messenger->messageIdent (mincluding, g1->getName ());

  obj = new Inclusion ();
  if (binclude)
    obj->includeGraph (getDCG (), nodeCall);
  else
    {
      SubGraphNode = obj->includeGraphAsSubGraph (getDCG (), nodeCall);
      NM->get_subGraphDefinitions (SubGraphNode, &ln, &lh);
      while (ptlINode_get_elem (&ln, &n))
	NM->addMark (n, DontOptimize);
    }

  delete obj;
  setClocksName ();
  subGraphsUnlink ();
  NM->removeNode (nodeCall);
  pK_messenger->messageEndOperVerbose (date);
  return pk_true;
}


/** 
    The partial definitions are rewritten using default operator. For example, \n
    x ::= E1 when C1, ... x ::= En when Cn \n
    is rewritten in \n
    x :=  (E1 when C1) default .... default (En when Cn).\n
    Since the clocks Ci are constrainsted to be exclusive by definition of partial definition ([Ei when Ci]*[Ej when Cj] is null for all i,j), 
    the order is not important in the default expression.
*/
void
DCG_basic::groupPartialDefinitionsOneLevel ()
{
  ptlINode lSubGraphs;
  ptINode aSubGraph;

  require_node_manager (NM);
  if (NM->PartialDefinitionsMerging ())
    if (NM->getSubGraphs (&lSubGraphs))
      {				// voir si on peut optimiser.
	// NM -> subGraphUnlinkNodesFromClocks(lSubGraphs);
	while (ptlINode_get_elem (&lSubGraphs, &aSubGraph))
	  {
	    NM->unlinkNodesFromClock (aSubGraph);
	    NM->set_interface (aSubGraph, NILINTERFACE);
	    NM->getSet_interface (aSubGraph);
	  }
      }
}


/**
   It splits the default Signal operator to a set of partial definitions. A definition
   - x := (E1 when C1) default ... default (En when Cn) 
   is rewitten into 
   - (| x ::= (E1 when C1) | ... | x ::= (En when Cn) |)

   By construction, in the graph the clocks C1...Cn are exclusive: [Ei when Ci]*[Ej when Cj] is null for all i,j

   @return "true" if some "default" have been splited, "false" otherwise.
*/
boolean
DCG_basic::splitDefaultDefinitionsOneLevel ()
{
  return getNodeManager ()->DefaultSpliting ();
}


/**
   It reinforces dependencies:
   the "iteration Precedences" build during the production of the graph are transformed in data dependence using auxillary signals.
   @return "true" if some reinforcing have been done, "false" oherwise.
   \see iterationPrecedences
*/
boolean
DCG_basic::dependenceReinforcingOneLevel ()
{
  if (!iterationPrecedences ())
    return pk_false;
  buildClockProcessInterfaceOneLevel ();
  return pk_true;
}

/**
   This method transforms the node calls (process, action, node,...) of the graph to node calls to node model by 
   reinforcing  the dependence relation: each input of a node precedes each output of that node. 
   Such a reinforcing may introduce circuits in the graph.
   @return "true" if some reinforcing have been done, "false" toherwise.
   \see iterationPrecedences
*/
boolean
DCG_basic::transformModelReferences2NodeReferencesOneLevel ()
{
  // It may add circuits !!!
  return getNodeManager ()->setsbDCplusDependencesForNodeCalls ();
}


/**
   It produces the morphim of the current graph according to the definitions assigned of the Signal operators given in a table (ht).
   When bSimul is "true", the current graph is modified (selected booleans become outputs).
  
   @return "true" if the operation succeeds, "false" otherwise.

   \ref MorphismDefinedInTable "For more details..."
*/
boolean DCG_basic::applyTabularMorphismOneLevel (ptHomomorphismTable ht, ptDCG_basic * gresult, boolean bSimul)
{
  pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("DCG_basic::applyTabularMorphismOneLevel"));
  return pk_false;
}




/** 
    Flatenning the subgraphs of the graph. The dependences are propagated.
    For each sub-graph, it calls NodeManager::subGraphFlattenning()
*/
void
DCG_basic::breakCPGOneLevel ()
{
  ptlINode lSubGraphs;
  ptINode aSubGraph;

  require_node_manager (NM);
  if (NM->getSubGraphs (&lSubGraphs))
    while (ptlINode_get_elem (&lSubGraphs, &aSubGraph))
      subGraphFlattenning (aSubGraph);
}

/**
   NOT IMPLEMENTED (in development).
   
   It removes the enumerated types of the current graph. 
   The references are modified: T#value access value is replaced by the associated constant.
*/
boolean
DCG_basic::changeEnumerated2IntegersOneLevel ()
{
  ptlinstances ltypes;
  ptexpcal elem;
  boolean b;
  TypeType TT;
  require_expr_manager (EM);

  if (getUserTypesOneLevel (&ltypes))
    {
      b = pk_false;
      while (!b && ptlinstances_get_next (&ltypes, &elem))
	{
	  TT = EM->getExpType (elem);
	  if (TypeTestEnumerationClass (TT))
	    b = pk_true;
	}
      if (b)
	pK_messenger->globalMessageStr (codeNotYetImplemented, string2repstr ("DCG_basic::changeEnumerated2IntegersOneLevel"));
    }
  return pk_true;
}


/**
   @return "true" if the cycles are printed, "false" otherwise.
   A cycle is seen as a composition of set of dependences/equations (see unparseTree::mkTreeCycles()).
*/
boolean
DCG_basic::unparseCyclesOneLevel ()
{
  ptlPointer cycles, PseudoCycles;
  ntrees treecycle;

  cycles = getCycles ();
  PseudoCycles = getPseudoCycles ();
  if ((cycles == NILLPOINTER) && (PseudoCycles == NILLPOINTER))
    return pk_false;
  treecycle = pK_Graph2SignalTreeBuilder->mkTreeCycles (getDCG (), cycles, PseudoCycles);
  if (IS_NILTREE (treecycle))
    return pk_false;
  SignalUnparseCycles (pk_false, getLevelName (), treecycle);
  return pk_true;
}


/** 
    sub-graph to graph translation.
    @return the new graph obtained by the copy of the nodes and the clock hierarchy of the sub-graph (aSubGraph). 
    The sub-hierarchy is exported if exportClockSystem.
    The name of the extacted graph is fixed by taking into account the pragmas assigned to the subgraph (cluster, runon ,...)
    When the parameter breduce is "true" the new formula may be simplified in construction (elementary simpliciations are applied.
*/
ptDCG_basic
DCG_basic::_copySubgraph2Graph_ (ptINode aSubGraph, boolean bExportClockSubGraph, boolean exportClockSystem, boolean breduce)
{
  ptDCG g;
  ptSubGraphExtract o;
  ptident id;
  ptlpragmas lp;
  ptpragma aPrag;
  ptrepstr EnonceValue;
  ntrees tEnonce;

  //TODO reBuildCH is a GDC_basic function that find in CPG the CH elements (as done from LAST)
  // copySubgraph2Graph is a  DCG_basic function
  // that generates a DCG_basic graph that has .. but an empty CH and call reBuildCH

  if (!bExportClockSubGraph) exportClockSystem = pk_false;
  require_node_manager (NM);
  o = new SubGraphExtract ();
  g = o->SubGraph2Graph (getDCG (), aSubGraph, bExportClockSubGraph, exportClockSystem, breduce);
  g->addAttribute (EXTRACTED_FROM, (attributeValue) this);
  /* Setting the name */
  if (!NM->get_clusterName (aSubGraph, &id))
    {
      lp = g->getPragmas ();
      if (getPragmaKeyword (lp, PRAGMA_RUN_ON, &aPrag) || getPragmaKeyword (lp, PRAGMA_SELECTED_OTHERS_PARTITION, &aPrag))
	{
	  tEnonce = getPragmaStatement (aPrag);
	  EnonceValue = nstringof (nchild (1, tEnonce));
	  EnonceValue = ExtractString (EnonceValue);
	  id = createSuffIdent (getName (), EnonceValue);
	}
      else
	{
	  id = newIdent_stringSuffixed (getName (), cstExtract);
	  id = getNameManager ()->new_exclusiveIdent (id, 1);
	}
    }
  g->setName (id);
  delete o;
  if (exportClockSystem) return new DCG_basic (g, pk_false, pk_false);  // ioStep and Step are already set.
  return new DCG_basic (g, pk_false, pk_true);
}


_VIRTUAL_ void
DCG_basic::buildClockProcessInterfaceOneLevel ()
{
  // it should compute the interface of the node.
}




/**
   Assertion management.
   It rewrites the assertions of the model (See assertions2events(), assertions2events::rewAssertions()).

   \ref Booleans2Events "For more details..."
*/
boolean DCG_basic::transformAssertions2EventsOneLevel ()
{
  ptassertions2events arew;
  boolean b;

  arew = new assertions2events (getDCG ());
  b = arew->rewAssertions ();
  delete arew;
  return b;
}



/**
   Removing the renamings (i.e. x:=y) from the graph.
*/
boolean DCG_basic::removeSignalEqualitiesOneLevel ()
{
  typeMark
    m;
  boolean
    b,
    bapplied;

  constantEvaluation ();
  m = getMark ();
  b = pk_true;
  bapplied = pk_false;
  ptDCG
    vgdc = getDCG ();
  while (b)
    {
      b = pk_false;
      if (vgdc->markRenamming (m))
	{
	  b = pk_true;
	  vgdc->substituteMarkedVariables (m);
	}
      if (vgdc->markScalarConstantSignal (m))
	{
	  b = pk_true;
	  vgdc->substituteMarkedVariables (m);
	}
      if (b)
	bapplied = pk_true;
    }
  unmarkObjects (m);
  if (bapplied)
    {
      updatePragmas ();
      delaysRewriting ();
    }
  return bapplied;
}

/** 
    It performs the resolution of the clock systems using a triangularization technic. 
    A BDD-based data structure is used.
    Here, only the study of static properties is perfomed. They allow to characterized the set of states 
    in which the automaton associated to a program can evolve, independently of initial values, and 
    the set of the transitions between these states. 
    The result is a forest of clock trees.
    @param ilevelmax (in) depth of solving for boolean reductions. 

    Dynamic properties can be studied using Sigali tool (see DCG_basic::buildF3Abstraction).
*/
void
DCG_basic::ClockCalculusOneLevel (int ilevelmax)
{
  int date;
  boolean bBasic;

  require_expr_manager (EM);
  require_clock_manager (CKM);
  if (ilevelmax > 0)
    {
      date = pK_messenger->messageIdent (clockCalculus_announce, getName ());
      bBasic = (getFormatLevel () == DCG_BASIC);
      CKM->systemSolving_init ();
      do
	{
	  CKM->clockCalculus (ilevelmax);
	  EM->rewritingExprEH ();
	}
      while (CKM->updateHierarchy ());
      CKM->systemSolving_postlude ();
      getDCG ()->updateClocksGraph ();
      pK_messenger->messageEndOperVerbose (date);
      if (bBasic) setFormatLevel (DCG_POLY);
      addHierarchyPragmas ();
    }
}


/**
   @return the clock of the dependence prev -> elem.
   prev, elem may be nodes/clocks.
   For clocks, it returns the undefined the clock.
*/
ptinsth DCG_basic::getDependenceClock (pointer prev, pointer elem)
{
  ptObjectCommonInfos x;
  objectType objtype;
  ptINode n1, n2;
  ptdependence aDep;
  ptldependence li;

  require_clock_manager (CKM);
  require_node_manager (NM);

  x = (ptObjectCommonInfos) prev;
  objtype = x->getwhoami ();
  if (objtype != nodeObject)
    return (ptinsth) prev;
  n1 = (ptINode) prev;
  x = (ptObjectCommonInfos) elem;
  objtype = x->getwhoami ();
  if (objtype != nodeObject)
    return (ptinsth) elem;
  n2 = (ptINode) elem;
  li = NM->get_successors (n1);
  if (!li->exist_dependence (n1, n2, &aDep))
    return CKM->getUndefClock ();
  return aDep->getClockDependence ();
}

/**
   @return "true" currently.
   It will return "true" when the product of the clock of the cycle is the null clock (it is a "null clock cycle"), "false" otherwise (cycle)
*/
boolean DCG_basic::isFalseCycle (ptlPointer aCycle)
{
  ptinsth h, hcycle;
  pointer prev, velem;

  require_clock_manager (CKM);
  require_node_manager (NM);

  hcycle = CKM->getUndefClock ();
  ptlPointer_get_elemNext (&aCycle, &prev);
  while (ptlPointer_get_elemNext (&aCycle, &velem))
    {
      h = getDependenceClock (prev, velem);
      hcycle = CKM->makeInf (hcycle, h);
      prev = velem;
    }
  ClockCalculus ();
  hcycle = CKM->getRepresentativeFormula (hcycle);
  return CKM->satisfiesNullClock (hcycle);
}

void
DCG_basic::analyzeCyclesOneLevel ()
{
  ptlPointer nlcycles, lcycles, lfcycles;
  pointer aCycle;

  lcycles = getCycles ();
  if (lcycles != NILLPOINTER)
    {
      nlcycles = lcycles;
      lfcycles = NILLPOINTER;
      while (ptlPointer_get_elemNext (&lcycles, &aCycle))
	if (isFalseCycle ((ptlPointer) aCycle))
	  {
	    ptlPointer_remove_elem (&nlcycles, aCycle);
	    lfcycles = ptlPointer_add (lfcycles, aCycle);
	  }
      if (lfcycles != NILLPOINTER)
	{
	  pK_messenger->message (infoPseudoCycle);
	  setPseudoCycles (lfcycles);
	  setCycles (nlcycles);
	}
    }
}

LTypes DCG_basic::selectTypes (SelectedClassType clType)
{
  LTypes
    ltypes;
  ptlinstances
    lt;
  ptexpcal
    elem;
  TypeType
    TT,
    TT1;

  // TO BE REMOVED (LBesnard) !!!!
  require_expr_manager (EM);
  if (clType == FiacreTypes)	// what is it ?
    {
      ltypes.push_back ((TypeType) PureType);
      ltypes.push_back ((TypeType) BooleanType);
      ltypes.push_back ((TypeType) IntegerType);
      ltypes.push_back ((TypeType) ShortType);
      ltypes.push_back ((TypeType) LongType);
      // Selecting the enumerated types.
      if (getUserTypesOneLevel (&lt))
	while (ptlinstances_get_next (&lt, &elem))
	  {
	    TT = EM->getExpType (elem);
	    if (TypeTestNamedClass (TT))
	      TT1 = NamedTypeGetTerm (TT);
	    else
	      TT1 = TT;
	    if (TypeTestEnumerationClass (TT1))
	      ltypes.push_back (TT);
	  }
    }
  return ltypes;
}


/**
   \copydoc DCG::normalDefaults()
*/
boolean
DCG_basic::normalDefaultsOneLevel ()
{
  boolean b;
  b = getDCG ()->normalDefaults ();
  //  if (b) ClockCalculus();
  return b;
}


/**
   Adds pragmas specific to the hierarchy:
   - DefinedClockHierarchy is added if the program is endochronous and without constraints.
*/
void
DCG_basic::addHierarchyPragmas ()
{
  if (isEndochronous ())
    if (!hasUnsolvedClockConstraintsOneLevel ())
      addPragma (newPredefinedPragma (PRAGMA_DEFINED_CLOCK_HIERARCHY, NILOBJPRAGMA, NILTREE));
}

/**
   @return the clock, upper bound of h and the clocks of the nodes of the operand list (ln).
*/
ptinsth
DCG_basic::upperboundof (ptlINode ln, ptinsth h)
{
  ptlINode unusedNodes;
  ptINode n;
  ptlinsth unusedClocks;

  require_clock_manager (CKM);
  require_node_manager (NM);
  unusedNodes = getUnused (&unusedClocks);
  while (ptlINode_get_elem (&ln, &n))
    if (!ptlINode_isIn (unusedNodes, n))
      if (! NM->is_partialDefinition(n))
	h = CKM->makeSup (h, NM->get_nodeClock (n));
  return h;
}


/**
   It computes the IO step clock of the program.
*/
void
DCG_basic::computeIOstep ()
{
  ptinsth h;
  ptlINode ln;

  require_clock_manager (CKM);
  require_node_manager (NM);
  h = CKM->getUndefClock ();
  if (NM->getInputs (&ln))
    h = upperboundof (ln, h);
  if (NM->getOutputs (&ln))
    h = upperboundof (ln, h);
  if (NM->getExternalCalls (&ln))
    {
      h = upperboundof (ln, h);
      ptlINode_free (ln);
    }
  setIOstep (h);
}


/**
   It computes the step clock of the program (the upper bound of all the computation of the program).
   Currently it is not computed.
*/
void
DCG_basic::computeStep ()
{
  ptlINode ln;
  require_clock_manager (CKM);
  require_node_manager (NM);
  if (NM->getAllNodes (&ln))
    setStep (upperboundof (ln, CKM->getUndefClock ()));
}


/**
   It computes some control clocks: ioStep, Step.
   The clockcalculus functionality must be called after this call.
*/
void
DCG_basic::computeControlClocks ()
{
  computeIOstep ();
  computeStep ();
}


/**
   -- Testing--
  It takes into account the utility clocks (if computed) for solving the system.
  @return "true if the system has been mofified, "false" otherwise.
*/
boolean DCG_basic::setUtilityClocksAsClockOneLevel ()
{
  if (isGraphOfExternalProcess ())
    return pk_false;
  if (!getDCG ()->setUtilityClocksAsClock ())
    return pk_false;
  return pk_true;
}

/**
   It computes the assignmnt clocks of the node of the graph.
*/
boolean
DCG_basic::computeAssignmentClockOneLevel ()
{
  // pK_AssignmentClock is a predefined object of the virtual machine.
  pK_messenger->messageIdent (messAssignmentClock, getName ());
  pK_AssignmentClock->init (getDCG ());
  pK_AssignmentClock->apply ();
  return pk_true;
}


/**
   It computes the assignmnt clocks of the node of the graph.
*/
boolean
DCG_basic::computeClockWithoutDefaultConstantValueClockOneLevel ()
{
  // pK_AssignmentClock is a predefined object of the virtual machine.
  pK_messenger->messageIdent (messAssignmentClock, getName ());
  pK_AssignmentClock->init (getDCG ());
  pK_AssignmentClock->computeClockWithoutDefaultConstantValue ();
  return pk_true;
}

/**
   It computes the utility clocks of the nodes of the graph.
*/
boolean
DCG_basic::computeProductiveClockOneLevel ()
{
  pK_messenger->messageIdent (messUtilityClock, getName ());
  pK_UtilityClock->init (getDCG ());
  pK_UtilityClock->apply ();
  return pk_true;
}




/**
   Each (partial) definition x:=E becomes x::=E when hu(x) | x::=defaultvalue x$1 where hu(x) denotes the utility clock of x.
   The graph is not modified if the computeProductiveClock() method has not been applied.
*/
boolean
DCG_basic::rewUsingProductiveClockOneLevel ()
{
  boolean b;
  b = getDCG ()->rewUsingProductiveClock ();
  return b;
}


void DCG_basic::changePragmas(ptexpcal vprevInput, ptexpcal vNewInput, int PramaId)
{ 
  ptpragma v;
  ptlpragmas lpragmas;
  ptrepstr name;
  ptObjPragma obj;
  ptlinstances l;
  ntrees PragmaStatement;

  lpragmas = getPragmas();
  while (ptlpragmas_get_next(&lpragmas, &v))
    {
      if (isPragmaKeyword(v, PramaId))
	{
	  getPragma(v, &name, &obj, &PragmaStatement);
	  l = (ptlinstances) obj;
	  if (ptlinstances_isin_remove_exp(&l, vprevInput)) l =  ptlinstances_insert(l, vNewInput);
	  setPragmaObjects(v, (ptObjPragma) l);
	}
    }
}


/** 
    Getting a graph (g)
    - the current when it is a process-graph 
    - or the process-graph of a module when the module is reduced to one graph.

    @return "true" if such a graph is selected, "false" otherwise.
*/
boolean DCG_basic::getOnlyIfOneGraph(ptDCG * g)
{
  ptlDCG lg;
  pointer elem;

  if (!getModuleGraphs(&lg))
    {
      *g = getDCG();
      return pk_true;
    }
  // module: only if one graph.
  if (!ptlPointer_get_elemNext(&lg, &elem))
    return pk_false;
  if (ptlPointer_get_elemNext(&lg, &elem))
    return pk_false;
  *g = (ptDCG) elem;
  return pk_true;
}

boolean DCG_basic::computeIOTransitiveClosureOneLevel()
{
  if (isGraphOfExternalProcess()) return pk_true;
  pK_messenger->messageIdent(infoTransitiveClosure, getName());
  buildIOTransitiveClosure();
  pK_messenger->setcputime();
  return pK_messenger->getErrorCode();
}
