/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    It transforms a multi-root hierarchy of each process to a single root hierarchy (parametrization of the clock system).
    The processes become endochronous. The interface of the processes may be modified (some input booleans are added for the parametrization), in this 
    case the returned boolean (modified) is set to the value "true", otherwise it is set to the value "false".
    
    @return the value "true" if the operation is performed, "false" otherwise.
*/
  _STATIC_ boolean DCG_endo::parameterizeMasterClocks (ptDCG_poly gpoly, boolean * modified, ptDCG_endo * gp)
{
  if (!gpoly->DCG_poly::parameterizeMasterClocks (pk_true, modified)) return pk_false;
  (*gp) = new DCG_endo (gpoly);
  if (!(*gp)->setArrayClocks ()) return pk_false; 
  (*gp)->buildClockProcessInterface();
  return pk_true;
}

DCG_endo::DCG_endo (ptDCG_poly gpoly):DCG_poly (gpoly)
{
  setFormatLevel (DCG_ENDO);
}

DCG_endo::DCG_endo ()
{
}


/** 
   This  method can be applied to a process.
   @return true if the set of IO signals is event free and the clock of every IO  signal is either the root or a function of IO boolean signals
 */
boolean
DCG_endo::is_makeBool_safe ()
{
  ptlinsth ClocksBuiltOnIOClock, ClocksNotBuiltOnIOClock, lhtodeclare;
  ptlinstances lsig, lins, louts, appendSigs;
  ptinsth vh;
  ptexpcal vsig, le, ls, par;
  ptlINode ln;
  ptDCG vg;
  ptINode aNode;
  ptident id;
  TypeType vDeclType;

  require_managers (EM, NM, CKM);
  lins = getInputs ();
  louts = getOutputs ();
  lsig = lins;
  // set of input signals is event free
  while (ptlinstances_get_next (&lsig, &vsig))
    if (EM->IsEvent (vsig))
      {
	/* an input can not moved from boolean to event ...
	  vDeclType = EM-> getDeclaredType(vsig); if ( vDeclType == BooleanType) EM-> setType(vsig, BooleanType); else */
	return pk_false;
      }
  // Set of output signals is event free, but the event signals declared as booleans are accepted.
  // Recall: boolean always true are set to event type by type systhesis...
  lsig = louts;
  while (ptlinstances_get_next (&lsig, &vsig))
    if (EM->IsEvent (vsig))
      { 
	vDeclType = EM-> getDeclaredType(vsig);
	if ( vDeclType == BooleanType) EM-> setType(vsig, BooleanType); else return pk_false;
      }
  // Idem for the external call.
  if (NM->getProcCalls (&ln))
    while (ptlINode_get_elem (&ln, &aNode))
      {
	NM->get_procCall_arg (aNode, &id, &par, &le, &ls);
	while (EM->getNaryArgs (&ls, &vsig))
	  if (EM->IsEvent (vsig))
	    return pk_false;
      }
  // Idem for the external node call.
  if (NM->getNodeCalls (&ln))
    while (ptlINode_get_elem (&ln, &aNode))
      {
	NM->get_nodeCall_arg (aNode, &vg, &par, &le, &ls);
	if (vg->isGraphOfExternalProcess ())
	  while (EM->getNaryArgs (&ls, &vsig))
	    if (EM->IsEvent (vsig))
	      return pk_false;
      }
  // set of IO signals is event free, verifying the clocks: no outputs are added for the abstraction.
  /* 
     vh = buildIOClockSubHierarchyLoc(pk_false, &lhtodeclare, &appendSigs);
     if (appendSigs != NILLINSTANCES) return pk_false;
     ClocksBuiltOnIOClock = IOSplitExportedClocks(lhtodeclare, &ClocksNotBuiltOnIOClock);
     return (ClocksNotBuiltOnIOClock == NILLINSTH);
   */
  return pk_true;
}

/** 
   This  method can be applied to a graph of a module (set of processes) or a process.
   \copydoc addExceptionsForConstraintsOneLevel
   @return true if the operation is done, false when the graph is a DC+ one.
 */
boolean
DCG_endo::addExceptionsForConstraints ()
{
  // printf("DEBUG entering DCG_endo::addExceptionsForConstraints \n"); fflush(stdout);
  for_all_graphs_do (addExceptionsForConstraintsOneLevel ());
  // printf("DEBUG  exiting DCG_endo::addExceptionsForConstraints \n"); fflush(stdout);
  return pk_true;
}

/**
   Specialisation.
   Nothing to do for a DCG_endo.
*/
void
DCG_endo::parameterizeMasterClocks ()
{
}


/** 
    This method can be applied to a graph of a module (set of processes) or a process.
    It performs the interformat DCGPoly (that contains event objects, forest of trees of clocks) to DCGBool 
    (without event objects, one tree of boolean clocks) translation.

    @return the value "true" if the operation is performed, "false" otherwise.
    The setTick() method, that completes the graph to an endochronous one, is assumed to be applied before this call.
*/
boolean
DCG_endo::transformToBooleanControl (boolean mkSubForNosub)
{
  ptDCG g;
  pointer elem;
  boolean bModule, bapply;
  ptlDCG lg, lg1;
  typeMark vmyMark;
  ptrepstr codeOp;

  if (isVerboseMode ())
    {
      codeOp = string2repstr (csttransformToBooleanControl);
      pK_messenger->messagestr (mbeginOperName, codeOp);
    }
  bModule = getModuleGraphs (&lg);
  if (!bModule)
    ptlPointer_insert_elem (&lg, (pointer) getDCG ());

  vmyMark = getMark ();
  bapply = pk_true;
  while (bapply && ptlPointer_get_elemNext (&lg, &elem))
    {
      g = (ptDCG) elem;
      if (!g->satisfiesMarked (vmyMark))
	{
	  g->addMark (vmyMark);
	  if (isVerboseMode ())
	    pK_messenger->messageIdent (mbeginOper, g->getName ());
	  init (g);
	  gw->transformToBooleanControlOneLevel (mkSubForNosub);
	  if (g->getReferencedGraphs (&lg1))
	    lg = ptlPointer_fusion (lg, lg1);
	  if (isVerboseMode ())
	    pK_messenger->message (mendOper);
	}
      ptlPointer_remove_elem (&lg, elem);
    }
  resetMark (vmyMark);

  if (bModule && bapply)
    setFormatLevel (DCG_BOOL);
  if (isVerboseMode ())
    pK_messenger->messagestr (mendOperName, codeOp);
  return bapply;
}

/** 
    It performs the user-partitionning by analyzing the pragma "RunOn i " where i denotes a string. 
    Nodes annotated by the same string value are in the same class or sub-graph. 
    @param ln (out) is a list of "sub-graph" nodes produced by the partitionning.

    @return "true" if the operation succeeds (all objects are localized), "false" otherwise.

    \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module (set of processes).
 
*/
boolean
DCG_endo::partitionningRunOn (ptlINode * ln)
{
  ptRunOnMark aRunOnMark;
  ptrepstr codeOp;
  boolean b;

  // Take care of shared signal (no reference to ?)
  require_node_manager (NM);
  codeOp = string2repstr (cstRunOnPartition);
  pK_messenger->messagestr (messOperName, codeOp);
  makeNewInputs();
  NodesInterfaceCalculus ();
  aRunOnMark = new RunOnMark (getDCG ());
  b = aRunOnMark->partitionningNodes (ln);
  if (!b)
    pK_messenger->message (infoRunOnNotFound);
  delete aRunOnMark;
  buildClockProcessInterface ();
  if (isVerboseMode ())
    pK_messenger->messagestr (mendOperName, codeOp);
  return b;
}
