

/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_

/**
   It returns in lnodes the ?lnodes + the nodes of the hierarchy with h as root.
   It returns in lclocks the ?lclocks + the clocks of the hierarchy with vclk as root.
*/
void
DCG::getHierarchyNodesClocks (ptinsth vclk, ptlINode * lnodes, ptlinsth * lclocks)
{
  ptlINode ln;
  ptINode node;
  ptlinsth lh;
  ptinsth h;

  require_managers (EM, NM, CKM);
  ln = CKM->getNodesAssignedTo (vclk);
  // the I/O nodes are filtered.
  while (ptlINode_get_elem (&ln, &node))
    if (!NM->is_ioNode (node))
      ptlINode_insert_elem (lnodes, node);

  lh = CKM->getChildren (vclk);
  (*lclocks) = ptlinsth_sorted_fusion (*lclocks, lh);
  while (ptlinsth_get_next (&lh, &h))
    getHierarchyNodesClocks (h, lnodes, lclocks);
}


/** Static cycle detection.

 The static path calculus does not take into account clocks of the dependencies. 
 When a cycle is detected a Signal program
 foo_CYC.SIG is printed. foo_CYC.SIG contains the cycle desciption(s)
 @return "true" if the current graph contains some cycle, "false" otherwise.
 @param berrmess (in) when it is "true", an error is sent to the user,otherwise a warning.
 \ref Closure "For more details..."
*/
boolean
DCG::hasStaticCycleOneLevel (boolean berrmess, boolean bKeepCycles)
{
  boolean b;
  ntrees treecycle;
  ptident id;
  ptCircuitCalculus mkparc;

  cycles = NILLPOINTER;
  PseudoCycles = NILLPOINTER;
  if (isGraphOfExternalProcess ())
    return pk_false;
  id = getName ();
  pK_messenger->messageIdent (infoopt, id);

  mkparc = new CircuitCalculus (this);
  b = mkparc->circuit_detection ();
  pK_messenger->setcputime ();
  if (b)
    {
      if (berrmess)
	pK_messenger->errorTreeIdent (errcycledep, NILTREE, id);
      if (bKeepCycles)
	{
	  cycles = mkparc->getCycles ();
	  if (cycles == NILLPOINTER)
	    pK_messenger->error (errcompilerbug);
	}
    }
  delete mkparc;

  return b;
}



/** 
    Delay Rewriting:
    - x := x $1 init v0, x is the constant signal v0.
    - x := E $1 init v0, E is a valued signal equal to the value v0, x is the constant signal v0.
    - x := y $I init v0 (I a signal, assumed bounded), I is constant to n, is rewritten in x := y $ n init v0 or x := y (n=0)
    @return "true" if the graph is modified, "false" otherwise.
*/
boolean
DCG::delaysRewriting1 ()
{
  ptlINode lNodes;
  ptINode node;
  ptexpcal e, vopg, vopd, del, win, vinit, x;
  compilerInteger lb1, lb2, ub1, ub2;
  compilerInteger nd, nw;
  ptinsth h;
  boolean toTest, b, res;

  require_managers (EM, NM, CKM);
  res = pk_false;
  if (NM->getDelays (&lNodes))
    while (ptlINode_get_elem (&lNodes, &node))
      {
	NM->get_delay_arg (node, &vopg, &vopd, &del, &win, &vinit);
	b = pk_false;
	toTest = pk_false;
	if (vopg == vopd)
	  {
	    nd = -1;
	    if (!EM->isNILEXPCAL (del))
	      EM->getIntegerValue (del, &nd);
	    nw = -1;
	    if (!EM->isNILEXPCAL (win))
	      EM->getIntegerValue (win, &nw);
	    if ((nd == 1) && (nw <= 0))
	      {
		e = EM->getInitialValue (vopg);
		b = pk_true;
	      }
	  }
	else if (EM->getBounds (vopd, &lb1, &ub1) && EM->getBounds (vinit, &lb2, &ub2))
	  {
	    lb1 = (lb1 < lb2 ? lb1 : lb2);
	    ub1 = (ub1 > ub2 ? ub1 : ub2);
	    if (lb1 == ub1)
	      {
		b = pk_true;
		e = EM->mkIntegerConstant (lb1);
	      }
	    EM->setBounds (vopg, lb1, ub1);
	  }
	else
	  toTest = !EM->isNILEXPCAL (del);

	if (b)
	  {
	    h = NM->get_nodeClock (node);
	    NM->resetNodeForModification (node);
	    NM->modifiedDefinition (vopg, e, CKM->getInitClock (), h);
	    EM->setInitialValue (vopg, NILEXPCAL);
	    ptlINode_insert_elem (&lNodesAtInitClock, node);
	    res = pk_true;
	  }

	if (toTest)
	  {
	    if (!EM->IsConstant (del))
	      {
		if (EM->getBounds (del, &lb1, &ub1))
		  {		// bounds are assumed > 0
		    if (lb1 == ub1)
		      {
			if (lb1 == 0)
			  x = vopd;
			else
			  {
			  BackAnnotationType t = NM->getBackAnnotation (node);
			    x = EM->mkBinaryExp (opinit, vopd, vinit, t);
			    x = EM->mkTernaryExp (opret, x, EM->mkIntegerConstant (lb1), win, t);
			  };
			h = EM->getClock (vopd);
			NM->modifiedDefinition (vopg, x, h, h);
			res = pk_true;
		      }
		  }
	      }
	  }
      }
  return res;
}

/** 
    It applies the rule: for x := x $N init V, y := y $N init V, and x^=y then x=y

   @return "true" if the graph is modified, "false" otherwise.
*/
boolean
DCG::delayUnification ()
{
  ptlINode ln, ln1;
  ptINode node, node1;
  ptexpcal vopg, vopd, del, win, vinit;
  ptexpcal vopg1, vopd1, del1, win1, vinit1;
  ptinsth h, h1;
  boolean bfirst, res;

  require_managers (EM, NM, CKM);
  res = pk_false;
  if (NM->getDelays (&ln))
    while (ptlINode_get_elem (&ln, &node))
      {
	NM->get_delay_arg (node, &vopg, &vopd, &del, &win, &vinit);
	if (EM->isEquivalent (vopg, vopd, pk_false))	// x:= x $ 
	  {
	    h = EM->getClock (vopg);
	    ln1 = ln;
	    bfirst = pk_true;
	    while (ptlINode_get_elem (&ln1, &node1))
	      {
		NM->get_delay_arg (node1, &vopg1, &vopd1, &del1, &win1, &vinit1);
		h1 = EM->getClock (vopg1);
		if (h1 == h)	// synchronous delays.
		  if (EM->isEquivalent (vopg1, vopd1, pk_false))	// y:= y $ 
		    if ((EM->isEquivalent (del, del1, pk_true)) && EM->isEquivalent (win, win1, pk_true))
		      if (EM->isEquivalent (vinit, vinit1, pk_true))	// same initial value.
			{
			  NM->resetNodeForModification (node1);
			  NM->modifiedDefinition (vopg1, vopg, h, h);
			  if (bfirst)
			    ln = ln1;
			  res = pk_true;
			}
		bfirst = pk_false;
	      }
	  }
      }
  return res;
}

/**
   It marks the operand clock substituable if it is an undersampling, and not an output event.
*/
void
DCG::setSubstitutable (ptinsth h)
{
  boolean found;
  ptlinstances lsig;
  ptexpcal aSig;

  require_clock_manager (CKM);
  require_expr_manager (EM);

  if (CKM->satisfiesSub (h))
    if (CKM->getConditionSub (h) != NILEXPCAL)	// affine clock.
      {
	// if it exists an output of event type, the clock is not substitutable.
	lsig = CKM->getSynchronousSignals (h);
	found = pk_false;
	while (!found && ptlinstances_get_next (&lsig, &aSig))
	  if (EM->IsSignal (aSig))	// filtering the identifier of struct/bundle
	    found = EM->IsEvent (aSig) && EM->IsOutputFlow (aSig);

	if (!found)
	  CKM->setSubstitutable (h);
      }
}


ptlinsth DCG::getIODependencesClocks (ptdependence aDep, ptlinsth lh)
{
  ptINode vFrom, vTo;
  ptexpcal sig;
  ptinsth h;
  boolean b;

  require_managers (EM, NM, CKM);
  h = CKM->getRepresentativeFormula (aDep->getDependence (&vFrom, &vTo));
  if (CKM->satisfiesPredefinedClock (h))
    return lh;
  if (!NM->getDependenceNodeLabel (vFrom, &sig))
    return lh;
  if (!(EM->IsInputFlow (sig) || EM->IsOutputFlow (sig)))
    return lh;
  if (!NM->getDependenceNodeLabel (vTo, &sig))
    return lh;
  if (!(EM->IsInputFlow (sig) || EM->IsOutputFlow (sig)))
    return lh;

  return ptlinsth_sorted_insert_elem_bool (lh, h, &b);
}


void
DCG::nodeGroupingByClocks (ptlINode ln, ptlINode * lsubGraphs)
{
  ptINode n, aNode;
  ptinsth vh, vclk;
  boolean found;
  ptlINode ln1, lnc;
  ptexpcal vlabel;
  ptlinsth lhc;
  require_managers (EM, NM, CKM);
  while (ptlINode_get_elem (&ln, &aNode))
    {
      if (NM->is_setOfPartialDefinitions (aNode))
	nodeGroupingByClocks (NM->get_setOfPartialDefinitions_arg (aNode), lsubGraphs);
      else
	{
	  vclk = NM->get_nodeClock (aNode);
	  // Searching if a group as vclk clock is already in lsubGraphs.
	  ln1 = *lsubGraphs;
	  found = pk_false;
	  while (!found && ptlINode_get_elem (&ln1, &n))
	    if (NM->get_subGraphDefinitions (n, &lnc, &lhc))
	      if (ptlinsth_get_next (&lhc, &vh))
		found = CKM->eqClock (vh, vclk);

	  if (!found)
	    {
	      vlabel = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), vclk , NILBACKANNOTATION, pk_true);
	      n = NM->mk_node_subgraph (vlabel,  NILBACKANNOTATION);
	      ptlINode_insert_elem (lsubGraphs, n);
	      NM->subGraphInsertClock (n, vclk);
	    }
	  NM->subGraphInsertNode (n, aNode);
	}
    }
}

/**
   It merges the equations defined by the same function call, in the operand list (ln) : (| x:=f(E) | y := f(E) |) and f a function is rewritten (| x:=f(E) | y := x |).
   @return "true" if the graph is modified, "false" otherwise.
*/
boolean DCG::functionUnification (ptlINode ln)
{
  ptlINode ln1;
  ptINode node1, n;
  boolean bfirst, b;
  ptident id, id1;
  ptexpcal vout1, vout, params, OrderedInputs, OrderedOutputs, params1, OrderedInputs1, OrderedOutputs1, OrderedOutputsRef;
  ptinsth h; 
  ptDCG vg1, vg;

  require_node_manager (NM);
  require_expr_manager (EM);

  b = pk_false;
  while (ptlINode_get_elem (&ln, &n))
    if (NM->get_safetyProperty (n) == opsafe)
      {
	if (NM->isNodeCall (n))
	  {
	    NM->get_nodeCall_arg (n, &vg, &params, &OrderedInputs, &OrderedOutputs);
	    id = vg->getName ();
	  }
	else
	  NM->get_procCall_arg (n, &id, &params, &OrderedInputs, &OrderedOutputs);

	ln1 = ln;
	bfirst = pk_true;
	while (ptlINode_get_elem (&ln1, &node1))
	  {
	    if (NM->get_safetyProperty (node1) == opsafe)
	      {
		if (NM->isNodeCall (node1))
		  {
		    NM->get_nodeCall_arg (node1, &vg1, &params1, &OrderedInputs1, &OrderedOutputs1);
		    id1 = vg1->getName ();
		  }
		else
		  NM->get_procCall_arg (node1, &id1, &params1, &OrderedInputs1, &OrderedOutputs1);

		if (eq_idents (id1, id) && EM->isEquivalent (params1, params, pk_true) && EM->isEquivalent (OrderedInputs1, OrderedInputs, pk_true))
		  {
		    OrderedOutputsRef = OrderedOutputs;
		    while (EM->getNaryArgs (&OrderedOutputs1, &vout1))
		      {
			EM->getNaryArgs (&OrderedOutputsRef, &vout);
			NM->resetNodeForModification (node1);
			h = EM->getClock (vout1);
			NM->modifiedDefinition (vout1, vout, h, h);
			b = pk_true;
			if (bfirst)
			  ln = ln1;
		      }
		  }
	      }
	    bfirst = pk_false;
	  }
      }
  return b;
}


/**
   @return true is the graph is an event clocked level, false otherwise.
*/
boolean DCG::isEventClockedFormat ()
{
  GraphLevelType
    tg = getFormatLevel ();
  return ((tg == DCG_BASIC) || (tg == DCG_POLY) || (tg == DCG_ENDO));
}


void
DCG::interfaceReset (ptlinsth lh)
{
  ptinsth vh;
  require_clock_manager (CKM);
  while (ptlinsth_get_next (&lh, &vh))
    {
      vh = CKM->getRepresentativeFormula (vh);
      CKM->resetSubGraphInterface (vh);	// To be changed: use the list of the nodes of the graph.
      interface_free (CKM->getInterfaceSubHierarchy (vh));
    }
}



