
/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return a DCG_poly object with all clocks solved for a DCG_basic graph (gbasic).
*/
    _STATIC_ ptDCG_poly DCG_poly::DCG_basic2poly(ptDCG_basic gbasic)
{
  gbasic->ClockCalculus();
  return new DCG_poly(gbasic);
}

/**
   @return a DCG_poly object with all clocks solved for a DCG_basic graph (gbasic).
   @param vd (in) depth of resolution for boolean reduction.
*/
_STATIC_ ptDCG_poly DCG_poly::DCG_basic2poly(ptDCG_basic gbasic, int vd)
{
  gbasic->ClockCalculus(vd);
  return new DCG_poly(gbasic);
}

/**
   @return a DCG_poly object with all clocks solved for a DCG_basic graph (gbasic).
   @param vd (in) depth of resolution for boolean reduction.
*/
_STATIC_ ptDCG_poly DCG_poly::new_prePoly(ptDCG aDCG)
{
  ptDCG_poly g = new DCG_poly();
  g->setDCG(aDCG);
  return g;
}

 DCG_poly::DCG_poly():DCG_basic()
{
}

 DCG_poly::DCG_poly(ptDCG_basic gbasic):DCG_basic(gbasic)
{
  gbasic->setFormatLevel(DCG_POLY);
  // gbasic->getDCG()->markNodesFreeConditions();
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    It realizes elementary optimizations by deleting objects of the graph which are 
    not used to produce (resp. consume) outputs (resp. inputs) of the program. 
*/
boolean DCG_poly::nonProductiveSignalElimination()
{
  DCG_basic::nonProductiveSignalElimination();
  buildClockProcessInterface();
  return pk_true;
}


/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    It realizes elementary optimizations by deleting objects of the graph which are 
    not used to produce (resp. consume) outputs (resp. inputs) of the program. 
*/
boolean DCG_poly::nonProductiveSignalElimination(boolean withClockDepedence)
{
  DCG_basic::nonProductiveSignalElimination(withClockDepedence);
  buildClockProcessInterface();
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    For each graph of the application,

    \copydoc buildClockProcessInterfaceOneLevel
*/
boolean DCG_poly::buildClockProcessInterface()
{
  for_all_graphs_do(buildClockProcessInterfaceOneLevel());
  setExternalReferences();
  return pk_true;
}

/**
   This method is used for completing the clock hierarchy to have a endochronous program.
   It adds for all root clock of the system a node call (such that C_a :=_PRESENT_a()) with clk as clock.
   It adds induced constraints (ie a ^ when C_a).
   After each adding, the clock calculus is applied that may modify the set of roots.
   \see addBooleanInput, addPresentNode
*/
void DCG_poly::iterateNodeInputEndochronization(ptinsth clk, boolean bAddinputs)
{
  ptlinsth lh;
  ptinsth vh;
  boolean bcontinue;
  ptDCG vdcg;

  //LPT should be replaced by building of a container CP that calls P
  //LPT should be completed by function calls instead of signals

  require_clock_manager(CKM);
  // The variables are visited first.
  bcontinue = pk_true;
  vdcg = getDCG();
  while (bcontinue)
    {
      bcontinue = pk_false;
      lh = CKM->getlhroots();
      while ((!bcontinue) && ptlinsth_get_next(&lh, &vh))
	if (vh == CKM->getRepresentativeFormula(vh))
	  if ((vh != clk) && CKM->satisfiesVari(vh))
	    {
	      // if (bAddinputs)
	      addBooleanInput(vh, clk, vdcg->ChoiceNameBooleanClock(vh));
	      // else addPresentNode(vh, clk, NILIDENT, vdcg->ChoiceNameBooleanClock(vh));
	      ClockCalculusOneLevel(2);
	      bcontinue = pk_true;
	    }
    }
  // then the other roots of the hierarchy.
  bcontinue = pk_true;
  while (bcontinue)
    {
      bcontinue = pk_false;
      lh = CKM->getlhroots();
      while ((!bcontinue) && ptlinsth_get_next(&lh, &vh))
	if (vh == CKM->getRepresentativeFormula(vh))
	  if (vh != clk)
	    if (!CKM->satisfiesSub(vh))
	      {
		// if (bAddinputs)
		addBooleanInput(vh, clk, vdcg->ChoiceNameBooleanClock(vh));
		// else addPresentNode(vh, clk, NILIDENT, vdcg->ChoiceNameBooleanClock(vh));
		ClockCalculusOneLevel(2);
		bcontinue = pk_true;
	      }
    }
}

/** 
    It builds an abstract trees representation of the abstraction of the program:
    - I/O data dependences (assumed available, the I/O transitive closure has been computed)
    - I/O clock relations.

    This abstraction is useful for separated compiling
    @return the "black Box" or the "greyBox" abstraction representation of the current graph.
    \see{phylumClustering}
    \warning (\em restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
ntrees DCG_poly::produceInterfaceAbstraction()
{
  return pK_Graph2SignalTreeBuilder->graphAbstraction2SignalTree(getDCG());
}

/** 
    It builds an external representation of the exclusive clocks of the graph.
    It assumes that the functionality computeClockExclusions() has been already applied.
    @return the Signal abstract tree external representation of the exclusive clocks of the graph. 
    This tree can be printed using the \see{decsig} function.
*/
ntrees DCG_poly::produceExclusiveClocks()
{
  return pK_Graph2SignalTreeBuilder->exclusiveClocks(getDCG());
}

/** 
    It performs the separating of booleans calculus and other calculus.

    @param booleans (out) denotes the sub-graph of booleans calculus.
    @param others (out) denotes the sub-graph of non-booleans calculus.

    \warning (\em restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_poly::separateBooleansOthers(ptINode * booleans, ptINode * others)
{
  ptmkBoolOther mkparc;
  ptlINode lSubGraphs;
  ptrepstr codeOp;

  codeOp = string2repstr(cstBoolOtherPartition);
  pK_messenger->messagestr(messOperName, codeOp);

  undersamplingOnSignal(pk_false, pk_false);
  mkparc = new mkBoolOther(getDCG());
  lSubGraphs = mkparc->apply();
  finalyzeBinPart(lSubGraphs, booleans, others);
  return pk_true;
}

/** 
    It performs the separating of states nodes  and other nodes.
    @param states (out) denotes the sub-graph of state variables.
    @param others (out) denotes the sub-graph of static variables.

    \warning (\em restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_poly::separateStatesOthers(ptINode * states, ptINode * others)
{
  ptmkStatesOther mkparc;
  ptlINode lSubGraphs;
  ptrepstr codeOp;

  codeOp = string2repstr(cstStateOtherPartition);
  pK_messenger->messagestr(messOperName, codeOp);
  mkparc = new mkStatesOther(getDCG());
  lSubGraphs = mkparc->apply();
  finalyzeBinPart(lSubGraphs, states, others);
  return pk_true;
}

/** 
    It performs the separating of boolean/integer nodes  and other nodes.
    @param BooleansIntegers (out) denotes the sub-graph of variables with boolean or int type.
    @param others (out) denotes the sub-graph of non-state variables.

    \warning (\em restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_poly::separateBooleansIntegersOthers(ptINode * BooleansIntegers, ptINode * others)
{
  ptmkBooleanBoundedIntegerOther mkparc;
  ptlINode lSubGraphs;

  undersamplingOnSignal(pk_false, pk_false);
  mkparc = new mkBooleanBoundedIntegerOther(getDCG());
  lSubGraphs = mkparc->apply();
  finalyzeBinPart(lSubGraphs, BooleansIntegers, others);
  return pk_true;
}

/** 
    It performs the separating of states nodes  and other nodes.
    @param finiteTypesSubGraph (out) denotes the sub-graph of variables with a finite type (booleans, integers and enumerated types).
    @param others (out) denotes the sub-graph of non-finite type variables.

    \warning (\em restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_poly::separateFiniteTypesOthers(ptINode * finiteTypesSubGraph, ptINode * others)
{
  ptmkFiniteTypesOther mkparc;
  ptlINode lSubGraphs;

  undersamplingOnSignal(pk_false, pk_false);
  mkparc = new mkFiniteTypesOther(getDCG());
  lSubGraphs = mkparc->apply();
  finalyzeBinPart(lSubGraphs, finiteTypesSubGraph, others);
  return pk_true;
}

/** 
    It performs the separating of states nodes  and other nodes.
    @param selectedTypesSubGraph (out) denotes the sub-graph of variables with a finite type (booleans, integers and enumerated types).
    @param others (out) denotes the sub-graph of non-finite type variables.

    \warning (\em restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean DCG_poly::separateSelectedTypesOthers(SelectedClassType clType, ptINode * selectedTypesSubGraph, ptINode * others)
{
  ptmkSelectedTypesOther mkparc;
  ptlINode lSubGraphs;

  undersamplingOnSignal(pk_false, pk_false);
  mkparc = new mkSelectedTypesOther(getDCG(), selectTypes(clType));
  lSubGraphs = mkparc->apply();
  finalyzeBinPart(lSubGraphs, selectedTypesSubGraph, others);
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    For each graph of the application, this method builds a set of "delay" trees.
    All the signals build on the same signal (X) using the delay/window operators
    can be represented as a tree (X as root) when initializations are compatible. 
    This delay-tree can be used for:

    - unifying signals ; for example, # (x $ 1 init 0) $1 init 0 # is equivalent to # x $ 2 init [0,0] #;
    - multiplexing of variables by code generators.

    @return "true" if the operation succeeds, "false" otherwise (cycle).

    \see{memoryOptimizations}
*/
boolean DCG_poly::delayTreeBuild()
{
  for_all_graphs_do_br(delayTreeBuildOneLevel(), b);
  return b;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc setIntrinsicClocksOneLevel

    After, the \ref buildClockProcessInterface "Interface calculus" is also applied.
*/
boolean DCG_poly::setIntrinsicClocks()
{
  for_all_graphs_do(setIntrinsicClocksOneLevel());
  buildClockProcessInterface();
  return pk_true;
}

/** 
    This method can be applied to a graph of a process or a module.
    \copydoc computeClockExclusionsOneLevel
*/
void DCG_poly::computeClockExclusions(int ecLevel)
{
  for_all_graphs_do(computeClockExclusionsOneLevel(ecLevel));
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    @param bstrong : at pk_true, the clock of the dependences are taken into account.
    \copydoc buildIOClockSubHierarchyOneLevel
*/
boolean DCG_poly::buildIOClockSubHierarchy(boolean bstrong)
{
  for_all_graphs_do(buildIOClockSubHierarchyOneLevel(bstrong));
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc setDelayMemoryOverloadingOneLevel
*/
boolean DCG_poly::setDelayMemoryOverloading(boolean bclu)
{
  for_all_graphs_do(setDelayMemoryOverloadingOneLevel(bclu));
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc memoryOptimizationsOneLevel
*/
boolean DCG_poly::memoryOptimizations(boolean mutliplexDelays)
{
  for_all_graphs_do(memoryOptimizationsOneLevel(mutliplexDelays));
  return pk_true;
}


/** 
    This method can be applied to a graph of a module (set of processes) or a process.
    \copydoc computeIOTransitiveClosureOneLevel
*/
boolean DCG_poly::computeIOTransitiveClosure()
{
  setClockDependences();
  for_all_graphs_do_br(computeIOTransitiveClosureOneLevel(), breturn);
  return breturn;
}

/** 
    @return a copy of the current graph (defined as a morphism).
    \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
ptDCG_poly DCG_poly::copy()
{
  ptDCG_poly gCopy;
  ptDCG_basic gBasicCopy;

  gBasicCopy = DCG_basic::copy();
  gCopy = DCG_poly::DCG_basic2poly(gBasicCopy);
  gCopy->setIntrinsicClocks();
  return gCopy;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    \copydoc expressionExpandOneLevel
*/
boolean DCG_poly::expressionExpand()
{
  groupPartialDefinitions();
  splitDataStructure();
  for_all_graphs_do(expressionExpandOneLevel());
  return pk_true;
}

/** 
    It prints informations on the clock hierarchy.
    See hierarchyStatistics().
*/
boolean DCG_poly::printCHStatistics()
{
  hierarchyStatistics(getDCG());
  return pk_true;
}

/** 
    It calls ClockCalculus(int) with a default value for the depth of solving.
*/
boolean DCG_poly::ClockCalculus()
{
  DCG_basic::ClockCalculus(5);
  setIntrinsicClocks();
  return pk_true;
}

/** 
    This  method can be applied to a graph of a module (set of processes) or a process.
    It transforms a multi-root hierarchy of each process to a single root hierarchy (parametrization of the clock system).
    The processes become endochronous. The interface of the processes may be modified (some input booleans are added for the parametrization), in this 
    case the returned boolean (modified) is set to the value "true", otherwise it is set to the value "false".
    
    @return the value "true" if the operation is performed, "false" otherwise.
*/
boolean DCG_poly::parameterizeMasterClocks(boolean bAddinputs, boolean * modified)
{
  ptDCG g;
  pointer elem;
  boolean bapply;
  ptlDCG lg, lg1;
  ptrepstr codeOp;
  typeMark vmyMark;

  codeOp = string2repstr(cstSetTick);
  if (isVerboseMode())
    pK_messenger->messagestr(mbeginOperName, codeOp);
  vmyMark = getMark();
  if (!getModuleGraphs(&lg))
    ptlPointer_insert_elem(&lg, (pointer) getDCG());
  bapply = pk_true;
  *modified = pk_false;
  // lg : list of DCG main graphs; creating tick for such graphs 
  while (bapply && ptlPointer_get_elemNext(&lg, &elem))
    {
      g = (ptDCG) elem;
      if (!g->satisfiesMarked(vmyMark))
	{
	  g->addMark(vmyMark);
	  init(g);		// set gw <- g
	  gw->parameterizeMasterClocksOneLevel(bAddinputs);
	  if (g->getReferencedGraphs(&lg1))
	    lg = ptlPointer_fusion(lg, lg1);
	}
    }
  resetMark(vmyMark);
  if (isVerboseMode())
    pK_messenger->messagestr(mendOperName, codeOp);
  return bapply;
}

/**
   @return true if the DCG (for a module all the DCG) is endochronous, "false" otherwise.
*/
boolean DCG_poly::isEndochronous()
{
  pointer elem;
  ptDCG g;
  ptlDCG lg;

  if (!getModuleGraphs(&lg))
    ptlPointer_insert_elem(&lg, (pointer) getDCG());

  while (ptlPointer_get_elemNext(&lg, &elem))
    {
      g = (ptDCG) elem;
      if (!g->isEndochronous())
	return pk_false;
    }
  return pk_true;
}


/** 
    This  method can be applied to a graph or a module.
    \copydoc dependenceReinforcingOneLevel
*/
boolean DCG_poly::dependenceReinforcing()
{
  for_all_graphs_do_br(dependenceReinforcingOneLevel(), b);
  return b;
}

/**
   It produces the morphim of the current graph according to the definitions 
   assigned of the Signal operators given in the ht table. 
   When bSimul is "true", the current graph is modified (selected booleans become outputs).
  
   @return "true" if the operation succeeds, "false" otherwise.

   \ref MorphismDefinedInTable "For more details..."
*/
boolean DCG_poly::applyTabularMorphism(ptHomomorphismTable ht, ptDCG_poly * gresult, boolean bSimul)
{
  ptDCG_poly gm;
  ptDCG_poly gCost;
  ptHomomorphism o;
  typeMark vmyMark, usedMark;
  boolean bok;
  ptlinstances lsig;
  ptDCG vdcg;

  //PLG: applyTabularMorphismOneLevel is a DCG_basic function
  //PLG: applyTabularMorphism strctly apply a DCG morphism
  //PLG check node calls to subgraph
  //PLG after applyTabularMorphism the declarations of the resulting AST should be controled
  // and if necessary missing declarations printed by  UnparseUndefinedModelAbstractions

  require_managers(EM, NM, CKM);
  vdcg = getDCG();

  // keeping logical operators in conditions used in extract.
  vdcg->undersamplingOnSignal(pk_false, pk_true);

  lsig = NILLINSTANCES;
  lsig = vdcg->getSignalsRefinClocks(CKM->getlhroots(), lsig);
  lsig = vdcg->getSignalsRefinClocks(CKM->getlhplaced(), lsig);
  // selecting
  vmyMark = getMark();
  lsig = vdcg->booleanNodeSelecting(lsig, vmyMark);
  // Copy selected.
  gCost = copySelected(pk_true, vmyMark, string2ident(cstCost), &usedMark);
  gCost->maskOutputs();

  // morphism
  o = new Homomorphism();
  bok = o->apply(getDCG(), gCost->getDCG(), usedMark, ht);
  delete o;

  if (!bok)
    {
      pK_messenger->error(errProfiling);
      resetMark(vmyMark);
      return pk_false;
    }
  gCost->setIntrinsicClocks();
  if (bSimul)
    {
      vdcg->declareSignalsAsOutputs(lsig);	// it replaces unmarkedSignalsBecomesOutputs(vmyMark);
      buildClockProcessInterface();
    }
  pK_messenger->printString(" Profiling/bSimul done ");
  resetMark(vmyMark);

  /* The morphism has the same clock structure -> don't rewrite it.
     pK_messenger->printString(" profiling/removeSignalEqualitiesOneLevel");
     gCost->removeSignalEqualitiesOneLevel();
   */
  gm = new_prePoly(new DCG());
  gm->setKind(opmodu);
  gm->addLocalGraph(gCost->getDCG());
  gm->setName(makeIdent_string_ident("M_", gCost->getName()));
  (*gresult) = gm;
  if (bSimul)
    {
      addAttribute(ATTRIBUTE_SUFFIX, (attributeValue) string2ident("Simu"));
      unparse();
      removeAttribute(ATTRIBUTE_SUFFIX);
    }
  return pk_true;
}

/** 
    This  method can be applied to a graph or a module.
    \copydoc rewDefinitionsOneLevel()
*/
void DCG_poly::rewDefinitions()
{
  for_all_graphs_do(rewDefinitionsOneLevel());
}

/** 
    It computes the interface abstraction of the operand graph.
*/
void DCG_poly::buildInterfaceAbstraction()
{
  if (getFormatLevel() < DCG_BOOL)
    computeIOTransitiveClosureOneLevel();
  buildIOClockSubHierarchyOneLevel(pk_true);
  nonProductiveSignalElimination();
}

/** 
    It performs the following partitionning (called input train or phylum Clustering) :
    <em>
    Two nodes are in the same set if and only if they depends on the same subset of inputs signals of the graph.
    </em> 
    The graph is modified. The nodes are clustered into sub-graphs.
    @return the value "false" if the operation failed, "true" otherwise.
    
    \warning 
    - (\e restriction) This method can be applied to a graph of a process, not to a graph of a module (set of processes).

    - The graph must be in DCGBool sub-format. --NO NO NO, it must be applied to a DCGPoly.
*/
boolean DCG_poly::phylumClustering()
{
  ptInputsTrainClusters mkparc;
  ptlINode lSubGraphs;
  ptrepstr codeOp;
  ptDCG g;

  codeOp = string2repstr(cstInputTrain);
  // Restriction to one graph.
  if (!getOnlyIfOneGraph(&g))
    {
      pK_messenger->globalMessageStr(codeNotYetImplemented, string2repstr(cstForModule));
      return pk_false;
    }

  require_node_manager_of(g, NM);
  mkparc = new InputsTrainClusters(g);
  lSubGraphs = mkparc->apply();
  delete mkparc;

  NM->setDependencesLabels();
  return pk_true;
}

/**
   \copydoc DCG_basic::undersamplingOnSignal
*/
boolean DCG_poly::undersamplingOnSignal(boolean bkeepCondIO, boolean keepLogicalOperators)
{
  if (!DCG_basic::undersamplingOnSignal(bkeepCondIO, keepLogicalOperators))
    return pk_false;
  setIntrinsicClocks();
  return pk_true;
}

/**
   Creates a container for the reduction of the communications of the input signals of the operand subgraph(aSubGraph).
   When bOnlyAssignmentClock, the reduction is applied using the assignment clock, otherwise the utility clock is also used.
   It is assumed that these clocks have been comptuded before the call to this method.
   \See {Container, Container.h}

*/
boolean DCG_poly::reduceComm(ptINode aSubGraph, boolean bOnlyAssignmentClock)
{
  return pK_Containers->reduceComm(this, aSubGraph, bOnlyAssignmentClock);
}

/**
   It updates the dependences for the composed nodes (array, subgraph, bundle) then calls the clockcalculus in silent mode.
   \see {DCG_basic::setComposedNodesDependences(), ClockCalculus()}
*/
void DCG_poly::setComposedNodesDependences()
{
  boolean b;
  DCG_basic::setComposedNodesDependences();
  // The previous operation produces new clocks.
  b = pK_messenger->setSilentModeMessagesTo(pk_true);
  ClockCalculus();
  pK_messenger->setSilentModeMessagesTo(b);
}

/**
   It builds a automaton by valuations of boolean state variables of the operand list (lSelectedStateVar).

   The resulting automaton is assigned to the graph, 1use getAutomaton() method to get it.

   Assumed: an element of l is a boolean state variable.
   Restriction: to the delays

*/
void DCG_poly::buildAutomaton(ptlinstances lSelectedStateVar)
{
  ptDCG2Automaton ptDCG2Automaton_obj;

  ptDCG2Automaton_obj = new DCG2Automaton();
  ptDCG2Automaton_obj->build(this, lSelectedStateVar);
  unparseAutomaton("FOO");
}

/**
   It calls buildAutomaton(ptlinstances l) with l = {the boolean state variables of the graph}
*/
void DCG_poly::buildAutomaton()
{
  ptlinstances lStateVar;

  undersamplingOnSignal(pk_true, pk_false);
  booleanCells2kernel();
  lStateVar = NILLINSTANCES;
  mkListOfStatesVariables(getRootClock(), pk_false, &lStateVar);
  buildAutomaton(lStateVar);
}

/**
   It calls buildAutomaton(ptlinstances l) with l = {aStateVar}.
*/
void DCG_poly::buildAutomaton(ptexpcal aStateVar)
{
  ptDCG2Automaton ptDCG2Automaton_obj;

  ptDCG2Automaton_obj = new DCG2Automaton();
  ptDCG2Automaton_obj->build(this, ptlinstances_push_end(NILLINSTANCES, aStateVar));
  unparseAutomaton("FOO");
}



/**
   It must not be used (PRESENTLY IN TEST).
   \copydoc buildAbstractionOneLevel
   \todo buildAbstraction() must be finalized.
*/
boolean DCG_poly::buildAbstraction(ptDCG_poly * gResult)
{
  ptlDCG lg;
  ptrepstr codeOp;
  boolean b;

  b = pk_true;
  codeOp = string2repstr(cstTestAbstraction);
  if (isVerboseMode())
    pK_messenger->messagestr(mbeginOperName, codeOp);
  if (!getModuleGraphs(&lg))
    *gResult = buildAbstractionOneLevel();
  return b;
}

ptlinsth DCG_poly::getLocalMaxima()
{
  require_clock_manager(CKM);
  return CKM->getChildren (getRootClock ());
}



/** 
    @return "true" if the graph is a polyEndochronous one, "false" otherwise.
    Tests: 
       - If the hierarchy = a set of tree without ^- operator as root and there are no constraints on theses roots.
       - If the constraints on placed clocks in different trees have free variables as root of their hierarchy.
       
       ! reqController is set to true when multi-roots with constrainst on placed clocks in different trees, false otherwise.
         It is undefined when the method returns false.
 */
boolean DCG_poly::isPolyEndoChronous(boolean *reqController)
{
  // Added for POLYENDO.
  ptlinsth lh1, lh2, lh,lhunused;
  ptinsth vhrac1, vhrac2, vh, vh1, vh2;
  ptlINode lnUnused;
  boolean b;
  int vdepth;

  require_managers(EM, NM, CKM);

  // Try to remove the not formulae.
  CKM->ReduceMinusFormulae(3);
  setClocksName();
  setSubstituableClocks();

  if (! isStaticCycleFree(pk_true, pk_false)) return pk_false;

  *reqController = pk_false;
  lh = CKM->initSystemForModification ();

  /* Test 1: the hierarchy must be a set of trees without ^- operator 
     as root, and there are no constraints on the roots. */
  lnUnused = getUnused(&lhunused);
  b = pk_true;
  while ( b && ptlinsth_get_next(&lh, &vh))
    {
      vh = CKM->getRepresentativeFormula(vh);
      if (!ptlinsth_isin_elem(lhunused, vh))
	{
	  if (CKM->getMinusArgs(vh, &vh1, &vh2))
	    b = pk_false;
	  else
	    {
	      vh1 = CKM->getNextFormula (vh, pk_true);
	      if (vh1 != vh) b = pk_false;
	    }
	}
    }
  if (b)
    {
      b = ! CKM->isaRecursiveSystem(); // It replaces the following test:
      /*** Removed May 2019.
	   Test 2: verifying that the constraints on placed clocks in several 
	   trees have free variables as root of their hierarchy.
	   if (CKM->getConstraintsOnPlacedClocks (&lh1, &lh2))
	   {
	   while ( b && ptlinsth_get_next (&lh1, &vh1))
	   { ptlinsth_get_next (&lh2, &vh2);
   	     vhrac1 = CKM->getRoot (vh1, &vdepth);
	     vhrac2 = CKM->getRoot (vh2, &vdepth);
	     if ( vhrac1 != vhrac2 )
	     {
	     if (! CKM->satisfiesVari (vhrac1) || ! CKM->satisfiesVari (vhrac2))
	     b = pk_false; else *reqController = pk_true;
	     } } } */
    }
  CKM->finalizeClockHierarchy ();
  if (b) addPragma (newPredefinedPragma (PRAGMA_POLYCHRONOUS_PROCESS, NILOBJPRAGMA, NILTREE));
  return b;
}

void DCG_poly::addRendezVousForPlacedConstraint(ptinsth vh1, ptinsth vh2)
{
  ptinsth vhrac1, vhrac2;
  int vdepth;
  require_managers(EM, NM, CKM);

  vhrac1 = CKM->getRoot (vh1, &vdepth);
  vhrac2 = CKM->getRoot (vh2, &vdepth);
  if ( vhrac1 != vhrac2 )
    {
      reorganizeNodesForRendezVous(vh1, vhrac1, vh2, vhrac2);
      addRendezVous(vh1, vh2);
    }
}


/**
   Adding a "rendez-vous" for placed clocks not in the same hierarchy and declared as equal.

   Assumed: isPolyEndoChronous() has been called before this call.
 */
void DCG_poly::addRendezVousForPlacedConstraints()
{
  // Added for POLYENDO.
  ptlinsth lh1, lh2, lhEq, lhvisited2, lhvisited1;
  ptinsth  vh1, vhEq;

  require_managers(EM, NM, CKM);
  EM->setUnifyExpressionMode(pk_false);
  if (CKM->getConstraintsOnPlacedClocks (&lh1, &lh2))
    { 
      lhvisited1 = NILLINSTH; 
      lhvisited2 = NILLINSTH;
      CKM->initSystemForModification ();
      while( ptlinsth_get_next (&lh1, &vh1))
	{
	  lhEq = CKM->getClocksPlacedEquivTo(vh1);
	  while ( ptlinsth_get_next (&lhEq, &vhEq))
	    {
	      if (! ptlinsth_search_couple( vh1, vhEq, lhvisited1, lhvisited2))
		{
		  addRendezVousForPlacedConstraint(vh1, vhEq);
		  lhvisited1 = ptlinsth_insert_elem (lhvisited1, vh1);
		  lhvisited2 = ptlinsth_insert_elem (lhvisited2,vhEq);
		}
	    }
	}  
      ptlinsth_free(lhvisited1);     
      ptlinsth_free(lhvisited2);
      CKM->finalizeClockHierarchy();
      buildClockProcessInterface();
    }
  
}



/**
   It adds a rendez-vous for the constraints vh1 = vh2 where these clocks are 
   assumed to be placed and not in the same clock hierarchy.
    - For vh1 it adds the definitions, assigned to vh1: 
        (| (| n11 :: Send(rdv1) | n12::receive(rdv2) |) | n11 -> n12 | rdv1 ^= rdv2 ^= vh1 |)
        and for each node N assigned to vh1 (excluding the Send/receive) n12 -> N
    - For vh2 it adds the definitions, assigned to vh2:
        (| (| n21 :: Send(rdv2) | n22::receive(rdv1) |) | n21 -> n22 | rdv1 ^= rdv2  ^= vh2 |)
        and for each node N assigned to vh2 (excluding the Send/receive) n22 -> N
	
 */
void DCG_poly::addRendezVous(ptinsth vh1,  ptinsth vh2)
{
  // Added for POLYENDO.
  ptexpcal vlab, lexp, vsig1, vsig2;
  ptINode n11, n12, n21, n22;
  ptident idRDV, idSend, idReceive;
  TypeType TT;

  require_managers(EM, NM, CKM);

  idRDV = string2ident("RDV");
  idSend = getSendId();
  idReceive = getReceiveId();
  // TT = BooleanType;
  TT = IntegerType;
  vsig1 = EM->mkIdentExp(idRDV, tsignal, TT, NM->mk_node_undefined (), vh1, NILBACKANNOTATION);
  lexp = EM->mkNaryExp (oplident, vsig1, NILEXPCAL, NILBACKANNOTATION);

  n11 = NM->mkIntrinsicUnsafeAt(idSend, NILEXPCAL, NILEXPCAL, lexp, vh1, NILBACKANNOTATION);
  vlab = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), vh1, NILBACKANNOTATION, pk_false);
  NM->set_label (n11, vlab);
  EM->setNode(vlab, n11);

  n22 = NM->mkIntrinsicUnsafeAt(idReceive,NILEXPCAL, lexp, NILEXPCAL, vh2, NILBACKANNOTATION);
  vlab = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), vh2, NILBACKANNOTATION, pk_false);
  NM->set_label (n22, vlab);
  EM->setNode(vlab, n22);

  vsig2 = EM->mkIdentExp(idRDV, tsignal, TT, NM->mk_node_undefined (), vh2, NILBACKANNOTATION);
  lexp = EM->mkNaryExp (oplident, vsig2, NILEXPCAL, NILBACKANNOTATION);
  n21 = NM->mkIntrinsicUnsafeAt(idSend, NILEXPCAL, NILEXPCAL, lexp, vh2, NILBACKANNOTATION);
  vlab = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), vh2, NILBACKANNOTATION, pk_false);
  NM->set_label (n21, vlab);
  EM->setNode(vlab, n21);
  
 n12 = NM->mkIntrinsicUnsafeAt(idReceive, NILEXPCAL, lexp, NILEXPCAL, vh1,  NILBACKANNOTATION);
 vlab = EM->mkLabelIdent (getLabelId (), NM->mk_node_undefined (), vh1, NILBACKANNOTATION, pk_false);
 NM->set_label (n12, vlab);
 EM->setNode(vlab, n12);
}

