/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
   http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PRIVATE_ static ptDCG_poly gw = NILDCG_POLY;

#include "DCGs_Macros.c"

void
 DCG_poly::init(ptDCG g)
{
  if (gw == NILDCG_POLY)
    gw = new_prePoly(g);
  else
    gw->setDCG(g);
}

_PROTECTED_
/**
   Specialisation.
   @copydoc DCG_basic::copySelected()
*/
    ptDCG_poly DCG_poly::copySelected(boolean exportHierarchy, typeMark vMark, ptident id, typeMark * usedMark)
{
  ptDCG_basic gbasic = DCG_basic::copySelected(exportHierarchy, vMark, id, usedMark);
  return DCG_basic2poly(gbasic);
}

/**
   Recursively write each elementary process "PP" that has a subexpression "Ei" as "(| PP' | xi:=Ei|) where xi"
   such that xi is a new identifier and PP' is PP in which all occurences of "Ei" are replaced by "xi"
   \ref Graph2Kernel "For more details..."
*/
void DCG_poly::expressionExpandOneLevel()
{
  undersamplingOnSignal(pk_false, pk_false);
  DCG_basic::expressionExpandOneLevel();
}

/** 
    This methods adds boolean inputs for a list of signals (list). 
    All added signals are synchronized to vhtick.
    For a signal (SIG)
    - whose the clock is a root of a tree in the clock system 
    - and this clock is a free variable in the system

    a boolean signal (BHSIG) is added in the graph as an input. 
    Then, the following synchronizations are added in the system :
    - vhtick ^= BHSIG, 
    - when BHSIG ^= ^SIG. 
    After each insertion, the clock calculus is applied.

    \note The order of the list is important. The parametrized system depends on this order.
*/
void DCG_poly::extendedSpecification(ptinsth vhtick, ptlinstances list, boolean bAddInput)
{
  ptexpcal vsig;
  ptinsth vh;

  require_clock_manager(CKM);
  require_expr_manager(EM);
  list = ptlinstances_copy(list);
  while (ptlinstances_get_next(&list, &vsig))
    {
      vh = CKM->getRepresentativeFormula(EM->getClock(vsig));
      if (CKM->isFreeClock(vh))
	{
	  if (bAddInput)
	    addBooleanInput(vh, vhtick, namBooleanClockfrom(EM->getOriginalName(vsig)));
	  else
	    {
	      // id = EM->getOriginalName (vsig);
	      // addPresentNode (vh, vhtick, id, namBooleanClockfrom (id));
	      addPresentNode(vh, vhtick, vsig);	// id, namBooleanClockfrom (id));
	    }
	  ClockCalculusOneLevel(2);
	}
    }
}

/**
   For each undersampling clock ([C]) of lhtodeclare, 
   if C is not an input of the graph, C is added in the outputs of the graph; it is also added to lappSignals
   - otherwise, C is added to lappSignals

   It can be used for exporting some booleans in an abstration of a model.
   @return "true" if the list lappSignals is modified, "false" otherwise.
*/
_VIRTUAL_ boolean DCG_poly::buildClockExportation(ptlinsth lhtodeclare, ptlinstances * lappSignals)
{
  ptinsth vh;
  ptexpcal vExport, vsig;
  ptlinstances l, louts, ls, curlist, lins;
  boolean bpos, badded;
  ptlinsth lh;

  require_managers(EM, NM, CKM);
  louts = getOutputs();
  lins = getInputs();
  badded = pk_false;

  lh = lhtodeclare;
  while (ptlinsth_get_next(&lhtodeclare, &vh))
    if (!CKM->satisfiesTickClock(vh) && CKM->satisfiesSub(vh))
      {
	vExport = EM->getExprExportBooleanAbtraction(CKM->getConditionSub(vh), NILLINSTANCES, &bpos);
	ls = EM->getSignalsReferencedIn(vExport);
	l = ls;
	while (ptlinstances_get_next(&ls, &vsig))
	  if (!ptlinstances_isin_exp(lins, vsig) && !ptlinstances_isin_exp(louts, vsig))
	    if (ptlinstances_isin_insert_exp(lappSignals, &curlist, vsig))
	      badded = pk_true;
	ptlinstances_free(l);
      }
  return pk_false;
}


/** 
    It creates the tree of delayed signals for the current graph.
    \ref States "For more details..."
*/
boolean DCG_poly::delayTreeBuildOneLevel()
{
  ptmkDelayTree o;
  boolean b;

  o = new mkDelayTree(getDCG());
  b = o->delayTreeBuilder();
  if (!b)
    {
      pK_messenger->error(errbcldelay);
      pK_messenger->error(errcompilerbug);
    }
  delete o;
  return b;
}

/**
   This method calculates the "profile" (inputs,outputs, parameters)
   for each node of the clock hierarchy. Each node of the hierarchy is seen as a Signal process.
   \ref ProfileBuilder "For more details..."
*/
void DCG_poly::buildClockProcessInterfaceOneLevel()
{
  int date;
  ptProfileBuilder profileBuilder;

  date = pK_messenger->messageBeginOperVerbose(infoInterfaceCalculus);
  interfaceReset();
  setClocksName();
  setSubstituableClocks();
  profileBuilder = new ProfileBuilder(getDCG());
  profileBuilder->interfaceCalculus();
  profileBuilder->finalize();
  pK_messenger->messageEndOperVerbose(date);
}

/**
   It builts the clock environment of a program (tick, iostep, step, Idle).
   During this, constraint informations are stored in the current graph structure.

   It completes the hierarchy (forest of clock trees) to obtain a tree. A fictitious root
   is appended for an exochronous process (DCGPoly format level). 

   iostep, step, Idle not implemented

   \warning (\e restriction) Currently, only "tick" clock is built.
*/
void DCG_poly::setIntrinsicClocksOneLevel()
{
  int date;

  require_clock_manager(CKM);
  date = pK_messenger->tracing(mbegclockEnvBuilding, ident2repstr(getName()));
  CKM->finalizeClockHierarchy();
  storeNullClockInfo();
  storeInitClockInfo();

  /*----------------------- todo ------ */
  /*  h = buildIdleClock(); */
  /*  h = buildIoStepClock(); */

  getNodeManager()->removeDependenceNullClock();
  pK_messenger->messageEndOperVerbose(date);
}

/**
   Computes the exclusive relation over clocks in the current clock hierarchy
 
   The result is assigned to each clock: a predefined attribute, called EXCLUSIVE_CLOCKS defined in the
   file $GraphVM_export/CompilerAttributes.h, is associated with each clock, the value of the attribute gives the list 
   of its exclusive clocks.

   \ref ClockCalculusMatrix "For more details..." 
*/
void DCG_poly::computeClockExclusionsOneLevel(int ecLevel)
{
  ptexclusiveClocks ecObj;
  int date;
  ptinsth vh;
  ptlinsth lh;
  typeMark vmyMark;
  ptlINode ln;
  ptINode n;

  //----------------------------------------------------------------------------------------
  //LPT: NEW NEW NEW
  //PLG suggested name explicitClockExclusions 11/01/25
  //PLG set clock exclusions as H*(k1+k2+.....) in clock null formulas

  require_clock_manager(CKM);
  require_node_manager(NM);
  if (!isGraphOfExternalProcess())
    {
      date = pK_messenger->messageIdent(mbegExcluClocks, getName());

      if (ecLevel == 2)
	vmyMark = UNDEFINED_MARK;	// convention.
      else
	{
	  // Selecting the clocks with associated signals 
	  vmyMark = getMark();
	  lh = CKM->getlhroots();
	  while (ptlinsth_get_next(&lh, &vh))
	    if (!CKM->satisfiesUndefClock(vh) && !CKM->satisfiesInitClock(vh))
	      if (CKM->getSynchronousSignals(vh) != NILLINSTANCES)
		CKM->addMark(vh, vmyMark);
	  lh = CKM->getlhplaced();
	  while (ptlinsth_get_next(&lh, &vh))
	    if (CKM->getSynchronousSignals(vh) != NILLINSTANCES)
	      CKM->addMark(vh, vmyMark);
	  // the clocks referenced in the expressions ?
	  if (ecLevel == 1)
	    {
	      if (NM->getAllNodes(&ln))
		while (ptlINode_get_elem(&ln, &n))
		  {
		    lh = NM->getInodelInputClocks(n);
		    while (ptlinsth_get_next(&lh, &vh))
		      CKM->addMark(vh, vmyMark);
		  }
	    }
	}
      ecObj = new exclusiveClocks(getDCG());
      ecObj->apply(vmyMark);
      delete ecObj;
      if (ecLevel != 2)
	CKM->resetMarkAll(vmyMark);
      pK_messenger->messageEndOperVerbose(date);
    }
}

/** 
    It computes the projection of the clock system 
    - on the clocks of the inputs/outputs signals of the graph 
    - and, if bclkdep, the clocks of the input/output data dependencies. In this case, this operation supposes that the transitive 
    closure functionality has been applied.
    \see{computeIOTransitiveClosure, buildClockExportation(), }
*/
ptinsth DCG_poly::buildIOClockSubHierarchyLoc(boolean bclkdep, ptlinsth * lhtodeclare, ptlinstances * appendSigs)
{
  ptlinsth lh, lh1;
  ptinsth vhemb;

  require_clock_manager(CKM);
  require_expr_manager(EM);

  lh = EM->getClockOfSignals(getInputs());
  lh = ptlinsth_sorted_fusion(lh, EM->getClockOfSignals(getOutputs()));
  if (bclkdep)
    lh = ptlinsth_sorted_fusion(lh, getIODependencesClocks());
  vhemb = CKM->getBranchingClocks(lh);
  lh1 = CKM->CompleteUpstreamParametrization(lh, vhemb);
  *lhtodeclare = lh1;

  *appendSigs = NILLINSTANCES;
  while (buildClockExportation(lh1, appendSigs))
    {
      lh = ptlinsth_sorted_fusion(lh, EM->getClockOfSignals(*appendSigs));
      if (bclkdep)
	lh = ptlinsth_sorted_fusion(lh, getIODependencesClocks());
      vhemb = CKM->getBranchingClocks(lh);
      lh1 = CKM->CompleteUpstreamParametrization(lh, vhemb);
      *lhtodeclare = ptlinsth_sorted_fusion(*lhtodeclare, lh1);
    }

  return vhemb;
}

/** 
    It computes the projection of the clock system on the clocks of the inputs/outputs 
    signals of the graph and (if bclkdep) the clocks of the input/output data dependencies. In this case, 
    this operation supposes that the transitive closure functionality has been applied.
    \see{computeIOTransitiveClosure, buildClockExportation(), }
*/
void DCG_poly::buildIOClockSubHierarchyOneLevel(boolean bclkdep)
{
  ptlinsth ClocksBuiltOnIOClock, ClocksNotBuiltOnIOClock, lhtodeclare;
  ptinsth vhemb;
  ptlinstances appendSigs;

  require_clock_manager(CKM);
  require_expr_manager(EM);

  undersamplingOnSignal(pk_true, pk_false);
  pK_messenger->messageIdent(mbegprojection, getName());

  vhemb = buildIOClockSubHierarchyLoc(bclkdep, &lhtodeclare, &appendSigs);
  ClocksBuiltOnIOClock = IOSplitExportedClocks(lhtodeclare, &ClocksNotBuiltOnIOClock);
  // setIOstep(vhemb);
  CKM->setProjectedSystem(ClocksBuiltOnIOClock, ClocksNotBuiltOnIOClock, appendSigs);
}

/** 
    Assumed : the method delayTreeBuild() must be applied before.
    The graph is reinforced (so, it must be applied before code generation partitionning).
    For a one-delay node ZX := X $1 , it tries to aggregate the delay (ZX) with the delayed signal (X).
    It looks at the successors of ZX : S1,...,Sn and it adds a dependence from Si to X.
    If such reinforcing does not introduce a circuit in the current graph, the signals X and ZX can
    be merge in the same memory (the references to the previous value of X are generqted before the update of the memory).
    In this case, the signal ZX is annotated with this predefined attribute called MEMORY_OPTIMIZED. 
    This attribute is analyzed by code generators. 

    \see{delayTreeBuild}

    \ref Optimisations "For more details..."
*/
void DCG_poly::setDelayMemoryOverloadingOneLevel(boolean bclu)
{
  ptmkaggregationsAndMultiplexings mkparc;

  if (!isGraphOfExternalProcess())
    {
      mkparc = new mkaggregationsAndMultiplexings(getDCG());
      mkparc->delaysMemoryOptimisation(bclu);
      delete mkparc;
    }
}

/** 
    Applying aggregation of when/cell definitions and multiplexing of variables (delays).

    Its realizes optimizations related to the memory size for code generation using
    - the "when/cell" aggregations : two signals that can be represented in 
    the same area are annotated using the predefined attribute AGGREGATTRIBUTE. 
    This attribute is analyzed by code generators. 
    //PLG: WHY not Signal transformation with x::= E and x  when h1, x when h2,...

    - the multiplexings of variables : signals built on the same signal using the "window"
    and "delay" operators are represented in the same area if the initial values are compatible. 

    Array and circular management of indexes are used
    for the optimization. The method delayTreeBuild() must be applied before.
 
    @param mutliplexDelays (in) with the value "true" a signal and the delays built on it are multiplexed, 
    "false" they are managed by two variables if there is only one delay on the signal
    In other words, for zx := x $1, and it is the only delay on x, if the signals are scalars, they are never
    multiplexed in the same area even if mutliplexDelays is pk_true; otherwise they are managed by two variables.

    \see{delayTreeBuild}

    \ref Optimisations "For more details..."
   Deprecated
*/
void DCG_poly::memoryOptimizationsOneLevel(boolean mutliplexDelays)
{
  ptmkaggregationsAndMultiplexings mkparc;

  // Deprecated
  //----------------------------------------------------------------------------------------
  // SPLIT SPLIT between all elementary methodes moveCell2PartialDef, multiplexExclusiveSignals, 
  // memoryOptimizationsOneLevel is a DCG_poly function 
  // x:=E1 and y:=E2 can be represented by the same variable xy:=E1 default E2 is E1 and E2 have disjoint clocks and no delay

  if (!isGraphOfExternalProcess())
    {
      mkparc = new mkaggregationsAndMultiplexings(getDCG());
      //  When/cell optimisations and delayed memories multiplexing.
      if (!mkparc->aggregationsAndMultiplexing(mutliplexDelays))
	pK_messenger->error(errrecsystem);
      delete mkparc;
    }
}

/** 
    It calculates the transitive closure reduced to input/output signals of the graph.
    The results is annotations assigned to i/o signals 
    using the CLOSUREATTRIBUTE attribute defined in the file incl.dir/CompilerAttributes.h
    @return the value "true" if the translation is executed, "false" otherwise 
    (internal problems in this case).

    \ref Closure "For more details..."
*/
boolean DCG_poly::computeIOTransitiveClosureOneLevel()
{
  if (isGraphOfExternalProcess()) return pk_true;
  pK_messenger->messageIdent(infoTransitiveClosure, getName());
  DCG_basic::buildIOTransitiveClosure();
  ClockCalculus();
  pK_messenger->setcputime();
  return pK_messenger->getErrorCode();
}

/**
   Building the tick of the program by parameterizing the clock system.
   When the current graph is not endochronous, this transformation induces modifications on the interface 
   of the graph by adding input booleans.\n
   Order : 
   - the clock of an free input is fixed before the clock of an output. 
   - in a subset (inputs or outputs), the syntactic declaration order is used.

   When a signal is added, the clock calculus is applied to obtain an endochonous system.
   Last, when the graph is not endochronous input signals are created for each root of the clock hierarchy (arbitrary order).
   @return "true" if the resulting graph is endochonous, "false" otherwise.
*/
boolean DCG_poly::parameterizeMasterClocksOneLevel(boolean bAddinputs)
{
  ptinsth vh, vhroot;
  boolean modified, b;
  ptlinsth lh;

  require_clock_manager(CKM);
  CKM->initSystemForModification();
  if (isEndochronous())
    {
      modified = pk_false;
      b = pk_true;
    }
  else
    {
      modified = pk_true;
      pK_messenger->message(mextendedSpec);
      vhroot = getRootClock();	/* fictitious root */
      setCompleteClock(CKM->getUndefClock ()); // Added 20 Feb 2017 LBesnard xxxxx
      lh = CKM->getChildren(vhroot);
      CKM->resetClocksAsRoot(lh);
      /* todo : add the use of local maxima.
         CKM->OrderBuilding();
         CKM->OrderReduce();
         if (!CKM->getMaximalClock(&vh)) 
       */
      vh = CKM->new_variFormulaRootClock();

      /* vh is the Tick */
      extendedSpecification(vh, getInputs(), bAddinputs);
      extendedSpecification(vh, getOutputs(), bAddinputs);
      b = isEndochronous();
      if (!b)
	{
	  iterateNodeInputEndochronization(vh, true /*bAddinputs */ );
	  b = isEndochronous();
	}
    }
  setIntrinsicClocksOneLevel();	/* set vh as the tick clock of the program */
  if (modified)
    {
      setClocksName();
      setSubstituableClocks();
      computeControlClocks();	// iostep, ...
      ClockCalculus();
    }
  if (b)
    addHierarchyPragmas();
  return b;
}

/**
   It rewrites the definitions (not the dynamical ones) of the graph, by substiting a reference to a signal by the expression of definition of this signal.
   It is useful to remove cycles in the graph. For example \n
   (| x := a when c default y  \n
    | y := x when c default b  \n
    |) \n
    is rewritten in \n
   (| x := a when c default b  \n
    | y := x when c default b  \n
    |) \n
    by this method (the substitution of y by its definition in the definition of x, removes the cycle).
    \see { DCG::rewDefinitions()  }
*/
void DCG_poly::rewDefinitionsOneLevel()
{
  require_managers(EM, NM, CKM);

  getDCG()->rewDefinitions();
  ClockCalculus();
  EM->rewritingExprEH();
}

/**
   Rewrite the boolean cells to the kernel language.
*/
void DCG_poly::booleanCells2kernel()
{
  ptlINode ln;
  ptINode n;
  boolean b;

  require_managers(EM, NM, CKM);
  if (NM->getEquations(&ln))
    {
      b = pk_false;
      while (ptlINode_get_elem(&ln, &n))
	if (NM->is_cellNode(n) && EM->IsBoolean(NM->get_leftOperand(n)))
	  {
	    NM->cell2kernel(n);
	    b = pk_true;
	  }

      if (b)
	{
	  b = pK_messenger->setSilentModeMessagesTo(pk_true);
	  ClockCalculus();
	  b = pK_messenger->setSilentModeMessagesTo(b);
	}
    }
}

/**
   It generates the graph of the abstraction of the current DCG.
   Not fully implemented (in development).
*/
ptDCG_poly DCG_poly::buildAbstractionOneLevel()
{
  ptDCG_poly g;
  typeMark vmyMark, usedMark;
  ptDCGColouringAbstraction o;

  // selecting
  vmyMark = getMark();
  o = new ColouringAbstraction(getDCG());
  o->apply(vmyMark);
  delete o;
  // Copy selected.
  g = copySelected(pk_true, vmyMark, string2ident("test"), &usedMark);
  g->setIntrinsicClocks();
  return g;
}

/**
   @return "true" if a node (vnode) is referenced only in the hierarchy with vhroot as root, "false" otherwise.
 */
boolean DCG_poly::isFromOtherHierarchy(ptINode vnode, ptinsth vhroot, ptinsth vh)
{
  //  Added for POLYENDO
  ptINode vFrom, vTo;
  ptldependence ld;
  ptdependence aDep;
  ptinsth vhemb, vh1;
  ptlinsth lh;

  require_managers(EM, NM, CKM);

  if ( NM-> is_delay(vnode))
    {
      ptexpcal resret, sigret, vret, vfen, vinit;
      NM-> get_delay_arg(vnode, &resret, &sigret, &vret, &vfen, &vinit);
      vhemb = CKM->getBranching(EM->getClock(sigret), vhroot);
      return (vhemb == vhroot);
    }
  
  // using dependencies
  ld = NM->get_predecessors(vnode);
  while (ld->get_next(&aDep, &ld))
    {
      aDep->getDependence(&vFrom, &vTo);
      vh1 = NM->get_nodeClock(vFrom);
      if ( vh1 != vh ) 
	{
	  vhemb = CKM->getBranching(vh1, vhroot);
	  if (vhemb != vhroot)
	    return pk_false;
	}
    }

  lh = NM->get_clockPredecessors(vnode);
  while (ptlinsth_get_next(&lh, &vh1))
    if ( vh1 != vh) 
      {
	vhemb = CKM->getBranching(vh1, vhroot);
	if (vhemb != vhroot)
	  return pk_false;
      }
  return pk_true;
  
}



/**
   @return the nodes assigned to a clock (vh) that are referenced only in the hierarchy dominated by vhroot.
   Assumed: vh is assumed not to be in the hierarchy with vhroot as root.
*/
void DCG_poly::moveNodes(ptinsth vhFrom, ptinsth vhFromRoot, ptinsth vhTo, ptinsth vhrootTo)
{
  // Added for POLYENDO
  ptlINode ln;
  ptINode vnode;
  boolean onContinue;

  require_managers(EM, NM, CKM);
  onContinue= pk_true;
  while (onContinue) {
    onContinue= pk_false;
    ln = CKM->getNodesAssignedTo(vhFrom);
    while (ptlINode_get_elem(&ln, &vnode))
      /* if (isReferencedOnlyInHierarchy(vnode, vhrootTo))
	{
	  moveNode(vnode, vhFrom, vhTo);
	  onContinue= pk_true;
	}
	else */
	if (isFromOtherHierarchy(vnode, vhrootTo, vhFrom))
	  {
	    moveNode(vnode, vhFrom, vhTo);
	    onContinue= pk_true;
	  }
  }
}


/**
   It moves the nodes of the list lnToMove1 from a clock (vh1) to another one (vh2).
   Assumed: the nodes of lnToMove1 are assumed to be assigned to vh1.
   The clocks of the signals are also updated.
 */
void DCG_poly::moveNode(ptINode n, ptinsth vh1, ptinsth vh2)
{
  // Added for POLYENDO
  ptlinstances lsigori, lsig;
  ptexpcal vsig;

  require_managers(EM, NM, CKM);
  lsigori = CKM->getSynchronousSignals (vh1);
  NM->moveNodeToClock(n, vh2);
  lsig = NM->getInodelOutputFlows(n);
  while (ptlinstances_get_next(&lsig, &vsig))
    if (ptlinstances_isin_exp(lsigori, vsig))
      {
	CKM->removeSignalFromClass (vh1, vsig);
	EM->setClock(vsig, vh2);
      }
}


/**
   Assumed : 
   - vh1, clock of a hierachy with vhrac1 as root
   - vh2, clock of a hierachy with vhrac2 as root
   It moves:
   - the nodes assigned to vh1 to vh2 if such a node is only referenced in the hierarchy with vhrac2 as root.
   - the nodes assigned to vh2 to vh1 if such a node is only referenced in the hierarchy with vhrac1 as root.
   It changes:
   - the references to the clock vh1 by vh2 in the hierarchy with vhrac2 as root
   - the references to the clock vh2 by vh1 in the hierarchy with vhrac1 as root
*/
void DCG_poly::reorganizeNodesForRendezVous(ptinsth vh1, ptinsth vhrac1, ptinsth vh2, ptinsth vhrac2)
{
  //  Added for POLYENDO.
  ptinsth vh1c, vh2c;

  require_managers(EM, NM, CKM);
  moveNodes(vh1, vhrac1, vh2, vhrac2);
  moveNodes(vh2, vhrac2, vh1, vhrac1);

  changeClockReferences(vhrac1, vh2, vh1);
  changeClockReferences(vhrac2, vh1, vh2);

  if (CKM->satisfiesSub(vh1) && CKM->satisfiesSub(vh2))
    {
      CKM->getComplExtraction(vh1, &vh1c);
      CKM->getComplExtraction(vh2, &vh2c);
      changeClockReferences(vhrac1, vh2c, vh1c);
      changeClockReferences(vhrac2, vh1c, vh2c);
      } 
}

/**
   It replaces the references to vh1 by vh2 in the hierarchy with vhroot as root.
 */
void DCG_poly::changeClockReferences(ptinsth vhroot, ptinsth vh1, ptinsth vh2)
{
  //  Added for POLYENDO.
  ptlINode ln;
  ptINode n;
  ptlinsth lh;
  ptinsth vh;

  require_managers(EM, NM, CKM);
  vhroot =  CKM->getRepresentativeFormula(vhroot);
  ln = CKM->getNodesAssignedTo(vhroot);
  while (ptlINode_get_elem(&ln, &n))
    {
      NM->replace(n, vh1, vh2);
    }
  // Visiting the sub-hierarchy of vhroot.
  lh = CKM->getChildren (vhroot);
  while (ptlinsth_get_next (&lh, &vh))
    {
      changeClockReferences(vh, vh1, vh2);
    }
}


/**
   @return "true" if a node (vnode) is referenced only in the hierarchy with vhroot as root, "false" otherwise.
boolean DCG_poly::isReferencedOnlyInHierarchy(ptINode vnode, ptinsth vhroot)
{
  //  Added for POLYENDO
  ptINode vFrom, vTo;
  ptldependence ld;
  ptdependence aDep;
  ptinsth vhemb, vh1;
  ptlinsth lh;

  return pk_false;

  require_managers(EM, NM, CKM);
  ld = NM->get_successors(vnode);

  while (ld->get_next(&aDep, &ld))
    {
      aDep->getDependence(&vFrom, &vTo);
      vh1 = NM->get_nodeClock(vTo);
      vhemb = CKM->getBranching(vh1, vhroot);
      if (vhemb != vhroot)
	return pk_false;
    }

  lh = NM->get_clockSuccessors(vnode);
  while (ptlinsth_get_next(&lh, &vh1))
    {
      vhemb = CKM->getBranching(vh1, vhroot);
      if (vhemb != vhroot)
	return pk_false;
    }
  return pk_true;
}
*/

