/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   @return "true" if identifier (id) is not in the table of the manager and adds it, "false" otherwise.
*/
boolean NameManager::insertInIdentNameTable(ptident id)
{
  ptelemHashTable p;
  return !findOrInsert(id, &p);
}

/**
   @return id_X where X is a non negative integer > v, such that id_X
     is not in the name manager table
   id_X  is added to the name manager table
*/
ptident NameManager::new_exclusiveIdent(ptident id, integer v)
{
  integer suffix;
  ptident idorig;

  idorig = id;
  suffix = v;
  id = makeIdent_ident_int(idorig, suffix);
  while (!insertInIdentNameTable(id)) {
    suffix = suffix + 1;
    id = makeIdent_ident_int(idorig, suffix);
  }
  return id;
}

/**
   @return name where
   - name=id if id is not in the name manager table;
   - name=id_X, where X is a non negative integer such that id_X
     is not in the name manager table
   name is added to the name manager table
*/
ptident NameManager::exclusiveIdent(ptident id)
{
  integer suffix;
  ptident idorig;

  if (insertInIdentNameTable(id))
    return id;

  idorig = id;
  suffix = 0;
  id = makeIdent_ident_int(idorig, suffix);
  while (!insertInIdentNameTable(id)) {
    suffix = suffix + 1;
    id = makeIdent_ident_int(idorig, suffix);
  }
  return id;
}

/**
   @return a new identifier (id_X), built on 
      a prefix (name or predefined name used for types if name is the undefined name)
      and an integer (X). This integer is greater or equal to 0 and id_X is not in the table of the 
      manager. id_X is added to the manager.
*/
ptident NameManager::mkNameType(ptident name)
{
  if (name == idNil)
    name = TypeId;
  return exclusiveIdent(name);
}

/**
   @return a new identifier (id_X), built on a prefix (name or predefined name used for indexes if name is the undefined name)
   and an integer (X) greater or equal to 0 and id_X is not in the table of the manager. id_X is added to the manager.
*/
ptident NameManager::mkNameIndex(ptident name)
{
  // unused
  if (name == idNil)
    name = IndexId;
  return exclusiveIdent(name);
}

/**
   @return a new identifier (id_X), built on a prefix (name or predefined name used for constants if name is the undefined name)
   and an integer (X) greater or equal to 0 and id_X is not in the table of the 
   manager. id_X is added to the manager.
*/
ptident NameManager::mkNameConst(ptident name)
{
  //  unused
  if (name == idNil)
    name = getConstantId();
  return exclusiveIdent(name);
}

/**
   @return a new identifier (id_X), built on a prefix (name or predefined name used for labels if name is the undefined name)
   and an integer (X) greater or equal to 0 and id_X is not in the table of the 
   manager. id_X is added to the manager.
*/
ptident NameManager::mkNameLabel(ptident name)
{
  //  unused
  if (name == idNil)
    name = LabelId;
  return exclusiveIdent(name);
}

/**
   @return a new identifier (id_X), built on a prefix (name or predefined name used for signals if name is the undefined name)
   and an integer (X) greater or equal to 0 and id_X is not in the table of the 
   manager. id_X is added to the manager.
*/
ptident NameManager::mkIdentName(ptident id)
{
  if (insertInIdentNameTable(id))
    { 
      if (id != idNil)
	{
	  return id;
	}
    }
  id= exclusiveIdent(id);
  return id;
}

/**
   @return a new identifier (id_X), built on the string name
*/
ptident NameManager::mkIdentName(char *name)
{
  return mkIdentName(string2ident(name));
}

/**
   @return a new identifier (id_X), built on a prefix (name or predefined name used for state variables if name is the undefined name)
   and an integer (X) greater or equal to 0 and id_X is not in the table of the 
   manager. id_X is added to the manager.
*/
ptident NameManager::mkIdentNameState(ptident id)
{
  ptident idprev;
  idprev = newIdent_concIdentIdent(id, PreviousId);
  return mkIdentName(idprev);
}

/**
   @return a new clock identifier made of the standard clock string followed by an integer
*/
ptident NameManager::new_pureClockIdent(integer i)
{
  if (i < 0)
    i = 0;
  return new_exclusiveIdent(ClockId, i);
  // return makeIdent_ident_int_1(ClockId, i);	// Testing Automata...
}

/**
   @return the identifier TT_id (if b), FF_id (if not b).
*/
ptident NameManager::new_clockSamplingIdent(ptident id, boolean b)
{
  ptident pre;

  pre = getIdUndersampling(b);
  id = newIdent_concIdentIdent(pre, id);
  return exclusiveIdent(id);
}

/**
   @return the the current model name and then sets it to id.
*/
ptident NameManager::setCurrentModelName(ptident id)
{
  ptident x;

  x = curModelName;
  curModelName = id;
  return x;
}

/**
   Sets the ContextSubExpr identifier to id.
   @return the value of the ContextSubExpr identifier before the call.
 */
ptident NameManager::setContextSubExpr(ptident id)
{
  ptident x;

  x = contextSubExpr;
  contextSubExpr = id;
  // to always add an suffix for each identifier of the sub expression
  insertInIdentNameTable(contextSubExpr);
  return x;
}

/**
   @return the value of the context identifier for sub-expression building and fixes it to the name associated with the second operand of a when operator.
*/
ptident NameManager::setContextSubExprCondition()
{
  return setContextSubExpr(newIdent_concIdentIdent(curModelName, ConditionId));
}

/**
   @return the value of the ContextSubExpr identifier and sets it to id.
*/
noperators NameManager::setContextSubExpr(noperators op)
{
  noperators xop;

  xop = opContextSubExpr;
  opContextSubExpr = op;
  return xop;
}

ptident NameManager::getContextSubExprOperator()
{
  ptident id;
  if ((opContextSubExpr == opmemo) || (opContextSubExpr == opfen) || (opContextSubExpr == opret) || (opContextSubExpr == opvar))
    id = contextSubExpr;
  else if (!getNameAssociatedWithOperator(opContextSubExpr, &id))
    id = contextSubExpr;
  insertInIdentNameTable(id); // to have an integer suffix for auxilliary created signal.
  return id;
}

/**
   @return the current context operator (managed during the graph building phase).
*/
noperators NameManager::getOpContextSubExpr()
{
  return opContextSubExpr;
}

ptident NameManager::getConstantId()
{
  return newIdent_concIdentIdent(curModelName, ConstId);
}

ptident NameManager::mkPresentId(ptident id)
{
  return exclusiveIdent(makeIdent_ident_ident(PresentId, id));
}

/** Constructor.
    Some identifiers are  reserved for code generators.
*/
NameManager::NameManager()
{
  idNil = getIdentNil();
  curModelName = idNil;
  contextSubExpr = idNil;
  opContextSubExpr = opnil;
  // reserved identifiers (code generators)
  insertInIdentNameTable(string2ident("_ii_"));
  insertInIdentNameTable(string2ident("_jj_"));
  insertInIdentNameTable(string2ident("_kk_"));
  insertInIdentNameTable(getMicroStepName());
  // reserved for flat mode printer.
  insertInIdentNameTable(getConstraintsId());

  TypeId = getTypeId();
  IndexId = getIndexId();
  ConstId = getConstId();
  LabelId = getLabelId();
  PreviousId = getPreviousId();
  ClockId = getClockId();
  ConditionId = getConditionId();
  PresentId = getPresentId();

}

/**
   Destructor.
*/
NameManager::~NameManager()
{
}
