/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_
/**
   Construtor of an interface for printing some text form from a tree.
   @param comBeg (in) text for the begining of a comment.
   @param comEnd (in) text for the end of a comment.
   @param comInCom (in) text for the begining of a comment in a comment.
   @param comInComEnd (in) text for the end of a comment in a comment.
   @param errSymbol (in) text to print on error.
   @param holoSymb (in) text to print when the depth of printing is reached.
*/
  intePrinter::intePrinter (ptrepstr comBeg, ptrepstr comEnd, ptrepstr comInCom, ptrepstr comInComEnd, ptrepstr errSymbol, ptrepstr holoSymb)
{
  comcomstr = comBeg;
  comendstr = comEnd;
  comincomstr = comInCom;
  combegstr = comInComEnd;
  errsymbol = errSymbol;
  symbolholo = holoSymb;
  setPrinterContext (pk_false);	// default.
  inteprint_initialize ();
}


intePrinter::~intePrinter ()
{
}

/** 
    It initialises the printer interface in "graphical" mode when bgraphic, or "textual" mode otherwise.
    @return the value of the mode before this call.
*/
boolean
intePrinter::setPrinterContext (boolean bgraphic)
{
  boolean x;
  x = graphicContext;
  // bgraphic = pk_false;
  if (bgraphic)
    commentlevel = 1;
  else
    commentlevel = 0;
  graphicContext = bgraphic;
  return x;
}

/** 
    @return the size (number of characters) of an identifier (p).
*/
int
intePrinter::sigidentsize (ptident p)
{
  int lbg, nbempty;

  lbg = identsize_getnbempty (p, 1, &nbempty);
  return lbg + nbempty;
}


/** It prints the text of the beginning of a comment. */
void
intePrinter::intedebcomprint (void)
{
  sigpristr (comcomstr);
}



/** It prints the text of the end of a comment. */
void
intePrinter::intefincomprint (void)
{
  sigpristr (comendstr);
}


/**
   @return "true" in graphical context, "false" otherwise.
   In graphical context, the comments are not printed.
*/
boolean
intePrinter::snflagon (void)
{
  return (boolean) (!graphicContext);
}



/**
   It prints the holophaste symbol. It is used when the depth of printing is achieved.
*/
void
intePrinter::printholo (void)
{
  sigpristr (symbolholo);
}



/**
   It prints the error symbol for an error in a tree structure.
*/
void
intePrinter::printerror (void)
{
  sigpristr (errsymbol);
}



/** It prints a keyword (tkw) and sets the insertion mode to "separate". */
void
intePrinter::printkey (int tkw)
{
  sigpristr (strTableGet (basekeyword, tkw));
  snglue (pk_false);
}



/* It prints a keyword (tkw) followed by a space.*/
void
intePrinter::printkeysp (int tkw)
{
  printkey (tkw);
  printspace ();
}

/* It prints a keyword (tkw) followed by a space.*/
void
intePrinter::printkeyspline (int tkw)
{
  printkeysp(tkw);
  snline ();
}



/* It prints a space before a keyword (tkw).*/
void
intePrinter::printspkey (int tkw)
{
  printspace ();
  printkey (tkw);
}



/** It prints a keyword (tkw). */
void
intePrinter::printalphakey (int tkw)
{
  inteconsstrprint (strTableGet (basekeyword, tkw));
}



/**
  It prints a keyword (tkw) preceded or not by a space character according to the current mode (concatenation or separation), 
  followed by a psace character.
*/
void
intePrinter::printalphakeysp (int tkw)
{
  separate ();
  sigpristr (strTableGet (basekeyword, tkw));
  printspace ();
}



/**
  It prints a keyword (tkw) preceded or not by a space character according to the current mode (concatenation or separation).
*/
void
intePrinter::printcspkey (int tkw)
{
  separate ();
  printkey (tkw);
}



/**
   It prints a space character followed by a keyword (tkw) and a space character.
*/
void
intePrinter::printspalphakey (int tkw)
{
  snspace ();
  sigpristr (strTableGet (basekeyword, tkw));
  printspace ();
}



/**
   It prints a keyword (tkw) followed by a space character and adds a mark (if level > 0).  
*/
void
intePrinter::printkeymark (int tkw, int level)
{
  printkey (tkw);
  snspace ();
  snmark (level);
}

/**
   It prints a space character before calling \ref printkeymark.
*/
void
intePrinter::printspkeymark (int tkw, int level)
{
  snspace ();
  printkeymark (tkw, level);
}



/**
  It prints a keyword (tkw) preceded or not by a space character according to the current mode (concatenation or separation).
  It is followed by a space character. Then, it adds a mark (if level > 0).
*/
void
intePrinter::printalphakeymark (int tkw, int level)
{
  printalphakeysp (tkw);
  snmark (level);
}



/** It prints a keyword (tkw) and adds a mark (if level > 0).
 */
void
intePrinter::printgluekeymark (int tkw, int level)
{
  printkey (tkw);
  snmark (level);
}



/**
   It removes a mark (if level > 0) and prints a keyword (tkw) preceded by a space character.
*/
void
intePrinter::printkeyrelease (int tkw, int level)
{
  snrelease (level);
  snspace ();
  printkey (tkw);
}

/**
   It removes a mark (if level > 0) and prints a keyword (tkw).
*/
void
intePrinter::printgluekeyrelease (int tkw, int level)
{
  snrelease (level);
  printkey (tkw);
}



/**
   It prints a keyword (tkw) and removes a mark (if level > 0).
*/
void
intePrinter::printreleasegluekey (int tkw, int level)
{
  printkey (tkw);
  snrelease (level);
}



/**
   It prints a keyword (tkw) at the beginning of a line (level>0), otherwise preceded by a space character.
*/
void
intePrinter::printlkey (int tkw, int level)
{
  if (level > 0)
    snline ();
  else
    snspace ();
  printkey (tkw);
}



/**
   It prints a keyword (tkw) at the beginning of a line (level>0), otherwise preceded by a space character.
   It is followed by a space character.
*/
void
intePrinter::printlkeysp (int tkw, int level)
{
  if (level > 0)
    snline ();
  else
    snspace ();
  printkeysp (tkw);
}



/**
   It increases  the number of attribute level (commentlevel).
   @param b0 (out) prefix of comments
   @param bend (out) postfix of comments
   @param b1 (out) prefix/postfix for a comment in a comment.
   @return "true" if there are overlaped comments, "false" otherwise
*/
boolean
intePrinter::open_unparsePrintableTreeAttribute (ptrepstr * b0, ptrepstr * bend, ptrepstr * b1)
{
  commentlevel = commentlevel + 1;
  if (commentlevel > 1)
    {
      *b0 = comincomstr;
      *bend = comendstr;
      return pk_true;
    }
  *b0 = comcomstr;
  *b1 = combegstr;
  *bend = comendstr;

  return pk_false;
}



/** 
    It decreases  the number of attribute level (commentlevel).
*/
void
intePrinter::close_unparsePrintableTreeAttribute (void)
{
  commentlevel = commentlevel - 1;
}



/**
   It prints a string (b0) followed by a new line if ol.
*/
void
intePrinter::reset_unparsePrintableTreeAttribute (ptrepstr b0, boolean ol)
{
  if (ol)
    {
      //      snline();
      sigwristr (b0);
      snline ();
    }
  else
    sigpristr (b0);
}

/**
  oMode : output mode.
  cbuf : buffer used for printing.
  cname : filename when oMode is file writing, nilstr otherwise (unused).
  effFileName (out) : when the output is a file, it contains the effective main (dir/cname) of the file, by default it is set to cname.
  @return true if the printing can be done, false otherwise.
*/
boolean
intePrinter::open_bufferMakePathName (ptrepbuf cbuf, ptrepstr cname, const char *fileSuff, ptrepstr * effFileName)
{
  return inteprint_open_bufferNewName (cbuf, cname, fileSuff, effFileName);
}

/**
  oMode : output mode.
  cbuf : buffer used for printing.
  cname : filename when oMode is file writing, nilstr otherwise (unused).
  effFileName (out) : when the output is a file, it contains the effective main (dir/cname) of the file, by default it is set to cname.
  @return true if the printing can be done, false otherwise.
*/
boolean
intePrinter::open_bufferDevice (ptrepbuf cbuf, printModeType oMode)
{
  if (oMode == ottyout)
    {
      inteprint_open_stdOut (cbuf, 0);
      return pk_true;
    }
  //  if (oMode == ottygraph) { inteprint_open_stdOut(cbuf, 1); return  pk_true;} obsolete mode.
  if (oMode == ographical)
    {
      inteprint_open_stdOut (cbuf, 2);
      return pk_true;
    }
  return pk_false;
}

/**
   It loads a table defined in the file \$Signal_bin/\<signame\>.kw.
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean
intePrinter::load_KWLangTable (const char *signame)
{
  return tabkwload (signame, &basekeyword);
}


/** Signal printing finalizing. */
void
intePrinter::intpriend (void)
{
}
