/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/
_PROTECTED_
/**
   It finalizes the compiling of the module represented by aModule:
     - labels cannot be declared in a module;
     - Signals cannot be declared in a module;
     - private models are removed.
*/
    _VIRTUAL_ void
 Trees2Kernel::moduleFinalize(ptrepfi aModule)
{
  ptlrepid loc, l;
  ntrees treeRef;
  ptrepfi vfi;
  integer nb_private, i, n;

  // Restrictions : labels/signals cannot be declared in a module
  if ((kl_getLabelsModel(aModule) != NILLREPID) || (kl_getLocalFlowsModel(aModule) != NILLREPID))
    {
      mkklErrorAnnotTree(errModuleWithLocalSignals, kl_getTreeReference(aModule));
      kl_setLabelsModel(aModule, NILLREPID);
    }

  // PRIVATE models are deleted.
  loc = kl_getLocModels(aModule);
  n = ptlrepid_size(loc);
  if (n != 0)
    {
      l = loc;
      nb_private = 0;
      for (i = 1; i <= n; i++)
	{
	  vfi = getModelRepid(l);
	  treeRef = kl_getTreeReference(vfi);
	  l = succid(l);
	  if (STree_isPrivateDeclaration(treeRef))
	    {
	      loc = removerepid(loc, i - nb_private);
	      nb_private++;
	    }
	}
      kl_setLocModels(aModule, loc);
      setDclInModule(loc, aModule);
    }
  setDclInModule(kl_getConstantsTypesModel(aModule), aModule);
}

/** 
    @return the sub-list of a list of models (lsModel) composed of the models annotated by the pragma "main", or 
    or all the models when it does not exist one annotated with the pragma.
*/
ptlrepid Trees2Kernel::selectMainProcesses(ptlrepid lsModel, ntrees x)
{
  integer i, n;
  ptrepfi aModel;
  ptlrepid l, list;
  ptpragma aPragma;
  ntrees treeDecl, root;
  ptklExprProc expfi;

  if (lsModel == NILLREPID)
    return NILLREPID;
  list = NILLREPID;
  l = lsModel;
  n = ptlrepid_size(l);
  for (i = 1; i <= n; i++)
    {
      aModel = getModelRepid(l);
      if (getPragmaKeyword(kl_getPragmaModel(aModel), PRAGMA_MAIN, &aPragma))
	list = addRepid(list, l);
      l = succid(l);
    }
  if (list != NILLREPID)
    return list;
  // il ne faut pas compiler les processus provenant des importations.
  l = lsModel;
  n = ptlrepid_size(l);
  for (i = 1; i <= n; i++)
    {
      aModel = getModelRepid(l);
      treeDecl = kl_getBodyModel(aModel, &expfi);
      if (STree_getRootTree(treeDecl, &root))
	if (root == x)
	  list = addRepid(list, l);
      l = succid(l);
    }

  if (list != NILLREPID)
    return list;
  return lsModel;
}

/**
   Compiling a Signal module.
   @return the kernel tree representation associated with a abstract tree module (aModule).
   @param aModule (in) a module represented as an abstract tree produced by the Signal syntactic analyzer.
   @param varid (out) stack associated with the name of the compiled module.
*/
_VIRTUAL_ ptrepfi Trees2Kernel::moduleCompiling(ntrees aModule, ptIdStack * varid)
{
  ntrees aProcess, ldeclarations;
  ptident id;
  ptrepfi vfi;
  boolean cc;

  // Using a process to apply the same algorithms.
  id = copy_ident(nchild(1, aModule));
  ldeclarations = STreeModel_getLocalDeclarations(aModule);	// Todo : take into account the pragmas ?
  aProcess = STree_mkSkeletonModel(id, opprocess);
  STree_setLocalDeclarations(aProcess, ldeclarations);
  cc = pK_messenger->getErrorCode();
  cSpec(aProcess, varid, &vfi);
  modelCompiling(vfi);
  kl_setBodyModel(vfi, NILEXPFIL);
  STree_setLocalDeclarations(aModule, ldeclarations);
  kl_setTreeReference(vfi, aModule);
  moduleFinalize(vfi);
  kl_setError(vfi, pK_messenger->getErrorCode());
  pK_messenger->setErrorCode(cc);
  pushModule(vfi);
  return vfi;
}

/**
   Compiling a list of Signal module (lmodules).
   @return the kernel tree representation of the compiled list.
   @param lmodules a list of modules represented as an abstract tree produced by the Signal syntactic analyzer.
   @param oneModule (out) "true" if the size of the list is equal to 1, "false" otherwise. 
   In fact, the Signal grammar accepts several modules in a file, but our implementation restricts to one 
   module by file.
*/
ptlrepid Trees2Kernel::modulesListCompiling(ntrees lmodules, boolean * oneModule)
{
  integer i, n;
  ptlrepid lvfi;
  ptrepfi vfi;
  ptIdStack varid;

  n = ntreearity(lmodules);
  (*oneModule) = (n == 1);
  lvfi = NILLREPID;
  for (i = 1; i <= n; i++)
    {
      vfi = moduleCompiling(nchild(i, lmodules), &varid);
      appendProcess(&lvfi, varid, vfi);
    }
  return lvfi;
}

/** Loading a Signal library from a file (strfilename).
    @return "true" if the operation succeeds, "false" otherwise.
*/
boolean Trees2Kernel::parseLibrary(ptrepstr strfilename, ntrees * module)
{
  boolean cc, b;

  cc = pK_messenger->getErrorCode();
  b = sysparser(without_phylum, repstr2string(strfilename), module);
  if (!b)
    pK_messenger->errorStr(errNotModule, strfilename);
  pK_messenger->setErrorCode(cc);
  return b;
}

/**
   Signal intrinsics.

   Pops the modules, stored in "moduleInstrinsics" field of the class, of Signal intrinsics.
*/
void Trees2Kernel::popIntrinsics()
{
  ptlrepid linstrinsics;
  integer n, i;

  if (moduleInstrinsics == NILREPFI)
    return;
  popModels(kl_getLocModels(moduleInstrinsics));
  linstrinsics = kl_getModules(moduleInstrinsics);

  n = ptlrepid_size(linstrinsics);
  for (i = 1; i <= n; i++)
    {
      popModels(kl_getLocModels(getModelRepid(linstrinsics)));
      linstrinsics = succid(linstrinsics);
    }
  moduleInstrinsics = NILREPFI;
}

/**
   Signal intrinsics.
   
   This method annotates the Signal intrinsics by a predefined attribute (see setInstrinsicAttribute()).
*/
void Trees2Kernel::AttributeIntrinsics()
{
  ptlrepid linstrinsics, l;
  integer n, i;

  /*   The predefined Signal intrinsics are previously compiled in loadIntrinsicLibrary() method 
     and stored in the "moduleInstrinsics" field of the class.
     The models are pushed during the compiling in the moduleCompiling() method. */

  if (moduleInstrinsics != NILREPFI)
    {
      linstrinsics = kl_getLocModels(moduleInstrinsics);
      setInstrinsicAttribute(linstrinsics);
      linstrinsics = kl_getModules(moduleInstrinsics);
      n = ptlrepid_size(linstrinsics);
      for (i = 1; i <= n; i++)
	{
	  l = kl_getLocModels(getModelRepid(linstrinsics));
	  setInstrinsicAttribute(l);
	  linstrinsics = succid(linstrinsics);
	}
    }
}

/**
   Signal intrinsics.

   Compiling a library of intrinsics (Signal module) represented as an abstract tree (vlib) produced by the Signal parser.
   It stores the result in the field moduleInstrinsics  of the Trees2Kernel class.
*/
_VIRTUAL_ void Trees2Kernel::loadIntrinsicLibrary(ntrees vlib)
{
  boolean nonvide, intrinsicLoading_old;
  ptlrepid lvfi;

  intrinsicLoading_old = intrinsicLoading;
  intrinsicLoading = pk_true;
  moduleInstrinsics = NILREPFI;

  lvfi = modulesListCompiling(vlib, &nonvide);
  moduleInstrinsics = getModelRepid(lvfi);	// The fisrt of the list.
  intrinsicLoading = intrinsicLoading_old;
}

#define suffixeLibrary1 ".LIB"
#define suffixeLibrary2 ".SIG"

/**
   @return "true" if it exists a file idNameFile.LIB or idNameFile.SIG using the SIGNAL_LIBRARY_PATH to found the absolute adress of this file.
   "false" otherwise.
   idNameFile.LIB is prior on idNameFile.SIG.
*/
boolean Trees2Kernel::getFileNameLibrary(ntrees idNameFile, ptrepstr * fileName)
{
  ptrepstr dirname;

  makeRepstr_ident_string(idNameFile, suffixeLibrary1, fileName);
  if (!sysGetDirnameLibrary(*fileName, &dirname, fileName))
    {
      makeRepstr_ident_string(idNameFile, suffixeLibrary2, fileName);
      if (!sysGetDirnameLibrary(*fileName, &dirname, fileName))
	{
	  pK_messenger->errorIdent(errModuleNotFound, idNameFile);
	  return pk_false;
	}
    }
  return pk_true;
}

/**
   Loading (compiling) the SIGNAL program (assumed to be a library module) defined in a file (idNameFile). 
   It uses the SIGNAL_LIBRARY_PATH to found the absolute adress of the file.
   It calls the Signal parser and modulesListCompiling() method to produce the kernel tree representation (aModule) 
   from the abstrat tree (produced by the parser).
   @return "true" if the operation succeeds, "false" otherwise.
   \note: (Restriction) the Signal grammar accepts several modules in a file, but our implementation restricts to one 
   module by file.
*/
_VIRTUAL_ boolean Trees2Kernel::loadLibraryFile(ntrees aImportedModule, ptrepfi * aModule)
{
  ptlrepid lvfi;
  ntrees treeModule;
  boolean b;
  ptrepstr fileName;

  if (!getFileNameLibrary(aImportedModule, &fileName))
    return pk_false;

  if (GUIContext || (loadedLibraries.find(fileName) == loadedLibraries.end()))
    {
      if (!parseLibrary(fileName, &treeModule))
	return pk_false;
      if (noper(treeModule) != oplmodu)
	{
	  pK_messenger->errorStr(errNotModule, fileName);
	  return pk_false;
	}

      loadedLibraries[fileName] = NILREPFI;	// en cours.

      lvfi = modulesListCompiling(treeModule, &b);
      if (!b)
	mkklWarning(warlmodurestrict);
      *aModule = getModelRepid(lvfi);	// The fisrt of the list.
      loadedLibraries[fileName] = *aModule;
    }
  else if (loadedLibraries[fileName] == NILREPFI)
    {
      // cycle in referenced libraries
      pK_messenger->warningident(warModuleCycle, (ptident) aImportedModule);
      return pk_false;
    }
  else
    {
      *aModule = loadedLibraries[fileName];
      pushModule(*aModule);
    }
  return pk_true;
}

/**
   Similar to loadLibraryFile(). It was also the entry point for compiling the "use M.P" syntatic struture which 
   allows to select a model (P) in a module (M). This strucutre has bee removed from Signal grammar.

   @param aImportedModule (in) the name of an imported object (module).
   @param  aModule (out) the kernel tree representation (aModule) produced by loadLibraryFile() method.
   @return "true" if the operation succeeds, "false" otherwise.
*/
_VIRTUAL_ boolean Trees2Kernel::importOfModule(ntrees aImportedModule, ptrepfi * aModule)
{
  noperators op;

  op = noper(aImportedModule);
  if ((op == opnom) || (op == opident))
    {
      if (loadLibraryFile(aImportedModule, aModule))
	{
	  if (!eq_idents(aImportedModule, kl_getNameModel(*aModule)))
	    pK_messenger->warningident(warModuleNotDefined, aImportedModule);
	  return pk_true;
	}
      return pk_false;
    }
  printf("DEBUG Trees2Kernel::importOfModule not yet 1, operator is %i\n", op);
  fflush(stdout);
  mkklErrorAnnotTree(errnotimple, aImportedModule);
  return pk_false;
}

/**
   Compiling a reference to modules (Signal Syntax : use M1, ..Mn).
   @param limports (in) list of names of modules (ie M1,...,Mn)
   @param lvfi (in/out) list of modules, in kernel tree representation, in which models are added taking account  visibility rules.
*/
_VIRTUAL_ void Trees2Kernel::importOfModules(ntrees limports, ptlrepid * lvfi)
{
  ntrees save;
  integer i, n;
  ptrepfi vfi;
  ptlrepid xlrepid;

  save = operande;
  n = ntreearity(limports);
  for (i = 1; i <= n; i++)
    if (importOfModule(nchild(i, limports), &vfi))
      appendrepfi(lvfi, NILPILE, vfi, &xlrepid);
  operande = save;
}
