/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_



/** The traceability mode of a grap (g) is set to m for the trace of the rewrittings. */
void
pKServicesAPI::setKeepTrace(ptDCG_basic g, boolean m)
{
  g->setKeepTrace(m);
}


/**
    It prints in the file foo_ABSTRAT.SIG, where foo is the name of the operand graph, the abstract tree of the interface abstraction of g.
    It calls produceInterfaceAbstraction(), then prints the abstraction in a Signal Textual form.
*/
void
pKServicesAPI::unparseAbstraction(ptDCG_poly g)
{
  ntrees treeclos;
  treeclos = g->produceInterfaceAbstraction();
  SignalUnparse(treeclos, infoclos);
}

/** 
    Entry point from SignalGUI, InteEclipse  
    It applies DCG_basic::unparse() method to g. It takes into account the context call.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::unparse(ptDCG_basic g)
{
  ptDCG_basic g1;
  boolean b;

  g1 = mkEnvForUnparse(g);
  b = g1->unparse();
  resetEnvForUnparse(g, g1);
  return b;
}

/** 
    It applies DCG_basic::unparse() method to g. It takes into account the context call.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::unparse(ptDCG_basic g, const char *optionSuf)
{
  ptDCG_basic g1;
  boolean b;
  g1 = mkEnvForUnparse(g);
  b = g1->unparse(optionSuf);
  resetEnvForUnparse(g, g1);
  return b;
}

/**
   It performs the translation of a graph with a tree of boolean clocks to a flat form 
   in which there is a tree of boolean clocks reduces to one level.
   When btra is "true", the resulting graph of a transformation is unparsed.

   It applies 
   - DCGPoly2DCGBool)
   - DCG_poly::nonProductiveSignalElimination() 
   - DCG_flat::flattenCH() 

   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::DCGPoly2DCGFlat(ptDCG_poly vgraph, ptDCG_flat * gf, boolean btra)
{
  ptDCG_bool gb;

  if (!DCGPoly2DCGBool(vgraph, pk_true, &gb, btra))
    return pk_false;
  gb->nonProductiveSignalElimination();
  if (!DCG_flat::flattenCH(gb, gf))
    return pk_false;
  (*gf)->nonProductiveSignalElimination();
  if (btra)
    unparse(gb);
  return pk_true;
}

/**
   @return the result of DCG_basic::breakCPG() method applied to g. 
*/
boolean pKServicesAPI::breakCPG(ptDCG_basic g)
{
  return g->breakCPG();
}

/**
   Profiling operation for the operand graph (gPoly).
   It applies loadProfilingModule() then mkMorphism() 
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::applyTabularMorphism(ptDCG_poly gPoly)
{
  ptDCG_bool gBool;
  boolean b;

  if (!loadProfilingModule())
    return pk_false;
  if (!DCGPoly2DCGBool(gPoly, pk_true, &gBool, pk_false))
    return pk_false;
  b = mkMorphism(gBool, ht);
  delete gBool;
  return b;
}

/**
   This is the entry point of the code generation for SYnDEx tool.
   This  method can be applied to a graph of a module (reduced to one processes) or a process.
   It generates, using the SyndexService() service, the code for the Syndex code, for the operand DCG_flat graph.
   @return "true" if the operation succeeds, "false" otherwise.

   \see {http://www-rocq.inria.fr/syndex}
*/
boolean pKServicesAPI::export_SynDEx(ptDCG_flat g, boolean asMain)
{
  ptDCG vDCG;
  ptDCG_flat gf;
  ptlDCG lg;
  ptSyndexService gen;
  boolean b;

  if (hasUnsolvedClockConstraints(g))
    return pk_false;
  // taking the graph of a module when it is reduced to one graph.
  b = pk_true;
  if (g->getKind() == opmodu)
    {
      g->getModuleGraphs(&lg);
      ptlPointer_get_elemNext(&lg, (pointer *) & vDCG);
      if (lg != NILLDCG)
	b = pk_false;
      gf = new DCG_flat();
      gf->setDCG(vDCG);
    }
  else
    gf = g;
  if (b)
    {
      b = pk_false;
      gen = new SyndexService(gf, asMain);
      if (gen->translation())
	if (gen->produce_tree())
	  {
	    gen->print();
	    b = pk_true;
	  }
      delete gen;
    }
  return b;
}

/**
   This is the entry point of the code generation for Geneauto/P.
   This generator has been removed in Nov 2017. 
   It was specific to the P language.
*/
boolean pKServicesAPI::export_Geneauto(ptDCG_basic g)
{
  printf ("##INFO## Signal to P (Geneauto is not implemented in this version).\n");
  return false;
}

/** 
    It applies DCG_basic::buildClockProcessInterface() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::buildClockProcessInterface(ptDCG_basic g)
{
  return g->buildClockProcessInterface();
}

/** 
    It applies DCG_basic::nonProductiveSignalElimination() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::nonProductiveSignalElimination(ptDCG_basic g)
{
  return g->nonProductiveSignalElimination();
}

/** 
    It applies DCG_basic::removeSignalClones() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::removeSignalClones(ptDCG_basic g)
{
  return g->removeSignalClones();
}

/** 
    It applies DCG_basic::substituteExpr2Var() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::substituteExpr2Var(ptDCG_basic g, integer maxRef)
{
  return g->substituteExpr2Var(maxRef);
}

/** 
    It applies DCG_poly::buildIOClockSubHierarchy() method to g.
    if bClockOfDependences it takes into account the clock of the I/O dependencies.

    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::buildIOClockSubHierarchy(ptDCG_poly g, boolean bClockOfDependences)
{
  return g->buildIOClockSubHierarchy(bClockOfDependences);
}

/** 
    It applies DCG_basic::partitionningRunOn() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::partitionningRunOn(ptDCG_basic g, ptlINode * ln)
{
  return g->partitionningRunOn(ln);
}

/** 
    It applies DCG_poly::computeIOTransitiveClosure() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::computeIOTransitiveClosure(ptDCG_basic g)
{
 ptDCG_basic g1;
 boolean b;

  g1 = mkEnvForUnparse(g);
  b = g1->computeIOTransitiveClosure();
  resetEnvForUnparse(g, g1);
  return b;
}

/** 
    It applies DCG_basic::isStaticCycleFree() method to g.
    @param bPrintMessOnCycle (in) an error is signaled when it is pk_true.
    @param bKeepCycles (in) when it is "true" he cycles are kept in the Graph structure.
    The cycles are unparsed if bKeepCycles.
    @return "true" if the operation succeeds (no cycles), "false" otherwise.
 */
boolean pKServicesAPI::isStaticCycleFree(ptDCG_poly g, boolean bPrintMessOnCycle, boolean bKeepCycles)
{
  boolean b;

  b = g->isStaticCycleFree(bPrintMessOnCycle, bKeepCycles);
  if (bKeepCycles && !b)
    g->unparseCycles();
  return b;
}

/** 
    It applies DCG_poly::phylumClustering() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
 */
boolean pKServicesAPI::phylumClustering(ptDCG_poly g)
{
  return g->phylumClustering();
}

/** 
    @return the result of DCG_basic::produce_tree() method applied to g.
 */
ntrees pKServicesAPI::produce_tree(ptDCG_basic g)
{
  return g->produce_tree();
}

/** 
    @return the result of DCG_basic::produce_tree() method applied to g.
 */
ntrees pKServicesAPI::produce_tree(ptDCG_basic g, char *optionSuf)
{
  return g->produce_tree(optionSuf);
}

/** 
    Similar to pKServicesAPI::produce_tree() but here the hierarchy is flattened and the expressions are decomposed in elementary ones if kernelMode.
    @return the result of DCG_basic::produce_tree() method applied to g.
 */
ntrees pKServicesAPI::produceFlat(ptDCG_basic g, boolean kernelMode)
{
  boolean b, b1;
  ntrees vt;

  b = pK_Graph2SignalTreeBuilder->setFlatMode(pk_true);
  if (kernelMode)
    b1 = pK_Graph2SignalTreeBuilder->setKernelExprMode(pk_true);
  vt = mkEnvForUnparse(g)->produce_tree();
  if (kernelMode)
    b1 = pK_Graph2SignalTreeBuilder->setKernelExprMode(b1);
  b = pK_Graph2SignalTreeBuilder->setFlatMode(b);
  return vt;
}

/** 
    @return the result of DCG_basic::produceSubgraph() method applied to g.
 */
ntrees pKServicesAPI::produceSubgraph(ptDCG_basic g, ptINode aSubGraph)
{
  return g->produceSubgraph(aSubGraph);
}

/** 
    @return the result of DCG_poly::produceInterfaceAbstraction() method applied to g.
 */
ntrees pKServicesAPI::produceInterfaceAbstraction(ptDCG_poly g)
{
  return g->produceInterfaceAbstraction();
}

/**
    It applies DCG_poly::computeClockExclusions() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
*/
void pKServicesAPI::computeClockExclusions(ptDCG_poly g, integer vlevel)
{
  ntrees t;
  g->computeClockExclusions(vlevel);
  t = g->produceExclusiveClocks();
  SignalUnparse(t, infoExclusiveClocks);
}

/** 
    It applies DCG_basic::moveIn2OutDelays() method to g.
    @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::moveIn2OutDelays(ptDCG_basic g)
{
  return g->moveIn2OutDelays();
}

/** 
    @return the result of DCG_poly::separateBooleansOthers() method  applied to g.
*/
boolean pKServicesAPI::separateBooleansOthers(ptDCG_poly g, ptINode * booleans, ptINode * others)
{
  return g->separateBooleansOthers(booleans, others);
}

/** 
    @return the result of DCG_poly::separateStatesOthers() method  applied to g.
*/
boolean pKServicesAPI::separateStatesOthers(ptDCG_poly g, ptINode * states, ptINode * others)
{
  return g->separateStatesOthers(states, others);
}

/** 
    @return the result of DCG_basic::hasUnsolvedClockConstraints() method  applied to g.
*/
boolean pKServicesAPI::hasUnsolvedClockConstraints(ptDCG_basic g)
{
  return g->hasUnsolvedClockConstraints();
}

/**
   @return the result of DCG_basic::checkExplicitTyping() method  applied to g.
*/
boolean pKServicesAPI::checkExplicitTyping(ptDCG_basic g)
{
  return g->checkExplicitTyping();
}

/**
   @return the result of DCG_basic::setClockDependences() method  applied to g.
*/
boolean pKServicesAPI::setClockDependences(ptDCG_basic g)
{
  return g->setClockDependences();
}

/**
   @return the result of DCG_poly::setIntrinsicClocks() method applied to g.
*/
boolean pKServicesAPI::setIntrinsicClocks(ptDCG_poly g)
{
  return g->setIntrinsicClocks();
}

/**
   @return the result of DCG_bool::addExceptionsForConstraints() method applied to g. 
*/
boolean pKServicesAPI::addExceptionsForConstraints(ptDCG_bool g)
{
  return g->addExceptionsForConstraints();
}

/**
   @return the result of DCG_basic::expressionExpand() method applied to g. 
*/
boolean pKServicesAPI::expressionExpand(ptDCG_basic g)
{
  return g->expressionExpand();
}

/**
   @return the result of DCG_basic::changeEnumerated2Integers() method applied to g. 
*/
boolean pKServicesAPI::changeEnumerated2Integers(ptDCG_basic g)
{
  return g->changeEnumerated2Integers();
}

/**
   @return the result of DCG_basic::splitDefaultDefinitions() method applied to g. 
*/
boolean pKServicesAPI::splitDefaultDefinitions(ptDCG_basic g)
{
  return g->splitDefaultDefinitions();
}

/**
   @return the result of DCG_basic::splitDataStructure() method applied to g. 
*/
boolean pKServicesAPI::splitDataStructure(ptDCG_basic g)
{
  return g->splitDataStructure();
}

/**
   @return the result of DCG_basic::dependenceReinforcing() method applied to g. 
*/
boolean pKServicesAPI::dependenceReinforcing(ptDCG_basic g)
{
  return g->dependenceReinforcing();
}

/**
   @return the result of DCG_basic::transformModelReferences2NodeReferences() method applied to g. 
*/
boolean pKServicesAPI::transformModelReferences2NodeReferences(ptDCG_basic g)
{
  return g->transformModelReferences2NodeReferences();
}

/**
   @return the result of DCG_basic::groupPartialDefinitions() method applied to g. 
*/
boolean pKServicesAPI::groupPartialDefinitions(ptDCG_basic g)
{
  return g->groupPartialDefinitions();
}

/**
   @return the result of DCG_basic::splitDefaultDefinitions() method applied to g. 
*/
boolean pKServicesAPI::normalDefaults(ptDCG_basic g)
{
  return g->normalDefaults();
}


/**
   @return the result of DCG_basic::printCHStatistics() method applied to g. 
*/
boolean pKServicesAPI::printCHStatistics(ptDCG_basic g)
{
  return g->printCHStatistics();
}

/**
   It computes the assignement clock of the operand graph (g).
*/
void pKServicesAPI::computeAssignmentClock(ptDCG_poly g)
{
  g->computeAssignmentClock();
  clockCalculus(g);
}

/**
   It computes the assignement clock of the operand graph (g).
*/
void pKServicesAPI::computeClockWithoutDefaultConstantValue(ptDCG_poly g)
{
  g->computeClockWithoutDefaultConstantValue();
  clockCalculus(g);
}

/**
   It computes the productive clock (utility clocks) of the operand graph (g).
*/
void pKServicesAPI::computeProductiveClock(ptDCG_poly g)
{
  g->computeProductiveClock();
  clockCalculus(g);
}

/**
   It optimizes the operand graph (g) using the productive clocks (computed by the computeProductiveClock) method.
   The nodes are moved to their productive clock.
*/
void pKServicesAPI::rewUsingProductiveClock(ptDCG_poly g)
{
  g->rewUsingProductiveClock();
  clockCalculus(g);
}

/**
   @return the result of DCG_basic::removeSignalEqualities() method applied to g. 
*/
boolean pKServicesAPI::removeSignalEqualities(ptDCG_basic g)
{
  return g->removeSignalEqualities();
}

/**
   @return the result of DCG_poly::setDelayMemoryOverloading() method applied to g. 
*/
boolean pKServicesAPI::setDelayMemoryOverloading(ptDCG_poly g)
{
  return g->setDelayMemoryOverloading(pk_false); // a revoir Loïc
}

/**
   @return the result of DCG_basic::transformBooleanClocks2Events() method applied to g. 
*/
boolean pKServicesAPI::transformBooleanClocks2Events(ptDCG_basic g)
{
  return g->transformBooleanClocks2Events();
}

/**
   @return the result of DCG_poly::delayTreeBuild() method applied to g. 
*/
boolean pKServicesAPI::delayTreeBuild(ptDCG_poly g)
{
  return g->delayTreeBuild();
}

/**
   @return the result of DCG_basic::expandModelReference() method applied to g. 
*/
boolean pKServicesAPI::expandModelReference(ptDCG_basic g, boolean bAsSubGraph)
{
  return g->expandModelReference(bAsSubGraph);
}

/** 
    It applies DCG_poly::memoryOptimizations() on vgraph, parametrized by bdelayMultiplexing.
*/
void pKServicesAPI::memoryOptimizations(ptDCG_poly vgraph, boolean bdelayMultiplexing)
{
  vgraph->memoryOptimizations(bdelayMultiplexing);
}

/**
   Macro definition: boolean/others partitionning applied to a graph (vgraph).\n
   It applies 
   - DCG_poly::separateBooleansOthers() on vgraph that  separates the boolean nodes from the other nodes of the graph
   - then binPartitionningEnd().
*/
void pKServicesAPI::separateBooleansOthers(ptDCG_poly vgraph)
{
  ptINode booleansSubGraph, othersSubGraph;

  vgraph->separateBooleansOthers(&booleansSubGraph, &othersSubGraph);
  vgraph->subGraphFlattenning(booleansSubGraph);
  vgraph->buildClockProcessInterface();
  vgraph->nodeGroupingByClocks(othersSubGraph);
}

/**
   Macro definition: boolean/others partitionning applied to a graph (vgraph).\n
   It applies 
   - DCG_poly::separateBooleansOthers() on vgraph that  separates the boolean nodes from the other nodes of the graph
   - then binPartitionningEnd().
*/
void pKServicesAPI::separateBooleansOthers_old(ptDCG_poly vgraph)
{
  ptINode booleansSubGraph, othersSubGraph;

  vgraph->separateBooleansOthers(&booleansSubGraph, &othersSubGraph);
  binPartitionningEnd(vgraph, booleansSubGraph, othersSubGraph);
  exportPartition(vgraph, booleansSubGraph);
  exportPartition(vgraph, othersSubGraph);
}

/**
   Macro definition: booleans+integers/others partitionning applied to a graph (vgraph).\n
   It applies 
   - DCG_poly::separateBooleansIntegersOthers() on vgraph that separates the nodes with boolean or integer types from the other nodes of the graph
   - then binPartitionningEnd().
*/
void pKServicesAPI::separateBooleansIntegersOthers(ptDCG_poly vgraph)
{
  ptINode boolintSubGraph, othersSubGraph;

  vgraph->separateBooleansIntegersOthers(&boolintSubGraph, &othersSubGraph);
  binPartitionningEnd(vgraph, boolintSubGraph, othersSubGraph);
}

/**
   Macro definition: Finite types/others partitionning applied to a graph (vgraph).\n
   It applies 
   - DCG_poly::separateBooleansIntegersOthers() on vgraph that separates the nodes with a finite type (bool+int+enumerated) from the other nodes of the graph
   - then binPartitionningEnd().
*/
void pKServicesAPI::separateFiniteTypesOthers(ptDCG_poly vgraph)
{
  ptINode finiteTypesSubGraph, othersSubGraph;

  vgraph->separateFiniteTypesOthers(&finiteTypesSubGraph, &othersSubGraph);
  binPartitionningEnd(vgraph, finiteTypesSubGraph, othersSubGraph);
}

/**
   Macro definition: states/others partitionning applied to a graph (vgraph).\n
   It applies 
   - DCG_poly::separateStatesOthers() on vgraph that separates the state nodes from the other nodes of the graph
   - then binPartitionningEnd()
*/
void pKServicesAPI::separateStatesOthers(ptDCG_poly vgraph)
{
  ptINode statesSubGraph, othersSubGraph;

  vgraph->separateStatesOthers(&statesSubGraph, &othersSubGraph);
  binPartitionningEnd(vgraph, statesSubGraph, othersSubGraph);
  exportPartition(vgraph, statesSubGraph);
  exportPartition(vgraph, othersSubGraph);
}

/** 
    Macro definition: separated compiling for a graph (vgraph).\n
    It unparses, in the file foo_ABSTRACTION.SIG where foo is the name of the graph, the abstraction of a graph (vgraph).\n
    It applies the following methods:
     - separatedCompilingCommon()
     - DCG_poly::buildIOClockSubHierarchy()
     - DCG_poly::nonProductiveSignalElimination()
*/
boolean pKServicesAPI::InterfaceAbstraction(ptDCG_poly vgraph)
{
  vgraph->buildInterfaceAbstraction();
  unparseAbstraction(vgraph);
  return pk_true;
}

/** 
    @return the DCG_poly graph of the operand graph(basicGraph).
    The solving is applied with vd for the depth of resolution (boolean reductions).
*/
boolean pKServicesAPI::DCGBasic2DCGPoly(ptDCG_basic basicGraph, integer vd, ptDCG_poly * polyGraph)
{
  (*polyGraph) = DCG_poly::DCG_basic2poly(basicGraph, vd);
  setIntrinsicClocks(*polyGraph);
  if (!(*polyGraph)->delayTreeBuild())
    return pk_false;
  if (!(*polyGraph)->setArrayClocks())
    return pk_false;
  (*polyGraph)->setComposedNodesDependences();
  return pk_true;
}

/**
   It produces the endochronous DCG (gendo) of the poly DCG (polyGraph).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::DCGPoly2DCGEndo(ptDCG_poly polyGraph, ptDCG_endo * gendo)
{
  boolean modified;
  return DCG_endo::parameterizeMasterClocks(polyGraph, &modified, gendo);
}

/**
   It produces the boolean DCG (gBool) of the endo DCG (endoGraph).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::DCGEndo2DCGBool(ptDCG_endo endoGraph, boolean mkSubForNosub, ptDCG_bool * gBool)
{
  return DCG_bool::transformToBooleanControl(endoGraph, mkSubForNosub, gBool);
}

/**
   It produces the flat DCG (gFlat) of the bool DCG (boolGraph).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::DCGBool2DCGFlat(ptDCG_bool boolGraph, ptDCG_flat * gFlat)
{
  return DCG_flat::flattenCH(boolGraph, gFlat);
}

/* Pseudo cycles ? */
void pKServicesAPI::analyzeStaticCycles(ptDCG_endo gendo)
{
  if (!(gendo)->isStaticCycleFree(pk_false, pk_true))
    {
      gendo->analyzeCycles();	// detection of the "null clock cycles"
      gendo->setIntrinsicClocks();
      gendo->unparseCycles();
    }
}

/**
   It produces the boolean DCG (gBool) of the poly DCG (polyGraph).
   When btra is "true", the resulting graph of a transformation is unparsed.
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::DCGPoly2DCGBool(ptDCG_poly polyGraph, boolean mkSubForNosub, ptDCG_bool * gBool, boolean btra)
{
  ptDCG_endo gEndo;
  boolean b;

  if (!DCGPoly2DCGEndo(polyGraph, &gEndo)) return pk_false;
  if (btra) unparse(gEndo, "");
  b = DCG_bool::transformToBooleanControl(gEndo, mkSubForNosub, gBool);
  if (btra) unparse(*gBool, "");
  delete gEndo;
  return b;
}

/**
    This is the entry point of the code generation for the SIGALI prover.
    The code is generated in the file foo.z3z (foo is the name of the process).
    Moreover, if some specific intrinsic functions of SIGALI (see library $SignalLib_Sigali/SIGALI.SIG),
    are referenced in the program, the file foo_CMD.z3z is generated. It contains : 
    - some initialization commands of SIGALI tool and 
    - the set of properties that have to be proved and/or synthesized.
    \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean pKServicesAPI::SigaliPrintExtraction(ptDCG_basic g, boolean btrace)
{
  ptSigaliService gen;
  boolean b;

  gen = new SigaliService(g);
  b = gen->translation(btrace);
  if (b) gen->print();
  delete gen;
  return pk_true;
}

/**
    This is the entry point of the code generation for the SMT format..
    The code is generated in the file foo_decls.smt and foo_Defs.smt (foo is the name of the process).
    \warning (\e restriction) This method can be applied to a graph of a process, not to a graph of a module.
*/
boolean pKServicesAPI::SMTPrintExtraction(ptDCG_basic g, boolean genInterLevelTrace, boolean btrace)
{
  ptSMTService gen;
  ptDCG_basic g1;
  boolean b;

  g1 = mkEnvForUnparse(g);
  g1 ->setClocksName();
  gen = new SMTService(g);
  b = gen->translation(genInterLevelTrace, btrace);
  if (b) gen->print();
  delete gen;
  resetEnvForUnparse(g, g1);
  return pk_true;
}


/**
    This is the entry point of the code generation for the Fiacre export (access to Tina, Cadp,...).
    The code is generated in the file foo.fcr (foo is the name of the process).
*/
boolean pKServicesAPI::FiacrePrintExtraction(ptDCG_poly g)
{
  ptFiacreService gen;
  boolean b;

  gen = new FiacreService(g);
  b = gen->translation();
  if (b) gen->print();
  delete gen;
  return b;
}

/**
   Macro definition.\n
   @param vgraph (in) the graph on which the operation is applied. It is assumed without circuit.
   @param btotally (in) must be pk_true for dynamic schedulling (cluster+threads), "false" otherwise.
   @param balwaysCode (in) must be pk_true when the sequentializing operation is asked even if the graph contains constraints.
   @param bcheck (in)  with the value "true", dynamical tests are added for constraints.
   @param bdelayMultiplexing (in) with the value 
   - "true", all the delays (except for the only one delay on x, zx :=x $1, for x a scalar) built on the same signal are multiplexed.
   - "false", if there is only one delay on a signal x, zx :=x $1, x and zx are not multiplexed (whatever its type is).

   @return the value "true" if the graph is sequentialized, "false" otherwise.

   It applies:
   - DCG_bool::checkExplicitTyping() that verifies that the graph is ready for code generation (signals are well-typed, ...). 
   - DCG_bool::memoryOptimizations() 
   - and finally,  DCG_sch::totallyOrder() that sequentializes the nodes of the graph.   
*/
boolean pKServicesAPI::Serializing(ptDCG_bool vgraph, boolean btotally, boolean balwaysCode, boolean bcheck, boolean bdelayMultiplexing, ptDCG_sch * gSeq)
{
  if (!vgraph->checkExplicitTyping())
    if (!balwaysCode) return pk_false;
  vgraph->memoryOptimizations(bdelayMultiplexing);
  return DCG_sch::Ordering(vgraph, btotally, bcheck, gSeq);
}

/** 
    @return true if the operation succeeds,"false" otherwise.
    It applies the clockcalculus for a DCGPoly graph.
*/
boolean pKServicesAPI::clockCalculus(ptDCG_poly polyGraph)
{
  polyGraph->ClockCalculus();
  return setIntrinsicClocks(polyGraph);
}

/**
   It generates C code for a DCGSch graph (gSeq). tgencode specifies the type of code generation:
   - 1 : sequential code generation
   - 2 : sequential code generation with tasks
   - 3 : sequential code generation for separated compiling
   - 4 : distributed code generation.
*/
boolean pKServicesAPI::genCodeC(ptDCG_sch gSeq, integer tgencode)
{
  ptCCodeGenerator aCCodeGenerator;

  aCCodeGenerator = new CCodeGenerator(gSeq);
  if (tgencode == 1)
    aCCodeGenerator->seq(pk_true, pk_true, pk_true);
  else if (tgencode == 2)
    aCCodeGenerator->tasks();
  else if (tgencode == 3)
    aCCodeGenerator->separate();
  else if (tgencode == 4)
    aCCodeGenerator->dist();
  else;
  delete aCCodeGenerator;
  return pk_true;
}

/**
   Similar to pKServicesAPI::genCodeC() for C++ language.
*/
boolean pKServicesAPI::genCodeCPP(ptDCG_sch gSeq, integer tgencode)
{
  ptCPPCodeGenerator aCPPCodeGenerator;

  aCPPCodeGenerator = new CPPCodeGenerator(gSeq);
  if (tgencode == 1)
    aCPPCodeGenerator->seq(pk_true, pk_true, pk_true);
  else if (tgencode == 2)
    aCPPCodeGenerator->tasks();
  else if (tgencode == 3)
    aCPPCodeGenerator->separate();
  else if (tgencode == 4)
    aCPPCodeGenerator->dist();
  else;
  delete aCPPCodeGenerator;
  return pk_true;
}

/**
   Similar to pKServicesAPI::genCodeC() for Java language.
*/
boolean pKServicesAPI::genCodeJava(ptDCG_sch gSeq, integer tgencode)
{
  ptjavaCodeGenerator aJavaCodeGenerator;

  aJavaCodeGenerator = new javaCodeGenerator(gSeq);
  if (tgencode == 1)
    aJavaCodeGenerator->seq(pk_true, pk_true, pk_true);
  else if (tgencode == 2)
    aJavaCodeGenerator->tasks(pk_true, pk_true, pk_true);	// different C/CPP, why ?
  else if (tgencode == 3)
    aJavaCodeGenerator->separate();
  else if (tgencode == 4)
    aJavaCodeGenerator->dist();
  else;
  delete aJavaCodeGenerator;
  return pk_true;
}

/**
   It generates the Lustre program that represents the Signal program characterized by its graph (g).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::export_Lustre(ptDCG_bool g)
{
  ptLustreService gen;
  boolean b;

  gen = new LustreService(g);
  b = pk_false;
  if (gen->translation())
    if (gen->produce_tree())
      {
	gen->print();
	b = pk_true;
      }
  delete gen;
  return b;
}

/**
   It generates the dot representation of the Signal program characterized by its graph (g).
   @return "true" if the operation succeeds, "false" otherwise.
*/
boolean pKServicesAPI::export_Dot(ptDCG_basic g)
{
  ptDotService gen;
  boolean b;

  gen = new DotService(g);
  b = pk_false;
  if (gen->translation())
    if (gen->produce_tree())
      {
	gen->print();
	b = pk_true;
      }
  delete gen;
  return b;
}

/**
   Using the utility clocks and the assignment clocks...
   Creates a container for the reduction of the communications of the signals exchanged between the operand list of subgraphs(lSubgraph).
   TO BE COMPLETED.
*/
boolean pKServicesAPI::reducingCommunications(ptDCG_endo gEndo, ptlINode lSubgraph, boolean bOnlyAssignmentClock)
{
  ptlINode ln;
  ptINode aSubGraph;
  boolean bres;

  bres = pk_true;
  computeAssignmentClock(gEndo);
  computeClockWithoutDefaultConstantValue(gEndo);
  computeProductiveClock(gEndo);

  // Computing the utility clock of the signals for each subgraph (a signal may be referenced in several sub-graphs).
  ln = lSubgraph;
  while (ptlINode_get_elem(&ln, &aSubGraph))
    gEndo->computeProductiveClock(aSubGraph);
  clockCalculus(gEndo);

  /** Reducing the communications using the utility clocks and the assigment clocks. */
  ln = lSubgraph;
  while (bres && ptlINode_get_elem(&ln, &aSubGraph))
    if (!gEndo->reduceComm(aSubGraph, bOnlyAssignmentClock))
      bres = pk_false;

  clockCalculus(gEndo);
  buildClockProcessInterface(gEndo);

  return bres;
}




/**
   @return the result of DCG_poly::buildAbstraction() method applied to g. 
   It must not be used (PRESENTLY IN TEST).
*/
boolean pKServicesAPI::buildAbstraction(ptDCG_poly g, ptDCG_poly * gResult)
{
  return g->buildAbstraction(gResult);
}



/**
   Similar to pKServicesAPI::genCodeC() for VHDL language.
*/
boolean pKServicesAPI::genCodeVHDL(ptDCG_sch gSeq)
{
 ptvhdlCodeGenerator avhdlCodeGenerator;

  avhdlCodeGenerator = new vhdlCodeGenerator(gSeq);
  avhdlCodeGenerator->seq(pk_true, pk_true, pk_true);
  return pk_true;
}


/**
   Similar to pKServicesAPI::genCodeC()  for Verilog language.
*/
boolean pKServicesAPI::genCodeVerilog(ptDCG_sch gSeq)
{
  ptVLCodeGenerator aVLCodeGenerator;

  aVLCodeGenerator = new VLCodeGenerator(gSeq);
  aVLCodeGenerator->seq(pk_true, pk_true, pk_true);
  return pk_true;
}
