/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#define TRACE(s)

_PROTECTED_
/**
   It prints an external representation (Signal code) of a sub-graph (aSubGraph) of a graph (vgraph).
*/
  void
pKServicesAPI::printSubGraph (ptDCG_basic vgraph, ptINode aSubGraph)
{
  ntrees Process;
  Process = vgraph->produceSubgraph (aSubGraph);
  SignalUnparse (Process, infosubgraph);
}

/**
   It prints an external representation (Signal code) of a list of sub-graphs (lSubGraph) of a graph (vgraph).
   It calls printSubGraph() for each.
*/
void
pKServicesAPI::printSubGraphs (ptDCG_basic vgraph, ptlINode lSubgraph)
{
  ptINode aSubGraph;
  while (ptlINode_get_elem (&lSubgraph, &aSubGraph))
    printSubGraph (vgraph, aSubGraph);
}




/**
   Specific to the morphism functionality.
   @return 
   - "true" if the internal file "MORPHISM.table" is found in the environment. It defines the translation (morphism) associated with the Signal operators.
   aModule is the graph of this module, one graph by operator.
   - "false" otherwise, aModule is not modified
*/
boolean pKServicesAPI::mkMorphismModule (ptDCG_poly * aModule)
{
  integer status;
  boolean b, x, saveB;
  char * fileNameSrc, * fileNameSrcC;
  ptlDCG lg;
  ptDCG gDCG;
  ptDCG_basic gb;
  ptDCG_poly gPoly;

  // if (GUIContext) return pk_false;   /* GUI -- A completer -- */  
  fileNameSrc = (char *) "MORPHISM.table";
  if (!existFile (fileNameSrc, &fileNameSrcC))
    {
      pK_messenger->globalMessageStr (errFileNotFound, string2repstr (fileNameSrc));
      return pk_false;
    }
  pK_messenger->globalMessageStr (infoLoading, string2repstr (fileNameSrc));
  x = pK_messenger->setSilentModeMessagesTo (pk_true);
  saveB = BModifyOutputDirectory;
  BModifyOutputDirectory = pk_false;
  status = SIGNALFiles2DCG (fileNameSrcC, (char *) "", (char *) "", pk_false, &gDCG);
  BModifyOutputDirectory = saveB;
  b = (status == polychronyNoError);
  if (b)
    if (!gDCG->getModuleGraphs (&lg))
      {
	x = pK_messenger->setSilentModeMessagesTo (x);
	pK_messenger->globalMessageStr (errFileNotaModule, string2repstr (fileNameSrc));
	b = pk_false;
      }
    else
      {
	gb = new DCG_basic (gDCG); // (gDCG, pk_true);
	gPoly = DCG_poly::DCG_basic2poly (gb);
	gPoly->setIntrinsicClocks ();
	x = pK_messenger->setSilentModeMessagesTo (x);
	(*aModule) = gPoly;
      }
  return b;
}

/**
   Specific to the morphism functionality.
   It adds some use of libraries
    - use STD, COSTBIB, Cost_Context"
    - use XXX, COSTBIB, Cost_Context" where XXX is the value of the pragma "PROCESSOR_TYPE" associated with a graph (og)
    .
    in a tree (hierarchy). 

    Assumed : hierarchy is the result of the mophism operation (profiling) applied to the graph og.
*/
void
pKServicesAPI::addLibraryProfiling (ptDCG_bool og, ntrees hierarchy)
{
  ntrees usedLib, lusedLib, ContextModule, lids, commonLib;
  ptrepstr fn;
  ptident idContextModule;
  ptpragma valPragma;
  ptlpragmas prags;

  prags = og->getPragmas ();
  if (getPragmaKeyword (prags, PRAGMA_PROCESSOR_TYPE, &valPragma))
    usedLib = repstr2ident (strFirstStmtsValue (valPragma));
  else
    usedLib = string2ident ("STD");
  commonLib = string2ident ("COSTBIB");
  lusedLib = LL_make_limpo_list (commonLib);
  lusedLib = ntreeList_insertFirst (usedLib, lusedLib);
  idContextModule = string2ident ("Cost_Context");
  lusedLib = ntreeList_insertFirst (copy_ident (idContextModule), lusedLib);
  ContextModule = STree_mkSkeletonModel (idContextModule, opmodu);
  lids = LL_make_lident_list (string2ident (H_TYPECODING_CST));
  STree_importDeclsFromTo (lids, hierarchy, ContextModule);
  STree_removeDeclsFrom (lids, hierarchy);
  STree_addLocalDeclaration (hierarchy, lusedLib);
  fn = nilstr;
  SignalUnparseFile (ContextModule, fn, infoMorphismResult2);
}

/**
   Specific to the morphism functionality.\n
   @return "true" if the operation succeeds, "false" otherwise.
   - It loads (mkMorphismModule()) the predefined morphim table called "MORPHISM.table",
   - When this operation succeeds, it builds an operator table (ht object; attribute member of the pKServicesAPI object)
   - When this operation succeeds, it signals that the table is loaded (MLoaded is set to "true")
*/
boolean pKServicesAPI::loadProfilingModule ()
{
  ptDCG_poly aModule;

  if (MLoaded)
    return pk_true;
  if (!mkMorphismModule (&aModule))
    return pk_false;
  ht = new HomomorphismTable ();
  if (!ht->mkOperatorTable (aModule->getDCG ()))
    return pk_false;
  MLoaded = pk_true;
  return pk_true;
}



/**
   Specific to the morphism functionality.
   @return "true" if the operation succeeds, "false" otherwise.
   It applies a morphism defined by a table (ht) on a graph vgraph.
   The table gives the translation associated to each Signal operator.
*/
boolean pKServicesAPI::mkMorphism (ptDCG_bool vgraph, ptHomomorphismTable ht)
{
  ptDCG_poly g1;
  boolean b;
  ntrees hierarchy;
  ptident id;
  boolean bsim;

  bsim = pk_true;
  // vgraph->expressionExpand();
  vgraph->unparse ();
  b = vgraph->applyTabularMorphism (ht, &g1, bsim);
  if (b)
    {
      vgraph->unparse ();
      g1->setFormatLevel (vgraph->getFormatLevel ());
      g1->nonProductiveSignalElimination ();
      id = g1->getName ();
      hierarchy = g1->produce_tree ();
      STree_setNameModel (hierarchy, id);
      addLibraryProfiling (vgraph, hierarchy);
      SignalUnparse (hierarchy, infoMorphismResult1);
    }
  return pk_true;
}

/**
   Assumed: P1 and P2 are sub-graphs of the operand graph (vgraph).
   - it applies DCG_poly::encapsulatingClockSystem() on vgraph; It produces the synchronisation part (P3) of vgraph
   - it applies DCG_poly::updateInterfacesSubGraphs() on (P3,P1) and (P3,P2)
   - and calls printSubGraph() for each Pi

   At the end, the graph represents the program (| P1() | P2() | P3() |)
  */
void
pKServicesAPI::binPartitionningEnd (ptDCG_poly vgraph, ptINode P1, ptINode P2)
{
  ptINode ckSystemSubGraph;
  ckSystemSubGraph = vgraph->encapsulatingClockSystem ();
  vgraph->updateInterfacesSubGraphs (ckSystemSubGraph, P1);
  vgraph->updateInterfacesSubGraphs (ckSystemSubGraph, P2);
  vgraph->buildClockProcessInterface ();
  TRACE (printSubGraph (vgraph, ckSystemSubGraph));
  TRACE (printSubGraph (vgraph, P1));
  TRACE (printSubGraph (vgraph, P2));
}

/**
   The hierarchy of a basic graph is completed to apply the same methods than a Poly graph (without applying the clock calculus):
   a virtual root is created as root of the hierarchy and all the clocks becomes a child of this vitual root.
   
   The resetEnvForUnparse() must be applied to remove the virtual root.
*/
ptDCG_basic
pKServicesAPI::mkEnvForUnparse (ptDCG_basic vgraph)
{
  if (vgraph->getFormatLevel () != DCG_BASIC)
    return vgraph;
  ptDCG_poly g = DCG_poly::new_prePoly (vgraph->getDCG ());
  setIntrinsicClocks (g);
  return g;
}

/**
   Reset the operations of setIntrinsicClocks(), previously applied in mkEnvForUnparse().
*/
void
pKServicesAPI::resetEnvForUnparse (ptDCG_basic g, ptDCG_basic g1)
{
  if (g != g1)
    g1->initGraphForModifications ();
}


/**
   Export (Signal file) the subgraph (vPart) of a graph (vgraph).
   The name of the file is the name associated with the sub-graph.

   Assumed: vPart denotes a subgraph of vgraph.
   
   Useful for verifying the correctness of the partitionings.
*/
void
pKServicesAPI::exportPartition (ptDCG_poly vgraph, ptINode vPart)
{
  ptDCG_basic gRes;
  ptDCG_poly gPoly;

  gRes = vgraph->cloneSubgraph2Graph (vPart, pk_false, pk_true);
  gPoly = new DCG_poly (gRes);	// The hierarchy is already built.
  gPoly->setIntrinsicClocks ();
  gPoly->setName (vgraph->getName (vPart));
  unparse (gPoly);
}


