/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

_PUBLIC_ pKServicesAPImkGraph::pKServicesAPImkGraph()
{
  BModifyOutputDirectory = pk_true;
  MLoaded = pk_false;
  optGenLis = pk_false;
  traceabilityMode = pk_false;
}

/**
   Destructor.
*/
pKServicesAPImkGraph::~pKServicesAPImkGraph()
{
}

/**
   It defines an environment variable (eVar) by an associated data (vVar).
   It can be used to avoid the access to the execution environment (shell variables).
   \note for the initialisation of the machine (method pKServicesAPImkGraph::initialize()), the 
   shell variable \b Signal_bin must be known.
*/
void
 pKServicesAPImkGraph::setEnv(char *eVar, char *vVar)
{
  sysSetEnvDef(eVar, vVar);
}

/**
   @return the data associated with a key (vstr). This value is
   - the data previously set by a call to the method setEnv(),
   - the data given by the execution environment: the key (vstr) is seen as a shell variable,
   - the null string otherwise.
*/
char *pKServicesAPImkGraph::getEnv(char *vstr)
{
  return sysGetEnvDef(vstr);
}

/**
   It defines a messenger option 
   - for code=1, it defines the "statistic mode"  to val (ie .when it is pk_true, time infos are visible).
   - for code=2, it defines the "warning mode" to val (ie when it is pk_true, warnings are visible).
   - for code=3, it defines the "debug mode" to val (ie when it is pk_true, tracing operations are enable).
*/

void pKServicesAPImkGraph::setMessengerOption(integer code, boolean val)
{
  pK_messenger->setOptionMessenger(code, val);
}

/**
   Initializing the use of the API.
   @param modeCall (in) it gives the tool context: the batch Compiler, the original graphical user interface, or the eclipse Plugin.

 If you don't want use the shell environment variables, 
   you must initialize them, using pKServicesAPImkGraph::setEnv() method, before calling this method.
   The following variables must be set:
   - Signal_bin: it must be defined by the address of directory that contains the table of keywords (SIGNAL.kw, ...), the tables of messages (messignal.*): in fact, it is the directory that contains the binaries of the batch Signal compiler of Polychrony.
   
   The followings are optional:
   - pK_VERSION Polychrony version
   - T_MEM_ARB  size of  the tree memory  (integer value)
   - T_MEM_CHAIN size of  the string memory (integer value)
   - SIGNAL_LIBRARY_PATH the paths used for finding Signal Libraries (with the same syntax than PATH).
   - SHELL_TypeTableMaxSize : the size of the table of types (will be removed).

   It calls userbegin() followed by commonini().
   used in BatchCompiler::initialize, InteEclipse...pKServices_g
*/
void pKServicesAPImkGraph::pK_initialize(pKContextType modeCall)
{
  ToolContext = modeCall;
  userbegin_init();
  commonini();
}

/**
   Finalyzing the API.
*/
void pKServicesAPImkGraph::finalyze()
{
  pK_SignalPrinter->end();
  stringMemoryFinalize();
  ntreesMemoryFinalize();
}

void pKServicesAPImkGraph::reset()
{
  userbegin_reset();
}

/** 
    The traceability mode is set to m.
    @return the traceability mode before the call.
*/
boolean pKServicesAPImkGraph::setTraceabilityMode(boolean m)
{
  boolean pmode;

  pmode = traceabilityMode;
  traceabilityMode = m;
  return pmode;
}

/**
   Analyzes a file in Intermedaite Signal Format and generate a tree.
   @return polychronyNoError if the resultt is a correct model or module tree
   @return polychronySyntaxErrorProg otherwise
*/
integer pKServicesAPImkGraph::IntermediateFormat2AbstractTree(char *srcfile, ntrees * abstracTree)
{
  boolean isAModelOrAModule;

  if (SignalFormatParseSrcfile(srcfile, abstracTree, &isAModelOrAModule))
    {
      if (isAModelOrAModule)
	return polychronyNoError;
      pK_messenger->error(ERR_NOT_COMPILABLE);
      if (ISNOT_NILTREE(*abstracTree))
	pK_messenger->errorTree(ERR_NOT_COMPILABLE, *abstracTree);
    }
  unparse_SignalLis(*abstracTree, NILTREE, pk_true);
  return polychronySyntaxErrorProg;
}

/**
   Analyzes a file and generate a tree.
   @return polychronyNoError if the resultt is a correct model or module tree
   @return polychronySyntaxErrorProg otherwise
*/
integer pKServicesAPImkGraph::File2AbstractTree(char *srcfile, ntrees * abstracTree)
{
  boolean isAModelOrAModule;
  if (SIGNALParseSrcfile(srcfile, abstracTree, &isAModelOrAModule))
    {
      if (isAModelOrAModule)
	return polychronyNoError;
      pK_messenger->error(ERR_NOT_COMPILABLE);
      if (ISNOT_NILTREE(*abstracTree))
	pK_messenger->errorTree(ERR_NOT_COMPILABLE, *abstracTree);
    }
  unparse_SignalLis(*abstracTree, NILTREE, pk_true);
  return polychronySyntaxErrorProg;
}

/**
   \see SIGNALFiles2DCG(char *, char *, char*, boolean, ptDCG *)
   Moreover, the table of messages for the parser is loaded if loadMessageTable.
*/
integer pKServicesAPImkGraph::parameterFile2abstractTree(ntrees termAbstractTree, char *paramfile, ntrees * paramAbstractTree)
{
  if (STree_getNbParameters(termAbstractTree) == 0)
    {
      pK_messenger->globalMessage(ProgramWithoutParameter);
      SET_NILTREE(*paramAbstractTree);
      return polychronyNoError;
    }
  if (SIGNALParseParamfile(paramfile, paramAbstractTree))
    return polychronyNoError;
  unparse_SignalLis(termAbstractTree, (*paramAbstractTree), pk_true);
  return polychronySyntaxErrorParam;
}

/**
   \see SIGNALFiles2DCG(char *, char *, char*, boolean, ptDCG *)
   Moreover, the table of messages for the parser is loaded if loadMessageTable.
*/
integer
    pKServicesAPImkGraph::SIGNALFiles2DCG(boolean loadMessageTable, boolean generateDCG, char *srcfile, char *paramfile, char *dirName,
					  boolean LisOption, ptDCG * vgraph)
{
  ntrees mpp, mparam;
  boolean bc, cc1, cc2;
  integer status, messTableOld;

  if (loadMessageTable)
    messTableOld = pK_messenger->gmessload(parsermess);
  cc1 = SIGNALParseSrcfile(srcfile, &mpp, &bc);
  if (!cc1)
    {
      // Parsing error: the result parsed tree is printed (test June 2013 Lbesnard).
      unparse_SignalLis(mpp, NILTREE, pk_true);
    }

  cc2 = SIGNALParseParamfile(paramfile, &mparam);
  if ((strlen(paramfile) != 0) && STree_getNbParameters(mpp) == 0)
    {
      if ((ISNOT_NILTREE(mparam)) && (noper(mparam) == oplepara))
	{
	  pK_messenger->globalMessageStr(ProgramWithoutParameter, string2repstr("Error"));
	  bc = pk_false;
	}
      else
	{
	  cc2 = pk_true;
	  pK_messenger->globalMessageStr(ProgramWithoutParameter, string2repstr("Warning"));
	}
    }

  if (bc)
    {
      if (generateDCG)
	status = SIGNALTree2DCG(mpp, cc1, mparam, cc2, dirName, LisOption, vgraph);
      else
	status = polychronyNoError;
    }
  else
    status = polychronySyntaxErrorProg;

  if (loadMessageTable)
    pK_messenger->gmessrestore(messTableOld);
  return status;
}

_PUBLIC_
/**
   Abstract tree representation to graph translation (entry point for the GUI integrated compiler).

   The following steps are applied : 
     - translating to kernel-tree representation.
     - production the graph.
   @param mpp (in) : abstract tree of the compiled program (produced by the parser).
   @param cc1 (in) : returned error-code of the parser for mpp.
   @param mparam (in) : abstract tree of the parameters (produced by the parser).
   @param cc2 (in) : returned error-code of the parser for mparam.
   @param dirName (in) : the complete name of the directory where to generate the compiling result.
   @param LisOption (in) : when it is "true", it produces the annotated original Signal program (X_LIS.SIG file) when errors or not.
   @param vgraph (out) : the graph of an instance of the model defined in srcfile.
   @return a non null value when errors are detected, 0 otherwise.   
*/
    integer pKServicesAPImkGraph::SIGNALTree2DCG(ntrees mpp, boolean cc1, ntrees mparam, boolean cc2, char *dirName, boolean LisOption,
						 ptDCG * vgraph)
{
  integer status;
  ptident nomprogsig;
  ptrepstr dir;

  dir = string2repstr(dirName);
  optGenLis = LisOption;

  cc = pk_true;
  if (!cc1)
    return polychronySyntaxErrorProg;
  pK_messenger->gmessload(mkklmess);
  if (!pK_Tree2Kernel->initialize())
    return polychronyInitError;
  status = polychronyNoError;
  this_termAbstractTree = mpp;
  this_paramAbstractTree = mparam;
  status = makeModel(mpp, &nomprogsig);
  // setting output directory
  if (BModifyOutputDirectory)
    {
      if (dir == nilstr)
	dir = ident2repstr(nomprogsig);
      sys_set_userOutputDir(dir);
    }

  if (!cc)
    unparse_SignalLis(mpp, this_paramAbstractTree, pk_true);
  // else
  // Il faudrait intervenir dans la premiere passe 
  // (cc retourné devrait tenir compte des bibliotheques)
  if (printReferencedLibrary(pKServices_vfi, pk_false))
    {
      cc = pk_false;
      status = polychronySolvingRefError;
    }
  if (cc && cc2)
    {
      status = makeInstanceModel(mpp, mparam, vgraph);
      if (optGenLis)
	{
	  unparse_SignalLis(mpp, this_paramAbstractTree, pk_false);
	  printReferencedLibrary(pKServices_vfi, pk_true);
	}
    }
  else if (!cc2)
    status = polychronySyntaxErrorParam;
  else
    status = polychronySolvingRefError;

  pK_Tree2Kernel->finalize();
  return status;
}

/**
   Abstract tree representation to graph translation (entry point for the GUI integrated compiler).

   The following steps are applied : 
     - translating to kernel-tree representation.
     - production the graph.
   @param mpp (in) : abstract tree of the compiled program (produced by the parser).
   @param mparam (in) : abstract tree of the parameters (produced by the parser).
   @param dirName (in) : the complete name of the directory where to generate the compiling result.
   @param LisOption (in) : when it is "true", it produces the annotated original Signal program (X_LIS.SIG file) when errors or not.
   @param vgraph (out) : the graph of an instance of the model defined in srcfile.
   @return a non null value when errors are detected, 0 otherwise.   
*/
integer pKServicesAPImkGraph::SIGNALTree2DCG(ntrees termAbstracTree, ntrees paramAbstractTree, boolean LisOption, ptDCG * vgraph)
{
  integer status;
  ptident nomprogsig;

  optGenLis = LisOption;

  cc = pk_true;
  pK_messenger->gmessload(mkklmess);
  if (!pK_Tree2Kernel->initialize())
    return polychronyInitError;
  status = polychronyNoError;
  this_termAbstractTree = termAbstracTree;
  this_paramAbstractTree = paramAbstractTree;
  initcallTrees2Kernel();
  status = makeModel(termAbstracTree, &nomprogsig);
  if (printReferencedLibrary(pKServices_vfi, pk_false))
    {
      cc = pk_false;
      unparse_SignalLis(termAbstracTree, this_paramAbstractTree, pk_true);
      status = polychronySolvingRefError;
    }
  else
    {
      status = makeInstanceModel(termAbstracTree, paramAbstractTree, vgraph);
      if (status == polychronyNoError)
	{
	  if (LisOption)
	    {
	      unparse_SignalLis(termAbstracTree, this_paramAbstractTree, pk_false);
	      printReferencedLibrary(pKServices_vfi, pk_true);
	    }
	}
      else
	unparse_SignalLis(termAbstracTree, this_paramAbstractTree, pk_true);
    }

  pK_Tree2Kernel->finalize();
  return status;
}

/**
   Signal Textual form to graph translation.

   The following steps are applied : 
     - Parsing program and parameters
     - translating to kernel-tree representation (SIGNALTree2DCG function)
     - production the graph (SIGNALTree2DCG function)
   @param srcfile (in) : name of the file which contains the program.
   @param paramfile (in) : name of the file which contains the parameters.
   @param dirName (in) : the complete name of the directory where to generate the compiling result.
   @param LisOption (in) : when it is "true", it produces the annotated original Signal program (X_LIS.SIG file) when errors or not.

   @param vgraph (out) : the graph of an instance of the model defined in srcfile.
   @return a non null value when errors are detected, 0 otherwise.   
*/
integer pKServicesAPImkGraph::SIGNALFiles2DCG(char *srcfile, char *paramfile, char *dirName, boolean LisOption, ptDCG * vgraph)
{
  return SIGNALFiles2DCG(pk_true, pk_true, srcfile, paramfile, dirName, LisOption, vgraph);
}
